/* Generated by Cython 0.29.20 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\npy_math.h"
        ],
        "include_dirs": [
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include"
        ],
        "name": "flydra_core._reconstruct_utils",
        "sources": [
            "flydra_core/_reconstruct_utils.pyx"
        ]
    },
    "module_name": "flydra_core._reconstruct_utils"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_20"
#define CYTHON_HEX_VERSION 0x001D14F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flydra_core___reconstruct_utils
#define __PYX_HAVE_API__flydra_core___reconstruct_utils
/* Early includes */
#include <math.h>
#include "numpy/npy_math.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "flydra_core\\_reconstruct_utils.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper;

/* "flydra_core/_reconstruct_utils.pyx":56
 *     return ReconstructHelper(*args,**kw)
 * 
 * cdef class ReconstructHelper:             # <<<<<<<<<<<<<<
 *     cdef readonly double fc1, fc2, cc1, cc2
 *     cdef readonly double fc1p, fc2p, cc1p, cc2p
 */
struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper {
  PyObject_HEAD
  double fc1;
  double fc2;
  double cc1;
  double cc2;
  double fc1p;
  double fc2p;
  double cc1p;
  double cc2p;
  double k1;
  double k2;
  double k3;
  double p1;
  double p2;
  double alpha_c;
  int simple;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* PyExec.proto */
static PyObject* __Pyx_PyExec3(PyObject*, PyObject*, PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyExec2(PyObject*, PyObject*);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.math' */

/* Module declarations from 'flydra_core._reconstruct_utils' */
static PyTypeObject *__pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper = 0;
static double __pyx_v_11flydra_core_18_reconstruct_utils_cinf;
static float __pyx_v_11flydra_core_18_reconstruct_utils_MINIMUM_ECCENTRICITY;
static int __pyx_v_11flydra_core_18_reconstruct_utils_gave_water_warning;
#define __Pyx_MODULE_NAME "flydra_core._reconstruct_utils"
extern int __pyx_module_is_main_flydra_core___reconstruct_utils;
int __pyx_module_is_main_flydra_core___reconstruct_utils = 0;

/* Implementation of 'flydra_core._reconstruct_utils' */
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_KeyError;
static const char __pyx_k_0[] = "0";
static const char __pyx_k_A[] = "A";
static const char __pyx_k_P[] = "P";
static const char __pyx_k_Q[] = "Q";
static const char __pyx_k_X[] = "X";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_u[] = "u";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_ET[] = "ET";
static const char __pyx_k_k1[] = "k1";
static const char __pyx_k_k2[] = "k2";
static const char __pyx_k_k3[] = "k3";
static const char __pyx_k_kw[] = "kw";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_p1[] = "p1";
static const char __pyx_k_p2[] = "p2";
static const char __pyx_k_p3[] = "p3";
static const char __pyx_k_p4[] = "p4";
static const char __pyx_k_vt[] = "vt";
static const char __pyx_k_xl[] = "xl";
static const char __pyx_k_xy[] = "xy";
static const char __pyx_k_yl[] = "yl";
static const char __pyx_k_K11[] = "K11";
static const char __pyx_k_K12[] = "K12";
static const char __pyx_k_K13[] = "K13";
static const char __pyx_k_K22[] = "K22";
static const char __pyx_k_K23[] = "K23";
static const char __pyx_k__17[] = "\n";
static const char __pyx_k__29[] = "*";
static const char __pyx_k_cc1[] = "cc1";
static const char __pyx_k_cc2[] = "cc2";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_dot[] = "dot";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_fc1[] = "fc1";
static const char __pyx_k_fc2[] = "fc2";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_kc1[] = "kc1";
static const char __pyx_k_kc2[] = "kc2";
static const char __pyx_k_kc3[] = "kc3";
static const char __pyx_k_kc4[] = "kc4";
static const char __pyx_k_svd[] = "svd";
static const char __pyx_k_tag[] = "tag";
static const char __pyx_k_Pmat[] = "Pmat";
static const char __pyx_k_allA[] = "allA";
static const char __pyx_k_area[] = "area";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cc1p[] = "cc1p";
static const char __pyx_k_cc2p[] = "cc2p";
static const char __pyx_k_dist[] = "dist";
static const char __pyx_k_dual[] = "dual";
static const char __pyx_k_elem[] = "elem";
static const char __pyx_k_fc1p[] = "fc1p";
static const char __pyx_k_fc2p[] = "fc2p";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_find[] = "find";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_rect[] = "rect";
static const char __pyx_k_row3[] = "row3";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_x_kk[] = "x_kk";
static const char __pyx_k_y_kk[] = "y_kk";
static const char __pyx_k_K11_s[] = "K11 = %s;\n";
static const char __pyx_k_K12_s[] = "K12 = %s;\n";
static const char __pyx_k_K13_s[] = "K13 = %s;\n";
static const char __pyx_k_K21_s[] = "K21 = %s;\n";
static const char __pyx_k_K22_s[] = "K22 = %s;\n";
static const char __pyx_k_K23_s[] = "K23 = %s;\n";
static const char __pyx_k_K31_s[] = "K31 = %s;\n";
static const char __pyx_k_K32_s[] = "K32 = %s;\n";
static const char __pyx_k_K33_s[] = "K33 = %s;\n";
static const char __pyx_k_all2d[] = "all2d";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_fname[] = "fname";
static const char __pyx_k_get_K[] = "get_K";
static const char __pyx_k_isnan[] = "isnan";
static const char __pyx_k_kc1_s[] = "kc1 = %s;\n";
static const char __pyx_k_kc2_s[] = "kc2 = %s;\n";
static const char __pyx_k_kc3_s[] = "kc3 = %s;\n";
static const char __pyx_k_kc4_s[] = "kc4 = %s;\n";
static const char __pyx_k_kc5_s[] = "kc5 = %s;\n";
static const char __pyx_k_new_x[] = "new_x";
static const char __pyx_k_new_y[] = "new_y";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_recon[] = "recon";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_slope[] = "slope";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_cam_id[] = "cam_id";
static const char __pyx_k_find3d[] = "find3d";
static const char __pyx_k_get_Kp[] = "get_Kp";
static const char __pyx_k_helper[] = "helper";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_k3_0_0[] = "k3 != 0.0";
static const char __pyx_k_linalg[] = "linalg";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_n_cams[] = "n_cams";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_orig_x[] = "orig_x";
static const char __pyx_k_orig_y[] = "orig_y";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_Lcoords[] = "Lcoords";
static const char __pyx_k_alpha_c[] = "alpha_c";
static const char __pyx_k_cam_ids[] = "cam_ids";
static const char __pyx_k_environ[] = "environ";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_new_xyw[] = "new_xyw";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_allclose[] = "allclose";
static const char __pyx_k_builtins[] = "__builtins__";
static const char __pyx_k_comments[] = "comments";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_iselement[] = "iselement";
static const char __pyx_k_least_err[] = "least_err";
static const char __pyx_k_mean_dist[] = "mean_dist";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_undistort[] = "undistort";
static const char __pyx_k_xy_values[] = "xy_values";
static const char __pyx_k_Error_on_P[] = "Error on P";
static const char __pyx_k_SubElement[] = "SubElement";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_cam_id2idx[] = "cam_id2idx";
static const char __pyx_k_comments_s[] = "comments = '%s';\n";
static const char __pyx_k_good_A_idx[] = "good_A_idx";
static const char __pyx_k_max_n_cams[] = "max_n_cams";
static const char __pyx_k_numpy_dual[] = "numpy.dual";
static const char __pyx_k_with_water[] = "with_water";
static const char __pyx_k_LinAlgError[] = "LinAlgError";
static const char __pyx_k_Pmat_fastnx[] = "Pmat_fastnx";
static const char __pyx_k_SVD_error_P[] = "SVD error, P=";
static const char __pyx_k_bad_cam_ids[] = "bad_cam_ids";
static const char __pyx_k_best_n_cams[] = "best_n_cams";
static const char __pyx_k_flydra_core[] = "flydra_core";
static const char __pyx_k_value_tuple[] = "value_tuple";
static const char __pyx_k_STRICT_WATER[] = "STRICT_WATER";
static const char __pyx_k_alloc_n_cams[] = "alloc_n_cams";
static const char __pyx_k_cam_ids_used[] = "cam_ids_used";
static const char __pyx_k_eccentricity[] = "eccentricity";
static const char __pyx_k_find_best_3d[] = "find_best_3d";
static const char __pyx_k_get_nlparams[] = "get_nlparams";
static const char __pyx_k_n_iterations[] = "n_iterations";
static const char __pyx_k_full_matrices[] = "full_matrices";
static const char __pyx_k_X_for_least_err[] = "X_for_least_err";
static const char __pyx_k_found_data_dict[] = "found_data_dict";
static const char __pyx_k_common_variables[] = "common_variables";
static const char __pyx_k_missing_cam_data[] = "missing_cam_data";
static const char __pyx_k_ReconstructHelper[] = "ReconstructHelper";
static const char __pyx_k_least_error_was_f[] = "least error was %f";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_return_line_coords[] = "return_line_coords";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_MINIMUM_ECCENTRICITY[] = "MINIMUM_ECCENTRICITY";
static const char __pyx_k_cam_ids_and_points2d[] = "cam_ids_and_points2d";
static const char __pyx_k_cam_ids_for_least_err[] = "cam_ids_for_least_err";
static const char __pyx_k_flydra_core_undistort[] = "flydra_core.undistort";
static const char __pyx_k_non_linear_parameters[] = "non_linear_parameters";
static const char __pyx_k_xml_etree_ElementTree[] = "xml.etree.ElementTree";
static const char __pyx_k_NoAcceptablePointFound[] = "NoAcceptablePointFound";
static const char __pyx_k_least_err_by_n_cameras[] = "least_err_by_n_cameras";
static const char __pyx_k_make_ReconstructHelper[] = "make_ReconstructHelper";
static const char __pyx_k_cam_combinations_by_size[] = "cam_combinations_by_size";
static const char __pyx_k_ACCEPTABLE_DISTANCE_PIXELS[] = "ACCEPTABLE_DISTANCE_PIXELS";
static const char __pyx_k_HYPOTHESIS_TEST_n_cams_d_f[] = "HYPOTHESIS TEST - n_cams %d: %f";
static const char __pyx_k_ReconstructHelper_from_xml[] = "ReconstructHelper_from_xml";
static const char __pyx_k_STRICT_WATER_HYPOTHESIS_TEST[] = "STRICT_WATER_HYPOTHESIS_TEST";
static const char __pyx_k_flydra_core_common_variables[] = "flydra_core.common_variables";
static const char __pyx_k_flydra_core__reconstruct_utils[] = "flydra_core._reconstruct_utils";
static const char __pyx_k_HYPOTHESIS_TEST_mean_dist_f_for[] = "HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)";
static const char __pyx_k_make_ReconstructHelper_from_rad[] = "make_ReconstructHelper_from_rad_file";
static const char __pyx_k_reconstruct_utils_Hypothesis_te[] = "_reconstruct_utils: Hypothesis test intersection done without refraction correction. Result will be wrong. Set environment variable STRICT_WATER_HYPOTHESIS_TEST to raise an error rather than give this warning.";
static const char __pyx_k_skew_not_tested_with_non_simple[] = "skew not tested with non-simple distortion model";
static const char __pyx_k_HYPOTHESIS_TEST_least_err_ACCEPT[] = "HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f";
static const char __pyx_k_HYPOTHESIS_TEST_shorting_for_n_c[] = "HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)";
static const char __pyx_k_cannot_encode_non_simple_intrins[] = "cannot encode non-simple intrinsic model as a .rad file";
static const char __pyx_k_flydra_core__reconstruct_utils_p[] = "flydra_core\\_reconstruct_utils.pyx";
static const char __pyx_k_hypothesis_testing_algorithm__fi[] = "hypothesis_testing_algorithm__find_best_3d";
static const char __pyx_k_need_to_properly_roundtrip_alpha[] = "need to properly roundtrip alpha_c != 0";
static const char __pyx_k_water_and_hypothesis_testing_not[] = "water and hypothesis testing not yet implemented";
static PyObject *__pyx_kp_s_0;
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_ACCEPTABLE_DISTANCE_PIXELS;
static PyObject *__pyx_n_s_ET;
static PyObject *__pyx_kp_s_Error_on_P;
static PyObject *__pyx_kp_s_HYPOTHESIS_TEST_least_err_ACCEPT;
static PyObject *__pyx_kp_s_HYPOTHESIS_TEST_mean_dist_f_for;
static PyObject *__pyx_kp_s_HYPOTHESIS_TEST_n_cams_d_f;
static PyObject *__pyx_kp_s_HYPOTHESIS_TEST_shorting_for_n_c;
static PyObject *__pyx_n_s_K11;
static PyObject *__pyx_kp_s_K11_s;
static PyObject *__pyx_n_s_K12;
static PyObject *__pyx_kp_s_K12_s;
static PyObject *__pyx_n_s_K13;
static PyObject *__pyx_kp_s_K13_s;
static PyObject *__pyx_kp_s_K21_s;
static PyObject *__pyx_n_s_K22;
static PyObject *__pyx_kp_s_K22_s;
static PyObject *__pyx_n_s_K23;
static PyObject *__pyx_kp_s_K23_s;
static PyObject *__pyx_kp_s_K31_s;
static PyObject *__pyx_kp_s_K32_s;
static PyObject *__pyx_kp_s_K33_s;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_Lcoords;
static PyObject *__pyx_n_s_LinAlgError;
static PyObject *__pyx_n_s_MINIMUM_ECCENTRICITY;
static PyObject *__pyx_n_s_NoAcceptablePointFound;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_P;
static PyObject *__pyx_n_s_Pmat;
static PyObject *__pyx_n_s_Pmat_fastnx;
static PyObject *__pyx_n_s_Q;
static PyObject *__pyx_n_s_ReconstructHelper;
static PyObject *__pyx_n_s_ReconstructHelper_from_xml;
static PyObject *__pyx_n_s_STRICT_WATER;
static PyObject *__pyx_n_s_STRICT_WATER_HYPOTHESIS_TEST;
static PyObject *__pyx_kp_s_SVD_error_P;
static PyObject *__pyx_n_s_SubElement;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_X;
static PyObject *__pyx_n_s_X_for_least_err;
static PyObject *__pyx_kp_s__17;
static PyObject *__pyx_n_s__29;
static PyObject *__pyx_n_s_all2d;
static PyObject *__pyx_n_s_allA;
static PyObject *__pyx_n_s_allclose;
static PyObject *__pyx_n_s_alloc_n_cams;
static PyObject *__pyx_n_s_alpha;
static PyObject *__pyx_n_s_alpha_c;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_area;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_bad_cam_ids;
static PyObject *__pyx_n_s_best_n_cams;
static PyObject *__pyx_n_s_builtins;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cam_combinations_by_size;
static PyObject *__pyx_n_s_cam_id;
static PyObject *__pyx_n_s_cam_id2idx;
static PyObject *__pyx_n_s_cam_ids;
static PyObject *__pyx_n_s_cam_ids_and_points2d;
static PyObject *__pyx_n_s_cam_ids_for_least_err;
static PyObject *__pyx_n_s_cam_ids_used;
static PyObject *__pyx_kp_s_cannot_encode_non_simple_intrins;
static PyObject *__pyx_n_s_cc1;
static PyObject *__pyx_n_s_cc1p;
static PyObject *__pyx_n_s_cc2;
static PyObject *__pyx_n_s_cc2p;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_comments;
static PyObject *__pyx_kp_s_comments_s;
static PyObject *__pyx_n_s_common_variables;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_dist;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dot;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dual;
static PyObject *__pyx_n_s_eccentricity;
static PyObject *__pyx_n_s_elem;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_environ;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_fc1;
static PyObject *__pyx_n_s_fc1p;
static PyObject *__pyx_n_s_fc2;
static PyObject *__pyx_n_s_fc2p;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_find;
static PyObject *__pyx_n_s_find3d;
static PyObject *__pyx_n_s_find_best_3d;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_flydra_core;
static PyObject *__pyx_n_s_flydra_core__reconstruct_utils;
static PyObject *__pyx_kp_s_flydra_core__reconstruct_utils_p;
static PyObject *__pyx_n_s_flydra_core_common_variables;
static PyObject *__pyx_n_s_flydra_core_undistort;
static PyObject *__pyx_n_s_fname;
static PyObject *__pyx_n_s_found_data_dict;
static PyObject *__pyx_n_s_full_matrices;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_K;
static PyObject *__pyx_n_s_get_Kp;
static PyObject *__pyx_n_s_get_nlparams;
static PyObject *__pyx_n_s_good_A_idx;
static PyObject *__pyx_n_s_helper;
static PyObject *__pyx_n_s_hypothesis_testing_algorithm__fi;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_iselement;
static PyObject *__pyx_n_s_isnan;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_k1;
static PyObject *__pyx_n_s_k2;
static PyObject *__pyx_n_s_k3;
static PyObject *__pyx_kp_s_k3_0_0;
static PyObject *__pyx_n_s_kc1;
static PyObject *__pyx_kp_s_kc1_s;
static PyObject *__pyx_n_s_kc2;
static PyObject *__pyx_kp_s_kc2_s;
static PyObject *__pyx_n_s_kc3;
static PyObject *__pyx_kp_s_kc3_s;
static PyObject *__pyx_n_s_kc4;
static PyObject *__pyx_kp_s_kc4_s;
static PyObject *__pyx_kp_s_kc5_s;
static PyObject *__pyx_n_s_kw;
static PyObject *__pyx_n_s_least_err;
static PyObject *__pyx_n_s_least_err_by_n_cameras;
static PyObject *__pyx_kp_s_least_error_was_f;
static PyObject *__pyx_n_s_linalg;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_ReconstructHelper;
static PyObject *__pyx_n_s_make_ReconstructHelper_from_rad;
static PyObject *__pyx_n_s_max_n_cams;
static PyObject *__pyx_n_s_mean_dist;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_missing_cam_data;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_n_cams;
static PyObject *__pyx_n_s_n_iterations;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_s_need_to_properly_roundtrip_alpha;
static PyObject *__pyx_n_s_new_x;
static PyObject *__pyx_n_s_new_xyw;
static PyObject *__pyx_n_s_new_y;
static PyObject *__pyx_n_s_non_linear_parameters;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_numpy_dual;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_orig_x;
static PyObject *__pyx_n_s_orig_y;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_p1;
static PyObject *__pyx_n_s_p2;
static PyObject *__pyx_n_s_p3;
static PyObject *__pyx_n_s_p4;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_recon;
static PyObject *__pyx_kp_s_reconstruct_utils_Hypothesis_te;
static PyObject *__pyx_n_s_rect;
static PyObject *__pyx_n_s_return_line_coords;
static PyObject *__pyx_n_s_row3;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_kp_s_skew_not_tested_with_non_simple;
static PyObject *__pyx_n_s_slope;
static PyObject *__pyx_n_s_svd;
static PyObject *__pyx_n_s_tag;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_u;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_n_s_undistort;
static PyObject *__pyx_n_s_value_tuple;
static PyObject *__pyx_n_s_vt;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_kp_s_water_and_hypothesis_testing_not;
static PyObject *__pyx_n_s_with_water;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_x_kk;
static PyObject *__pyx_n_s_xl;
static PyObject *__pyx_n_s_xml_etree_ElementTree;
static PyObject *__pyx_n_s_xy;
static PyObject *__pyx_n_s_xy_values;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_y_kk;
static PyObject *__pyx_n_s_yl;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_make_ReconstructHelper_from_rad_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_2ReconstructHelper_from_xml(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_elem); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_4make_ReconstructHelper(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kw); /* proto */
static int __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_fc1, PyObject *__pyx_v_fc2, PyObject *__pyx_v_cc1, PyObject *__pyx_v_cc2, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2, PyObject *__pyx_v_alpha_c, PyObject *__pyx_v_fc1p, PyObject *__pyx_v_fc2p, PyObject *__pyx_v_cc1p, PyObject *__pyx_v_cc2p, PyObject *__pyx_v_k3); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2__reduce__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4add_element(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_parent); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6as_obj_for_json(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_8save_to_rad_file(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_fname, PyObject *__pyx_v_comments); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_10__richcmp__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_12get_K(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_14get_Kp(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_16get_nlparams(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_18undistort(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, double __pyx_v_x_kk, double __pyx_v_y_kk, int __pyx_v_n_iterations); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_20distort(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, double __pyx_v_xl, double __pyx_v_yl); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_22undistort_image(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_numpy_image); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_6hypothesis_testing_algorithm__find_best_3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_recon, PyObject *__pyx_v_found_data_dict, double __pyx_v_ACCEPTABLE_DISTANCE_PIXELS, int __pyx_v_debug, Py_ssize_t __pyx_v_max_n_cams, int __pyx_v_with_water); /* proto */
static PyObject *__pyx_tp_new_11flydra_core_18_reconstruct_utils_ReconstructHelper(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__20;
static PyObject *__pyx_slice__21;
static PyObject *__pyx_slice__27;
static PyObject *__pyx_slice__28;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
/* Late includes */

/* "flydra_core/_reconstruct_utils.pyx":32
 *     pass
 * 
 * def make_ReconstructHelper_from_rad_file(filename):             # <<<<<<<<<<<<<<
 *     params = {}
 *     exec(open(filename).read(),params)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_1make_ReconstructHelper_from_rad_file(PyObject *__pyx_self, PyObject *__pyx_v_filename); /*proto*/
static PyMethodDef __pyx_mdef_11flydra_core_18_reconstruct_utils_1make_ReconstructHelper_from_rad_file = {"make_ReconstructHelper_from_rad_file", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_1make_ReconstructHelper_from_rad_file, METH_O, 0};
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_1make_ReconstructHelper_from_rad_file(PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_ReconstructHelper_from_rad_file (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_make_ReconstructHelper_from_rad_file(__pyx_self, ((PyObject *)__pyx_v_filename));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_make_ReconstructHelper_from_rad_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_params = NULL;
  struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_helper = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_ReconstructHelper_from_rad_file", 0);

  /* "flydra_core/_reconstruct_utils.pyx":33
 * 
 * def make_ReconstructHelper_from_rad_file(filename):
 *     params = {}             # <<<<<<<<<<<<<<
 *     exec(open(filename).read(),params)
 *     if params['K12'] != 0:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_params = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":34
 * def make_ReconstructHelper_from_rad_file(filename):
 *     params = {}
 *     exec(open(filename).read(),params)             # <<<<<<<<<<<<<<
 *     if params['K12'] != 0:
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_open, __pyx_v_filename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_read); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyExec3(__pyx_t_1, __pyx_v_params, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":35
 *     params = {}
 *     exec(open(filename).read(),params)
 *     if params['K12'] != 0:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')
 *     helper = ReconstructHelper(
 */
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_K12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_t_3, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "flydra_core/_reconstruct_utils.pyx":36
 *     exec(open(filename).read(),params)
 *     if params['K12'] != 0:
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')             # <<<<<<<<<<<<<<
 *     helper = ReconstructHelper(
 *         params['K11'], params['K22'], params['K13'], params['K23'],
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 36, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":35
 *     params = {}
 *     exec(open(filename).read(),params)
 *     if params['K12'] != 0:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')
 *     helper = ReconstructHelper(
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":38
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')
 *     helper = ReconstructHelper(
 *         params['K11'], params['K22'], params['K13'], params['K23'],             # <<<<<<<<<<<<<<
 *         params['kc1'], params['kc2'], params['kc3'], params['kc4'])
 *     return helper
 */
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_K11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_K22); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_K13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_K23); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "flydra_core/_reconstruct_utils.pyx":39
 *     helper = ReconstructHelper(
 *         params['K11'], params['K22'], params['K13'], params['K23'],
 *         params['kc1'], params['kc2'], params['kc3'], params['kc4'])             # <<<<<<<<<<<<<<
 *     return helper
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_kc1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_kc2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_kc3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_kc4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "flydra_core/_reconstruct_utils.pyx":37
 *     if params['K12'] != 0:
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')
 *     helper = ReconstructHelper(             # <<<<<<<<<<<<<<
 *         params['K11'], params['K22'], params['K13'], params['K23'],
 *         params['kc1'], params['kc2'], params['kc3'], params['kc4'])
 */
  __pyx_t_10 = PyTuple_New(8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper), __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_helper = ((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":40
 *         params['K11'], params['K22'], params['K13'], params['K23'],
 *         params['kc1'], params['kc2'], params['kc3'], params['kc4'])
 *     return helper             # <<<<<<<<<<<<<<
 * 
 * def ReconstructHelper_from_xml(elem):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_helper));
  __pyx_r = ((PyObject *)__pyx_v_helper);
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":32
 *     pass
 * 
 * def make_ReconstructHelper_from_rad_file(filename):             # <<<<<<<<<<<<<<
 *     params = {}
 *     exec(open(filename).read(),params)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.make_ReconstructHelper_from_rad_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XDECREF((PyObject *)__pyx_v_helper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":42
 *     return helper
 * 
 * def ReconstructHelper_from_xml(elem):             # <<<<<<<<<<<<<<
 *     assert ET.iselement(elem)
 *     assert elem.tag == "non_linear_parameters"
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_3ReconstructHelper_from_xml(PyObject *__pyx_self, PyObject *__pyx_v_elem); /*proto*/
static PyMethodDef __pyx_mdef_11flydra_core_18_reconstruct_utils_3ReconstructHelper_from_xml = {"ReconstructHelper_from_xml", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_3ReconstructHelper_from_xml, METH_O, 0};
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_3ReconstructHelper_from_xml(PyObject *__pyx_self, PyObject *__pyx_v_elem) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ReconstructHelper_from_xml (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_2ReconstructHelper_from_xml(__pyx_self, ((PyObject *)__pyx_v_elem));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_2ReconstructHelper_from_xml(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_elem) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_name = NULL;
  struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_helper = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ReconstructHelper_from_xml", 0);

  /* "flydra_core/_reconstruct_utils.pyx":43
 * 
 * def ReconstructHelper_from_xml(elem):
 *     assert ET.iselement(elem)             # <<<<<<<<<<<<<<
 *     assert elem.tag == "non_linear_parameters"
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_iselement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_elem) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_elem);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 43, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":44
 * def ReconstructHelper_from_xml(elem):
 *     assert ET.iselement(elem)
 *     assert elem.tag == "non_linear_parameters"             # <<<<<<<<<<<<<<
 * 
 *     args = []
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_elem, __pyx_n_s_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_non_linear_parameters, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 44, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":46
 *     assert elem.tag == "non_linear_parameters"
 * 
 *     args = []             # <<<<<<<<<<<<<<
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:
 *         args.append( float(elem.find(name).text) )
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_args = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":47
 * 
 *     args = []
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:             # <<<<<<<<<<<<<<
 *         args.append( float(elem.find(name).text) )
 *     args = tuple(args)
 */
  __pyx_t_1 = __pyx_tuple__2; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 9) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 47, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":48
 *     args = []
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:
 *         args.append( float(elem.find(name).text) )             # <<<<<<<<<<<<<<
 *     args = tuple(args)
 *     helper = ReconstructHelper(*args)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_elem, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_text); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_args, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":47
 * 
 *     args = []
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:             # <<<<<<<<<<<<<<
 *         args.append( float(elem.find(name).text) )
 *     args = tuple(args)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":49
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:
 *         args.append( float(elem.find(name).text) )
 *     args = tuple(args)             # <<<<<<<<<<<<<<
 *     helper = ReconstructHelper(*args)
 *     return helper
 */
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_args, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":50
 *         args.append( float(elem.find(name).text) )
 *     args = tuple(args)
 *     helper = ReconstructHelper(*args)             # <<<<<<<<<<<<<<
 *     return helper
 * 
 */
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_helper = ((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":51
 *     args = tuple(args)
 *     helper = ReconstructHelper(*args)
 *     return helper             # <<<<<<<<<<<<<<
 * 
 * def make_ReconstructHelper(*args,**kw):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_helper));
  __pyx_r = ((PyObject *)__pyx_v_helper);
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":42
 *     return helper
 * 
 * def ReconstructHelper_from_xml(elem):             # <<<<<<<<<<<<<<
 *     assert ET.iselement(elem)
 *     assert elem.tag == "non_linear_parameters"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper_from_xml", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF((PyObject *)__pyx_v_helper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":53
 *     return helper
 * 
 * def make_ReconstructHelper(*args,**kw):             # <<<<<<<<<<<<<<
 *     return ReconstructHelper(*args,**kw)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_5make_ReconstructHelper(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_11flydra_core_18_reconstruct_utils_5make_ReconstructHelper = {"make_ReconstructHelper", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_18_reconstruct_utils_5make_ReconstructHelper, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_5make_ReconstructHelper(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kw = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_ReconstructHelper (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "make_ReconstructHelper", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kw = PyDict_Copy(__pyx_kwds); if (unlikely(!__pyx_v_kw)) return NULL;
    __Pyx_GOTREF(__pyx_v_kw);
  } else {
    __pyx_v_kw = NULL;
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_4make_ReconstructHelper(__pyx_self, __pyx_v_args, __pyx_v_kw);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kw);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_4make_ReconstructHelper(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kw) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_ReconstructHelper", 0);

  /* "flydra_core/_reconstruct_utils.pyx":54
 * 
 * def make_ReconstructHelper(*args,**kw):
 *     return ReconstructHelper(*args,**kw)             # <<<<<<<<<<<<<<
 * 
 * cdef class ReconstructHelper:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper), __pyx_v_args, __pyx_v_kw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":53
 *     return helper
 * 
 * def make_ReconstructHelper(*args,**kw):             # <<<<<<<<<<<<<<
 *     return ReconstructHelper(*args,**kw)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.make_ReconstructHelper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":63
 *     cdef readonly int simple
 * 
 *     def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,             # <<<<<<<<<<<<<<
 *                  fc1p=None, fc2p=None, cc1p=None, cc2p=None, k3=None,
 *                  ):
 */

/* Python wrapper */
static int __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__[] = "create instance of ReconstructHelper\n\n        ReconstructHelper(fc1, fc2, cc1, cc2, k1, k2, p1, p2 )\n        where:\n        fc - focal length\n        cc - camera center\n        k - radial distortion parameters (non-linear)\n        p - tangential distortion parameters (non-linear)\n        alpha_c - skew between X and Y pixel axes\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__;
#endif
static int __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fc1 = 0;
  PyObject *__pyx_v_fc2 = 0;
  PyObject *__pyx_v_cc1 = 0;
  PyObject *__pyx_v_cc2 = 0;
  PyObject *__pyx_v_k1 = 0;
  PyObject *__pyx_v_k2 = 0;
  PyObject *__pyx_v_p1 = 0;
  PyObject *__pyx_v_p2 = 0;
  PyObject *__pyx_v_alpha_c = 0;
  PyObject *__pyx_v_fc1p = 0;
  PyObject *__pyx_v_fc2p = 0;
  PyObject *__pyx_v_cc1p = 0;
  PyObject *__pyx_v_cc2p = 0;
  PyObject *__pyx_v_k3 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fc1,&__pyx_n_s_fc2,&__pyx_n_s_cc1,&__pyx_n_s_cc2,&__pyx_n_s_k1,&__pyx_n_s_k2,&__pyx_n_s_p1,&__pyx_n_s_p2,&__pyx_n_s_alpha_c,&__pyx_n_s_fc1p,&__pyx_n_s_fc2p,&__pyx_n_s_cc1p,&__pyx_n_s_cc2p,&__pyx_n_s_k3,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[8] = ((PyObject *)__pyx_int_0);

    /* "flydra_core/_reconstruct_utils.pyx":64
 * 
 *     def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,
 *                  fc1p=None, fc2p=None, cc1p=None, cc2p=None, k3=None,             # <<<<<<<<<<<<<<
 *                  ):
 *         """create instance of ReconstructHelper
 */
    values[9] = ((PyObject *)Py_None);
    values[10] = ((PyObject *)Py_None);
    values[11] = ((PyObject *)Py_None);
    values[12] = ((PyObject *)Py_None);
    values[13] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fc1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fc2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 1); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 2); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 3); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 4); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 5); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 6); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, 7); __PYX_ERR(0, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_alpha_c);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fc1p);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fc2p);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc1p);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc2p);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k3);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 63, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fc1 = values[0];
    __pyx_v_fc2 = values[1];
    __pyx_v_cc1 = values[2];
    __pyx_v_cc2 = values[3];
    __pyx_v_k1 = values[4];
    __pyx_v_k2 = values[5];
    __pyx_v_p1 = values[6];
    __pyx_v_p2 = values[7];
    __pyx_v_alpha_c = values[8];
    __pyx_v_fc1p = values[9];
    __pyx_v_fc2p = values[10];
    __pyx_v_cc1p = values[11];
    __pyx_v_cc2p = values[12];
    __pyx_v_k3 = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 8, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 63, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), __pyx_v_fc1, __pyx_v_fc2, __pyx_v_cc1, __pyx_v_cc2, __pyx_v_k1, __pyx_v_k2, __pyx_v_p1, __pyx_v_p2, __pyx_v_alpha_c, __pyx_v_fc1p, __pyx_v_fc2p, __pyx_v_cc1p, __pyx_v_cc2p, __pyx_v_k3);

  /* "flydra_core/_reconstruct_utils.pyx":63
 *     cdef readonly int simple
 * 
 *     def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,             # <<<<<<<<<<<<<<
 *                  fc1p=None, fc2p=None, cc1p=None, cc2p=None, k3=None,
 *                  ):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_fc1, PyObject *__pyx_v_fc2, PyObject *__pyx_v_cc1, PyObject *__pyx_v_cc2, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2, PyObject *__pyx_v_alpha_c, PyObject *__pyx_v_fc1p, PyObject *__pyx_v_fc2p, PyObject *__pyx_v_cc1p, PyObject *__pyx_v_cc2p, PyObject *__pyx_v_k3) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_k3);

  /* "flydra_core/_reconstruct_utils.pyx":76
 *         alpha_c - skew between X and Y pixel axes
 *         """
 *         self.fc1 = self.fc1p = fc1             # <<<<<<<<<<<<<<
 *         self.fc2 = self.fc2p = fc2
 *         self.cc1 = self.cc1p = cc1
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_fc1); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_v_self->fc1 = __pyx_t_1;
  __pyx_v_self->fc1p = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":77
 *         """
 *         self.fc1 = self.fc1p = fc1
 *         self.fc2 = self.fc2p = fc2             # <<<<<<<<<<<<<<
 *         self.cc1 = self.cc1p = cc1
 *         self.cc2 = self.cc2p = cc2
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_fc2); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_v_self->fc2 = __pyx_t_1;
  __pyx_v_self->fc2p = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":78
 *         self.fc1 = self.fc1p = fc1
 *         self.fc2 = self.fc2p = fc2
 *         self.cc1 = self.cc1p = cc1             # <<<<<<<<<<<<<<
 *         self.cc2 = self.cc2p = cc2
 *         self.k1 = k1
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_cc1); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_v_self->cc1 = __pyx_t_1;
  __pyx_v_self->cc1p = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":79
 *         self.fc2 = self.fc2p = fc2
 *         self.cc1 = self.cc1p = cc1
 *         self.cc2 = self.cc2p = cc2             # <<<<<<<<<<<<<<
 *         self.k1 = k1
 *         self.k2 = k2
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_cc2); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 79, __pyx_L1_error)
  __pyx_v_self->cc2 = __pyx_t_1;
  __pyx_v_self->cc2p = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":80
 *         self.cc1 = self.cc1p = cc1
 *         self.cc2 = self.cc2p = cc2
 *         self.k1 = k1             # <<<<<<<<<<<<<<
 *         self.k2 = k2
 *         if k3 is None:
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_k1); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 80, __pyx_L1_error)
  __pyx_v_self->k1 = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":81
 *         self.cc2 = self.cc2p = cc2
 *         self.k1 = k1
 *         self.k2 = k2             # <<<<<<<<<<<<<<
 *         if k3 is None:
 *             k3 = 0.0
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_k2); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_v_self->k2 = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":82
 *         self.k1 = k1
 *         self.k2 = k2
 *         if k3 is None:             # <<<<<<<<<<<<<<
 *             k3 = 0.0
 *         self.k3 = k3
 */
  __pyx_t_2 = (__pyx_v_k3 == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "flydra_core/_reconstruct_utils.pyx":83
 *         self.k2 = k2
 *         if k3 is None:
 *             k3 = 0.0             # <<<<<<<<<<<<<<
 *         self.k3 = k3
 *         self.p1 = p1
 */
    __Pyx_INCREF(__pyx_float_0_0);
    __Pyx_DECREF_SET(__pyx_v_k3, __pyx_float_0_0);

    /* "flydra_core/_reconstruct_utils.pyx":82
 *         self.k1 = k1
 *         self.k2 = k2
 *         if k3 is None:             # <<<<<<<<<<<<<<
 *             k3 = 0.0
 *         self.k3 = k3
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":84
 *         if k3 is None:
 *             k3 = 0.0
 *         self.k3 = k3             # <<<<<<<<<<<<<<
 *         self.p1 = p1
 *         self.p2 = p2
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_k3); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 84, __pyx_L1_error)
  __pyx_v_self->k3 = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":85
 *             k3 = 0.0
 *         self.k3 = k3
 *         self.p1 = p1             # <<<<<<<<<<<<<<
 *         self.p2 = p2
 *         self.alpha_c = alpha_c
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_p1); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_v_self->p1 = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":86
 *         self.k3 = k3
 *         self.p1 = p1
 *         self.p2 = p2             # <<<<<<<<<<<<<<
 *         self.alpha_c = alpha_c
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_p2); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_v_self->p2 = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":87
 *         self.p1 = p1
 *         self.p2 = p2
 *         self.alpha_c = alpha_c             # <<<<<<<<<<<<<<
 * 
 *         # "simple" means that the forward and back normalizations from
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_alpha_c); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_v_self->alpha_c = __pyx_t_1;

  /* "flydra_core/_reconstruct_utils.pyx":92
 *         # pixel coords to normalized coords are identical.
 * 
 *         self.simple=1             # <<<<<<<<<<<<<<
 *         if fc1p is not None and fc1p!=fc1:
 *             self.simple = 0
 */
  __pyx_v_self->simple = 1;

  /* "flydra_core/_reconstruct_utils.pyx":93
 * 
 *         self.simple=1
 *         if fc1p is not None and fc1p!=fc1:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.fc1p = fc1p
 */
  __pyx_t_2 = (__pyx_v_fc1p != Py_None);
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_fc1p, __pyx_v_fc1, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flydra_core/_reconstruct_utils.pyx":94
 *         self.simple=1
 *         if fc1p is not None and fc1p!=fc1:
 *             self.simple = 0             # <<<<<<<<<<<<<<
 *             self.fc1p = fc1p
 *         if fc2p is not None and fc2p!=fc2:
 */
    __pyx_v_self->simple = 0;

    /* "flydra_core/_reconstruct_utils.pyx":95
 *         if fc1p is not None and fc1p!=fc1:
 *             self.simple = 0
 *             self.fc1p = fc1p             # <<<<<<<<<<<<<<
 *         if fc2p is not None and fc2p!=fc2:
 *             self.simple = 0
 */
    __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_fc1p); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 95, __pyx_L1_error)
    __pyx_v_self->fc1p = __pyx_t_1;

    /* "flydra_core/_reconstruct_utils.pyx":93
 * 
 *         self.simple=1
 *         if fc1p is not None and fc1p!=fc1:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.fc1p = fc1p
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":96
 *             self.simple = 0
 *             self.fc1p = fc1p
 *         if fc2p is not None and fc2p!=fc2:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.fc2p = fc2p
 */
  __pyx_t_4 = (__pyx_v_fc2p != Py_None);
  __pyx_t_2 = (__pyx_t_4 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_fc2p, __pyx_v_fc2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flydra_core/_reconstruct_utils.pyx":97
 *             self.fc1p = fc1p
 *         if fc2p is not None and fc2p!=fc2:
 *             self.simple = 0             # <<<<<<<<<<<<<<
 *             self.fc2p = fc2p
 *         if cc1p is not None and cc1p!=cc1:
 */
    __pyx_v_self->simple = 0;

    /* "flydra_core/_reconstruct_utils.pyx":98
 *         if fc2p is not None and fc2p!=fc2:
 *             self.simple = 0
 *             self.fc2p = fc2p             # <<<<<<<<<<<<<<
 *         if cc1p is not None and cc1p!=cc1:
 *             self.simple = 0
 */
    __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_fc2p); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)
    __pyx_v_self->fc2p = __pyx_t_1;

    /* "flydra_core/_reconstruct_utils.pyx":96
 *             self.simple = 0
 *             self.fc1p = fc1p
 *         if fc2p is not None and fc2p!=fc2:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.fc2p = fc2p
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":99
 *             self.simple = 0
 *             self.fc2p = fc2p
 *         if cc1p is not None and cc1p!=cc1:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.cc1p = cc1p
 */
  __pyx_t_2 = (__pyx_v_cc1p != Py_None);
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_cc1p, __pyx_v_cc1, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __pyx_t_4;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flydra_core/_reconstruct_utils.pyx":100
 *             self.fc2p = fc2p
 *         if cc1p is not None and cc1p!=cc1:
 *             self.simple = 0             # <<<<<<<<<<<<<<
 *             self.cc1p = cc1p
 *         if cc2p is not None and cc2p!=cc2:
 */
    __pyx_v_self->simple = 0;

    /* "flydra_core/_reconstruct_utils.pyx":101
 *         if cc1p is not None and cc1p!=cc1:
 *             self.simple = 0
 *             self.cc1p = cc1p             # <<<<<<<<<<<<<<
 *         if cc2p is not None and cc2p!=cc2:
 *             self.simple = 0
 */
    __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_cc1p); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_v_self->cc1p = __pyx_t_1;

    /* "flydra_core/_reconstruct_utils.pyx":99
 *             self.simple = 0
 *             self.fc2p = fc2p
 *         if cc1p is not None and cc1p!=cc1:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.cc1p = cc1p
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":102
 *             self.simple = 0
 *             self.cc1p = cc1p
 *         if cc2p is not None and cc2p!=cc2:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.cc2p = cc2p
 */
  __pyx_t_4 = (__pyx_v_cc2p != Py_None);
  __pyx_t_2 = (__pyx_t_4 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_cc2p, __pyx_v_cc2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L14_bool_binop_done:;
  if (__pyx_t_3) {

    /* "flydra_core/_reconstruct_utils.pyx":103
 *             self.cc1p = cc1p
 *         if cc2p is not None and cc2p!=cc2:
 *             self.simple = 0             # <<<<<<<<<<<<<<
 *             self.cc2p = cc2p
 *         if not self.simple and self.alpha_c != 0:
 */
    __pyx_v_self->simple = 0;

    /* "flydra_core/_reconstruct_utils.pyx":104
 *         if cc2p is not None and cc2p!=cc2:
 *             self.simple = 0
 *             self.cc2p = cc2p             # <<<<<<<<<<<<<<
 *         if not self.simple and self.alpha_c != 0:
 *             raise NotImplementedError('skew not tested with non-simple distortion model')
 */
    __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_cc2p); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 104, __pyx_L1_error)
    __pyx_v_self->cc2p = __pyx_t_1;

    /* "flydra_core/_reconstruct_utils.pyx":102
 *             self.simple = 0
 *             self.cc1p = cc1p
 *         if cc2p is not None and cc2p!=cc2:             # <<<<<<<<<<<<<<
 *             self.simple = 0
 *             self.cc2p = cc2p
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":105
 *             self.simple = 0
 *             self.cc2p = cc2p
 *         if not self.simple and self.alpha_c != 0:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('skew not tested with non-simple distortion model')
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->simple != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L17_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->alpha_c != 0.0) != 0);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L17_bool_binop_done:;
  if (unlikely(__pyx_t_3)) {

    /* "flydra_core/_reconstruct_utils.pyx":106
 *             self.cc2p = cc2p
 *         if not self.simple and self.alpha_c != 0:
 *             raise NotImplementedError('skew not tested with non-simple distortion model')             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 106, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":105
 *             self.simple = 0
 *             self.cc2p = cc2p
 *         if not self.simple and self.alpha_c != 0:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('skew not tested with non-simple distortion model')
 * 
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":63
 *     cdef readonly int simple
 * 
 *     def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,             # <<<<<<<<<<<<<<
 *                  fc1p=None, fc2p=None, cc1p=None, cc2p=None, k3=None,
 *                  ):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k3);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":108
 *             raise NotImplementedError('skew not tested with non-simple distortion model')
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """this allows ReconstructHelper to be pickled"""
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2__reduce__[] = "this allows ReconstructHelper to be pickled";
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2__reduce__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2__reduce__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "flydra_core/_reconstruct_utils.pyx":110
 *     def __reduce__(self):
 *         """this allows ReconstructHelper to be pickled"""
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,             # <<<<<<<<<<<<<<
 *                 self.k1, self.k2, self.p1, self.p2, self.alpha_c,
 *                 self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fc1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->fc2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->cc1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->cc2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "flydra_core/_reconstruct_utils.pyx":111
 *         """this allows ReconstructHelper to be pickled"""
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,
 *                 self.k1, self.k2, self.p1, self.p2, self.alpha_c,             # <<<<<<<<<<<<<<
 *                 self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)
 *         return (make_ReconstructHelper, args)
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->k1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->k2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->p1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_self->p2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->alpha_c); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "flydra_core/_reconstruct_utils.pyx":112
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,
 *                 self.k1, self.k2, self.p1, self.p2, self.alpha_c,
 *                 self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)             # <<<<<<<<<<<<<<
 *         return (make_ReconstructHelper, args)
 * 
 */
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->fc1p); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyFloat_FromDouble(__pyx_v_self->fc2p); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_self->cc1p); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->cc2p); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = PyFloat_FromDouble(__pyx_v_self->k3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "flydra_core/_reconstruct_utils.pyx":110
 *     def __reduce__(self):
 *         """this allows ReconstructHelper to be pickled"""
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,             # <<<<<<<<<<<<<<
 *                 self.k1, self.k2, self.p1, self.p2, self.alpha_c,
 *                 self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)
 */
  __pyx_t_15 = PyTuple_New(14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 8, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_14);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_args = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":113
 *                 self.k1, self.k2, self.p1, self.p2, self.alpha_c,
 *                 self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)
 *         return (make_ReconstructHelper, args)             # <<<<<<<<<<<<<<
 * 
 *     def add_element(self, parent):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_make_ReconstructHelper); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_args);
  __pyx_t_15 = 0;
  __pyx_r = __pyx_t_14;
  __pyx_t_14 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":108
 *             raise NotImplementedError('skew not tested with non-simple distortion model')
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """this allows ReconstructHelper to be pickled"""
 *         args = (self.fc1, self.fc2, self.cc1, self.cc2,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":115
 *         return (make_ReconstructHelper, args)
 * 
 *     def add_element(self, parent):             # <<<<<<<<<<<<<<
 *         """add self as XML element to parent"""
 *         assert ET.iselement(parent)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_5add_element(PyObject *__pyx_v_self, PyObject *__pyx_v_parent); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4add_element[] = "add self as XML element to parent";
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_5add_element(PyObject *__pyx_v_self, PyObject *__pyx_v_parent) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_element (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4add_element(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), ((PyObject *)__pyx_v_parent));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4add_element(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_parent) {
  PyObject *__pyx_v_elem = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_element", 0);

  /* "flydra_core/_reconstruct_utils.pyx":117
 *     def add_element(self, parent):
 *         """add self as XML element to parent"""
 *         assert ET.iselement(parent)             # <<<<<<<<<<<<<<
 *         elem = ET.SubElement(parent,"non_linear_parameters")
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_iselement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_parent) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parent);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 117, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":118
 *         """add self as XML element to parent"""
 *         assert ET.iselement(parent)
 *         elem = ET.SubElement(parent,"non_linear_parameters")             # <<<<<<<<<<<<<<
 * 
 *         if self.k3 != 0.0:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SubElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_parent, __pyx_n_s_non_linear_parameters};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_parent, __pyx_n_s_non_linear_parameters};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_parent);
    __Pyx_GIVEREF(__pyx_v_parent);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_parent);
    __Pyx_INCREF(__pyx_n_s_non_linear_parameters);
    __Pyx_GIVEREF(__pyx_n_s_non_linear_parameters);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_s_non_linear_parameters);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_elem = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":120
 *         elem = ET.SubElement(parent,"non_linear_parameters")
 * 
 *         if self.k3 != 0.0:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('k3 != 0.0')
 * 
 */
  __pyx_t_4 = ((__pyx_v_self->k3 != 0.0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "flydra_core/_reconstruct_utils.pyx":121
 * 
 *         if self.k3 != 0.0:
 *             raise NotImplementedError('k3 != 0.0')             # <<<<<<<<<<<<<<
 * 
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 121, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":120
 *         elem = ET.SubElement(parent,"non_linear_parameters")
 * 
 *         if self.k3 != 0.0:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('k3 != 0.0')
 * 
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":123
 *             raise NotImplementedError('k3 != 0.0')
 * 
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p',
 *                      ]:
 */
  __pyx_t_1 = __pyx_tuple__5; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= 13) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":126
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p',
 *                      ]:
 *             e = ET.SubElement(elem, name)             # <<<<<<<<<<<<<<
 *             e.text = repr( getattr(self,name) )
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ET); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_SubElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_elem, __pyx_v_name};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_elem, __pyx_v_name};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_elem);
      __Pyx_GIVEREF(__pyx_v_elem);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_5, __pyx_v_elem);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_5, __pyx_v_name);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":127
 *                      ]:
 *             e = ET.SubElement(elem, name)
 *             e.text = repr( getattr(self,name) )             # <<<<<<<<<<<<<<
 * 
 *     def as_obj_for_json(self):
 */
    __pyx_t_2 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_Repr(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_text, __pyx_t_3) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":123
 *             raise NotImplementedError('k3 != 0.0')
 * 
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p',
 *                      ]:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":115
 *         return (make_ReconstructHelper, args)
 * 
 *     def add_element(self, parent):             # <<<<<<<<<<<<<<
 *         """add self as XML element to parent"""
 *         assert ET.iselement(parent)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.add_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elem);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":129
 *             e.text = repr( getattr(self,name) )
 * 
 *     def as_obj_for_json(self):             # <<<<<<<<<<<<<<
 *         result = {}
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7as_obj_for_json(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7as_obj_for_json(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("as_obj_for_json (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6as_obj_for_json(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6as_obj_for_json(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("as_obj_for_json", 0);

  /* "flydra_core/_reconstruct_utils.pyx":130
 * 
 *     def as_obj_for_json(self):
 *         result = {}             # <<<<<<<<<<<<<<
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":131
 *     def as_obj_for_json(self):
 *         result = {}
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 *             result[name] = getattr(self,name)
 */
  __pyx_t_1 = __pyx_tuple__6; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= 14) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 131, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":133
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 *             result[name] = getattr(self,name)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_name, __pyx_t_3) < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":131
 *     def as_obj_for_json(self):
 *         result = {}
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 *             result[name] = getattr(self,name)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":134
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 *             result[name] = getattr(self,name)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def save_to_rad_file( self, fname, comments=None ):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":129
 *             e.text = repr( getattr(self,name) )
 * 
 *     def as_obj_for_json(self):             # <<<<<<<<<<<<<<
 *         result = {}
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.as_obj_for_json", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":136
 *         return result
 * 
 *     def save_to_rad_file( self, fname, comments=None ):             # <<<<<<<<<<<<<<
 *         if not self.simple:
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_9save_to_rad_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_9save_to_rad_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fname = 0;
  PyObject *__pyx_v_comments = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("save_to_rad_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fname,&__pyx_n_s_comments,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fname)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_comments);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "save_to_rad_file") < 0)) __PYX_ERR(0, 136, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fname = values[0];
    __pyx_v_comments = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("save_to_rad_file", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.save_to_rad_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_8save_to_rad_file(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), __pyx_v_fname, __pyx_v_comments);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_8save_to_rad_file(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_fname, PyObject *__pyx_v_comments) {
  PyObject *__pyx_v_rad_fd = NULL;
  PyObject *__pyx_v_K = NULL;
  PyObject *__pyx_v_nlparams = NULL;
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_p1 = NULL;
  PyObject *__pyx_v_p2 = NULL;
  PyObject *__pyx_v_k3 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("save_to_rad_file", 0);
  __Pyx_INCREF(__pyx_v_comments);

  /* "flydra_core/_reconstruct_utils.pyx":137
 * 
 *     def save_to_rad_file( self, fname, comments=None ):
 *         if not self.simple:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 *         rad_fd = open(fname,'w')
 */
  __pyx_t_1 = ((!(__pyx_v_self->simple != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "flydra_core/_reconstruct_utils.pyx":138
 *     def save_to_rad_file( self, fname, comments=None ):
 *         if not self.simple:
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')             # <<<<<<<<<<<<<<
 *         rad_fd = open(fname,'w')
 *         K = self.get_K()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 138, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":137
 * 
 *     def save_to_rad_file( self, fname, comments=None ):
 *         if not self.simple:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 *         rad_fd = open(fname,'w')
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":139
 *         if not self.simple:
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 *         rad_fd = open(fname,'w')             # <<<<<<<<<<<<<<
 *         K = self.get_K()
 *         nlparams = self.get_nlparams()
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_fname);
  __Pyx_GIVEREF(__pyx_v_fname);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_fname);
  __Pyx_INCREF(__pyx_n_s_w);
  __Pyx_GIVEREF(__pyx_n_s_w);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_w);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rad_fd = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":140
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 *         rad_fd = open(fname,'w')
 *         K = self.get_K()             # <<<<<<<<<<<<<<
 *         nlparams = self.get_nlparams()
 *         k1, k2, p1, p2, k3 = nlparams
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_K); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_K = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":141
 *         rad_fd = open(fname,'w')
 *         K = self.get_K()
 *         nlparams = self.get_nlparams()             # <<<<<<<<<<<<<<
 *         k1, k2, p1, p2, k3 = nlparams
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_nlparams); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nlparams = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":142
 *         K = self.get_K()
 *         nlparams = self.get_nlparams()
 *         k1, k2, p1, p2, k3 = nlparams             # <<<<<<<<<<<<<<
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_nlparams))) || (PyList_CheckExact(__pyx_v_nlparams))) {
    PyObject* sequence = __pyx_v_nlparams;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 5)) {
      if (size > 5) __Pyx_RaiseTooManyValuesError(5);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 142, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 4); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 4); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
      for (i=0; i < 5; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
    __pyx_t_7 = PyObject_GetIter(__pyx_v_nlparams); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    for (index=0; index < 5; index++) {
      PyObject* item = __pyx_t_8(__pyx_t_7); if (unlikely(!item)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 5) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 142, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_v_k1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_k2 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_p1 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_p2 = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_k3 = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":143
 *         nlparams = self.get_nlparams()
 *         k1, k2, p1, p2, k3 = nlparams
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K11_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":144
 *         k1, k2, p1, p2, k3 = nlparams
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K12_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":145
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K13_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":146
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K21_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":147
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K22_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":148
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K23_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":149
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K31_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":150
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 *         rad_fd.write('\n')
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K32_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":151
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('\n')
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_K, __pyx_tuple__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_K33_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":152
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 *         rad_fd.write('\n')             # <<<<<<<<<<<<<<
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))
 *         rad_fd.write('kc2 = %s;\n'%repr(k2))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_s__17) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s__17);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":153
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 *         rad_fd.write('\n')
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))             # <<<<<<<<<<<<<<
 *         rad_fd.write('kc2 = %s;\n'%repr(k2))
 *         rad_fd.write('kc3 = %s;\n'%repr(p1))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_Repr(__pyx_v_k1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_kc1_s, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":154
 *         rad_fd.write('\n')
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))
 *         rad_fd.write('kc2 = %s;\n'%repr(k2))             # <<<<<<<<<<<<<<
 *         rad_fd.write('kc3 = %s;\n'%repr(p1))
 *         rad_fd.write('kc4 = %s;\n'%repr(p2))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_Repr(__pyx_v_k2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_kc2_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":155
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))
 *         rad_fd.write('kc2 = %s;\n'%repr(k2))
 *         rad_fd.write('kc3 = %s;\n'%repr(p1))             # <<<<<<<<<<<<<<
 *         rad_fd.write('kc4 = %s;\n'%repr(p2))
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_Repr(__pyx_v_p1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_kc3_s, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":156
 *         rad_fd.write('kc2 = %s;\n'%repr(k2))
 *         rad_fd.write('kc3 = %s;\n'%repr(p1))
 *         rad_fd.write('kc4 = %s;\n'%repr(p2))             # <<<<<<<<<<<<<<
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))
 *         rad_fd.write('\n')
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_Repr(__pyx_v_p2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_kc4_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":157
 *         rad_fd.write('kc3 = %s;\n'%repr(p1))
 *         rad_fd.write('kc4 = %s;\n'%repr(p2))
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))             # <<<<<<<<<<<<<<
 *         rad_fd.write('\n')
 *         if comments is not None:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_Repr(__pyx_v_k3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_kc5_s, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":158
 *         rad_fd.write('kc4 = %s;\n'%repr(p2))
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))
 *         rad_fd.write('\n')             # <<<<<<<<<<<<<<
 *         if comments is not None:
 *             comments = str(comments)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_kp_s__17) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s__17);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":159
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))
 *         rad_fd.write('\n')
 *         if comments is not None:             # <<<<<<<<<<<<<<
 *             comments = str(comments)
 *         rad_fd.write("comments = '%s';\n"%comments)
 */
  __pyx_t_1 = (__pyx_v_comments != Py_None);
  __pyx_t_9 = (__pyx_t_1 != 0);
  if (__pyx_t_9) {

    /* "flydra_core/_reconstruct_utils.pyx":160
 *         rad_fd.write('\n')
 *         if comments is not None:
 *             comments = str(comments)             # <<<<<<<<<<<<<<
 *         rad_fd.write("comments = '%s';\n"%comments)
 *         rad_fd.write('\n')
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_comments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_comments, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":159
 *         rad_fd.write('kc5 = %s;\n'%repr(k3))
 *         rad_fd.write('\n')
 *         if comments is not None:             # <<<<<<<<<<<<<<
 *             comments = str(comments)
 *         rad_fd.write("comments = '%s';\n"%comments)
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":161
 *         if comments is not None:
 *             comments = str(comments)
 *         rad_fd.write("comments = '%s';\n"%comments)             # <<<<<<<<<<<<<<
 *         rad_fd.write('\n')
 *         rad_fd.close()
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_comments_s, __pyx_v_comments); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":162
 *             comments = str(comments)
 *         rad_fd.write("comments = '%s';\n"%comments)
 *         rad_fd.write('\n')             # <<<<<<<<<<<<<<
 *         rad_fd.close()
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_kp_s__17) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s__17);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":163
 *         rad_fd.write("comments = '%s';\n"%comments)
 *         rad_fd.write('\n')
 *         rad_fd.close()             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self,other,op):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_rad_fd, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":136
 *         return result
 * 
 *     def save_to_rad_file( self, fname, comments=None ):             # <<<<<<<<<<<<<<
 *         if not self.simple:
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.save_to_rad_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rad_fd);
  __Pyx_XDECREF(__pyx_v_K);
  __Pyx_XDECREF(__pyx_v_nlparams);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_p1);
  __Pyx_XDECREF(__pyx_v_p2);
  __Pyx_XDECREF(__pyx_v_k3);
  __Pyx_XDECREF(__pyx_v_comments);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":165
 *         rad_fd.close()
 * 
 *     def __richcmp__(self,other,op):             # <<<<<<<<<<<<<<
 *         # cmp op
 *         #  < 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 165, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_10__richcmp__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_10__richcmp__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_v_isequal = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "flydra_core/_reconstruct_utils.pyx":174
 *         # >= 5
 * 
 *         if isinstance(other, ReconstructHelper):             # <<<<<<<<<<<<<<
 *             isequal = (numpy.allclose(self.get_K(),other.get_K()) and
 *                        numpy.allclose(self.get_Kp(),other.get_Kp()) and
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "flydra_core/_reconstruct_utils.pyx":175
 * 
 *         if isinstance(other, ReconstructHelper):
 *             isequal = (numpy.allclose(self.get_K(),other.get_K()) and             # <<<<<<<<<<<<<<
 *                        numpy.allclose(self.get_Kp(),other.get_Kp()) and
 *                        numpy.allclose(self.get_nlparams(),other.get_nlparams()) )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_allclose); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_K); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_get_K); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_5, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_5, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_t_7);
      __pyx_t_5 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_3 = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "flydra_core/_reconstruct_utils.pyx":176
 *         if isinstance(other, ReconstructHelper):
 *             isequal = (numpy.allclose(self.get_K(),other.get_K()) and
 *                        numpy.allclose(self.get_Kp(),other.get_Kp()) and             # <<<<<<<<<<<<<<
 *                        numpy.allclose(self.get_nlparams(),other.get_nlparams()) )
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_allclose); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_Kp); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_get_Kp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_10, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_10, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_3 = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "flydra_core/_reconstruct_utils.pyx":177
 *             isequal = (numpy.allclose(self.get_K(),other.get_K()) and
 *                        numpy.allclose(self.get_Kp(),other.get_Kp()) and
 *                        numpy.allclose(self.get_nlparams(),other.get_nlparams()) )             # <<<<<<<<<<<<<<
 *         else:
 *             isequal = False
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_allclose); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_nlparams); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_get_nlparams); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_9, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_9, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_10, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_10, __pyx_t_7);
      __pyx_t_9 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L4_bool_binop_done:;
    __pyx_v_isequal = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":174
 *         # >= 5
 * 
 *         if isinstance(other, ReconstructHelper):             # <<<<<<<<<<<<<<
 *             isequal = (numpy.allclose(self.get_K(),other.get_K()) and
 *                        numpy.allclose(self.get_Kp(),other.get_Kp()) and
 */
    goto __pyx_L3;
  }

  /* "flydra_core/_reconstruct_utils.pyx":179
 *                        numpy.allclose(self.get_nlparams(),other.get_nlparams()) )
 *         else:
 *             isequal = False             # <<<<<<<<<<<<<<
 * 
 *         if op in [1,2,5]:
 */
  /*else*/ {
    __Pyx_INCREF(Py_False);
    __pyx_v_isequal = Py_False;
  }
  __pyx_L3:;

  /* "flydra_core/_reconstruct_utils.pyx":181
 *             isequal = False
 * 
 *         if op in [1,2,5]:             # <<<<<<<<<<<<<<
 *             result = isequal
 *         elif op == 3:
 */
  __Pyx_INCREF(__pyx_v_op);
  __pyx_t_3 = __pyx_v_op;
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L8_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "flydra_core/_reconstruct_utils.pyx":182
 * 
 *         if op in [1,2,5]:
 *             result = isequal             # <<<<<<<<<<<<<<
 *         elif op == 3:
 *             result = not isequal
 */
    __Pyx_INCREF(__pyx_v_isequal);
    __pyx_v_result = __pyx_v_isequal;

    /* "flydra_core/_reconstruct_utils.pyx":181
 *             isequal = False
 * 
 *         if op in [1,2,5]:             # <<<<<<<<<<<<<<
 *             result = isequal
 *         elif op == 3:
 */
    goto __pyx_L7;
  }

  /* "flydra_core/_reconstruct_utils.pyx":183
 *         if op in [1,2,5]:
 *             result = isequal
 *         elif op == 3:             # <<<<<<<<<<<<<<
 *             result = not isequal
 *         else:
 */
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_op, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "flydra_core/_reconstruct_utils.pyx":184
 *             result = isequal
 *         elif op == 3:
 *             result = not isequal             # <<<<<<<<<<<<<<
 *         else:
 *             result = False
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_isequal); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 184, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_t_1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_result = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":183
 *         if op in [1,2,5]:
 *             result = isequal
 *         elif op == 3:             # <<<<<<<<<<<<<<
 *             result = not isequal
 *         else:
 */
    goto __pyx_L7;
  }

  /* "flydra_core/_reconstruct_utils.pyx":186
 *             result = not isequal
 *         else:
 *             result = False             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  /*else*/ {
    __Pyx_INCREF(Py_False);
    __pyx_v_result = Py_False;
  }
  __pyx_L7:;

  /* "flydra_core/_reconstruct_utils.pyx":187
 *         else:
 *             result = False
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def get_K(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":165
 *         rad_fd.close()
 * 
 *     def __richcmp__(self,other,op):             # <<<<<<<<<<<<<<
 *         # cmp op
 *         #  < 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_isequal);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":189
 *         return result
 * 
 *     def get_K(self):             # <<<<<<<<<<<<<<
 *         # See
 *         # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_13get_K(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_13get_K(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_K (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_12get_K(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_12get_K(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_v_K = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_K", 0);

  /* "flydra_core/_reconstruct_utils.pyx":192
 *         # See
 *         # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
 *         K = numpy.array((( self.fc1, self.alpha_c*self.fc1, self.cc1),             # <<<<<<<<<<<<<<
 *                          ( 0,        self.fc2,              self.cc2),
 *                          ( 0,        0,                     1       )))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->fc1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->alpha_c * __pyx_v_self->fc1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->cc1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":193
 *         # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
 *         K = numpy.array((( self.fc1, self.alpha_c*self.fc1, self.cc1),
 *                          ( 0,        self.fc2,              self.cc2),             # <<<<<<<<<<<<<<
 *                          ( 0,        0,                     1       )))
 *         return K
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->fc2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->cc2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":192
 *         # See
 *         # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
 *         K = numpy.array((( self.fc1, self.alpha_c*self.fc1, self.cc1),             # <<<<<<<<<<<<<<
 *                          ( 0,        self.fc2,              self.cc2),
 *                          ( 0,        0,                     1       )))
 */
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_tuple__18);
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_K = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":195
 *                          ( 0,        self.fc2,              self.cc2),
 *                          ( 0,        0,                     1       )))
 *         return K             # <<<<<<<<<<<<<<
 * 
 *     def get_Kp(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_K);
  __pyx_r = __pyx_v_K;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":189
 *         return result
 * 
 *     def get_K(self):             # <<<<<<<<<<<<<<
 *         # See
 *         # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.get_K", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_K);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":197
 *         return K
 * 
 *     def get_Kp(self):             # <<<<<<<<<<<<<<
 *         K = numpy.array((( self.fc1p, 0,                     self.cc1p),
 *                          ( 0,         self.fc2p,             self.cc2p),
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_15get_Kp(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_15get_Kp(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_Kp (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_14get_Kp(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_14get_Kp(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_v_K = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_Kp", 0);

  /* "flydra_core/_reconstruct_utils.pyx":198
 * 
 *     def get_Kp(self):
 *         K = numpy.array((( self.fc1p, 0,                     self.cc1p),             # <<<<<<<<<<<<<<
 *                          ( 0,         self.fc2p,             self.cc2p),
 *                          ( 0,         0,                     1       )))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->fc1p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->cc1p); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":199
 *     def get_Kp(self):
 *         K = numpy.array((( self.fc1p, 0,                     self.cc1p),
 *                          ( 0,         self.fc2p,             self.cc2p),             # <<<<<<<<<<<<<<
 *                          ( 0,         0,                     1       )))
 *         return K
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->fc2p); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->cc2p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":198
 * 
 *     def get_Kp(self):
 *         K = numpy.array((( self.fc1p, 0,                     self.cc1p),             # <<<<<<<<<<<<<<
 *                          ( 0,         self.fc2p,             self.cc2p),
 *                          ( 0,         0,                     1       )))
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __Pyx_INCREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__18);
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_K = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":201
 *                          ( 0,         self.fc2p,             self.cc2p),
 *                          ( 0,         0,                     1       )))
 *         return K             # <<<<<<<<<<<<<<
 * 
 *     def get_nlparams(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_K);
  __pyx_r = __pyx_v_K;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":197
 *         return K
 * 
 *     def get_Kp(self):             # <<<<<<<<<<<<<<
 *         K = numpy.array((( self.fc1p, 0,                     self.cc1p),
 *                          ( 0,         self.fc2p,             self.cc2p),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.get_Kp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_K);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":203
 *         return K
 * 
 *     def get_nlparams(self):             # <<<<<<<<<<<<<<
 *         return (self.k1, self.k2, self.p1, self.p2, self.k3)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_17get_nlparams(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_17get_nlparams(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nlparams (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_16get_nlparams(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_16get_nlparams(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_nlparams", 0);

  /* "flydra_core/_reconstruct_utils.pyx":204
 * 
 *     def get_nlparams(self):
 *         return (self.k1, self.k2, self.p1, self.p2, self.k3)             # <<<<<<<<<<<<<<
 * 
 *     def undistort(self, double x_kk, double y_kk, int n_iterations=5):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->k1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->k2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->p1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->p2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->k3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":203
 *         return K
 * 
 *     def get_nlparams(self):             # <<<<<<<<<<<<<<
 *         return (self.k1, self.k2, self.p1, self.p2, self.k3)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.get_nlparams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":206
 *         return (self.k1, self.k2, self.p1, self.p2, self.k3)
 * 
 *     def undistort(self, double x_kk, double y_kk, int n_iterations=5):             # <<<<<<<<<<<<<<
 *         """undistort 2D coordinate pair
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_19undistort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_18undistort[] = "undistort 2D coordinate pair\n\n        Iteratively performs an undistortion using camera intrinsic\n        parameters.\n\n        Implementation translated from CalTechCal.\n\n        See also the OpenCV reference manual, which has the equation\n        used.\n        ";
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_19undistort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x_kk;
  double __pyx_v_y_kk;
  int __pyx_v_n_iterations;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("undistort (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x_kk,&__pyx_n_s_y_kk,&__pyx_n_s_n_iterations,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_kk)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_kk)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("undistort", 0, 2, 3, 1); __PYX_ERR(0, 206, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n_iterations);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "undistort") < 0)) __PYX_ERR(0, 206, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_x_kk = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x_kk == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 206, __pyx_L3_error)
    __pyx_v_y_kk = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y_kk == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 206, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_n_iterations = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n_iterations == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 206, __pyx_L3_error)
    } else {
      __pyx_v_n_iterations = ((int)5);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("undistort", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 206, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.undistort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_18undistort(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), __pyx_v_x_kk, __pyx_v_y_kk, __pyx_v_n_iterations);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_18undistort(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, double __pyx_v_x_kk, double __pyx_v_y_kk, int __pyx_v_n_iterations) {
  double __pyx_v_xl;
  double __pyx_v_yl;
  double __pyx_v_xd;
  double __pyx_v_yd;
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_r_2;
  double __pyx_v_k_radial;
  double __pyx_v_delta_x;
  double __pyx_v_delta_y;
  CYTHON_UNUSED int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("undistort", 0);

  /* "flydra_core/_reconstruct_utils.pyx":226
 *         # undoradial.m / CalTechCal/normalize.m
 * 
 *         xd = ( x_kk - self.cc1 ) / self.fc1             # <<<<<<<<<<<<<<
 *         yd = ( y_kk - self.cc2 ) / self.fc2
 * 
 */
  __pyx_t_1 = (__pyx_v_x_kk - __pyx_v_self->cc1);
  if (unlikely(__pyx_v_self->fc1 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 226, __pyx_L1_error)
  }
  __pyx_v_xd = (__pyx_t_1 / __pyx_v_self->fc1);

  /* "flydra_core/_reconstruct_utils.pyx":227
 * 
 *         xd = ( x_kk - self.cc1 ) / self.fc1
 *         yd = ( y_kk - self.cc2 ) / self.fc2             # <<<<<<<<<<<<<<
 * 
 *         xd = xd - self.alpha_c * yd
 */
  __pyx_t_1 = (__pyx_v_y_kk - __pyx_v_self->cc2);
  if (unlikely(__pyx_v_self->fc2 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 227, __pyx_L1_error)
  }
  __pyx_v_yd = (__pyx_t_1 / __pyx_v_self->fc2);

  /* "flydra_core/_reconstruct_utils.pyx":229
 *         yd = ( y_kk - self.cc2 ) / self.fc2
 * 
 *         xd = xd - self.alpha_c * yd             # <<<<<<<<<<<<<<
 * 
 *         # comp_distortion_oulu.m
 */
  __pyx_v_xd = (__pyx_v_xd - (__pyx_v_self->alpha_c * __pyx_v_yd));

  /* "flydra_core/_reconstruct_utils.pyx":234
 * 
 *         # initial guess
 *         x = xd             # <<<<<<<<<<<<<<
 *         y = yd
 * 
 */
  __pyx_v_x = __pyx_v_xd;

  /* "flydra_core/_reconstruct_utils.pyx":235
 *         # initial guess
 *         x = xd
 *         y = yd             # <<<<<<<<<<<<<<
 * 
 *         for i from 0<=i<n_iterations:
 */
  __pyx_v_y = __pyx_v_yd;

  /* "flydra_core/_reconstruct_utils.pyx":237
 *         y = yd
 * 
 *         for i from 0<=i<n_iterations:             # <<<<<<<<<<<<<<
 *             r_2 = x*x + y*y
 *             k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))
 */
  __pyx_t_2 = __pyx_v_n_iterations;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "flydra_core/_reconstruct_utils.pyx":238
 * 
 *         for i from 0<=i<n_iterations:
 *             r_2 = x*x + y*y             # <<<<<<<<<<<<<<
 *             k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))
 *             delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)
 */
    __pyx_v_r_2 = ((__pyx_v_x * __pyx_v_x) + (__pyx_v_y * __pyx_v_y));

    /* "flydra_core/_reconstruct_utils.pyx":239
 *         for i from 0<=i<n_iterations:
 *             r_2 = x*x + y*y
 *             k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))             # <<<<<<<<<<<<<<
 *             delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)
 *             delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y
 */
    __pyx_v_k_radial = (1.0 + (__pyx_v_r_2 * (__pyx_v_self->k1 + (__pyx_v_r_2 * (__pyx_v_self->k2 + (__pyx_v_r_2 * __pyx_v_self->k3))))));

    /* "flydra_core/_reconstruct_utils.pyx":240
 *             r_2 = x*x + y*y
 *             k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))
 *             delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)             # <<<<<<<<<<<<<<
 *             delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y
 *             x = (xd-delta_x)/k_radial
 */
    __pyx_v_delta_x = ((((2.0 * __pyx_v_self->p1) * __pyx_v_x) * __pyx_v_y) + (__pyx_v_self->p2 * (__pyx_v_r_2 + ((2.0 * __pyx_v_x) * __pyx_v_x))));

    /* "flydra_core/_reconstruct_utils.pyx":241
 *             k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))
 *             delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)
 *             delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y             # <<<<<<<<<<<<<<
 *             x = (xd-delta_x)/k_radial
 *             y = (yd-delta_y)/k_radial
 */
    __pyx_v_delta_y = ((__pyx_v_self->p1 * (__pyx_v_r_2 + ((2.0 * __pyx_v_y) * __pyx_v_y))) + (((2.0 * __pyx_v_self->p2) * __pyx_v_x) * __pyx_v_y));

    /* "flydra_core/_reconstruct_utils.pyx":242
 *             delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)
 *             delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y
 *             x = (xd-delta_x)/k_radial             # <<<<<<<<<<<<<<
 *             y = (yd-delta_y)/k_radial
 * 
 */
    __pyx_t_1 = (__pyx_v_xd - __pyx_v_delta_x);
    if (unlikely(__pyx_v_k_radial == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 242, __pyx_L1_error)
    }
    __pyx_v_x = (__pyx_t_1 / __pyx_v_k_radial);

    /* "flydra_core/_reconstruct_utils.pyx":243
 *             delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y
 *             x = (xd-delta_x)/k_radial
 *             y = (yd-delta_y)/k_radial             # <<<<<<<<<<<<<<
 * 
 *         # undoradial.m
 */
    __pyx_t_1 = (__pyx_v_yd - __pyx_v_delta_y);
    if (unlikely(__pyx_v_k_radial == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 243, __pyx_L1_error)
    }
    __pyx_v_y = (__pyx_t_1 / __pyx_v_k_radial);
  }

  /* "flydra_core/_reconstruct_utils.pyx":247
 *         # undoradial.m
 * 
 *         xl = (self.fc1p)*x + (self.fc1p*self.alpha_c)*y + (self.cc1p)             # <<<<<<<<<<<<<<
 *         yl = (self.fc2p)*y + (self.cc2p)
 *         return (xl, yl)
 */
  __pyx_v_xl = (((__pyx_v_self->fc1p * __pyx_v_x) + ((__pyx_v_self->fc1p * __pyx_v_self->alpha_c) * __pyx_v_y)) + __pyx_v_self->cc1p);

  /* "flydra_core/_reconstruct_utils.pyx":248
 * 
 *         xl = (self.fc1p)*x + (self.fc1p*self.alpha_c)*y + (self.cc1p)
 *         yl = (self.fc2p)*y + (self.cc2p)             # <<<<<<<<<<<<<<
 *         return (xl, yl)
 * 
 */
  __pyx_v_yl = ((__pyx_v_self->fc2p * __pyx_v_y) + __pyx_v_self->cc2p);

  /* "flydra_core/_reconstruct_utils.pyx":249
 *         xl = (self.fc1p)*x + (self.fc1p*self.alpha_c)*y + (self.cc1p)
 *         yl = (self.fc2p)*y + (self.cc2p)
 *         return (xl, yl)             # <<<<<<<<<<<<<<
 * 
 *     def distort(self, double xl, double yl):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_xl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_yl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":206
 *         return (self.k1, self.k2, self.p1, self.p2, self.k3)
 * 
 *     def undistort(self, double x_kk, double y_kk, int n_iterations=5):             # <<<<<<<<<<<<<<
 *         """undistort 2D coordinate pair
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.undistort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":251
 *         return (xl, yl)
 * 
 *     def distort(self, double xl, double yl):             # <<<<<<<<<<<<<<
 *         """distort 2D coordinate pair"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_21distort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_20distort[] = "distort 2D coordinate pair";
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_21distort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_xl;
  double __pyx_v_yl;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("distort (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_xl,&__pyx_n_s_yl,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_xl)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_yl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("distort", 1, 2, 2, 1); __PYX_ERR(0, 251, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "distort") < 0)) __PYX_ERR(0, 251, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_xl = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_xl == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L3_error)
    __pyx_v_yl = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_yl == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("distort", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 251, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.distort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_20distort(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), __pyx_v_xl, __pyx_v_yl);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_20distort(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, double __pyx_v_xl, double __pyx_v_yl) {
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_r_2;
  double __pyx_v_term1;
  double __pyx_v_xd;
  double __pyx_v_yd;
  double __pyx_v_r_4;
  double __pyx_v_r_6;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("distort", 0);

  /* "flydra_core/_reconstruct_utils.pyx":256
 *         cdef double x, y, r_2, term1, xd, yd
 * 
 *         assert self.alpha_c==0 # there's no way the following code works otherwise             # <<<<<<<<<<<<<<
 * 
 *         x = ( xl - self.cc1p ) / self.fc1p
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_self->alpha_c == 0.0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 256, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":258
 *         assert self.alpha_c==0 # there's no way the following code works otherwise
 * 
 *         x = ( xl - self.cc1p ) / self.fc1p             # <<<<<<<<<<<<<<
 *         y = ( yl - self.cc2p ) / self.fc2p
 * 
 */
  __pyx_t_1 = (__pyx_v_xl - __pyx_v_self->cc1p);
  if (unlikely(__pyx_v_self->fc1p == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 258, __pyx_L1_error)
  }
  __pyx_v_x = (__pyx_t_1 / __pyx_v_self->fc1p);

  /* "flydra_core/_reconstruct_utils.pyx":259
 * 
 *         x = ( xl - self.cc1p ) / self.fc1p
 *         y = ( yl - self.cc2p ) / self.fc2p             # <<<<<<<<<<<<<<
 * 
 *         r_2 = x*x + y*y
 */
  __pyx_t_1 = (__pyx_v_yl - __pyx_v_self->cc2p);
  if (unlikely(__pyx_v_self->fc2p == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 259, __pyx_L1_error)
  }
  __pyx_v_y = (__pyx_t_1 / __pyx_v_self->fc2p);

  /* "flydra_core/_reconstruct_utils.pyx":261
 *         y = ( yl - self.cc2p ) / self.fc2p
 * 
 *         r_2 = x*x + y*y             # <<<<<<<<<<<<<<
 *         r_4 = r_2**2
 *         r_6 = r_2*r_4
 */
  __pyx_v_r_2 = ((__pyx_v_x * __pyx_v_x) + (__pyx_v_y * __pyx_v_y));

  /* "flydra_core/_reconstruct_utils.pyx":262
 * 
 *         r_2 = x*x + y*y
 *         r_4 = r_2**2             # <<<<<<<<<<<<<<
 *         r_6 = r_2*r_4
 *         term1 = self.k1*r_2 + self.k2*r_4 + self.k3*r_6
 */
  __pyx_v_r_4 = pow(__pyx_v_r_2, 2.0);

  /* "flydra_core/_reconstruct_utils.pyx":263
 *         r_2 = x*x + y*y
 *         r_4 = r_2**2
 *         r_6 = r_2*r_4             # <<<<<<<<<<<<<<
 *         term1 = self.k1*r_2 + self.k2*r_4 + self.k3*r_6
 * 
 */
  __pyx_v_r_6 = (__pyx_v_r_2 * __pyx_v_r_4);

  /* "flydra_core/_reconstruct_utils.pyx":264
 *         r_4 = r_2**2
 *         r_6 = r_2*r_4
 *         term1 = self.k1*r_2 + self.k2*r_4 + self.k3*r_6             # <<<<<<<<<<<<<<
 * 
 *         # OpenCV manual (chaper 6, "3D reconstruction", "camera
 */
  __pyx_v_term1 = (((__pyx_v_self->k1 * __pyx_v_r_2) + (__pyx_v_self->k2 * __pyx_v_r_4)) + (__pyx_v_self->k3 * __pyx_v_r_6));

  /* "flydra_core/_reconstruct_utils.pyx":274
 *         # consistent with his webpage and this below.
 * 
 *         xd = x + x*term1 + (2*self.p1*x*y + self.p2*(r_2+2*x**2))             # <<<<<<<<<<<<<<
 *         yd = y + y*term1 + (self.p1*(r_2+2*y**2) + 2*self.p2*x*y)
 * 
 */
  __pyx_v_xd = ((__pyx_v_x + (__pyx_v_x * __pyx_v_term1)) + ((((2.0 * __pyx_v_self->p1) * __pyx_v_x) * __pyx_v_y) + (__pyx_v_self->p2 * (__pyx_v_r_2 + (2.0 * pow(__pyx_v_x, 2.0))))));

  /* "flydra_core/_reconstruct_utils.pyx":275
 * 
 *         xd = x + x*term1 + (2*self.p1*x*y + self.p2*(r_2+2*x**2))
 *         yd = y + y*term1 + (self.p1*(r_2+2*y**2) + 2*self.p2*x*y)             # <<<<<<<<<<<<<<
 * 
 *         xd = (self.fc1)*xd + (self.cc1)
 */
  __pyx_v_yd = ((__pyx_v_y + (__pyx_v_y * __pyx_v_term1)) + ((__pyx_v_self->p1 * (__pyx_v_r_2 + (2.0 * pow(__pyx_v_y, 2.0)))) + (((2.0 * __pyx_v_self->p2) * __pyx_v_x) * __pyx_v_y)));

  /* "flydra_core/_reconstruct_utils.pyx":277
 *         yd = y + y*term1 + (self.p1*(r_2+2*y**2) + 2*self.p2*x*y)
 * 
 *         xd = (self.fc1)*xd + (self.cc1)             # <<<<<<<<<<<<<<
 *         yd = (self.fc2)*yd + (self.cc2)
 * 
 */
  __pyx_v_xd = ((__pyx_v_self->fc1 * __pyx_v_xd) + __pyx_v_self->cc1);

  /* "flydra_core/_reconstruct_utils.pyx":278
 * 
 *         xd = (self.fc1)*xd + (self.cc1)
 *         yd = (self.fc2)*yd + (self.cc2)             # <<<<<<<<<<<<<<
 * 
 *         return (xd, yd)
 */
  __pyx_v_yd = ((__pyx_v_self->fc2 * __pyx_v_yd) + __pyx_v_self->cc2);

  /* "flydra_core/_reconstruct_utils.pyx":280
 *         yd = (self.fc2)*yd + (self.cc2)
 * 
 *         return (xd, yd)             # <<<<<<<<<<<<<<
 * 
 *     def undistort_image(self, numpy_image ):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_xd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_yd); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":251
 *         return (xl, yl)
 * 
 *     def distort(self, double xl, double yl):             # <<<<<<<<<<<<<<
 *         """distort 2D coordinate pair"""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.distort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":282
 *         return (xd, yd)
 * 
 *     def undistort_image(self, numpy_image ):             # <<<<<<<<<<<<<<
 *         assert len(numpy_image.shape)==2
 *         assert numpy_image.dtype == numpy.uint8
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_23undistort_image(PyObject *__pyx_v_self, PyObject *__pyx_v_numpy_image); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_23undistort_image(PyObject *__pyx_v_self, PyObject *__pyx_v_numpy_image) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("undistort_image (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_22undistort_image(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self), ((PyObject *)__pyx_v_numpy_image));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_22undistort_image(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self, PyObject *__pyx_v_numpy_image) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_imnew = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("undistort_image", 0);

  /* "flydra_core/_reconstruct_utils.pyx":283
 * 
 *     def undistort_image(self, numpy_image ):
 *         assert len(numpy_image.shape)==2             # <<<<<<<<<<<<<<
 *         assert numpy_image.dtype == numpy.uint8
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_numpy_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!((__pyx_t_2 == 2) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 283, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":284
 *     def undistort_image(self, numpy_image ):
 *         assert len(numpy_image.shape)==2
 *         assert numpy_image.dtype == numpy.uint8             # <<<<<<<<<<<<<<
 * 
 *         f = self.fc1, self.fc2 # focal length
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_numpy_image, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 284, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":286
 *         assert numpy_image.dtype == numpy.uint8
 * 
 *         f = self.fc1, self.fc2 # focal length             # <<<<<<<<<<<<<<
 *         c = self.cc1, self.cc2 # center
 *         k = self.k1, self.k2, self.p1, self.p2, self.k3  # NL terms: r^2, r^4, tan1, tan2, r^6
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->fc1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->fc2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_f = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":287
 * 
 *         f = self.fc1, self.fc2 # focal length
 *         c = self.cc1, self.cc2 # center             # <<<<<<<<<<<<<<
 *         k = self.k1, self.k2, self.p1, self.p2, self.k3  # NL terms: r^2, r^4, tan1, tan2, r^6
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cc1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->cc2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_v_c = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":288
 *         f = self.fc1, self.fc2 # focal length
 *         c = self.cc1, self.cc2 # center
 *         k = self.k1, self.k2, self.p1, self.p2, self.k3  # NL terms: r^2, r^4, tan1, tan2, r^6             # <<<<<<<<<<<<<<
 * 
 *         assert self.simple
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->k1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->k2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->p1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->p2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->k3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_v_k = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":290
 *         k = self.k1, self.k2, self.p1, self.p2, self.k3  # NL terms: r^2, r^4, tan1, tan2, r^6
 * 
 *         assert self.simple             # <<<<<<<<<<<<<<
 *         assert self.alpha_c==0
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!(__pyx_v_self->simple != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 290, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":291
 * 
 *         assert self.simple
 *         assert self.alpha_c==0             # <<<<<<<<<<<<<<
 * 
 *         imnew = flydra_core.undistort.rect(numpy_image, f=f, c=c, k=k) # perform the undistortion
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_self->alpha_c == 0.0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 291, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":293
 *         assert self.alpha_c==0
 * 
 *         imnew = flydra_core.undistort.rect(numpy_image, f=f, c=c, k=k) # perform the undistortion             # <<<<<<<<<<<<<<
 *         imnew = imnew.astype(numpy.uint8)
 *         return imnew
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_undistort); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_rect); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_numpy_image);
  __Pyx_GIVEREF(__pyx_v_numpy_image);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_numpy_image);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_f, __pyx_v_f) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_c, __pyx_v_c) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_k, __pyx_v_k) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_imnew = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":294
 * 
 *         imnew = flydra_core.undistort.rect(numpy_image, f=f, c=c, k=k) # perform the undistortion
 *         imnew = imnew.astype(numpy.uint8)             # <<<<<<<<<<<<<<
 *         return imnew
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_imnew, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_uint8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF_SET(__pyx_v_imnew, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":295
 *         imnew = flydra_core.undistort.rect(numpy_image, f=f, c=c, k=k) # perform the undistortion
 *         imnew = imnew.astype(numpy.uint8)
 *         return imnew             # <<<<<<<<<<<<<<
 * 
 * cdef int gave_water_warning
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_imnew);
  __pyx_r = __pyx_v_imnew;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":282
 *         return (xd, yd)
 * 
 *     def undistort_image(self, numpy_image ):             # <<<<<<<<<<<<<<
 *         assert len(numpy_image.shape)==2
 *         assert numpy_image.dtype == numpy.uint8
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.undistort_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_imnew);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":57
 * 
 * cdef class ReconstructHelper:
 *     cdef readonly double fc1, fc2, cc1, cc2             # <<<<<<<<<<<<<<
 *     cdef readonly double fc1p, fc2p, cc1p, cc2p
 *     cdef readonly double k1, k2, k3, p1, p2
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fc1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.fc1.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fc2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.fc2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cc1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.cc1.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cc2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.cc2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":58
 * cdef class ReconstructHelper:
 *     cdef readonly double fc1, fc2, cc1, cc2
 *     cdef readonly double fc1p, fc2p, cc1p, cc2p             # <<<<<<<<<<<<<<
 *     cdef readonly double k1, k2, k3, p1, p2
 *     cdef readonly double alpha_c
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fc1p); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.fc1p.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->fc2p); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.fc2p.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cc1p); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.cc1p.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cc2p); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.cc2p.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":59
 *     cdef readonly double fc1, fc2, cc1, cc2
 *     cdef readonly double fc1p, fc2p, cc1p, cc2p
 *     cdef readonly double k1, k2, k3, p1, p2             # <<<<<<<<<<<<<<
 *     cdef readonly double alpha_c
 *     cdef readonly int simple
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->k1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.k1.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->k2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.k2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->k3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.k3.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->p1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.p1.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->p2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.p2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":60
 *     cdef readonly double fc1p, fc2p, cc1p, cc2p
 *     cdef readonly double k1, k2, k3, p1, p2
 *     cdef readonly double alpha_c             # <<<<<<<<<<<<<<
 *     cdef readonly int simple
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->alpha_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.alpha_c.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":61
 *     cdef readonly double k1, k2, k3, p1, p2
 *     cdef readonly double alpha_c
 *     cdef readonly int simple             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple___get__(((struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple___get__(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->simple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.ReconstructHelper.simple.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_reconstruct_utils.pyx":300
 * gave_water_warning = 0
 * 
 * def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,             # <<<<<<<<<<<<<<
 *                                                 double ACCEPTABLE_DISTANCE_PIXELS,
 *                                                 int debug=0, Py_ssize_t max_n_cams=5,
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_7hypothesis_testing_algorithm__find_best_3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_18_reconstruct_utils_6hypothesis_testing_algorithm__find_best_3d[] = "Use hypothesis testing algorithm to find best 3D point\n\n    Finds combination of cameras which uses the most number of cameras\n    while minimizing mean reprojection error. Algorithm used accepts\n    any camera combination with reprojection error less than the\n    variable ACCEPTABLE_DISTANCE_PIXELS.\n\n    ";
static PyMethodDef __pyx_mdef_11flydra_core_18_reconstruct_utils_7hypothesis_testing_algorithm__find_best_3d = {"hypothesis_testing_algorithm__find_best_3d", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_18_reconstruct_utils_7hypothesis_testing_algorithm__find_best_3d, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11flydra_core_18_reconstruct_utils_6hypothesis_testing_algorithm__find_best_3d};
static PyObject *__pyx_pw_11flydra_core_18_reconstruct_utils_7hypothesis_testing_algorithm__find_best_3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_recon = 0;
  PyObject *__pyx_v_found_data_dict = 0;
  double __pyx_v_ACCEPTABLE_DISTANCE_PIXELS;
  int __pyx_v_debug;
  Py_ssize_t __pyx_v_max_n_cams;
  int __pyx_v_with_water;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hypothesis_testing_algorithm__find_best_3d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_recon,&__pyx_n_s_found_data_dict,&__pyx_n_s_ACCEPTABLE_DISTANCE_PIXELS,&__pyx_n_s_debug,&__pyx_n_s_max_n_cams,&__pyx_n_s_with_water,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_recon)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_found_data_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hypothesis_testing_algorithm__find_best_3d", 0, 3, 6, 1); __PYX_ERR(0, 300, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ACCEPTABLE_DISTANCE_PIXELS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("hypothesis_testing_algorithm__find_best_3d", 0, 3, 6, 2); __PYX_ERR(0, 300, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_debug);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_n_cams);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_with_water);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "hypothesis_testing_algorithm__find_best_3d") < 0)) __PYX_ERR(0, 300, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_recon = values[0];
    __pyx_v_found_data_dict = values[1];
    __pyx_v_ACCEPTABLE_DISTANCE_PIXELS = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_ACCEPTABLE_DISTANCE_PIXELS == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_debug = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_debug == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
    } else {
      __pyx_v_debug = ((int)0);
    }
    if (values[4]) {
      __pyx_v_max_n_cams = __Pyx_PyIndex_AsSsize_t(values[4]); if (unlikely((__pyx_v_max_n_cams == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
    } else {
      __pyx_v_max_n_cams = ((Py_ssize_t)5);
    }
    if (values[5]) {
      __pyx_v_with_water = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_with_water == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L3_error)
    } else {
      __pyx_v_with_water = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hypothesis_testing_algorithm__find_best_3d", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 300, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.hypothesis_testing_algorithm__find_best_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_18_reconstruct_utils_6hypothesis_testing_algorithm__find_best_3d(__pyx_self, __pyx_v_recon, __pyx_v_found_data_dict, __pyx_v_ACCEPTABLE_DISTANCE_PIXELS, __pyx_v_debug, __pyx_v_max_n_cams, __pyx_v_with_water);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_18_reconstruct_utils_6hypothesis_testing_algorithm__find_best_3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_recon, PyObject *__pyx_v_found_data_dict, double __pyx_v_ACCEPTABLE_DISTANCE_PIXELS, int __pyx_v_debug, Py_ssize_t __pyx_v_max_n_cams, int __pyx_v_with_water) {
  int __pyx_v_n_cams;
  int __pyx_v_best_n_cams;
  int __pyx_v_i;
  int __pyx_v_missing_cam_data;
  double __pyx_v_alpha;
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_orig_x;
  double __pyx_v_orig_y;
  double __pyx_v_new_x;
  double __pyx_v_new_y;
  double __pyx_v_dist;
  double __pyx_v_mean_dist;
  double __pyx_v_least_err;
  PyObject *__pyx_v_svd = NULL;
  PyObject *__pyx_v_cam_ids = NULL;
  PyObject *__pyx_v_least_err_by_n_cameras = NULL;
  PyObject *__pyx_v_allA = NULL;
  PyObject *__pyx_v_bad_cam_ids = NULL;
  PyObject *__pyx_v_cam_id2idx = NULL;
  PyObject *__pyx_v_all2d = NULL;
  PyObject *__pyx_v_Pmat_fastnx = NULL;
  PyObject *__pyx_v_cam_id = NULL;
  PyObject *__pyx_v_value_tuple = NULL;
  PyObject *__pyx_v_xy_values = NULL;
  PyObject *__pyx_v_Pmat = NULL;
  PyObject *__pyx_v_row3 = NULL;
  PyObject *__pyx_v_cam_ids_for_least_err = NULL;
  PyObject *__pyx_v_X_for_least_err = NULL;
  PyObject *__pyx_v_cam_ids_used = NULL;
  PyObject *__pyx_v_good_A_idx = NULL;
  PyObject *__pyx_v_A = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_u = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_vt = NULL;
  PyObject *__pyx_v_X = NULL;
  PyObject *__pyx_v_new_xyw = NULL;
  PyObject *__pyx_v_cam_ids_and_points2d = NULL;
  PyObject *__pyx_v_xy = NULL;
  PyObject *__pyx_v_P = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_area = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_slope = NULL;
  PyObject *__pyx_v_eccentricity = NULL;
  PyObject *__pyx_v_p1 = NULL;
  PyObject *__pyx_v_p2 = NULL;
  PyObject *__pyx_v_p3 = NULL;
  PyObject *__pyx_v_p4 = NULL;
  PyObject *__pyx_v_Lcoords = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_Q = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  double __pyx_t_18;
  double __pyx_t_19;
  long __pyx_t_20;
  PyObject *(*__pyx_t_21)(PyObject *);
  int __pyx_t_22;
  int __pyx_t_23;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  int __pyx_t_31;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hypothesis_testing_algorithm__find_best_3d", 0);

  /* "flydra_core/_reconstruct_utils.pyx":321
 *     #cdef double *least_err_by_n_cameras # fake dict (index = key)
 * 
 *     svd = numpy.dual.svd # eliminate global name lookup             # <<<<<<<<<<<<<<
 * 
 *     cam_ids = recon.cam_ids # shorthand
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dual); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_svd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_svd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":323
 *     svd = numpy.dual.svd # eliminate global name lookup
 * 
 *     cam_ids = recon.cam_ids # shorthand             # <<<<<<<<<<<<<<
 *     max_n_cams = min(len(cam_ids), max_n_cams)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_recon, __pyx_n_s_cam_ids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cam_ids = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":324
 * 
 *     cam_ids = recon.cam_ids # shorthand
 *     max_n_cams = min(len(cam_ids), max_n_cams)             # <<<<<<<<<<<<<<
 * 
 *     # Initialize least_err_by_n_cameras to be infinity.  Note that
 */
  __pyx_t_3 = __pyx_v_max_n_cams;
  __pyx_t_4 = PyObject_Length(__pyx_v_cam_ids); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 324, __pyx_L1_error)
  if (((__pyx_t_3 < __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_max_n_cams = __pyx_t_5;

  /* "flydra_core/_reconstruct_utils.pyx":329
 *     # values at 0th and 1st index will always remain infinity.
 * 
 *     least_err_by_n_cameras = [cinf]*(max_n_cams+1) # allow 0-based indexing to last camera             # <<<<<<<<<<<<<<
 *     allA = numpy.zeros( (2*(len(cam_ids)+1),4), dtype=numpy.float64)
 *     bad_cam_ids = []
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_11flydra_core_18_reconstruct_utils_cinf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1 * (((__pyx_v_max_n_cams + 1)<0) ? 0:(__pyx_v_max_n_cams + 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (__pyx_v_max_n_cams + 1); __pyx_temp++) {
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_2, __pyx_temp, __pyx_t_1);
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_least_err_by_n_cameras = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":330
 * 
 *     least_err_by_n_cameras = [cinf]*(max_n_cams+1) # allow 0-based indexing to last camera
 *     allA = numpy.zeros( (2*(len(cam_ids)+1),4), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 *     bad_cam_ids = []
 *     cam_id2idx = {}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_v_cam_ids); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 330, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t((2 * (__pyx_t_5 + 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_4);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_allA = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":331
 *     least_err_by_n_cameras = [cinf]*(max_n_cams+1) # allow 0-based indexing to last camera
 *     allA = numpy.zeros( (2*(len(cam_ids)+1),4), dtype=numpy.float64)
 *     bad_cam_ids = []             # <<<<<<<<<<<<<<
 *     cam_id2idx = {}
 *     all2d = {}
 */
  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_bad_cam_ids = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":332
 *     allA = numpy.zeros( (2*(len(cam_ids)+1),4), dtype=numpy.float64)
 *     bad_cam_ids = []
 *     cam_id2idx = {}             # <<<<<<<<<<<<<<
 *     all2d = {}
 *     Pmat_fastnx = recon.Pmat # shorthand
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_cam_id2idx = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":333
 *     bad_cam_ids = []
 *     cam_id2idx = {}
 *     all2d = {}             # <<<<<<<<<<<<<<
 *     Pmat_fastnx = recon.Pmat # shorthand
 * 
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_all2d = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":334
 *     cam_id2idx = {}
 *     all2d = {}
 *     Pmat_fastnx = recon.Pmat # shorthand             # <<<<<<<<<<<<<<
 * 
 *     global gave_water_warning
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_recon, __pyx_n_s_Pmat); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_Pmat_fastnx = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":337
 * 
 *     global gave_water_warning
 *     if with_water:             # <<<<<<<<<<<<<<
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 */
  __pyx_t_9 = (__pyx_v_with_water != 0);
  if (__pyx_t_9) {

    /* "flydra_core/_reconstruct_utils.pyx":338
 *     global gave_water_warning
 *     if with_water:
 *         if STRICT_WATER:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 *         if not gave_water_warning:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_STRICT_WATER); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_t_9)) {

      /* "flydra_core/_reconstruct_utils.pyx":339
 *     if with_water:
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')             # <<<<<<<<<<<<<<
 *         if not gave_water_warning:
 *             warnings.warn('_reconstruct_utils: Hypothesis test intersection done '
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 339, __pyx_L1_error)

      /* "flydra_core/_reconstruct_utils.pyx":338
 *     global gave_water_warning
 *     if with_water:
 *         if STRICT_WATER:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 *         if not gave_water_warning:
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":340
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 *         if not gave_water_warning:             # <<<<<<<<<<<<<<
 *             warnings.warn('_reconstruct_utils: Hypothesis test intersection done '
 *                           'without refraction correction. Result will be wrong. '
 */
    __pyx_t_9 = ((!(__pyx_v_11flydra_core_18_reconstruct_utils_gave_water_warning != 0)) != 0);
    if (__pyx_t_9) {

      /* "flydra_core/_reconstruct_utils.pyx":341
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 *         if not gave_water_warning:
 *             warnings.warn('_reconstruct_utils: Hypothesis test intersection done '             # <<<<<<<<<<<<<<
 *                           'without refraction correction. Result will be wrong. '
 *                           'Set environment variable STRICT_WATER_HYPOTHESIS_TEST '
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_warnings); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_kp_s_reconstruct_utils_Hypothesis_te) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_reconstruct_utils_Hypothesis_te);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":345
 *                           'Set environment variable STRICT_WATER_HYPOTHESIS_TEST '
 *                           'to raise an error rather than give this warning.')
 *             gave_water_warning = 1             # <<<<<<<<<<<<<<
 *     for i,cam_id in enumerate(cam_ids):
 *         cam_id2idx[cam_id] = i
 */
      __pyx_v_11flydra_core_18_reconstruct_utils_gave_water_warning = 1;

      /* "flydra_core/_reconstruct_utils.pyx":340
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 *         if not gave_water_warning:             # <<<<<<<<<<<<<<
 *             warnings.warn('_reconstruct_utils: Hypothesis test intersection done '
 *                           'without refraction correction. Result will be wrong. '
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":337
 * 
 *     global gave_water_warning
 *     if with_water:             # <<<<<<<<<<<<<<
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":346
 *                           'to raise an error rather than give this warning.')
 *             gave_water_warning = 1
 *     for i,cam_id in enumerate(cam_ids):             # <<<<<<<<<<<<<<
 *         cam_id2idx[cam_id] = i
 * 
 */
  __pyx_t_10 = 0;
  if (likely(PyList_CheckExact(__pyx_v_cam_ids)) || PyTuple_CheckExact(__pyx_v_cam_ids)) {
    __pyx_t_8 = __pyx_v_cam_ids; __Pyx_INCREF(__pyx_t_8); __pyx_t_5 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_cam_ids); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_11 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 346, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_11(__pyx_t_8);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 346, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_10;
    __pyx_t_10 = (__pyx_t_10 + 1);

    /* "flydra_core/_reconstruct_utils.pyx":347
 *             gave_water_warning = 1
 *     for i,cam_id in enumerate(cam_ids):
 *         cam_id2idx[cam_id] = i             # <<<<<<<<<<<<<<
 * 
 *         # do we have incoming data?
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(PyDict_SetItem(__pyx_v_cam_id2idx, __pyx_v_cam_id, __pyx_t_2) < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":350
 * 
 *         # do we have incoming data?
 *         try:             # <<<<<<<<<<<<<<
 *             value_tuple = found_data_dict[cam_id]
 *         except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      /*try:*/ {

        /* "flydra_core/_reconstruct_utils.pyx":351
 *         # do we have incoming data?
 *         try:
 *             value_tuple = found_data_dict[cam_id]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             bad_cam_ids.append( cam_id )
 */
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_found_data_dict, __pyx_v_cam_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_value_tuple, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":350
 * 
 *         # do we have incoming data?
 *         try:             # <<<<<<<<<<<<<<
 *             value_tuple = found_data_dict[cam_id]
 *         except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L15_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":352
 *         try:
 *             value_tuple = found_data_dict[cam_id]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             bad_cam_ids.append( cam_id )
 *             continue # don't build this row
 */
      __pyx_t_15 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_15) {
        __Pyx_AddTraceback("flydra_core._reconstruct_utils.hypothesis_testing_algorithm__find_best_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_6, &__pyx_t_1) < 0) __PYX_ERR(0, 352, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_1);

        /* "flydra_core/_reconstruct_utils.pyx":353
 *             value_tuple = found_data_dict[cam_id]
 *         except KeyError:
 *             bad_cam_ids.append( cam_id )             # <<<<<<<<<<<<<<
 *             continue # don't build this row
 * 
 */
        __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_bad_cam_ids, __pyx_v_cam_id); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 353, __pyx_L10_except_error)

        /* "flydra_core/_reconstruct_utils.pyx":354
 *         except KeyError:
 *             bad_cam_ids.append( cam_id )
 *             continue # don't build this row             # <<<<<<<<<<<<<<
 * 
 *         # was a 2d point found?
 */
        goto __pyx_L17_except_continue;
        __pyx_L17_except_continue:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L14_try_continue;
      }
      goto __pyx_L10_except_error;
      __pyx_L10_except_error:;

      /* "flydra_core/_reconstruct_utils.pyx":350
 * 
 *         # do we have incoming data?
 *         try:             # <<<<<<<<<<<<<<
 *             value_tuple = found_data_dict[cam_id]
 *         except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L1_error;
      __pyx_L14_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      goto __pyx_L6_continue;
      __pyx_L15_try_end:;
    }

    /* "flydra_core/_reconstruct_utils.pyx":357
 * 
 *         # was a 2d point found?
 *         xy_values = value_tuple[:2]             # <<<<<<<<<<<<<<
 *         x,y = xy_values
 *         if npy_isnan(x):
 */
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_value_tuple, 0, 2, NULL, NULL, &__pyx_slice__20, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_xy_values, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":358
 *         # was a 2d point found?
 *         xy_values = value_tuple[:2]
 *         x,y = xy_values             # <<<<<<<<<<<<<<
 *         if npy_isnan(x):
 *             bad_cam_ids.append( cam_id )
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_xy_values))) || (PyList_CheckExact(__pyx_v_xy_values))) {
      PyObject* sequence = __pyx_v_xy_values;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 358, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_v_xy_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_17 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_17(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L18_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_6 = __pyx_t_17(__pyx_t_2); if (unlikely(!__pyx_t_6)) goto __pyx_L18_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_2), 2) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
      __pyx_t_17 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L19_unpacking_done;
      __pyx_L18_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_17 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 358, __pyx_L1_error)
      __pyx_L19_unpacking_done:;
    }
    __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_x = __pyx_t_18;
    __pyx_v_y = __pyx_t_19;

    /* "flydra_core/_reconstruct_utils.pyx":359
 *         xy_values = value_tuple[:2]
 *         x,y = xy_values
 *         if npy_isnan(x):             # <<<<<<<<<<<<<<
 *             bad_cam_ids.append( cam_id )
 *             continue # don't build this row
 */
    __pyx_t_9 = (npy_isnan(__pyx_v_x) != 0);
    if (__pyx_t_9) {

      /* "flydra_core/_reconstruct_utils.pyx":360
 *         x,y = xy_values
 *         if npy_isnan(x):
 *             bad_cam_ids.append( cam_id )             # <<<<<<<<<<<<<<
 *             continue # don't build this row
 * 
 */
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_bad_cam_ids, __pyx_v_cam_id); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 360, __pyx_L1_error)

      /* "flydra_core/_reconstruct_utils.pyx":361
 *         if npy_isnan(x):
 *             bad_cam_ids.append( cam_id )
 *             continue # don't build this row             # <<<<<<<<<<<<<<
 * 
 *         # Similar to code in reconstruct.Reconstructor.find3d()
 */
      goto __pyx_L6_continue;

      /* "flydra_core/_reconstruct_utils.pyx":359
 *         xy_values = value_tuple[:2]
 *         x,y = xy_values
 *         if npy_isnan(x):             # <<<<<<<<<<<<<<
 *             bad_cam_ids.append( cam_id )
 *             continue # don't build this row
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":365
 *         # Similar to code in reconstruct.Reconstructor.find3d()
 * 
 *         Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns             # <<<<<<<<<<<<<<
 *         row3 = Pmat[2,:]
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]
 */
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_Pmat_fastnx, __pyx_v_cam_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_Pmat, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":366
 * 
 *         Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns
 *         row3 = Pmat[2,:]             # <<<<<<<<<<<<<<
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]
 */
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_Pmat, __pyx_tuple__22); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_row3, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":367
 *         Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns
 *         row3 = Pmat[2,:]
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]             # <<<<<<<<<<<<<<
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]
 * 
 */
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_6, __pyx_v_row3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_Pmat, __pyx_tuple__23); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_i * 2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __Pyx_INCREF(__pyx_slice__21);
    __Pyx_GIVEREF(__pyx_slice__21);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_slice__21);
    __pyx_t_6 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_allA, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":368
 *         row3 = Pmat[2,:]
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]             # <<<<<<<<<<<<<<
 * 
 *         all2d[cam_id] = xy_values
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_row3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_Pmat, __pyx_tuple__24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_long(((__pyx_v_i * 2) + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_slice__21);
    __Pyx_GIVEREF(__pyx_slice__21);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_slice__21);
    __pyx_t_2 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_allA, __pyx_t_1, __pyx_t_6) < 0)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":370
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]
 * 
 *         all2d[cam_id] = xy_values             # <<<<<<<<<<<<<<
 * 
 *     cam_ids_for_least_err = {}
 */
    if (unlikely(PyDict_SetItem(__pyx_v_all2d, __pyx_v_cam_id, __pyx_v_xy_values) < 0)) __PYX_ERR(0, 370, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":346
 *                           'to raise an error rather than give this warning.')
 *             gave_water_warning = 1
 *     for i,cam_id in enumerate(cam_ids):             # <<<<<<<<<<<<<<
 *         cam_id2idx[cam_id] = i
 * 
 */
    __pyx_L6_continue:;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":372
 *         all2d[cam_id] = xy_values
 * 
 *     cam_ids_for_least_err = {}             # <<<<<<<<<<<<<<
 *     X_for_least_err = {}
 *     for n_cams from 2<=n_cams<=max_n_cams:
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_cam_ids_for_least_err = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":373
 * 
 *     cam_ids_for_least_err = {}
 *     X_for_least_err = {}             # <<<<<<<<<<<<<<
 *     for n_cams from 2<=n_cams<=max_n_cams:
 * 
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_X_for_least_err = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":374
 *     cam_ids_for_least_err = {}
 *     X_for_least_err = {}
 *     for n_cams from 2<=n_cams<=max_n_cams:             # <<<<<<<<<<<<<<
 * 
 *         # Calculate in least reprojection error starting with all
 */
  __pyx_t_5 = __pyx_v_max_n_cams;
  for (__pyx_v_n_cams = 2; __pyx_v_n_cams <= __pyx_t_5; __pyx_v_n_cams++) {

    /* "flydra_core/_reconstruct_utils.pyx":382
 *         # reprojection error.
 * 
 *         alpha = 1.0/n_cams             # <<<<<<<<<<<<<<
 * 
 *         # Can we short-circuit the rest of these computations?
 */
    if (unlikely(__pyx_v_n_cams == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 382, __pyx_L1_error)
    }
    __pyx_v_alpha = (1.0 / __pyx_v_n_cams);

    /* "flydra_core/_reconstruct_utils.pyx":385
 * 
 *         # Can we short-circuit the rest of these computations?
 *         if not npy_isinf(least_err_by_n_cameras[n_cams-1]): # if it's infinity, it must be n_cams 0 or 1.             # <<<<<<<<<<<<<<
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
 */
    __pyx_t_20 = (__pyx_v_n_cams - 1);
    __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, __pyx_t_20, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_8); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = ((!(npy_isinf(__pyx_t_19) != 0)) != 0);
    if (__pyx_t_9) {

      /* "flydra_core/_reconstruct_utils.pyx":387
 *         if not npy_isinf(least_err_by_n_cameras[n_cams-1]): # if it's infinity, it must be n_cams 0 or 1.
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:             # <<<<<<<<<<<<<<
 *                 if debug>5:
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
 */
      __pyx_t_20 = (__pyx_v_n_cams - 1);
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, __pyx_t_20, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_ACCEPTABLE_DISTANCE_PIXELS); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_8, __pyx_t_6, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_9) {

        /* "flydra_core/_reconstruct_utils.pyx":388
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
 *                 if debug>5:             # <<<<<<<<<<<<<<
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
 *                         n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)
 */
        __pyx_t_9 = ((__pyx_v_debug > 5) != 0);
        if (__pyx_t_9) {

          /* "flydra_core/_reconstruct_utils.pyx":390
 *                 if debug>5:
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
 *                         n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)             # <<<<<<<<<<<<<<
 *                 # and if the error for 1 less is too large, don't bother with more.
 *                 break
 */
          __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_n_cams); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_20 = (__pyx_v_n_cams - 1);
          __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, __pyx_t_20, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_8 = PyFloat_FromDouble(__pyx_v_ACCEPTABLE_DISTANCE_PIXELS); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 390, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_8);
          __pyx_t_1 = 0;
          __pyx_t_6 = 0;
          __pyx_t_8 = 0;

          /* "flydra_core/_reconstruct_utils.pyx":389
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
 *                 if debug>5:
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(             # <<<<<<<<<<<<<<
 *                         n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)
 *                 # and if the error for 1 less is too large, don't bother with more.
 */
          __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_HYPOTHESIS_TEST_shorting_for_n_c, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (__Pyx_PrintOne(0, __pyx_t_8) < 0) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "flydra_core/_reconstruct_utils.pyx":388
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
 *                 if debug>5:             # <<<<<<<<<<<<<<
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
 *                         n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)
 */
        }

        /* "flydra_core/_reconstruct_utils.pyx":392
 *                         n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)
 *                 # and if the error for 1 less is too large, don't bother with more.
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         for cam_ids_used in recon.cam_combinations_by_size[n_cams]:
 */
        goto __pyx_L22_break;

        /* "flydra_core/_reconstruct_utils.pyx":387
 *         if not npy_isinf(least_err_by_n_cameras[n_cams-1]): # if it's infinity, it must be n_cams 0 or 1.
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:             # <<<<<<<<<<<<<<
 *                 if debug>5:
 *                     print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":385
 * 
 *         # Can we short-circuit the rest of these computations?
 *         if not npy_isinf(least_err_by_n_cameras[n_cams-1]): # if it's infinity, it must be n_cams 0 or 1.             # <<<<<<<<<<<<<<
 *             # If we've calculated error for 2 less than n_cams
 *             if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":394
 *                 break
 * 
 *         for cam_ids_used in recon.cam_combinations_by_size[n_cams]:             # <<<<<<<<<<<<<<
 *             missing_cam_data = 0 #False
 *             good_A_idx = []
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_recon, __pyx_n_s_cam_combinations_by_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_8, __pyx_v_n_cams, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_8 = __pyx_t_2; __Pyx_INCREF(__pyx_t_8); __pyx_t_3 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 394, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_11(__pyx_t_8);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 394, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_cam_ids_used, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":395
 * 
 *         for cam_ids_used in recon.cam_combinations_by_size[n_cams]:
 *             missing_cam_data = 0 #False             # <<<<<<<<<<<<<<
 *             good_A_idx = []
 *             if debug>5:
 */
      __pyx_v_missing_cam_data = 0;

      /* "flydra_core/_reconstruct_utils.pyx":396
 *         for cam_ids_used in recon.cam_combinations_by_size[n_cams]:
 *             missing_cam_data = 0 #False
 *             good_A_idx = []             # <<<<<<<<<<<<<<
 *             if debug>5:
 *                 assert len(cam_ids_used)==n_cams
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_good_A_idx, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":397
 *             missing_cam_data = 0 #False
 *             good_A_idx = []
 *             if debug>5:             # <<<<<<<<<<<<<<
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:
 */
      __pyx_t_9 = ((__pyx_v_debug > 5) != 0);
      if (__pyx_t_9) {

        /* "flydra_core/_reconstruct_utils.pyx":398
 *             good_A_idx = []
 *             if debug>5:
 *                 assert len(cam_ids_used)==n_cams             # <<<<<<<<<<<<<<
 *             for cam_id in cam_ids_used:
 *                 if cam_id in bad_cam_ids:
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(!Py_OptimizeFlag)) {
          __pyx_t_4 = PyObject_Length(__pyx_v_cam_ids_used); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 398, __pyx_L1_error)
          if (unlikely(!((__pyx_t_4 == __pyx_v_n_cams) != 0))) {
            PyErr_SetNone(PyExc_AssertionError);
            __PYX_ERR(0, 398, __pyx_L1_error)
          }
        }
        #endif

        /* "flydra_core/_reconstruct_utils.pyx":397
 *             missing_cam_data = 0 #False
 *             good_A_idx = []
 *             if debug>5:             # <<<<<<<<<<<<<<
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":399
 *             if debug>5:
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *                 if cam_id in bad_cam_ids:
 *                     missing_cam_data = 1 #True
 */
      if (likely(PyList_CheckExact(__pyx_v_cam_ids_used)) || PyTuple_CheckExact(__pyx_v_cam_ids_used)) {
        __pyx_t_2 = __pyx_v_cam_ids_used; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
        __pyx_t_21 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_cam_ids_used); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_21 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 399, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_21)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_21(__pyx_t_2);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 399, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":400
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:
 *                 if cam_id in bad_cam_ids:             # <<<<<<<<<<<<<<
 *                     missing_cam_data = 1 #True
 *                     break
 */
        __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_cam_id, __pyx_v_bad_cam_ids, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 400, __pyx_L1_error)
        __pyx_t_22 = (__pyx_t_9 != 0);
        if (__pyx_t_22) {

          /* "flydra_core/_reconstruct_utils.pyx":401
 *             for cam_id in cam_ids_used:
 *                 if cam_id in bad_cam_ids:
 *                     missing_cam_data = 1 #True             # <<<<<<<<<<<<<<
 *                     break
 *                 else:
 */
          __pyx_v_missing_cam_data = 1;

          /* "flydra_core/_reconstruct_utils.pyx":402
 *                 if cam_id in bad_cam_ids:
 *                     missing_cam_data = 1 #True
 *                     break             # <<<<<<<<<<<<<<
 *                 else:
 *                     i = cam_id2idx[cam_id]
 */
          goto __pyx_L30_break;

          /* "flydra_core/_reconstruct_utils.pyx":400
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:
 *                 if cam_id in bad_cam_ids:             # <<<<<<<<<<<<<<
 *                     missing_cam_data = 1 #True
 *                     break
 */
        }

        /* "flydra_core/_reconstruct_utils.pyx":404
 *                     break
 *                 else:
 *                     i = cam_id2idx[cam_id]             # <<<<<<<<<<<<<<
 *                     good_A_idx.extend( (i*2, i*2+1) )
 *             if missing_cam_data == 1:
 */
        /*else*/ {
          __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_cam_id2idx, __pyx_v_cam_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_v_i = __pyx_t_10;

          /* "flydra_core/_reconstruct_utils.pyx":405
 *                 else:
 *                     i = cam_id2idx[cam_id]
 *                     good_A_idx.extend( (i*2, i*2+1) )             # <<<<<<<<<<<<<<
 *             if missing_cam_data == 1:
 *                 continue
 */
          __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_i * 2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 405, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_16 = __Pyx_ListComp_Append(__pyx_v_good_A_idx, __pyx_t_6); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = __Pyx_PyInt_From_long(((__pyx_v_i * 2) + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 405, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_23 = __Pyx_PyList_Append(__pyx_v_good_A_idx, __pyx_t_6); if (unlikely(__pyx_t_23 == ((int)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          (void)((__pyx_t_16 | __pyx_t_23));
        }

        /* "flydra_core/_reconstruct_utils.pyx":399
 *             if debug>5:
 *                 assert len(cam_ids_used)==n_cams
 *             for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *                 if cam_id in bad_cam_ids:
 *                     missing_cam_data = 1 #True
 */
      }
      __pyx_L30_break:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":406
 *                     i = cam_id2idx[cam_id]
 *                     good_A_idx.extend( (i*2, i*2+1) )
 *             if missing_cam_data == 1:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __pyx_t_22 = ((__pyx_v_missing_cam_data == 1) != 0);
      if (__pyx_t_22) {

        /* "flydra_core/_reconstruct_utils.pyx":407
 *                     good_A_idx.extend( (i*2, i*2+1) )
 *             if missing_cam_data == 1:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # find 3D point
 */
        goto __pyx_L26_continue;

        /* "flydra_core/_reconstruct_utils.pyx":406
 *                     i = cam_id2idx[cam_id]
 *                     good_A_idx.extend( (i*2, i*2+1) )
 *             if missing_cam_data == 1:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":410
 * 
 *             # find 3D point
 *             A = allA[good_A_idx,:]             # <<<<<<<<<<<<<<
 *             u,d,vt=svd(A)
 *             X = vt[-1,:]/vt[-1,3] # normalize
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_good_A_idx);
      __Pyx_GIVEREF(__pyx_v_good_A_idx);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_good_A_idx);
      __Pyx_INCREF(__pyx_slice__21);
      __Pyx_GIVEREF(__pyx_slice__21);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_slice__21);
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_allA, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_A, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":411
 *             # find 3D point
 *             A = allA[good_A_idx,:]
 *             u,d,vt=svd(A)             # <<<<<<<<<<<<<<
 *             X = vt[-1,:]/vt[-1,3] # normalize
 * 
 */
      __Pyx_INCREF(__pyx_v_svd);
      __pyx_t_2 = __pyx_v_svd; __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_A) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_A);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 411, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_24 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_17 = Py_TYPE(__pyx_t_24)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_17(__pyx_t_24); if (unlikely(!__pyx_t_2)) goto __pyx_L33_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_1 = __pyx_t_17(__pyx_t_24); if (unlikely(!__pyx_t_1)) goto __pyx_L33_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 2; __pyx_t_7 = __pyx_t_17(__pyx_t_24); if (unlikely(!__pyx_t_7)) goto __pyx_L33_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_24), 3) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
        __pyx_t_17 = NULL;
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
        goto __pyx_L34_unpacking_done;
        __pyx_L33_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
        __pyx_t_17 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 411, __pyx_L1_error)
        __pyx_L34_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_u, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_vt, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":412
 *             A = allA[good_A_idx,:]
 *             u,d,vt=svd(A)
 *             X = vt[-1,:]/vt[-1,3] # normalize             # <<<<<<<<<<<<<<
 * 
 *             # calculate reprojection error
 */
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_vt, __pyx_tuple__25); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_vt, __pyx_tuple__26); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_X, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":415
 * 
 *             # calculate reprojection error
 *             mean_dist = 0.0             # <<<<<<<<<<<<<<
 *             for cam_id in cam_ids_used:
 *                 orig_x,orig_y = all2d[cam_id]
 */
      __pyx_v_mean_dist = 0.0;

      /* "flydra_core/_reconstruct_utils.pyx":416
 *             # calculate reprojection error
 *             mean_dist = 0.0
 *             for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *                 orig_x,orig_y = all2d[cam_id]
 *                 Pmat = Pmat_fastnx[cam_id]
 */
      if (likely(PyList_CheckExact(__pyx_v_cam_ids_used)) || PyTuple_CheckExact(__pyx_v_cam_ids_used)) {
        __pyx_t_1 = __pyx_v_cam_ids_used; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
        __pyx_t_21 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cam_ids_used); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_21 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 416, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_21)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 416, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 416, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 416, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 416, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_21(__pyx_t_1);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 416, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":417
 *             mean_dist = 0.0
 *             for cam_id in cam_ids_used:
 *                 orig_x,orig_y = all2d[cam_id]             # <<<<<<<<<<<<<<
 *                 Pmat = Pmat_fastnx[cam_id]
 *                 new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d
 */
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_all2d, __pyx_v_cam_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 417, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_2);
          #else
          __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_24 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_24);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_17 = Py_TYPE(__pyx_t_24)->tp_iternext;
          index = 0; __pyx_t_6 = __pyx_t_17(__pyx_t_24); if (unlikely(!__pyx_t_6)) goto __pyx_L37_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 1; __pyx_t_2 = __pyx_t_17(__pyx_t_24); if (unlikely(!__pyx_t_2)) goto __pyx_L37_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_24), 2) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
          __pyx_t_17 = NULL;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          goto __pyx_L38_unpacking_done;
          __pyx_L37_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          __pyx_t_17 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 417, __pyx_L1_error)
          __pyx_L38_unpacking_done:;
        }
        __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_orig_x = __pyx_t_19;
        __pyx_v_orig_y = __pyx_t_18;

        /* "flydra_core/_reconstruct_utils.pyx":418
 *             for cam_id in cam_ids_used:
 *                 orig_x,orig_y = all2d[cam_id]
 *                 Pmat = Pmat_fastnx[cam_id]             # <<<<<<<<<<<<<<
 *                 new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d
 *                 new_x, new_y = new_xyw[0:2]/new_xyw[2]
 */
        __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_Pmat_fastnx, __pyx_v_cam_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_XDECREF_SET(__pyx_v_Pmat, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":419
 *                 orig_x,orig_y = all2d[cam_id]
 *                 Pmat = Pmat_fastnx[cam_id]
 *                 new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d             # <<<<<<<<<<<<<<
 *                 new_x, new_y = new_xyw[0:2]/new_xyw[2]
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_dot); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_Pmat, __pyx_v_X};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 419, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_Pmat, __pyx_v_X};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 419, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_24 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 419, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_24);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_Pmat);
          __Pyx_GIVEREF(__pyx_v_Pmat);
          PyTuple_SET_ITEM(__pyx_t_24, 0+__pyx_t_10, __pyx_v_Pmat);
          __Pyx_INCREF(__pyx_v_X);
          __Pyx_GIVEREF(__pyx_v_X);
          PyTuple_SET_ITEM(__pyx_t_24, 1+__pyx_t_10, __pyx_v_X);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_24, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 419, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_new_xyw, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":420
 *                 Pmat = Pmat_fastnx[cam_id]
 *                 new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d
 *                 new_x, new_y = new_xyw[0:2]/new_xyw[2]             # <<<<<<<<<<<<<<
 * 
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
 */
        __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_new_xyw, 0, 2, NULL, NULL, &__pyx_slice__27, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_new_xyw, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_24 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_24))) || (PyList_CheckExact(__pyx_t_24))) {
          PyObject* sequence = __pyx_t_24;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 420, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          #else
          __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 420, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 420, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_2 = PyObject_GetIter(__pyx_t_24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          __pyx_t_17 = Py_TYPE(__pyx_t_2)->tp_iternext;
          index = 0; __pyx_t_6 = __pyx_t_17(__pyx_t_2); if (unlikely(!__pyx_t_6)) goto __pyx_L39_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 1; __pyx_t_7 = __pyx_t_17(__pyx_t_2); if (unlikely(!__pyx_t_7)) goto __pyx_L39_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_2), 2) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
          __pyx_t_17 = NULL;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L40_unpacking_done;
          __pyx_L39_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_17 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 420, __pyx_L1_error)
          __pyx_L40_unpacking_done:;
        }
        __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_new_x = __pyx_t_18;
        __pyx_v_new_y = __pyx_t_19;

        /* "flydra_core/_reconstruct_utils.pyx":422
 *                 new_x, new_y = new_xyw[0:2]/new_xyw[2]
 * 
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)             # <<<<<<<<<<<<<<
 *                 mean_dist = mean_dist + dist*alpha
 *             if debug>5:
 */
        __pyx_v_dist = sqrt((pow((__pyx_v_orig_x - __pyx_v_new_x), 2.0) + pow((__pyx_v_orig_y - __pyx_v_new_y), 2.0)));

        /* "flydra_core/_reconstruct_utils.pyx":423
 * 
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
 *                 mean_dist = mean_dist + dist*alpha             # <<<<<<<<<<<<<<
 *             if debug>5:
 *                 print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))
 */
        __pyx_v_mean_dist = (__pyx_v_mean_dist + (__pyx_v_dist * __pyx_v_alpha));

        /* "flydra_core/_reconstruct_utils.pyx":416
 *             # calculate reprojection error
 *             mean_dist = 0.0
 *             for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *                 orig_x,orig_y = all2d[cam_id]
 *                 Pmat = Pmat_fastnx[cam_id]
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":424
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
 *                 mean_dist = mean_dist + dist*alpha
 *             if debug>5:             # <<<<<<<<<<<<<<
 *                 print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))
 * 
 */
      __pyx_t_22 = ((__pyx_v_debug > 5) != 0);
      if (__pyx_t_22) {

        /* "flydra_core/_reconstruct_utils.pyx":425
 *                 mean_dist = mean_dist + dist*alpha
 *             if debug>5:
 *                 print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))             # <<<<<<<<<<<<<<
 * 
 *             least_err = least_err_by_n_cameras[n_cams]
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_mean_dist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_24 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_cam_ids_used); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_24);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_24);
        __pyx_t_1 = 0;
        __pyx_t_24 = 0;
        __pyx_t_24 = __Pyx_PyString_Format(__pyx_kp_s_HYPOTHESIS_TEST_mean_dist_f_for, __pyx_t_7); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 425, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__Pyx_PrintOne(0, __pyx_t_24) < 0) __PYX_ERR(0, 425, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":424
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
 *                 mean_dist = mean_dist + dist*alpha
 *             if debug>5:             # <<<<<<<<<<<<<<
 *                 print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))
 * 
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":427
 *                 print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))
 * 
 *             least_err = least_err_by_n_cameras[n_cams]             # <<<<<<<<<<<<<<
 *             if mean_dist < least_err:
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 */
      __pyx_t_24 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, __pyx_v_n_cams, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_24); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 427, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_v_least_err = __pyx_t_19;

      /* "flydra_core/_reconstruct_utils.pyx":428
 * 
 *             least_err = least_err_by_n_cameras[n_cams]
 *             if mean_dist < least_err:             # <<<<<<<<<<<<<<
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used
 */
      __pyx_t_22 = ((__pyx_v_mean_dist < __pyx_v_least_err) != 0);
      if (__pyx_t_22) {

        /* "flydra_core/_reconstruct_utils.pyx":429
 *             least_err = least_err_by_n_cameras[n_cams]
 *             if mean_dist < least_err:
 *                 least_err_by_n_cameras[n_cams] = mean_dist             # <<<<<<<<<<<<<<
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used
 *                 X_for_least_err[n_cams] = X[:3]
 */
        __pyx_t_24 = PyFloat_FromDouble(__pyx_v_mean_dist); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 429, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        if (unlikely(__Pyx_SetItemInt(__pyx_v_least_err_by_n_cameras, __pyx_v_n_cams, __pyx_t_24, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0)) __PYX_ERR(0, 429, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":430
 *             if mean_dist < least_err:
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used             # <<<<<<<<<<<<<<
 *                 X_for_least_err[n_cams] = X[:3]
 * 
 */
        __pyx_t_24 = __Pyx_PyInt_From_int(__pyx_v_n_cams); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        if (unlikely(PyDict_SetItem(__pyx_v_cam_ids_for_least_err, __pyx_t_24, __pyx_v_cam_ids_used) < 0)) __PYX_ERR(0, 430, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":431
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used
 *                 X_for_least_err[n_cams] = X[:3]             # <<<<<<<<<<<<<<
 * 
 *     # now we have the best estimate for 2 views, 3 views, ...
 */
        __pyx_t_24 = __Pyx_PyObject_GetSlice(__pyx_v_X, 0, 3, NULL, NULL, &__pyx_slice__28, 0, 1, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_24);
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_n_cams); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely(PyDict_SetItem(__pyx_v_X_for_least_err, __pyx_t_7, __pyx_t_24) < 0)) __PYX_ERR(0, 431, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":428
 * 
 *             least_err = least_err_by_n_cameras[n_cams]
 *             if mean_dist < least_err:             # <<<<<<<<<<<<<<
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":394
 *                 break
 * 
 *         for cam_ids_used in recon.cam_combinations_by_size[n_cams]:             # <<<<<<<<<<<<<<
 *             missing_cam_data = 0 #False
 *             good_A_idx = []
 */
      __pyx_L26_continue:;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __pyx_L22_break:;

  /* "flydra_core/_reconstruct_utils.pyx":434
 * 
 *     # now we have the best estimate for 2 views, 3 views, ...
 *     best_n_cams = 2             # <<<<<<<<<<<<<<
 *     least_err = least_err_by_n_cameras[2]
 *     if debug>5:
 */
  __pyx_v_best_n_cams = 2;

  /* "flydra_core/_reconstruct_utils.pyx":435
 *     # now we have the best estimate for 2 views, 3 views, ...
 *     best_n_cams = 2
 *     least_err = least_err_by_n_cameras[2]             # <<<<<<<<<<<<<<
 *     if debug>5:
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 */
  __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, 2, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_8); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_least_err = __pyx_t_19;

  /* "flydra_core/_reconstruct_utils.pyx":436
 *     best_n_cams = 2
 *     least_err = least_err_by_n_cameras[2]
 *     if debug>5:             # <<<<<<<<<<<<<<
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 *     mean_dist = least_err
 */
  __pyx_t_22 = ((__pyx_v_debug > 5) != 0);
  if (__pyx_t_22) {

    /* "flydra_core/_reconstruct_utils.pyx":437
 *     least_err = least_err_by_n_cameras[2]
 *     if debug>5:
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)             # <<<<<<<<<<<<<<
 *     mean_dist = least_err
 *     if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):
 */
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_least_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = PyFloat_FromDouble(__pyx_v_ACCEPTABLE_DISTANCE_PIXELS); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_24);
    __pyx_t_8 = 0;
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyString_Format(__pyx_kp_s_HYPOTHESIS_TEST_least_err_ACCEPT, __pyx_t_7); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__Pyx_PrintOne(0, __pyx_t_24) < 0) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":436
 *     best_n_cams = 2
 *     least_err = least_err_by_n_cameras[2]
 *     if debug>5:             # <<<<<<<<<<<<<<
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 *     mean_dist = least_err
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":438
 *     if debug>5:
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 *     mean_dist = least_err             # <<<<<<<<<<<<<<
 *     if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):
 *         raise NoAcceptablePointFound('least error was %f'%least_err)
 */
  __pyx_v_mean_dist = __pyx_v_least_err;

  /* "flydra_core/_reconstruct_utils.pyx":439
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 *     mean_dist = least_err
 *     if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):             # <<<<<<<<<<<<<<
 *         raise NoAcceptablePointFound('least error was %f'%least_err)
 * 
 */
  __pyx_t_22 = ((!((__pyx_v_least_err < __pyx_v_ACCEPTABLE_DISTANCE_PIXELS) != 0)) != 0);
  if (unlikely(__pyx_t_22)) {

    /* "flydra_core/_reconstruct_utils.pyx":440
 *     mean_dist = least_err
 *     if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):
 *         raise NoAcceptablePointFound('least error was %f'%least_err)             # <<<<<<<<<<<<<<
 * 
 *     for n_cams from 3 <= n_cams <= max_n_cams:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_NoAcceptablePointFound); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_least_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_least_error_was_f, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_24 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_24, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __PYX_ERR(0, 440, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":439
 *         print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
 *     mean_dist = least_err
 *     if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):             # <<<<<<<<<<<<<<
 *         raise NoAcceptablePointFound('least error was %f'%least_err)
 * 
 */
  }

  /* "flydra_core/_reconstruct_utils.pyx":442
 *         raise NoAcceptablePointFound('least error was %f'%least_err)
 * 
 *     for n_cams from 3 <= n_cams <= max_n_cams:             # <<<<<<<<<<<<<<
 *         least_err = least_err_by_n_cameras[n_cams]
 *         if debug>5:
 */
  __pyx_t_5 = __pyx_v_max_n_cams;
  for (__pyx_v_n_cams = 3; __pyx_v_n_cams <= __pyx_t_5; __pyx_v_n_cams++) {

    /* "flydra_core/_reconstruct_utils.pyx":443
 * 
 *     for n_cams from 3 <= n_cams <= max_n_cams:
 *         least_err = least_err_by_n_cameras[n_cams]             # <<<<<<<<<<<<<<
 *         if debug>5:
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 */
    __pyx_t_24 = __Pyx_GetItemInt_List(__pyx_v_least_err_by_n_cameras, __pyx_v_n_cams, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_24); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_v_least_err = __pyx_t_19;

    /* "flydra_core/_reconstruct_utils.pyx":444
 *     for n_cams from 3 <= n_cams <= max_n_cams:
 *         least_err = least_err_by_n_cameras[n_cams]
 *         if debug>5:             # <<<<<<<<<<<<<<
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 *         if npy_isinf(least_err):
 */
    __pyx_t_22 = ((__pyx_v_debug > 5) != 0);
    if (__pyx_t_22) {

      /* "flydra_core/_reconstruct_utils.pyx":445
 *         least_err = least_err_by_n_cameras[n_cams]
 *         if debug>5:
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)             # <<<<<<<<<<<<<<
 *         if npy_isinf(least_err):
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 */
      __pyx_t_24 = __Pyx_PyInt_From_int(__pyx_v_n_cams); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_7 = PyFloat_FromDouble(__pyx_v_least_err); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_24);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_7);
      __pyx_t_24 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_HYPOTHESIS_TEST_n_cams_d_f, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PrintOne(0, __pyx_t_7) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":444
 *     for n_cams from 3 <= n_cams <= max_n_cams:
 *         least_err = least_err_by_n_cameras[n_cams]
 *         if debug>5:             # <<<<<<<<<<<<<<
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 *         if npy_isinf(least_err):
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":446
 *         if debug>5:
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 *         if npy_isinf(least_err):             # <<<<<<<<<<<<<<
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:
 */
    __pyx_t_22 = (npy_isinf(__pyx_v_least_err) != 0);
    if (__pyx_t_22) {

      /* "flydra_core/_reconstruct_utils.pyx":447
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 *         if npy_isinf(least_err):
 *             break # if we don't have e.g. 4 cameras, we won't have 5             # <<<<<<<<<<<<<<
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:
 *             mean_dist = least_err
 */
      goto __pyx_L46_break;

      /* "flydra_core/_reconstruct_utils.pyx":446
 *         if debug>5:
 *             print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
 *         if npy_isinf(least_err):             # <<<<<<<<<<<<<<
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":448
 *         if npy_isinf(least_err):
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:             # <<<<<<<<<<<<<<
 *             mean_dist = least_err
 *             best_n_cams = n_cams
 */
    __pyx_t_22 = ((__pyx_v_least_err < __pyx_v_ACCEPTABLE_DISTANCE_PIXELS) != 0);
    if (__pyx_t_22) {

      /* "flydra_core/_reconstruct_utils.pyx":449
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:
 *             mean_dist = least_err             # <<<<<<<<<<<<<<
 *             best_n_cams = n_cams
 * 
 */
      __pyx_v_mean_dist = __pyx_v_least_err;

      /* "flydra_core/_reconstruct_utils.pyx":450
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:
 *             mean_dist = least_err
 *             best_n_cams = n_cams             # <<<<<<<<<<<<<<
 * 
 *     # now calculate final values
 */
      __pyx_v_best_n_cams = __pyx_v_n_cams;

      /* "flydra_core/_reconstruct_utils.pyx":448
 *         if npy_isinf(least_err):
 *             break # if we don't have e.g. 4 cameras, we won't have 5
 *         if least_err < ACCEPTABLE_DISTANCE_PIXELS:             # <<<<<<<<<<<<<<
 *             mean_dist = least_err
 *             best_n_cams = n_cams
 */
    }
  }
  __pyx_L46_break:;

  /* "flydra_core/_reconstruct_utils.pyx":453
 * 
 *     # now calculate final values
 *     cam_ids_used = cam_ids_for_least_err[best_n_cams]             # <<<<<<<<<<<<<<
 *     if with_water:
 *         # Even though (for speed reasons) we did not use proper
 */
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_best_n_cams); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_cam_ids_for_least_err, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_XDECREF_SET(__pyx_v_cam_ids_used, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":454
 *     # now calculate final values
 *     cam_ids_used = cam_ids_for_least_err[best_n_cams]
 *     if with_water:             # <<<<<<<<<<<<<<
 *         # Even though (for speed reasons) we did not use proper
 *         # refraction-correct code above, we now recompute X using
 */
  __pyx_t_22 = (__pyx_v_with_water != 0);
  if (__pyx_t_22) {

    /* "flydra_core/_reconstruct_utils.pyx":458
 *         # refraction-correct code above, we now recompute X using
 *         # refraction.
 *         cam_ids_and_points2d = []             # <<<<<<<<<<<<<<
 *         for cam_id in cam_ids_used:
 *             xy = found_data_dict[cam_id][:2]
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_cam_ids_and_points2d = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":459
 *         # refraction.
 *         cam_ids_and_points2d = []
 *         for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *             xy = found_data_dict[cam_id][:2]
 *             cam_ids_and_points2d.append(( cam_id, xy ))
 */
    if (likely(PyList_CheckExact(__pyx_v_cam_ids_used)) || PyTuple_CheckExact(__pyx_v_cam_ids_used)) {
      __pyx_t_1 = __pyx_v_cam_ids_used; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cam_ids_used); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 459, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_1);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 459, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":460
 *         cam_ids_and_points2d = []
 *         for cam_id in cam_ids_used:
 *             xy = found_data_dict[cam_id][:2]             # <<<<<<<<<<<<<<
 *             cam_ids_and_points2d.append(( cam_id, xy ))
 * 
 */
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_found_data_dict, __pyx_v_cam_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 460, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_24 = __Pyx_PyObject_GetSlice(__pyx_t_7, 0, 2, NULL, NULL, &__pyx_slice__20, 0, 1, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 460, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_xy, __pyx_t_24);
      __pyx_t_24 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":461
 *         for cam_id in cam_ids_used:
 *             xy = found_data_dict[cam_id][:2]
 *             cam_ids_and_points2d.append(( cam_id, xy ))             # <<<<<<<<<<<<<<
 * 
 *         X = recon.find3d(cam_ids_and_points2d,
 */
      __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_INCREF(__pyx_v_cam_id);
      __Pyx_GIVEREF(__pyx_v_cam_id);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_v_cam_id);
      __Pyx_INCREF(__pyx_v_xy);
      __Pyx_GIVEREF(__pyx_v_xy);
      PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_v_xy);
      __pyx_t_23 = __Pyx_PyList_Append(__pyx_v_cam_ids_and_points2d, __pyx_t_24); if (unlikely(__pyx_t_23 == ((int)-1))) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":459
 *         # refraction.
 *         cam_ids_and_points2d = []
 *         for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *             xy = found_data_dict[cam_id][:2]
 *             cam_ids_and_points2d.append(( cam_id, xy ))
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":463
 *             cam_ids_and_points2d.append(( cam_id, xy ))
 * 
 *         X = recon.find3d(cam_ids_and_points2d,             # <<<<<<<<<<<<<<
 *                          undistort=False, # points are already undistorted
 *                          return_line_coords = False,
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_recon, __pyx_n_s_find3d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_v_cam_ids_and_points2d);
    __Pyx_GIVEREF(__pyx_v_cam_ids_and_points2d);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_v_cam_ids_and_points2d);

    /* "flydra_core/_reconstruct_utils.pyx":464
 * 
 *         X = recon.find3d(cam_ids_and_points2d,
 *                          undistort=False, # points are already undistorted             # <<<<<<<<<<<<<<
 *                          return_line_coords = False,
 *                          )
 */
    __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_undistort, Py_False) < 0) __PYX_ERR(0, 464, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":465
 *         X = recon.find3d(cam_ids_and_points2d,
 *                          undistort=False, # points are already undistorted
 *                          return_line_coords = False,             # <<<<<<<<<<<<<<
 *                          )
 *     else:
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return_line_coords, Py_False) < 0) __PYX_ERR(0, 464, __pyx_L1_error)

    /* "flydra_core/_reconstruct_utils.pyx":463
 *             cam_ids_and_points2d.append(( cam_id, xy ))
 * 
 *         X = recon.find3d(cam_ids_and_points2d,             # <<<<<<<<<<<<<<
 *                          undistort=False, # points are already undistorted
 *                          return_line_coords = False,
 */
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_24, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_X, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":454
 *     # now calculate final values
 *     cam_ids_used = cam_ids_for_least_err[best_n_cams]
 *     if with_water:             # <<<<<<<<<<<<<<
 *         # Even though (for speed reasons) we did not use proper
 *         # refraction-correct code above, we now recompute X using
 */
    goto __pyx_L50;
  }

  /* "flydra_core/_reconstruct_utils.pyx":468
 *                          )
 *     else:
 *         X = X_for_least_err[best_n_cams]             # <<<<<<<<<<<<<<
 *     del X_for_least_err
 * 
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_best_n_cams); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_X_for_least_err, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_X, __pyx_t_7);
    __pyx_t_7 = 0;
  }
  __pyx_L50:;

  /* "flydra_core/_reconstruct_utils.pyx":469
 *     else:
 *         X = X_for_least_err[best_n_cams]
 *     del X_for_least_err             # <<<<<<<<<<<<<<
 * 
 *     # calculate line3d
 */
  __Pyx_DECREF(__pyx_v_X_for_least_err);
  __pyx_v_X_for_least_err = NULL;

  /* "flydra_core/_reconstruct_utils.pyx":472
 * 
 *     # calculate line3d
 *     P = []             # <<<<<<<<<<<<<<
 *     for cam_id in cam_ids_used:
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_P = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":473
 *     # calculate line3d
 *     P = []
 *     for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 */
  if (likely(PyList_CheckExact(__pyx_v_cam_ids_used)) || PyTuple_CheckExact(__pyx_v_cam_ids_used)) {
    __pyx_t_7 = __pyx_v_cam_ids_used; __Pyx_INCREF(__pyx_t_7); __pyx_t_5 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_cam_ids_used); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 473, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_8); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_8); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_11(__pyx_t_7);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 473, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":474
 *     P = []
 *     for cam_id in cam_ids_used:
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]             # <<<<<<<<<<<<<<
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 *                 P.append( (p1,p2,p3,p4) )
 */
    __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_found_data_dict, __pyx_v_cam_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
      PyObject* sequence = __pyx_t_8;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 9)) {
        if (size > 9) __Pyx_RaiseTooManyValuesError(9);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 474, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_24 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 3); 
        __pyx_t_25 = PyTuple_GET_ITEM(sequence, 4); 
        __pyx_t_26 = PyTuple_GET_ITEM(sequence, 5); 
        __pyx_t_27 = PyTuple_GET_ITEM(sequence, 6); 
        __pyx_t_28 = PyTuple_GET_ITEM(sequence, 7); 
        __pyx_t_29 = PyTuple_GET_ITEM(sequence, 8); 
      } else {
        __pyx_t_24 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 3); 
        __pyx_t_25 = PyList_GET_ITEM(sequence, 4); 
        __pyx_t_26 = PyList_GET_ITEM(sequence, 5); 
        __pyx_t_27 = PyList_GET_ITEM(sequence, 6); 
        __pyx_t_28 = PyList_GET_ITEM(sequence, 7); 
        __pyx_t_29 = PyList_GET_ITEM(sequence, 8); 
      }
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_25);
      __Pyx_INCREF(__pyx_t_26);
      __Pyx_INCREF(__pyx_t_27);
      __Pyx_INCREF(__pyx_t_28);
      __Pyx_INCREF(__pyx_t_29);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[9] = {&__pyx_t_24,&__pyx_t_1,&__pyx_t_6,&__pyx_t_2,&__pyx_t_25,&__pyx_t_26,&__pyx_t_27,&__pyx_t_28,&__pyx_t_29};
        for (i=0; i < 9; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 474, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[9] = {&__pyx_t_24,&__pyx_t_1,&__pyx_t_6,&__pyx_t_2,&__pyx_t_25,&__pyx_t_26,&__pyx_t_27,&__pyx_t_28,&__pyx_t_29};
      __pyx_t_30 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_30);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_17 = Py_TYPE(__pyx_t_30)->tp_iternext;
      for (index=0; index < 9; index++) {
        PyObject* item = __pyx_t_17(__pyx_t_30); if (unlikely(!item)) goto __pyx_L55_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_30), 9) < 0) __PYX_ERR(0, 474, __pyx_L1_error)
      __pyx_t_17 = NULL;
      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
      goto __pyx_L56_unpacking_done;
      __pyx_L55_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
      __pyx_t_17 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 474, __pyx_L1_error)
      __pyx_L56_unpacking_done:;
    }
    __pyx_t_19 = __pyx_PyFloat_AsDouble(__pyx_t_24); if (unlikely((__pyx_t_19 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_x = __pyx_t_19;
    __pyx_v_y = __pyx_t_18;
    __Pyx_XDECREF_SET(__pyx_v_area, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_slope, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_eccentricity, __pyx_t_25);
    __pyx_t_25 = 0;
    __Pyx_XDECREF_SET(__pyx_v_p1, __pyx_t_26);
    __pyx_t_26 = 0;
    __Pyx_XDECREF_SET(__pyx_v_p2, __pyx_t_27);
    __pyx_t_27 = 0;
    __Pyx_XDECREF_SET(__pyx_v_p3, __pyx_t_28);
    __pyx_t_28 = 0;
    __Pyx_XDECREF_SET(__pyx_v_p4, __pyx_t_29);
    __pyx_t_29 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":475
 *     for cam_id in cam_ids_used:
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):             # <<<<<<<<<<<<<<
 *                 P.append( (p1,p2,p3,p4) )
 *     if len(P) < 2:
 */
    __pyx_t_8 = PyFloat_FromDouble(__pyx_v_11flydra_core_18_reconstruct_utils_MINIMUM_ECCENTRICITY); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_29 = PyObject_RichCompare(__pyx_v_eccentricity, __pyx_t_8, Py_GT); __Pyx_XGOTREF(__pyx_t_29); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_29); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    if (__pyx_t_9) {
    } else {
      __pyx_t_22 = __pyx_t_9;
      goto __pyx_L58_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_isnan); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_28))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_28);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_28);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_28, function);
      }
    }
    __pyx_t_29 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_28, __pyx_t_8, __pyx_v_p1) : __Pyx_PyObject_CallOneArg(__pyx_t_28, __pyx_v_p1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_29); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_31 = ((!__pyx_t_9) != 0);
    __pyx_t_22 = __pyx_t_31;
    __pyx_L58_bool_binop_done:;
    if (__pyx_t_22) {

      /* "flydra_core/_reconstruct_utils.pyx":476
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 *                 P.append( (p1,p2,p3,p4) )             # <<<<<<<<<<<<<<
 *     if len(P) < 2:
 *         Lcoords = None
 */
      __pyx_t_29 = PyTuple_New(4); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_29);
      __Pyx_INCREF(__pyx_v_p1);
      __Pyx_GIVEREF(__pyx_v_p1);
      PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_v_p1);
      __Pyx_INCREF(__pyx_v_p2);
      __Pyx_GIVEREF(__pyx_v_p2);
      PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_v_p2);
      __Pyx_INCREF(__pyx_v_p3);
      __Pyx_GIVEREF(__pyx_v_p3);
      PyTuple_SET_ITEM(__pyx_t_29, 2, __pyx_v_p3);
      __Pyx_INCREF(__pyx_v_p4);
      __Pyx_GIVEREF(__pyx_v_p4);
      PyTuple_SET_ITEM(__pyx_t_29, 3, __pyx_v_p4);
      __pyx_t_23 = __Pyx_PyObject_Append(__pyx_v_P, __pyx_t_29); if (unlikely(__pyx_t_23 == ((int)-1))) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":475
 *     for cam_id in cam_ids_used:
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):             # <<<<<<<<<<<<<<
 *                 P.append( (p1,p2,p3,p4) )
 *     if len(P) < 2:
 */
    }

    /* "flydra_core/_reconstruct_utils.pyx":473
 *     # calculate line3d
 *     P = []
 *     for cam_id in cam_ids_used:             # <<<<<<<<<<<<<<
 *         x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":477
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 *                 P.append( (p1,p2,p3,p4) )
 *     if len(P) < 2:             # <<<<<<<<<<<<<<
 *         Lcoords = None
 *     else:
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_P); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_t_22 = ((__pyx_t_5 < 2) != 0);
  if (__pyx_t_22) {

    /* "flydra_core/_reconstruct_utils.pyx":478
 *                 P.append( (p1,p2,p3,p4) )
 *     if len(P) < 2:
 *         Lcoords = None             # <<<<<<<<<<<<<<
 *     else:
 *         P = numpy.array(P)
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_Lcoords = ((PyObject*)Py_None);

    /* "flydra_core/_reconstruct_utils.pyx":477
 *         if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
 *                 P.append( (p1,p2,p3,p4) )
 *     if len(P) < 2:             # <<<<<<<<<<<<<<
 *         Lcoords = None
 *     else:
 */
    goto __pyx_L60;
  }

  /* "flydra_core/_reconstruct_utils.pyx":480
 *         Lcoords = None
 *     else:
 *         P = numpy.array(P)             # <<<<<<<<<<<<<<
 *         try:
 *             u,d,vt=svd(P,full_matrices=True)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_29, __pyx_n_s_numpy); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_array); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __pyx_t_29 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_28))) {
      __pyx_t_29 = PyMethod_GET_SELF(__pyx_t_28);
      if (likely(__pyx_t_29)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_28);
        __Pyx_INCREF(__pyx_t_29);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_28, function);
      }
    }
    __pyx_t_7 = (__pyx_t_29) ? __Pyx_PyObject_Call2Args(__pyx_t_28, __pyx_t_29, __pyx_v_P) : __Pyx_PyObject_CallOneArg(__pyx_t_28, __pyx_v_P);
    __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF_SET(__pyx_v_P, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "flydra_core/_reconstruct_utils.pyx":481
 *     else:
 *         P = numpy.array(P)
 *         try:             # <<<<<<<<<<<<<<
 *             u,d,vt=svd(P,full_matrices=True)
 *         except numpy.linalg.LinAlgError, err:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "flydra_core/_reconstruct_utils.pyx":482
 *         P = numpy.array(P)
 *         try:
 *             u,d,vt=svd(P,full_matrices=True)             # <<<<<<<<<<<<<<
 *         except numpy.linalg.LinAlgError, err:
 *             print 'SVD error, P=',repr(P)
 */
        __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 482, __pyx_L61_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_P);
        __Pyx_GIVEREF(__pyx_v_P);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_P);
        __pyx_t_28 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 482, __pyx_L61_error)
        __Pyx_GOTREF(__pyx_t_28);
        if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_full_matrices, Py_True) < 0) __PYX_ERR(0, 482, __pyx_L61_error)
        __pyx_t_29 = __Pyx_PyObject_Call(__pyx_v_svd, __pyx_t_7, __pyx_t_28); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 482, __pyx_L61_error)
        __Pyx_GOTREF(__pyx_t_29);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_29))) || (PyList_CheckExact(__pyx_t_29))) {
          PyObject* sequence = __pyx_t_29;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 482, __pyx_L61_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_28 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_28 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_28);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          #else
          __pyx_t_28 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 482, __pyx_L61_error)
          __Pyx_GOTREF(__pyx_t_28);
          __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 482, __pyx_L61_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 482, __pyx_L61_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
          __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_27 = PyObject_GetIter(__pyx_t_29); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 482, __pyx_L61_error)
          __Pyx_GOTREF(__pyx_t_27);
          __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
          __pyx_t_17 = Py_TYPE(__pyx_t_27)->tp_iternext;
          index = 0; __pyx_t_28 = __pyx_t_17(__pyx_t_27); if (unlikely(!__pyx_t_28)) goto __pyx_L67_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_28);
          index = 1; __pyx_t_7 = __pyx_t_17(__pyx_t_27); if (unlikely(!__pyx_t_7)) goto __pyx_L67_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          index = 2; __pyx_t_8 = __pyx_t_17(__pyx_t_27); if (unlikely(!__pyx_t_8)) goto __pyx_L67_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_8);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_27), 3) < 0) __PYX_ERR(0, 482, __pyx_L61_error)
          __pyx_t_17 = NULL;
          __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
          goto __pyx_L68_unpacking_done;
          __pyx_L67_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
          __pyx_t_17 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 482, __pyx_L61_error)
          __pyx_L68_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_u, __pyx_t_28);
        __pyx_t_28 = 0;
        __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_vt, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":481
 *     else:
 *         P = numpy.array(P)
 *         try:             # <<<<<<<<<<<<<<
 *             u,d,vt=svd(P,full_matrices=True)
 *         except numpy.linalg.LinAlgError, err:
 */
      }

      /* "flydra_core/_reconstruct_utils.pyx":491
 *             raise
 *         else:
 *             P = vt[0,:] # P,Q are planes (take row because this is transpose(V))             # <<<<<<<<<<<<<<
 *             Q = vt[1,:]
 * 
 */
      /*else:*/ {
        __pyx_t_29 = __Pyx_PyObject_GetItem(__pyx_v_vt, __pyx_tuple__23); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 491, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __Pyx_DECREF_SET(__pyx_v_P, __pyx_t_29);
        __pyx_t_29 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":492
 *         else:
 *             P = vt[0,:] # P,Q are planes (take row because this is transpose(V))
 *             Q = vt[1,:]             # <<<<<<<<<<<<<<
 * 
 *             # directly to Pluecker line coordinates
 */
        __pyx_t_29 = __Pyx_PyObject_GetItem(__pyx_v_vt, __pyx_tuple__24); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 492, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __pyx_v_Q = __pyx_t_29;
        __pyx_t_29 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":495
 * 
 *             # directly to Pluecker line coordinates
 *             Lcoords = ( -(P[3]*Q[2]) + P[2]*Q[3],             # <<<<<<<<<<<<<<
 *                           P[3]*Q[1]  - P[1]*Q[3],
 *                         -(P[2]*Q[1]) + P[1]*Q[2],
 */
        __pyx_t_29 = __Pyx_GetItemInt(__pyx_v_P, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_Q, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = PyNumber_Multiply(__pyx_t_29, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = PyNumber_Negative(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_P, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_29 = __Pyx_GetItemInt(__pyx_v_Q, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __pyx_t_28 = PyNumber_Multiply(__pyx_t_7, __pyx_t_29); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_28);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
        __pyx_t_29 = PyNumber_Add(__pyx_t_8, __pyx_t_28); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":496
 *             # directly to Pluecker line coordinates
 *             Lcoords = ( -(P[3]*Q[2]) + P[2]*Q[3],
 *                           P[3]*Q[1]  - P[1]*Q[3],             # <<<<<<<<<<<<<<
 *                         -(P[2]*Q[1]) + P[1]*Q[2],
 *                         -(P[3]*Q[0]) + P[0]*Q[3],
 */
        __pyx_t_28 = __Pyx_GetItemInt(__pyx_v_P, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_28);
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_Q, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = PyNumber_Multiply(__pyx_t_28, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_P, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_28 = __Pyx_GetItemInt(__pyx_v_Q, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_28);
        __pyx_t_27 = PyNumber_Multiply(__pyx_t_8, __pyx_t_28); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_27);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
        __pyx_t_28 = PyNumber_Subtract(__pyx_t_7, __pyx_t_27); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 496, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_28);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":497
 *             Lcoords = ( -(P[3]*Q[2]) + P[2]*Q[3],
 *                           P[3]*Q[1]  - P[1]*Q[3],
 *                         -(P[2]*Q[1]) + P[1]*Q[2],             # <<<<<<<<<<<<<<
 *                         -(P[3]*Q[0]) + P[0]*Q[3],
 *                         -(P[2]*Q[0]) + P[0]*Q[2],
 */
        __pyx_t_27 = __Pyx_GetItemInt(__pyx_v_P, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_27);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_Q, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyNumber_Multiply(__pyx_t_27, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Negative(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_P, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_27 = __Pyx_GetItemInt(__pyx_v_Q, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_27);
        __pyx_t_26 = PyNumber_Multiply(__pyx_t_8, __pyx_t_27); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_26);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
        __pyx_t_27 = PyNumber_Add(__pyx_t_7, __pyx_t_26); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 497, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_27);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":498
 *                           P[3]*Q[1]  - P[1]*Q[3],
 *                         -(P[2]*Q[1]) + P[1]*Q[2],
 *                         -(P[3]*Q[0]) + P[0]*Q[3],             # <<<<<<<<<<<<<<
 *                         -(P[2]*Q[0]) + P[0]*Q[2],
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )
 */
        __pyx_t_26 = __Pyx_GetItemInt(__pyx_v_P, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_26);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_Q, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyNumber_Multiply(__pyx_t_26, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Negative(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_P, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_26 = __Pyx_GetItemInt(__pyx_v_Q, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_26);
        __pyx_t_25 = PyNumber_Multiply(__pyx_t_8, __pyx_t_26); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_25);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
        __pyx_t_26 = PyNumber_Add(__pyx_t_7, __pyx_t_25); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 498, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_26);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":499
 *                         -(P[2]*Q[1]) + P[1]*Q[2],
 *                         -(P[3]*Q[0]) + P[0]*Q[3],
 *                         -(P[2]*Q[0]) + P[0]*Q[2],             # <<<<<<<<<<<<<<
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )
 *             if npy_isnan(Lcoords[0]):
 */
        __pyx_t_25 = __Pyx_GetItemInt(__pyx_v_P, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_25);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_Q, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyNumber_Multiply(__pyx_t_25, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Negative(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_P, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_25 = __Pyx_GetItemInt(__pyx_v_Q, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_25);
        __pyx_t_2 = PyNumber_Multiply(__pyx_t_8, __pyx_t_25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __pyx_t_25 = PyNumber_Add(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 499, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_25);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":500
 *                         -(P[3]*Q[0]) + P[0]*Q[3],
 *                         -(P[2]*Q[0]) + P[0]*Q[2],
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )             # <<<<<<<<<<<<<<
 *             if npy_isnan(Lcoords[0]):
 *                 Lcoords = None
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_P, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_Q, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyNumber_Multiply(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyNumber_Negative(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_P, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_Q, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = PyNumber_Multiply(__pyx_t_8, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Add(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":495
 * 
 *             # directly to Pluecker line coordinates
 *             Lcoords = ( -(P[3]*Q[2]) + P[2]*Q[3],             # <<<<<<<<<<<<<<
 *                           P[3]*Q[1]  - P[1]*Q[3],
 *                         -(P[2]*Q[1]) + P[1]*Q[2],
 */
        __pyx_t_6 = PyTuple_New(6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 495, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_29);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_29);
        __Pyx_GIVEREF(__pyx_t_28);
        PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_28);
        __Pyx_GIVEREF(__pyx_t_27);
        PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_27);
        __Pyx_GIVEREF(__pyx_t_26);
        PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_26);
        __Pyx_GIVEREF(__pyx_t_25);
        PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_25);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_2);
        __pyx_t_29 = 0;
        __pyx_t_28 = 0;
        __pyx_t_27 = 0;
        __pyx_t_26 = 0;
        __pyx_t_25 = 0;
        __pyx_t_2 = 0;
        __pyx_v_Lcoords = ((PyObject*)__pyx_t_6);
        __pyx_t_6 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":501
 *                         -(P[2]*Q[0]) + P[0]*Q[2],
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )
 *             if npy_isnan(Lcoords[0]):             # <<<<<<<<<<<<<<
 *                 Lcoords = None
 *     ## c_lib.free(least_err_by_n_cameras)
 */
        __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v_Lcoords, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 501, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_18 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_18 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L63_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_22 = (npy_isnan(__pyx_t_18) != 0);
        if (__pyx_t_22) {

          /* "flydra_core/_reconstruct_utils.pyx":502
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )
 *             if npy_isnan(Lcoords[0]):
 *                 Lcoords = None             # <<<<<<<<<<<<<<
 *     ## c_lib.free(least_err_by_n_cameras)
 *     return X, Lcoords, cam_ids_used, mean_dist
 */
          __Pyx_INCREF(Py_None);
          __Pyx_DECREF_SET(__pyx_v_Lcoords, ((PyObject*)Py_None));

          /* "flydra_core/_reconstruct_utils.pyx":501
 *                         -(P[2]*Q[0]) + P[0]*Q[2],
 *                         -(P[1]*Q[0]) + P[0]*Q[1] )
 *             if npy_isnan(Lcoords[0]):             # <<<<<<<<<<<<<<
 *                 Lcoords = None
 *     ## c_lib.free(least_err_by_n_cameras)
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L66_try_end;
      __pyx_L61_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
      __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
      __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
      __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
      __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "flydra_core/_reconstruct_utils.pyx":483
 *         try:
 *             u,d,vt=svd(P,full_matrices=True)
 *         except numpy.linalg.LinAlgError, err:             # <<<<<<<<<<<<<<
 *             print 'SVD error, P=',repr(P)
 *             Lcoords = None
 */
      __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_2, &__pyx_t_25);
      __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_numpy); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 483, __pyx_L63_except_error)
      __Pyx_GOTREF(__pyx_t_26);
      __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_linalg); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 483, __pyx_L63_except_error)
      __Pyx_GOTREF(__pyx_t_27);
      __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
      __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_27, __pyx_n_s_LinAlgError); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 483, __pyx_L63_except_error)
      __Pyx_GOTREF(__pyx_t_26);
      __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
      __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_26);
      __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_2, __pyx_t_25);
      __pyx_t_6 = 0; __pyx_t_2 = 0; __pyx_t_25 = 0;
      if (__pyx_t_10) {
        __Pyx_AddTraceback("flydra_core._reconstruct_utils.hypothesis_testing_algorithm__find_best_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_25, &__pyx_t_2, &__pyx_t_6) < 0) __PYX_ERR(0, 483, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_25);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_err = __pyx_t_2;

        /* "flydra_core/_reconstruct_utils.pyx":484
 *             u,d,vt=svd(P,full_matrices=True)
 *         except numpy.linalg.LinAlgError, err:
 *             print 'SVD error, P=',repr(P)             # <<<<<<<<<<<<<<
 *             Lcoords = None
 *         except:
 */
        __pyx_t_26 = PyObject_Repr(__pyx_v_P); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 484, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_26);
        __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 484, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_27);
        __Pyx_INCREF(__pyx_kp_s_SVD_error_P);
        __Pyx_GIVEREF(__pyx_kp_s_SVD_error_P);
        PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_kp_s_SVD_error_P);
        __Pyx_GIVEREF(__pyx_t_26);
        PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_26);
        __pyx_t_26 = 0;
        if (__Pyx_Print(0, __pyx_t_27, 1) < 0) __PYX_ERR(0, 484, __pyx_L63_except_error)
        __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

        /* "flydra_core/_reconstruct_utils.pyx":485
 *         except numpy.linalg.LinAlgError, err:
 *             print 'SVD error, P=',repr(P)
 *             Lcoords = None             # <<<<<<<<<<<<<<
 *         except:
 *             print 'Error on P'
 */
        __Pyx_INCREF(Py_None);
        __pyx_v_Lcoords = ((PyObject*)Py_None);
        __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L62_exception_handled;
      }

      /* "flydra_core/_reconstruct_utils.pyx":486
 *             print 'SVD error, P=',repr(P)
 *             Lcoords = None
 *         except:             # <<<<<<<<<<<<<<
 *             print 'Error on P'
 *             print P
 */
      /*except:*/ {
        __Pyx_AddTraceback("flydra_core._reconstruct_utils.hypothesis_testing_algorithm__find_best_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_2, &__pyx_t_25) < 0) __PYX_ERR(0, 486, __pyx_L63_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_25);

        /* "flydra_core/_reconstruct_utils.pyx":487
 *             Lcoords = None
 *         except:
 *             print 'Error on P'             # <<<<<<<<<<<<<<
 *             print P
 *             raise
 */
        if (__Pyx_PrintOne(0, __pyx_kp_s_Error_on_P) < 0) __PYX_ERR(0, 487, __pyx_L63_except_error)

        /* "flydra_core/_reconstruct_utils.pyx":488
 *         except:
 *             print 'Error on P'
 *             print P             # <<<<<<<<<<<<<<
 *             raise
 *         else:
 */
        if (__Pyx_PrintOne(0, __pyx_v_P) < 0) __PYX_ERR(0, 488, __pyx_L63_except_error)

        /* "flydra_core/_reconstruct_utils.pyx":489
 *             print 'Error on P'
 *             print P
 *             raise             # <<<<<<<<<<<<<<
 *         else:
 *             P = vt[0,:] # P,Q are planes (take row because this is transpose(V))
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_25);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_2, __pyx_t_25);
        __pyx_t_6 = 0; __pyx_t_2 = 0; __pyx_t_25 = 0; 
        __PYX_ERR(0, 489, __pyx_L63_except_error)
      }
      __pyx_L63_except_error:;

      /* "flydra_core/_reconstruct_utils.pyx":481
 *     else:
 *         P = numpy.array(P)
 *         try:             # <<<<<<<<<<<<<<
 *             u,d,vt=svd(P,full_matrices=True)
 *         except numpy.linalg.LinAlgError, err:
 */
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L62_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
      __pyx_L66_try_end:;
    }
  }
  __pyx_L60:;

  /* "flydra_core/_reconstruct_utils.pyx":504
 *                 Lcoords = None
 *     ## c_lib.free(least_err_by_n_cameras)
 *     return X, Lcoords, cam_ids_used, mean_dist             # <<<<<<<<<<<<<<
 * 
 * ## def undistort_image( char* src, int srcstep,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_25 = PyFloat_FromDouble(__pyx_v_mean_dist); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_X);
  __Pyx_INCREF(__pyx_v_Lcoords);
  __Pyx_GIVEREF(__pyx_v_Lcoords);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_Lcoords);
  __Pyx_INCREF(__pyx_v_cam_ids_used);
  __Pyx_GIVEREF(__pyx_v_cam_ids_used);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_cam_ids_used);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_25);
  __pyx_t_25 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "flydra_core/_reconstruct_utils.pyx":300
 * gave_water_warning = 0
 * 
 * def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,             # <<<<<<<<<<<<<<
 *                                                 double ACCEPTABLE_DISTANCE_PIXELS,
 *                                                 int debug=0, Py_ssize_t max_n_cams=5,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_AddTraceback("flydra_core._reconstruct_utils.hypothesis_testing_algorithm__find_best_3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_svd);
  __Pyx_XDECREF(__pyx_v_cam_ids);
  __Pyx_XDECREF(__pyx_v_least_err_by_n_cameras);
  __Pyx_XDECREF(__pyx_v_allA);
  __Pyx_XDECREF(__pyx_v_bad_cam_ids);
  __Pyx_XDECREF(__pyx_v_cam_id2idx);
  __Pyx_XDECREF(__pyx_v_all2d);
  __Pyx_XDECREF(__pyx_v_Pmat_fastnx);
  __Pyx_XDECREF(__pyx_v_cam_id);
  __Pyx_XDECREF(__pyx_v_value_tuple);
  __Pyx_XDECREF(__pyx_v_xy_values);
  __Pyx_XDECREF(__pyx_v_Pmat);
  __Pyx_XDECREF(__pyx_v_row3);
  __Pyx_XDECREF(__pyx_v_cam_ids_for_least_err);
  __Pyx_XDECREF(__pyx_v_X_for_least_err);
  __Pyx_XDECREF(__pyx_v_cam_ids_used);
  __Pyx_XDECREF(__pyx_v_good_A_idx);
  __Pyx_XDECREF(__pyx_v_A);
  __Pyx_XDECREF(__pyx_v_u);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_vt);
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XDECREF(__pyx_v_new_xyw);
  __Pyx_XDECREF(__pyx_v_cam_ids_and_points2d);
  __Pyx_XDECREF(__pyx_v_xy);
  __Pyx_XDECREF(__pyx_v_P);
  __Pyx_XDECREF(__pyx_v_area);
  __Pyx_XDECREF(__pyx_v_slope);
  __Pyx_XDECREF(__pyx_v_eccentricity);
  __Pyx_XDECREF(__pyx_v_p1);
  __Pyx_XDECREF(__pyx_v_p2);
  __Pyx_XDECREF(__pyx_v_p3);
  __Pyx_XDECREF(__pyx_v_p4);
  __Pyx_XDECREF(__pyx_v_Lcoords);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_Q);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_11flydra_core_18_reconstruct_utils_ReconstructHelper(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_11flydra_core_18_reconstruct_utils_ReconstructHelper(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc1_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3fc2_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc1_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3cc2_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc1p(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc1p_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc2p(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4fc2p_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc1p(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc1p_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc2p(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4cc2p_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k1_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k2_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k3(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2k3_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_p1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p1_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_p2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2p2_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_alpha_c(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7alpha_c_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_simple(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_6simple_1__get__(o);
}

static PyMethodDef __pyx_methods_11flydra_core_18_reconstruct_utils_ReconstructHelper[] = {
  {"__reduce__", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_3__reduce__, METH_NOARGS, __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_2__reduce__},
  {"add_element", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_5add_element, METH_O, __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_4add_element},
  {"as_obj_for_json", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_7as_obj_for_json, METH_NOARGS, 0},
  {"save_to_rad_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_9save_to_rad_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_K", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_13get_K, METH_NOARGS, 0},
  {"get_Kp", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_15get_Kp, METH_NOARGS, 0},
  {"get_nlparams", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_17get_nlparams, METH_NOARGS, 0},
  {"undistort", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_19undistort, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_18undistort},
  {"distort", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_21distort, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper_20distort},
  {"undistort_image", (PyCFunction)__pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_23undistort_image, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_11flydra_core_18_reconstruct_utils_ReconstructHelper[] = {
  {(char *)"fc1", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc1, 0, (char *)0, 0},
  {(char *)"fc2", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc2, 0, (char *)0, 0},
  {(char *)"cc1", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc1, 0, (char *)0, 0},
  {(char *)"cc2", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc2, 0, (char *)0, 0},
  {(char *)"fc1p", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc1p, 0, (char *)0, 0},
  {(char *)"fc2p", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_fc2p, 0, (char *)0, 0},
  {(char *)"cc1p", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc1p, 0, (char *)0, 0},
  {(char *)"cc2p", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_cc2p, 0, (char *)0, 0},
  {(char *)"k1", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k1, 0, (char *)0, 0},
  {(char *)"k2", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k2, 0, (char *)0, 0},
  {(char *)"k3", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_k3, 0, (char *)0, 0},
  {(char *)"p1", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_p1, 0, (char *)0, 0},
  {(char *)"p2", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_p2, 0, (char *)0, 0},
  {(char *)"alpha_c", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_alpha_c, 0, (char *)0, 0},
  {(char *)"simple", __pyx_getprop_11flydra_core_18_reconstruct_utils_17ReconstructHelper_simple, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper = {
  PyVarObject_HEAD_INIT(0, 0)
  "flydra_core._reconstruct_utils.ReconstructHelper", /*tp_name*/
  sizeof(struct __pyx_obj_11flydra_core_18_reconstruct_utils_ReconstructHelper), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_11flydra_core_18_reconstruct_utils_ReconstructHelper, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_11__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_11flydra_core_18_reconstruct_utils_ReconstructHelper, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_11flydra_core_18_reconstruct_utils_ReconstructHelper, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_11flydra_core_18_reconstruct_utils_17ReconstructHelper_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_11flydra_core_18_reconstruct_utils_ReconstructHelper, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__reconstruct_utils(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__reconstruct_utils},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_reconstruct_utils",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 0, 1, 0},
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_ACCEPTABLE_DISTANCE_PIXELS, __pyx_k_ACCEPTABLE_DISTANCE_PIXELS, sizeof(__pyx_k_ACCEPTABLE_DISTANCE_PIXELS), 0, 0, 1, 1},
  {&__pyx_n_s_ET, __pyx_k_ET, sizeof(__pyx_k_ET), 0, 0, 1, 1},
  {&__pyx_kp_s_Error_on_P, __pyx_k_Error_on_P, sizeof(__pyx_k_Error_on_P), 0, 0, 1, 0},
  {&__pyx_kp_s_HYPOTHESIS_TEST_least_err_ACCEPT, __pyx_k_HYPOTHESIS_TEST_least_err_ACCEPT, sizeof(__pyx_k_HYPOTHESIS_TEST_least_err_ACCEPT), 0, 0, 1, 0},
  {&__pyx_kp_s_HYPOTHESIS_TEST_mean_dist_f_for, __pyx_k_HYPOTHESIS_TEST_mean_dist_f_for, sizeof(__pyx_k_HYPOTHESIS_TEST_mean_dist_f_for), 0, 0, 1, 0},
  {&__pyx_kp_s_HYPOTHESIS_TEST_n_cams_d_f, __pyx_k_HYPOTHESIS_TEST_n_cams_d_f, sizeof(__pyx_k_HYPOTHESIS_TEST_n_cams_d_f), 0, 0, 1, 0},
  {&__pyx_kp_s_HYPOTHESIS_TEST_shorting_for_n_c, __pyx_k_HYPOTHESIS_TEST_shorting_for_n_c, sizeof(__pyx_k_HYPOTHESIS_TEST_shorting_for_n_c), 0, 0, 1, 0},
  {&__pyx_n_s_K11, __pyx_k_K11, sizeof(__pyx_k_K11), 0, 0, 1, 1},
  {&__pyx_kp_s_K11_s, __pyx_k_K11_s, sizeof(__pyx_k_K11_s), 0, 0, 1, 0},
  {&__pyx_n_s_K12, __pyx_k_K12, sizeof(__pyx_k_K12), 0, 0, 1, 1},
  {&__pyx_kp_s_K12_s, __pyx_k_K12_s, sizeof(__pyx_k_K12_s), 0, 0, 1, 0},
  {&__pyx_n_s_K13, __pyx_k_K13, sizeof(__pyx_k_K13), 0, 0, 1, 1},
  {&__pyx_kp_s_K13_s, __pyx_k_K13_s, sizeof(__pyx_k_K13_s), 0, 0, 1, 0},
  {&__pyx_kp_s_K21_s, __pyx_k_K21_s, sizeof(__pyx_k_K21_s), 0, 0, 1, 0},
  {&__pyx_n_s_K22, __pyx_k_K22, sizeof(__pyx_k_K22), 0, 0, 1, 1},
  {&__pyx_kp_s_K22_s, __pyx_k_K22_s, sizeof(__pyx_k_K22_s), 0, 0, 1, 0},
  {&__pyx_n_s_K23, __pyx_k_K23, sizeof(__pyx_k_K23), 0, 0, 1, 1},
  {&__pyx_kp_s_K23_s, __pyx_k_K23_s, sizeof(__pyx_k_K23_s), 0, 0, 1, 0},
  {&__pyx_kp_s_K31_s, __pyx_k_K31_s, sizeof(__pyx_k_K31_s), 0, 0, 1, 0},
  {&__pyx_kp_s_K32_s, __pyx_k_K32_s, sizeof(__pyx_k_K32_s), 0, 0, 1, 0},
  {&__pyx_kp_s_K33_s, __pyx_k_K33_s, sizeof(__pyx_k_K33_s), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_Lcoords, __pyx_k_Lcoords, sizeof(__pyx_k_Lcoords), 0, 0, 1, 1},
  {&__pyx_n_s_LinAlgError, __pyx_k_LinAlgError, sizeof(__pyx_k_LinAlgError), 0, 0, 1, 1},
  {&__pyx_n_s_MINIMUM_ECCENTRICITY, __pyx_k_MINIMUM_ECCENTRICITY, sizeof(__pyx_k_MINIMUM_ECCENTRICITY), 0, 0, 1, 1},
  {&__pyx_n_s_NoAcceptablePointFound, __pyx_k_NoAcceptablePointFound, sizeof(__pyx_k_NoAcceptablePointFound), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 0, 1, 1},
  {&__pyx_n_s_Pmat, __pyx_k_Pmat, sizeof(__pyx_k_Pmat), 0, 0, 1, 1},
  {&__pyx_n_s_Pmat_fastnx, __pyx_k_Pmat_fastnx, sizeof(__pyx_k_Pmat_fastnx), 0, 0, 1, 1},
  {&__pyx_n_s_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 0, 1, 1},
  {&__pyx_n_s_ReconstructHelper, __pyx_k_ReconstructHelper, sizeof(__pyx_k_ReconstructHelper), 0, 0, 1, 1},
  {&__pyx_n_s_ReconstructHelper_from_xml, __pyx_k_ReconstructHelper_from_xml, sizeof(__pyx_k_ReconstructHelper_from_xml), 0, 0, 1, 1},
  {&__pyx_n_s_STRICT_WATER, __pyx_k_STRICT_WATER, sizeof(__pyx_k_STRICT_WATER), 0, 0, 1, 1},
  {&__pyx_n_s_STRICT_WATER_HYPOTHESIS_TEST, __pyx_k_STRICT_WATER_HYPOTHESIS_TEST, sizeof(__pyx_k_STRICT_WATER_HYPOTHESIS_TEST), 0, 0, 1, 1},
  {&__pyx_kp_s_SVD_error_P, __pyx_k_SVD_error_P, sizeof(__pyx_k_SVD_error_P), 0, 0, 1, 0},
  {&__pyx_n_s_SubElement, __pyx_k_SubElement, sizeof(__pyx_k_SubElement), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_X, __pyx_k_X, sizeof(__pyx_k_X), 0, 0, 1, 1},
  {&__pyx_n_s_X_for_least_err, __pyx_k_X_for_least_err, sizeof(__pyx_k_X_for_least_err), 0, 0, 1, 1},
  {&__pyx_kp_s__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 0, 1, 0},
  {&__pyx_n_s__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 0, 1, 1},
  {&__pyx_n_s_all2d, __pyx_k_all2d, sizeof(__pyx_k_all2d), 0, 0, 1, 1},
  {&__pyx_n_s_allA, __pyx_k_allA, sizeof(__pyx_k_allA), 0, 0, 1, 1},
  {&__pyx_n_s_allclose, __pyx_k_allclose, sizeof(__pyx_k_allclose), 0, 0, 1, 1},
  {&__pyx_n_s_alloc_n_cams, __pyx_k_alloc_n_cams, sizeof(__pyx_k_alloc_n_cams), 0, 0, 1, 1},
  {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_alpha_c, __pyx_k_alpha_c, sizeof(__pyx_k_alpha_c), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_area, __pyx_k_area, sizeof(__pyx_k_area), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_bad_cam_ids, __pyx_k_bad_cam_ids, sizeof(__pyx_k_bad_cam_ids), 0, 0, 1, 1},
  {&__pyx_n_s_best_n_cams, __pyx_k_best_n_cams, sizeof(__pyx_k_best_n_cams), 0, 0, 1, 1},
  {&__pyx_n_s_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cam_combinations_by_size, __pyx_k_cam_combinations_by_size, sizeof(__pyx_k_cam_combinations_by_size), 0, 0, 1, 1},
  {&__pyx_n_s_cam_id, __pyx_k_cam_id, sizeof(__pyx_k_cam_id), 0, 0, 1, 1},
  {&__pyx_n_s_cam_id2idx, __pyx_k_cam_id2idx, sizeof(__pyx_k_cam_id2idx), 0, 0, 1, 1},
  {&__pyx_n_s_cam_ids, __pyx_k_cam_ids, sizeof(__pyx_k_cam_ids), 0, 0, 1, 1},
  {&__pyx_n_s_cam_ids_and_points2d, __pyx_k_cam_ids_and_points2d, sizeof(__pyx_k_cam_ids_and_points2d), 0, 0, 1, 1},
  {&__pyx_n_s_cam_ids_for_least_err, __pyx_k_cam_ids_for_least_err, sizeof(__pyx_k_cam_ids_for_least_err), 0, 0, 1, 1},
  {&__pyx_n_s_cam_ids_used, __pyx_k_cam_ids_used, sizeof(__pyx_k_cam_ids_used), 0, 0, 1, 1},
  {&__pyx_kp_s_cannot_encode_non_simple_intrins, __pyx_k_cannot_encode_non_simple_intrins, sizeof(__pyx_k_cannot_encode_non_simple_intrins), 0, 0, 1, 0},
  {&__pyx_n_s_cc1, __pyx_k_cc1, sizeof(__pyx_k_cc1), 0, 0, 1, 1},
  {&__pyx_n_s_cc1p, __pyx_k_cc1p, sizeof(__pyx_k_cc1p), 0, 0, 1, 1},
  {&__pyx_n_s_cc2, __pyx_k_cc2, sizeof(__pyx_k_cc2), 0, 0, 1, 1},
  {&__pyx_n_s_cc2p, __pyx_k_cc2p, sizeof(__pyx_k_cc2p), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 0, 1, 1},
  {&__pyx_kp_s_comments_s, __pyx_k_comments_s, sizeof(__pyx_k_comments_s), 0, 0, 1, 0},
  {&__pyx_n_s_common_variables, __pyx_k_common_variables, sizeof(__pyx_k_common_variables), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_dist, __pyx_k_dist, sizeof(__pyx_k_dist), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dot, __pyx_k_dot, sizeof(__pyx_k_dot), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dual, __pyx_k_dual, sizeof(__pyx_k_dual), 0, 0, 1, 1},
  {&__pyx_n_s_eccentricity, __pyx_k_eccentricity, sizeof(__pyx_k_eccentricity), 0, 0, 1, 1},
  {&__pyx_n_s_elem, __pyx_k_elem, sizeof(__pyx_k_elem), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_environ, __pyx_k_environ, sizeof(__pyx_k_environ), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_fc1, __pyx_k_fc1, sizeof(__pyx_k_fc1), 0, 0, 1, 1},
  {&__pyx_n_s_fc1p, __pyx_k_fc1p, sizeof(__pyx_k_fc1p), 0, 0, 1, 1},
  {&__pyx_n_s_fc2, __pyx_k_fc2, sizeof(__pyx_k_fc2), 0, 0, 1, 1},
  {&__pyx_n_s_fc2p, __pyx_k_fc2p, sizeof(__pyx_k_fc2p), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_find, __pyx_k_find, sizeof(__pyx_k_find), 0, 0, 1, 1},
  {&__pyx_n_s_find3d, __pyx_k_find3d, sizeof(__pyx_k_find3d), 0, 0, 1, 1},
  {&__pyx_n_s_find_best_3d, __pyx_k_find_best_3d, sizeof(__pyx_k_find_best_3d), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core, __pyx_k_flydra_core, sizeof(__pyx_k_flydra_core), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core__reconstruct_utils, __pyx_k_flydra_core__reconstruct_utils, sizeof(__pyx_k_flydra_core__reconstruct_utils), 0, 0, 1, 1},
  {&__pyx_kp_s_flydra_core__reconstruct_utils_p, __pyx_k_flydra_core__reconstruct_utils_p, sizeof(__pyx_k_flydra_core__reconstruct_utils_p), 0, 0, 1, 0},
  {&__pyx_n_s_flydra_core_common_variables, __pyx_k_flydra_core_common_variables, sizeof(__pyx_k_flydra_core_common_variables), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core_undistort, __pyx_k_flydra_core_undistort, sizeof(__pyx_k_flydra_core_undistort), 0, 0, 1, 1},
  {&__pyx_n_s_fname, __pyx_k_fname, sizeof(__pyx_k_fname), 0, 0, 1, 1},
  {&__pyx_n_s_found_data_dict, __pyx_k_found_data_dict, sizeof(__pyx_k_found_data_dict), 0, 0, 1, 1},
  {&__pyx_n_s_full_matrices, __pyx_k_full_matrices, sizeof(__pyx_k_full_matrices), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_K, __pyx_k_get_K, sizeof(__pyx_k_get_K), 0, 0, 1, 1},
  {&__pyx_n_s_get_Kp, __pyx_k_get_Kp, sizeof(__pyx_k_get_Kp), 0, 0, 1, 1},
  {&__pyx_n_s_get_nlparams, __pyx_k_get_nlparams, sizeof(__pyx_k_get_nlparams), 0, 0, 1, 1},
  {&__pyx_n_s_good_A_idx, __pyx_k_good_A_idx, sizeof(__pyx_k_good_A_idx), 0, 0, 1, 1},
  {&__pyx_n_s_helper, __pyx_k_helper, sizeof(__pyx_k_helper), 0, 0, 1, 1},
  {&__pyx_n_s_hypothesis_testing_algorithm__fi, __pyx_k_hypothesis_testing_algorithm__fi, sizeof(__pyx_k_hypothesis_testing_algorithm__fi), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_iselement, __pyx_k_iselement, sizeof(__pyx_k_iselement), 0, 0, 1, 1},
  {&__pyx_n_s_isnan, __pyx_k_isnan, sizeof(__pyx_k_isnan), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_k1, __pyx_k_k1, sizeof(__pyx_k_k1), 0, 0, 1, 1},
  {&__pyx_n_s_k2, __pyx_k_k2, sizeof(__pyx_k_k2), 0, 0, 1, 1},
  {&__pyx_n_s_k3, __pyx_k_k3, sizeof(__pyx_k_k3), 0, 0, 1, 1},
  {&__pyx_kp_s_k3_0_0, __pyx_k_k3_0_0, sizeof(__pyx_k_k3_0_0), 0, 0, 1, 0},
  {&__pyx_n_s_kc1, __pyx_k_kc1, sizeof(__pyx_k_kc1), 0, 0, 1, 1},
  {&__pyx_kp_s_kc1_s, __pyx_k_kc1_s, sizeof(__pyx_k_kc1_s), 0, 0, 1, 0},
  {&__pyx_n_s_kc2, __pyx_k_kc2, sizeof(__pyx_k_kc2), 0, 0, 1, 1},
  {&__pyx_kp_s_kc2_s, __pyx_k_kc2_s, sizeof(__pyx_k_kc2_s), 0, 0, 1, 0},
  {&__pyx_n_s_kc3, __pyx_k_kc3, sizeof(__pyx_k_kc3), 0, 0, 1, 1},
  {&__pyx_kp_s_kc3_s, __pyx_k_kc3_s, sizeof(__pyx_k_kc3_s), 0, 0, 1, 0},
  {&__pyx_n_s_kc4, __pyx_k_kc4, sizeof(__pyx_k_kc4), 0, 0, 1, 1},
  {&__pyx_kp_s_kc4_s, __pyx_k_kc4_s, sizeof(__pyx_k_kc4_s), 0, 0, 1, 0},
  {&__pyx_kp_s_kc5_s, __pyx_k_kc5_s, sizeof(__pyx_k_kc5_s), 0, 0, 1, 0},
  {&__pyx_n_s_kw, __pyx_k_kw, sizeof(__pyx_k_kw), 0, 0, 1, 1},
  {&__pyx_n_s_least_err, __pyx_k_least_err, sizeof(__pyx_k_least_err), 0, 0, 1, 1},
  {&__pyx_n_s_least_err_by_n_cameras, __pyx_k_least_err_by_n_cameras, sizeof(__pyx_k_least_err_by_n_cameras), 0, 0, 1, 1},
  {&__pyx_kp_s_least_error_was_f, __pyx_k_least_error_was_f, sizeof(__pyx_k_least_error_was_f), 0, 0, 1, 0},
  {&__pyx_n_s_linalg, __pyx_k_linalg, sizeof(__pyx_k_linalg), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_ReconstructHelper, __pyx_k_make_ReconstructHelper, sizeof(__pyx_k_make_ReconstructHelper), 0, 0, 1, 1},
  {&__pyx_n_s_make_ReconstructHelper_from_rad, __pyx_k_make_ReconstructHelper_from_rad, sizeof(__pyx_k_make_ReconstructHelper_from_rad), 0, 0, 1, 1},
  {&__pyx_n_s_max_n_cams, __pyx_k_max_n_cams, sizeof(__pyx_k_max_n_cams), 0, 0, 1, 1},
  {&__pyx_n_s_mean_dist, __pyx_k_mean_dist, sizeof(__pyx_k_mean_dist), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_missing_cam_data, __pyx_k_missing_cam_data, sizeof(__pyx_k_missing_cam_data), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_n_cams, __pyx_k_n_cams, sizeof(__pyx_k_n_cams), 0, 0, 1, 1},
  {&__pyx_n_s_n_iterations, __pyx_k_n_iterations, sizeof(__pyx_k_n_iterations), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_s_need_to_properly_roundtrip_alpha, __pyx_k_need_to_properly_roundtrip_alpha, sizeof(__pyx_k_need_to_properly_roundtrip_alpha), 0, 0, 1, 0},
  {&__pyx_n_s_new_x, __pyx_k_new_x, sizeof(__pyx_k_new_x), 0, 0, 1, 1},
  {&__pyx_n_s_new_xyw, __pyx_k_new_xyw, sizeof(__pyx_k_new_xyw), 0, 0, 1, 1},
  {&__pyx_n_s_new_y, __pyx_k_new_y, sizeof(__pyx_k_new_y), 0, 0, 1, 1},
  {&__pyx_n_s_non_linear_parameters, __pyx_k_non_linear_parameters, sizeof(__pyx_k_non_linear_parameters), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_numpy_dual, __pyx_k_numpy_dual, sizeof(__pyx_k_numpy_dual), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_orig_x, __pyx_k_orig_x, sizeof(__pyx_k_orig_x), 0, 0, 1, 1},
  {&__pyx_n_s_orig_y, __pyx_k_orig_y, sizeof(__pyx_k_orig_y), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_p1, __pyx_k_p1, sizeof(__pyx_k_p1), 0, 0, 1, 1},
  {&__pyx_n_s_p2, __pyx_k_p2, sizeof(__pyx_k_p2), 0, 0, 1, 1},
  {&__pyx_n_s_p3, __pyx_k_p3, sizeof(__pyx_k_p3), 0, 0, 1, 1},
  {&__pyx_n_s_p4, __pyx_k_p4, sizeof(__pyx_k_p4), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_recon, __pyx_k_recon, sizeof(__pyx_k_recon), 0, 0, 1, 1},
  {&__pyx_kp_s_reconstruct_utils_Hypothesis_te, __pyx_k_reconstruct_utils_Hypothesis_te, sizeof(__pyx_k_reconstruct_utils_Hypothesis_te), 0, 0, 1, 0},
  {&__pyx_n_s_rect, __pyx_k_rect, sizeof(__pyx_k_rect), 0, 0, 1, 1},
  {&__pyx_n_s_return_line_coords, __pyx_k_return_line_coords, sizeof(__pyx_k_return_line_coords), 0, 0, 1, 1},
  {&__pyx_n_s_row3, __pyx_k_row3, sizeof(__pyx_k_row3), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_kp_s_skew_not_tested_with_non_simple, __pyx_k_skew_not_tested_with_non_simple, sizeof(__pyx_k_skew_not_tested_with_non_simple), 0, 0, 1, 0},
  {&__pyx_n_s_slope, __pyx_k_slope, sizeof(__pyx_k_slope), 0, 0, 1, 1},
  {&__pyx_n_s_svd, __pyx_k_svd, sizeof(__pyx_k_svd), 0, 0, 1, 1},
  {&__pyx_n_s_tag, __pyx_k_tag, sizeof(__pyx_k_tag), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_n_s_undistort, __pyx_k_undistort, sizeof(__pyx_k_undistort), 0, 0, 1, 1},
  {&__pyx_n_s_value_tuple, __pyx_k_value_tuple, sizeof(__pyx_k_value_tuple), 0, 0, 1, 1},
  {&__pyx_n_s_vt, __pyx_k_vt, sizeof(__pyx_k_vt), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_kp_s_water_and_hypothesis_testing_not, __pyx_k_water_and_hypothesis_testing_not, sizeof(__pyx_k_water_and_hypothesis_testing_not), 0, 0, 1, 0},
  {&__pyx_n_s_with_water, __pyx_k_with_water, sizeof(__pyx_k_with_water), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_x_kk, __pyx_k_x_kk, sizeof(__pyx_k_x_kk), 0, 0, 1, 1},
  {&__pyx_n_s_xl, __pyx_k_xl, sizeof(__pyx_k_xl), 0, 0, 1, 1},
  {&__pyx_n_s_xml_etree_ElementTree, __pyx_k_xml_etree_ElementTree, sizeof(__pyx_k_xml_etree_ElementTree), 0, 0, 1, 1},
  {&__pyx_n_s_xy, __pyx_k_xy, sizeof(__pyx_k_xy), 0, 0, 1, 1},
  {&__pyx_n_s_xy_values, __pyx_k_xy_values, sizeof(__pyx_k_xy_values), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_y_kk, __pyx_k_y_kk, sizeof(__pyx_k_y_kk), 0, 0, 1, 1},
  {&__pyx_n_s_yl, __pyx_k_yl, sizeof(__pyx_k_yl), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 346, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 352, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "flydra_core/_reconstruct_utils.pyx":36
 *     exec(open(filename).read(),params)
 *     if params['K12'] != 0:
 *         raise NotImplementedError('need to properly roundtrip alpha_c != 0')             # <<<<<<<<<<<<<<
 *     helper = ReconstructHelper(
 *         params['K11'], params['K22'], params['K13'], params['K23'],
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_need_to_properly_roundtrip_alpha); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "flydra_core/_reconstruct_utils.pyx":47
 * 
 *     args = []
 *     for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:             # <<<<<<<<<<<<<<
 *         args.append( float(elem.find(name).text) )
 *     args = tuple(args)
 */
  __pyx_tuple__2 = PyTuple_Pack(9, __pyx_n_s_fc1, __pyx_n_s_fc2, __pyx_n_s_cc1, __pyx_n_s_cc2, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_p1, __pyx_n_s_p2, __pyx_n_s_alpha_c); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "flydra_core/_reconstruct_utils.pyx":106
 *             self.cc2p = cc2p
 *         if not self.simple and self.alpha_c != 0:
 *             raise NotImplementedError('skew not tested with non-simple distortion model')             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_skew_not_tested_with_non_simple); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "flydra_core/_reconstruct_utils.pyx":121
 * 
 *         if self.k3 != 0.0:
 *             raise NotImplementedError('k3 != 0.0')             # <<<<<<<<<<<<<<
 * 
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_k3_0_0); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "flydra_core/_reconstruct_utils.pyx":123
 *             raise NotImplementedError('k3 != 0.0')
 * 
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p',
 *                      ]:
 */
  __pyx_tuple__5 = PyTuple_Pack(13, __pyx_n_s_fc1, __pyx_n_s_fc2, __pyx_n_s_cc1, __pyx_n_s_cc2, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_p1, __pyx_n_s_p2, __pyx_n_s_alpha_c, __pyx_n_s_fc1p, __pyx_n_s_fc2p, __pyx_n_s_cc1p, __pyx_n_s_cc2p); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "flydra_core/_reconstruct_utils.pyx":131
 *     def as_obj_for_json(self):
 *         result = {}
 *         for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',             # <<<<<<<<<<<<<<
 *                      'fc1p', 'fc2p', 'cc1p', 'cc2p']:
 *             result[name] = getattr(self,name)
 */
  __pyx_tuple__6 = PyTuple_Pack(14, __pyx_n_s_fc1, __pyx_n_s_fc2, __pyx_n_s_cc1, __pyx_n_s_cc2, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_k3, __pyx_n_s_p1, __pyx_n_s_p2, __pyx_n_s_alpha_c, __pyx_n_s_fc1p, __pyx_n_s_fc2p, __pyx_n_s_cc1p, __pyx_n_s_cc2p); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "flydra_core/_reconstruct_utils.pyx":138
 *     def save_to_rad_file( self, fname, comments=None ):
 *         if not self.simple:
 *             raise ValueError('cannot encode non-simple intrinsic model as a .rad file')             # <<<<<<<<<<<<<<
 *         rad_fd = open(fname,'w')
 *         K = self.get_K()
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_cannot_encode_non_simple_intrins); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "flydra_core/_reconstruct_utils.pyx":143
 *         nlparams = self.get_nlparams()
 *         k1, k2, p1, p2, k3 = nlparams
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "flydra_core/_reconstruct_utils.pyx":144
 *         k1, k2, p1, p2, k3 = nlparams
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_1); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "flydra_core/_reconstruct_utils.pyx":145
 *         rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_2); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "flydra_core/_reconstruct_utils.pyx":146
 *         rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_0); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "flydra_core/_reconstruct_utils.pyx":147
 *         rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "flydra_core/_reconstruct_utils.pyx":148
 *         rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 */
  __pyx_tuple__13 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "flydra_core/_reconstruct_utils.pyx":149
 *         rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_0); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "flydra_core/_reconstruct_utils.pyx":150
 *         rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
 *         rad_fd.write('\n')
 */
  __pyx_tuple__15 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_1); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "flydra_core/_reconstruct_utils.pyx":151
 *         rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
 *         rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
 *         rad_fd.write('K33 = %s;\n'%repr(K[2,2]))             # <<<<<<<<<<<<<<
 *         rad_fd.write('\n')
 *         rad_fd.write('kc1 = %s;\n'%repr(k1))
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "flydra_core/_reconstruct_utils.pyx":194
 *         K = numpy.array((( self.fc1, self.alpha_c*self.fc1, self.cc1),
 *                          ( 0,        self.fc2,              self.cc2),
 *                          ( 0,        0,                     1       )))             # <<<<<<<<<<<<<<
 *         return K
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_int_0, __pyx_int_0, __pyx_int_1); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "flydra_core/_reconstruct_utils.pyx":339
 *     if with_water:
 *         if STRICT_WATER:
 *             raise NotImplementedError('water and hypothesis testing not yet implemented')             # <<<<<<<<<<<<<<
 *         if not gave_water_warning:
 *             warnings.warn('_reconstruct_utils: Hypothesis test intersection done '
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_water_and_hypothesis_testing_not); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "flydra_core/_reconstruct_utils.pyx":357
 * 
 *         # was a 2d point found?
 *         xy_values = value_tuple[:2]             # <<<<<<<<<<<<<<
 *         x,y = xy_values
 *         if npy_isnan(x):
 */
  __pyx_slice__20 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__20)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__20);
  __Pyx_GIVEREF(__pyx_slice__20);

  /* "flydra_core/_reconstruct_utils.pyx":366
 * 
 *         Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns
 *         row3 = Pmat[2,:]             # <<<<<<<<<<<<<<
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]
 */
  __pyx_slice__21 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__21)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__21);
  __Pyx_GIVEREF(__pyx_slice__21);
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_int_2, __pyx_slice__21); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "flydra_core/_reconstruct_utils.pyx":367
 *         Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns
 *         row3 = Pmat[2,:]
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]             # <<<<<<<<<<<<<<
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_int_0, __pyx_slice__21); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "flydra_core/_reconstruct_utils.pyx":368
 *         row3 = Pmat[2,:]
 *         allA[ i*2, : ] = x*row3 - Pmat[0,:]
 *         allA[ i*2+1, :]= y*row3 - Pmat[1,:]             # <<<<<<<<<<<<<<
 * 
 *         all2d[cam_id] = xy_values
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_int_1, __pyx_slice__21); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "flydra_core/_reconstruct_utils.pyx":412
 *             A = allA[good_A_idx,:]
 *             u,d,vt=svd(A)
 *             X = vt[-1,:]/vt[-1,3] # normalize             # <<<<<<<<<<<<<<
 * 
 *             # calculate reprojection error
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_int_neg_1, __pyx_slice__21); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_int_neg_1, __pyx_int_3); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "flydra_core/_reconstruct_utils.pyx":420
 *                 Pmat = Pmat_fastnx[cam_id]
 *                 new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d
 *                 new_x, new_y = new_xyw[0:2]/new_xyw[2]             # <<<<<<<<<<<<<<
 * 
 *                 dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
 */
  __pyx_slice__27 = PySlice_New(__pyx_int_0, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__27)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__27);
  __Pyx_GIVEREF(__pyx_slice__27);

  /* "flydra_core/_reconstruct_utils.pyx":431
 *                 least_err_by_n_cameras[n_cams] = mean_dist
 *                 cam_ids_for_least_err[n_cams] = cam_ids_used
 *                 X_for_least_err[n_cams] = X[:3]             # <<<<<<<<<<<<<<
 * 
 *     # now we have the best estimate for 2 views, 3 views, ...
 */
  __pyx_slice__28 = PySlice_New(Py_None, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice__28)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__28);
  __Pyx_GIVEREF(__pyx_slice__28);

  /* "flydra_core/_reconstruct_utils.pyx":27
 * MINIMUM_ECCENTRICITY = flydra_core.common_variables.MINIMUM_ECCENTRICITY
 * 
 * STRICT_WATER = int(os.environ.get('STRICT_WATER_HYPOTHESIS_TEST','0'))             # <<<<<<<<<<<<<<
 * 
 * class NoAcceptablePointFound(Exception):
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_n_s_STRICT_WATER_HYPOTHESIS_TEST, __pyx_kp_s_0); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "flydra_core/_reconstruct_utils.pyx":32
 *     pass
 * 
 * def make_ReconstructHelper_from_rad_file(filename):             # <<<<<<<<<<<<<<
 *     params = {}
 *     exec(open(filename).read(),params)
 */
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_n_s_filename, __pyx_n_s_params, __pyx_n_s_helper); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_flydra_core__reconstruct_utils_p, __pyx_n_s_make_ReconstructHelper_from_rad, 32, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 32, __pyx_L1_error)

  /* "flydra_core/_reconstruct_utils.pyx":42
 *     return helper
 * 
 * def ReconstructHelper_from_xml(elem):             # <<<<<<<<<<<<<<
 *     assert ET.iselement(elem)
 *     assert elem.tag == "non_linear_parameters"
 */
  __pyx_tuple__33 = PyTuple_Pack(4, __pyx_n_s_elem, __pyx_n_s_args, __pyx_n_s_name_2, __pyx_n_s_helper); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_flydra_core__reconstruct_utils_p, __pyx_n_s_ReconstructHelper_from_xml, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 42, __pyx_L1_error)

  /* "flydra_core/_reconstruct_utils.pyx":53
 *     return helper
 * 
 * def make_ReconstructHelper(*args,**kw):             # <<<<<<<<<<<<<<
 *     return ReconstructHelper(*args,**kw)
 * 
 */
  __pyx_tuple__35 = PyTuple_Pack(2, __pyx_n_s_args, __pyx_n_s_kw); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_flydra_core__reconstruct_utils_p, __pyx_n_s_make_ReconstructHelper, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 53, __pyx_L1_error)

  /* "flydra_core/_reconstruct_utils.pyx":300
 * gave_water_warning = 0
 * 
 * def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,             # <<<<<<<<<<<<<<
 *                                                 double ACCEPTABLE_DISTANCE_PIXELS,
 *                                                 int debug=0, Py_ssize_t max_n_cams=5,
 */
  __pyx_tuple__37 = PyTuple_Pack(57, __pyx_n_s_recon, __pyx_n_s_found_data_dict, __pyx_n_s_ACCEPTABLE_DISTANCE_PIXELS, __pyx_n_s_debug, __pyx_n_s_max_n_cams, __pyx_n_s_with_water, __pyx_n_s_n_cams, __pyx_n_s_best_n_cams, __pyx_n_s_alloc_n_cams, __pyx_n_s_i, __pyx_n_s_missing_cam_data, __pyx_n_s_alpha, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_orig_x, __pyx_n_s_orig_y, __pyx_n_s_new_x, __pyx_n_s_new_y, __pyx_n_s_dist, __pyx_n_s_mean_dist, __pyx_n_s_least_err, __pyx_n_s_svd, __pyx_n_s_cam_ids, __pyx_n_s_least_err_by_n_cameras, __pyx_n_s_allA, __pyx_n_s_bad_cam_ids, __pyx_n_s_cam_id2idx, __pyx_n_s_all2d, __pyx_n_s_Pmat_fastnx, __pyx_n_s_cam_id, __pyx_n_s_value_tuple, __pyx_n_s_xy_values, __pyx_n_s_Pmat, __pyx_n_s_row3, __pyx_n_s_cam_ids_for_least_err, __pyx_n_s_X_for_least_err, __pyx_n_s_cam_ids_used, __pyx_n_s_good_A_idx, __pyx_n_s_A, __pyx_n_s_u, __pyx_n_s_d, __pyx_n_s_vt, __pyx_n_s_X, __pyx_n_s_new_xyw, __pyx_n_s_cam_ids_and_points2d, __pyx_n_s_xy, __pyx_n_s_P, __pyx_n_s_area, __pyx_n_s_slope, __pyx_n_s_eccentricity, __pyx_n_s_p1, __pyx_n_s_p2, __pyx_n_s_p3, __pyx_n_s_p4, __pyx_n_s_Lcoords, __pyx_n_s_err, __pyx_n_s_Q); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(6, 0, 57, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_flydra_core__reconstruct_utils_p, __pyx_n_s_hypothesis_testing_algorithm__fi, 300, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper.tp_dictoffset && __pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 56, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__.doc = __pyx_doc_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_11flydra_core_18_reconstruct_utils_17ReconstructHelper___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ReconstructHelper, (PyObject *)&__pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_ptype_11flydra_core_18_reconstruct_utils_ReconstructHelper = &__pyx_type_11flydra_core_18_reconstruct_utils_ReconstructHelper;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_reconstruct_utils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_reconstruct_utils(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__reconstruct_utils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__reconstruct_utils(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__reconstruct_utils(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  float __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_reconstruct_utils' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__reconstruct_utils(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_reconstruct_utils", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flydra_core___reconstruct_utils) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flydra_core._reconstruct_utils")) {
      if (unlikely(PyDict_SetItemString(modules, "flydra_core._reconstruct_utils", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "flydra_core/_reconstruct_utils.pyx":4
 * #cython: language_level=2
 * 
 * import numpy             # <<<<<<<<<<<<<<
 * import numpy.dual
 * inf = numpy.inf
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":5
 * 
 * import numpy
 * import numpy.dual             # <<<<<<<<<<<<<<
 * inf = numpy.inf
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy_dual, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":6
 * import numpy
 * import numpy.dual
 * inf = numpy.inf             # <<<<<<<<<<<<<<
 * 
 * cdef double cinf
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inf, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":9
 * 
 * cdef double cinf
 * cinf = inf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_11flydra_core_18_reconstruct_utils_cinf = __pyx_t_3;

  /* "flydra_core/_reconstruct_utils.pyx":12
 * 
 * 
 * import flydra_core.common_variables             # <<<<<<<<<<<<<<
 * import flydra_core.undistort
 * import xml.etree.ElementTree as ET
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_flydra_core_common_variables, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flydra_core, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":13
 * 
 * import flydra_core.common_variables
 * import flydra_core.undistort             # <<<<<<<<<<<<<<
 * import xml.etree.ElementTree as ET
 * import warnings
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_flydra_core_undistort, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flydra_core, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":14
 * import flydra_core.common_variables
 * import flydra_core.undistort
 * import xml.etree.ElementTree as ET             # <<<<<<<<<<<<<<
 * import warnings
 * import os
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__29);
  __Pyx_GIVEREF(__pyx_n_s__29);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__29);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_xml_etree_ElementTree, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ET, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":15
 * import flydra_core.undistort
 * import xml.etree.ElementTree as ET
 * import warnings             # <<<<<<<<<<<<<<
 * import os
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":16
 * import xml.etree.ElementTree as ET
 * import warnings
 * import os             # <<<<<<<<<<<<<<
 * 
 * from libc.math cimport sqrt
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":25
 * 
 * cdef float MINIMUM_ECCENTRICITY
 * MINIMUM_ECCENTRICITY = flydra_core.common_variables.MINIMUM_ECCENTRICITY             # <<<<<<<<<<<<<<
 * 
 * STRICT_WATER = int(os.environ.get('STRICT_WATER_HYPOTHESIS_TEST','0'))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_common_variables); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MINIMUM_ECCENTRICITY); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_11flydra_core_18_reconstruct_utils_MINIMUM_ECCENTRICITY = __pyx_t_4;

  /* "flydra_core/_reconstruct_utils.pyx":27
 * MINIMUM_ECCENTRICITY = flydra_core.common_variables.MINIMUM_ECCENTRICITY
 * 
 * STRICT_WATER = int(os.environ.get('STRICT_WATER_HYPOTHESIS_TEST','0'))             # <<<<<<<<<<<<<<
 * 
 * class NoAcceptablePointFound(Exception):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_environ); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STRICT_WATER, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":29
 * STRICT_WATER = int(os.environ.get('STRICT_WATER_HYPOTHESIS_TEST','0'))
 * 
 * class NoAcceptablePointFound(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_NoAcceptablePointFound, __pyx_n_s_NoAcceptablePointFound, (PyObject *) NULL, __pyx_n_s_flydra_core__reconstruct_utils, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_NoAcceptablePointFound, __pyx_t_1, __pyx_t_5, NULL, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NoAcceptablePointFound, __pyx_t_6) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":32
 *     pass
 * 
 * def make_ReconstructHelper_from_rad_file(filename):             # <<<<<<<<<<<<<<
 *     params = {}
 *     exec(open(filename).read(),params)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_18_reconstruct_utils_1make_ReconstructHelper_from_rad_file, NULL, __pyx_n_s_flydra_core__reconstruct_utils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_ReconstructHelper_from_rad, __pyx_t_1) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":42
 *     return helper
 * 
 * def ReconstructHelper_from_xml(elem):             # <<<<<<<<<<<<<<
 *     assert ET.iselement(elem)
 *     assert elem.tag == "non_linear_parameters"
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_18_reconstruct_utils_3ReconstructHelper_from_xml, NULL, __pyx_n_s_flydra_core__reconstruct_utils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ReconstructHelper_from_xml, __pyx_t_1) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":53
 *     return helper
 * 
 * def make_ReconstructHelper(*args,**kw):             # <<<<<<<<<<<<<<
 *     return ReconstructHelper(*args,**kw)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_18_reconstruct_utils_5make_ReconstructHelper, NULL, __pyx_n_s_flydra_core__reconstruct_utils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_ReconstructHelper, __pyx_t_1) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":298
 * 
 * cdef int gave_water_warning
 * gave_water_warning = 0             # <<<<<<<<<<<<<<
 * 
 * def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,
 */
  __pyx_v_11flydra_core_18_reconstruct_utils_gave_water_warning = 0;

  /* "flydra_core/_reconstruct_utils.pyx":300
 * gave_water_warning = 0
 * 
 * def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,             # <<<<<<<<<<<<<<
 *                                                 double ACCEPTABLE_DISTANCE_PIXELS,
 *                                                 int debug=0, Py_ssize_t max_n_cams=5,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_18_reconstruct_utils_7hypothesis_testing_algorithm__find_best_3d, NULL, __pyx_n_s_flydra_core__reconstruct_utils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hypothesis_testing_algorithm__fi, __pyx_t_1) < 0) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":565
 * ##                     dst[u*cn + i] = 0
 * 
 * find_best_3d = hypothesis_testing_algorithm__find_best_3d # old name             # <<<<<<<<<<<<<<
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_hypothesis_testing_algorithm__fi); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_best_3d, __pyx_t_1) < 0) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_reconstruct_utils.pyx":1
 * #emacs, this is -*-Python-*- mode             # <<<<<<<<<<<<<<
 * #cython: language_level=2
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init flydra_core._reconstruct_utils", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flydra_core._reconstruct_utils");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* PyExec */
static CYTHON_INLINE PyObject* __Pyx_PyExec2(PyObject* o, PyObject* globals) {
    return __Pyx_PyExec3(o, globals, NULL);
}
static PyObject* __Pyx_PyExec3(PyObject* o, PyObject* globals, PyObject* locals) {
    PyObject* result;
    PyObject* s = 0;
    char *code = 0;
    if (!globals || globals == Py_None) {
        globals = __pyx_d;
    } else if (!PyDict_Check(globals)) {
        PyErr_Format(PyExc_TypeError, "exec() arg 2 must be a dict, not %.200s",
                     Py_TYPE(globals)->tp_name);
        goto bad;
    }
    if (!locals || locals == Py_None) {
        locals = globals;
    }
    if (__Pyx_PyDict_GetItemStr(globals, __pyx_n_s_builtins) == NULL) {
        if (PyDict_SetItem(globals, __pyx_n_s_builtins, PyEval_GetBuiltins()) < 0)
            goto bad;
    }
    if (PyCode_Check(o)) {
        if (__Pyx_PyCode_HasFreeVars((PyCodeObject *)o)) {
            PyErr_SetString(PyExc_TypeError,
                "code object passed to exec() may not contain free variables");
            goto bad;
        }
        #if CYTHON_COMPILING_IN_PYPY || PY_VERSION_HEX < 0x030200B1
        result = PyEval_EvalCode((PyCodeObject *)o, globals, locals);
        #else
        result = PyEval_EvalCode(o, globals, locals);
        #endif
    } else {
        PyCompilerFlags cf;
        cf.cf_flags = 0;
        if (PyUnicode_Check(o)) {
            cf.cf_flags = PyCF_SOURCE_IS_UTF8;
            s = PyUnicode_AsUTF8String(o);
            if (!s) goto bad;
            o = s;
        #if PY_MAJOR_VERSION >= 3
        } else if (!PyBytes_Check(o)) {
        #else
        } else if (!PyString_Check(o)) {
        #endif
            PyErr_Format(PyExc_TypeError,
                "exec: arg 1 must be string, bytes or code object, got %.200s",
                Py_TYPE(o)->tp_name);
            goto bad;
        }
        #if PY_MAJOR_VERSION >= 3
        code = PyBytes_AS_STRING(o);
        #else
        code = PyString_AS_STRING(o);
        #endif
        if (PyEval_MergeCompilerFlags(&cf)) {
            result = PyRun_StringFlags(code, Py_file_input, globals, locals, &cf);
        } else {
            result = PyRun_String(code, Py_file_input, globals, locals);
        }
        Py_XDECREF(s);
    }
    return result;
bad:
    Py_XDECREF(s);
    return 0;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* PrintOne */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
