/* Generated by Cython 0.29.20 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\arrayobject.h",
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\arrayscalars.h",
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\ndarrayobject.h",
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\ndarraytypes.h",
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include\\numpy\\ufuncobject.h"
        ],
        "include_dirs": [
            "C:\\Users\\drand\\AppData\\Local\\Packages\\PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0\\LocalCache\\local-packages\\Python38\\site-packages\\numpy\\core\\include"
        ],
        "name": "flydra_core._flydra_tracked_object",
        "sources": [
            "flydra_core/_flydra_tracked_object.pyx"
        ]
    },
    "module_name": "flydra_core._flydra_tracked_object"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_20"
#define CYTHON_HEX_VERSION 0x001D14F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flydra_core___flydra_tracked_object
#define __PYX_HAVE_API__flydra_core___flydra_tracked_object
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "math.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "flydra_core\\_flydra_tracked_object.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
  "flydra_core\\_fastgeom.pxd",
  "flydra_core\\_pmat_jacobian.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "flydra_core/_flydra_tracked_object.pyx":51
 * cnan = np.nan
 * 
 * ctypedef int mybool             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "math.h":
 */
typedef int __pyx_t_11flydra_core_22_flydra_tracked_object_mybool;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple;
struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine;
struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian;
struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate;
struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data;

/* "flydra_core/_flydra_tracked_object.pyx":317
 *                 break
 * 
 *     cpdef calculate_a_posteriori_estimate(self,             # <<<<<<<<<<<<<<
 *                                         long frame,
 *                                         object data_dict,
 */
struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate {
  int __pyx_n;
  int debug1;
  int skip_data_association;
  PyObject *original_camns_and_idxs;
  PyObject *original_cam_ids_and_points2d;
};

/* "flydra_core/_flydra_tracked_object.pyx":492
 *         self.last_frameno_with_data = self.observations_frames[-1]
 * 
 *     cpdef _filter_data(self, object xhatminus, object Pminus,             # <<<<<<<<<<<<<<
 *                        object data_dict, object camn2cam_id,
 *                        int debug=0):
 */
struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data {
  int __pyx_n;
  int debug;
};

/* "_fastgeom.pxd":4
 * #cython: language_level=2
 * 
 * cdef class ThreeTuple:             # <<<<<<<<<<<<<<
 *     cdef readonly double a
 *     cdef readonly double b
 */
struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple {
  PyObject_HEAD
  double a;
  double b;
  double c;
};


/* "_fastgeom.pxd":9
 *     cdef readonly double c
 * 
 * cdef class PlueckerLine:             # <<<<<<<<<<<<<<
 *     cdef readonly ThreeTuple u
 *     cdef readonly ThreeTuple v
 */
struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine {
  PyObject_HEAD
  struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *u;
  struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *v;
};


/* "_pmat_jacobian.pxd":5
 * cimport numpy as np
 * 
 * cdef class PinholeCameraModelWithJacobian:             # <<<<<<<<<<<<<<
 *     cdef double P00, P01, P02, P03
 *     cdef double P10, P11, P12, P13
 */
struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian {
  PyObject_HEAD
  struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *__pyx_vtab;
  double P00;
  double P01;
  double P02;
  double P03;
  double P10;
  double P11;
  double P12;
  double P13;
  double P20;
  double P21;
  double P22;
  double P23;
  PyObject *pmat;
};


/* "flydra_core/_flydra_tracked_object.pyx":136
 *     return val
 * 
 * cdef class TrackedObject:             # <<<<<<<<<<<<<<
 *     """
 *     Track one object using a Kalman filter.
 */
struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_vtab;
  long current_frameno;
  long last_frameno_with_data;
  PY_LONG_LONG max_frames_skipped;
  __pyx_t_11flydra_core_22_flydra_tracked_object_mybool kill_me;
  __pyx_t_11flydra_core_22_flydra_tracked_object_mybool save_all_data;
  double area_threshold;
  double area_threshold_for_orientation;
  PyObject *reconstructor;
  PyObject *my_kalman;
  PyObject *distorted_pixel_euclidian_distance_accept;
  double max_variance;
  PyObject *ekf_observation_covariance_pixels;
  PyObject *frames;
  PyObject *xhats;
  PyObject *timestamps;
  PyObject *Ps;
  PyObject *MLE_position;
  PyObject *MLE_Lcoords;
  PyObject *observations_frames;
  PyObject *observations_2d;
  int disable_image_stat_gating;
  int orientation_consensus;
  PyObject *fake_timestamp;
  unsigned int obj_id;
  PyObject *ekf_kalman_A;
  PyObject *ekf_kalman_Q;
};



/* "_pmat_jacobian.pxd":5
 * cimport numpy as np
 * 
 * cdef class PinholeCameraModelWithJacobian:             # <<<<<<<<<<<<<<
 *     cdef double P00, P01, P02, P03
 *     cdef double P10, P11, P12, P13
 */

struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian {
  void (*evaluate_jacobian_at_)(struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *, double, double, double, double, double *, double *, double *, double *, double *, double *, double *, double *);
  PyObject *(*evaluate_jacobian_at)(struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *, PyArrayObject *, int __pyx_skip_dispatch);
  PyObject *(*evaluate)(struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *, PyArrayObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *__pyx_vtabptr_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian;


/* "flydra_core/_flydra_tracked_object.pyx":136
 *     return val
 * 
 * cdef class TrackedObject:             # <<<<<<<<<<<<<<
 *     """
 *     Track one object using a Kalman filter.
 */

struct __pyx_vtabstruct_11flydra_core_22_flydra_tracked_object_TrackedObject {
  PyObject *(*calculate_a_posteriori_estimate)(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *, long, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate *__pyx_optional_args);
  PyObject *(*_filter_data)(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data *__pyx_optional_args);
};
static struct __pyx_vtabstruct_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, long __pyx_v_frame, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_xhatminus, PyObject *__pyx_v_Pminus, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data *__pyx_optional_args); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'flydra_core._fastgeom' */
static PyTypeObject *__pyx_ptype_11flydra_core_9_fastgeom_ThreeTuple = 0;
static PyTypeObject *__pyx_ptype_11flydra_core_9_fastgeom_PlueckerLine = 0;

/* Module declarations from 'flydra_core._mahalanobis' */
static struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *(*__pyx_f_11flydra_core_12_mahalanobis_line_fit_3d)(struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine *, struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static double (*__pyx_f_11flydra_core_12_mahalanobis_dist2)(struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, PyObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'flydra_core._pmat_jacobian' */
static PyTypeObject *__pyx_ptype_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian = 0;

/* Module declarations from 'flydra_core._flydra_tracked_object' */
static PyTypeObject *__pyx_ptype_11flydra_core_22_flydra_tracked_object_TrackedObject = 0;
static double __pyx_v_11flydra_core_22_flydra_tracked_object_c_inf;
static double __pyx_v_11flydra_core_22_flydra_tracked_object_cnan;
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(PyObject *, PyArrayObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object___pyx_unpickle_TrackedObject__set_state(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_double_t = { "double_t", NULL, sizeof(__pyx_t_5numpy_double_t), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "flydra_core._flydra_tracked_object"
extern int __pyx_module_is_main_flydra_core___flydra_tracked_object;
int __pyx_module_is_main_flydra_core___flydra_tracked_object = 0;

/* Implementation of 'flydra_core._flydra_tracked_object' */
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_A[] = "A";
static const char __pyx_k_C[] = "C";
static const char __pyx_k_P[] = "P";
static const char __pyx_k_Q[] = "Q";
static const char __pyx_k_R[] = "R";
static const char __pyx_k_s[] = "%s\n";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_hx[] = "hx";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_ss[] = "ss";
static const char __pyx_k_EKF[] = "EKF";
static const char __pyx_k__10[] = " ";
static const char __pyx_k__11[] = "  ";
static const char __pyx_k__12[] = "\n";
static const char __pyx_k__13[] = "============--";
static const char __pyx_k__16[] = "--------";
static const char __pyx_k__19[] = "    ->";
static const char __pyx_k__22[] = "*";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_arr[] = "arr";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_eye[] = "eye";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_inv[] = "inv";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_P_k1[] = "P_k1";
static const char __pyx_k_camn[] = "camn";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_geom[] = "geom";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_sqrt[] = "sqrt";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_xhat[] = "xhat";
static const char __pyx_k_Pmean[] = "Pmean";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_camns[] = "camns";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_doing[] = "doing";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_isEKF[] = "isEKF";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_level[] = "level";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_to_hz[] = "to_hz";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_Pminus[] = "Pminus";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_cam_id[] = "  cam_id";
static const char __pyx_k_debug1[] = "debug1";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_find2d[] = "find2d";
static const char __pyx_k_find3d[] = "find3d";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kalman[] = "kalman";
static const char __pyx_k_linalg[] = "linalg";
static const char __pyx_k_newarr[] = "newarr";
static const char __pyx_k_obj_id[] = "obj_id";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pprint[] = "pprint";
static const char __pyx_k_pt_idx[] = "pt_idx";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_wateri[] = "wateri";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_closest[] = "closest";
static const char __pyx_k_distort[] = "distort";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_idx_d_s[] = "(idx %d) %s";
static const char __pyx_k_xhat_k1[] = "xhat_k1";
static const char __pyx_k_fastgeom[] = "_fastgeom";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_isfinite[] = "isfinite";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_tostring[] = "tostring";
static const char __pyx_k_data_dict[] = "data_dict";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_distorted[] = "distorted";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_initial_P[] = "initial_P";
static const char __pyx_k_initial_x[] = "initial_x";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_xhatminus[] = "xhatminus";
static const char __pyx_k_NO_LCOORDS[] = "NO_LCOORDS";
static const char __pyx_k_ThreeTuple[] = "ThreeTuple";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_acceptable[] = "       (acceptable)";
static const char __pyx_k_camn_order[] = "camn_order";
static const char __pyx_k_kalman_ekf[] = "kalman_ekf";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_camn2cam_id[] = "camn2cam_id";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_filter_data[] = "_filter_data";
static const char __pyx_k_flydra_core[] = "flydra_core";
static const char __pyx_k_mahalanobis[] = "_mahalanobis";
static const char __pyx_k_KalmanFilter[] = "KalmanFilter";
static const char __pyx_k_kalman_model[] = "kalman_model";
static const char __pyx_k_missing_data[] = "missing_data";
static const char __pyx_k_obs0_Lcoords[] = "obs0_Lcoords";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_TrackedObject[] = "TrackedObject";
static const char __pyx_k_filter_data_2[] = "_filter_data():";
static const char __pyx_k_obs0_position[] = "obs0_position";
static const char __pyx_k_reconstructor[] = "reconstructor";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_save_all_data[] = "save_all_data";
static const char __pyx_k_PT_TUPLE_IDX_X[] = "PT_TUPLE_IDX_X";
static const char __pyx_k_PT_TUPLE_IDX_Y[] = "PT_TUPLE_IDX_Y";
static const char __pyx_k_area_threshold[] = "area_threshold";
static const char __pyx_k_fake_timestamp[] = "fake_timestamp";
static const char __pyx_k_not_acceptable[] = "       (not acceptable)";
static const char __pyx_k_obs2d_hashable[] = "obs2d_hashable";
static const char __pyx_k_PT_TUPLE_IDX_P1[] = "PT_TUPLE_IDX_P1";
static const char __pyx_k_PT_TUPLE_IDX_P2[] = "PT_TUPLE_IDX_P2";
static const char __pyx_k_PT_TUPLE_IDX_P3[] = "PT_TUPLE_IDX_P3";
static const char __pyx_k_PT_TUPLE_IDX_P4[] = "PT_TUPLE_IDX_P4";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_flydra_core_geom[] = "flydra_core.geom";
static const char __pyx_k_line_from_HZline[] = "line_from_HZline";
static const char __pyx_k_line_from_points[] = "line_from_points";
static const char __pyx_k_PT_TUPLE_IDX_AREA[] = "PT_TUPLE_IDX_AREA";
static const char __pyx_k_data_descriptions[] = "data_descriptions";
static const char __pyx_k_PT_TUPLE_IDX_SLOPE[] = "PT_TUPLE_IDX_SLOPE";
static const char __pyx_k_best_match_idx_d_s[] = "best match idx %d (%s)";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_max_frames_skipped[] = "max_frames_skipped";
static const char __pyx_k_return_line_coords[] = "return_line_coords";
static const char __pyx_k_adskalman_adskalman[] = "adskalman.adskalman";
static const char __pyx_k_pmats_and_points_cov[] = "pmats_and_points_cov";
static const char __pyx_k_orientation_consensus[] = "orientation_consensus";
static const char __pyx_k_skip_data_association[] = "skip_data_association";
static const char __pyx_k_first_observation_idxs[] = "first_observation_idxs";
static const char __pyx_k_flydra_core_kalman_ekf[] = "flydra_core.kalman.ekf";
static const char __pyx_k_so_far_the_best_taking[] = "       (so far the best -- taking)";
static const char __pyx_k_first_observation_camns[] = "first_observation_camns";
static const char __pyx_k_get_model_with_jacobian[] = "get_model_with_jacobian";
static const char __pyx_k_original_camns_and_idxs[] = "original_camns_and_idxs";
static const char __pyx_k_PT_TUPLE_IDX_CUR_VAL_IDX[] = "PT_TUPLE_IDX_CUR_VAL_IDX";
static const char __pyx_k_max_variance_dist_meters[] = "max_variance_dist_meters";
static const char __pyx_k_predicted_2d_undistorted[] = "    predicted_2d (undistorted)";
static const char __pyx_k_PT_TUPLE_IDX_ECCENTRICITY[] = "PT_TUPLE_IDX_ECCENTRICITY";
static const char __pyx_k_PT_TUPLE_IDX_FRAME_PT_IDX[] = "PT_TUPLE_IDX_FRAME_PT_IDX";
static const char __pyx_k_PT_TUPLE_IDX_MEAN_VAL_IDX[] = "PT_TUPLE_IDX_MEAN_VAL_IDX";
static const char __pyx_k_disable_image_stat_gating[] = "disable_image_stat_gating";
static const char __pyx_k_step1__calculate_a_priori[] = "step1__calculate_a_priori";
static const char __pyx_k_pyx_unpickle_TrackedObject[] = "__pyx_unpickle_TrackedObject";
static const char __pyx_k_PT_TUPLE_IDX_SUMSQF_VAL_IDX[] = "PT_TUPLE_IDX_SUMSQF_VAL_IDX";
static const char __pyx_k_Lcoords_3D_body_orientation_s[] = "Lcoords (3D body orientation) : %s";
static const char __pyx_k_flydra_core_data_descriptions[] = "flydra_core.data_descriptions";
static const char __pyx_k_flydra_core_kalman_point_prob[] = "flydra_core.kalman.point_prob";
static const char __pyx_k_original_cam_ids_and_points2d[] = "original_cam_ids_and_points2d";
static const char __pyx_k_step2__calculate_a_posteriori[] = "step2__calculate_a_posteriori";
static const char __pyx_k_TRO_frames_0_r_observations_2d[] = "<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>";
static const char __pyx_k_area_threshold_for_orientation[] = "area_threshold_for_orientation";
static const char __pyx_k_d_observations_d_estimates_for[] = "%d observations, %d estimates for %s\n";
static const char __pyx_k_pt_area_f_cur_d_mean_1f_sumsqf[] = "(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)";
static const char __pyx_k_calculate_a_posteriori_estimate[] = "calculate_a_posteriori_estimate";
static const char __pyx_k_disable_image_stat_gating_not_s[] = "--disable-image-stat-gating not specified and no image statistics were saved.";
static const char __pyx_k_initial_acceleration_covariance[] = "initial_acceleration_covariance_estimate";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_s_removing_previous_observation[] = "%s removing previous observation (from frame %d)\n";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd5[] = "Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))";
static const char __pyx_k_distorted_1f_1f_pixel_dist_1f_ma[] = "distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)";
static const char __pyx_k_distorted_pixel_euclidian_distan[] = "distorted_pixel_euclidian_distance_accept";
static const char __pyx_k_ekf_observation_covariance_pixel[] = "ekf_observation_covariance_pixels";
static const char __pyx_k_flydra_core__flydra_tracked_obje[] = "flydra_core\\_flydra_tracked_object.pyx";
static const char __pyx_k_initial_position_covariance_esti[] = "initial_position_covariance_estimate";
static const char __pyx_k_initial_velocity_covariance_esti[] = "initial_velocity_covariance_estimate";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_position_MLE_used_camns_and_idxs[] = "position MLE, used_camns_and_idxs";
static const char __pyx_k_some_rough_negative_log_likeliho[] = "some_rough_negative_log_likelihood";
static const char __pyx_k_updating_for_d_frames_since_upda[] = "updating for %d frames since update";
static const char __pyx_k_will_kill_next_time_because_Pmea[] = "will kill next time because Pmean too large (%f > %f)";
static const char __pyx_k_will_kill_next_time_because_fram[] = "will kill next time because frames skipped (%ld > %ld, frame %ld)";
static const char __pyx_k_flydra_core__flydra_tracked_obje_2[] = "flydra_core._flydra_tracked_object";
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_n_s_EKF;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd5;
static PyObject *__pyx_n_s_KalmanFilter;
static PyObject *__pyx_kp_s_Lcoords_3D_body_orientation_s;
static PyObject *__pyx_n_s_NO_LCOORDS;
static PyObject *__pyx_n_s_P;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_AREA;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_ECCENTRICITY;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_P1;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_P2;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_P3;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_P4;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_SLOPE;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_X;
static PyObject *__pyx_n_s_PT_TUPLE_IDX_Y;
static PyObject *__pyx_n_s_P_k1;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pmean;
static PyObject *__pyx_n_s_Pminus;
static PyObject *__pyx_n_s_Q;
static PyObject *__pyx_n_s_R;
static PyObject *__pyx_kp_s_TRO_frames_0_r_observations_2d;
static PyObject *__pyx_n_s_ThreeTuple;
static PyObject *__pyx_n_s_TrackedObject;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__10;
static PyObject *__pyx_kp_s__11;
static PyObject *__pyx_kp_s__12;
static PyObject *__pyx_kp_s__13;
static PyObject *__pyx_kp_s__16;
static PyObject *__pyx_kp_s__19;
static PyObject *__pyx_n_s__22;
static PyObject *__pyx_kp_s_acceptable;
static PyObject *__pyx_n_s_adskalman_adskalman;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_area_threshold;
static PyObject *__pyx_n_s_area_threshold_for_orientation;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_arr;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_kp_s_best_match_idx_d_s;
static PyObject *__pyx_n_s_calculate_a_posteriori_estimate;
static PyObject *__pyx_kp_s_cam_id;
static PyObject *__pyx_n_s_camn;
static PyObject *__pyx_n_s_camn2cam_id;
static PyObject *__pyx_n_s_camn_order;
static PyObject *__pyx_n_s_camns;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_closest;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_kp_s_d_observations_d_estimates_for;
static PyObject *__pyx_n_s_data_descriptions;
static PyObject *__pyx_n_s_data_dict;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_debug1;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_direction;
static PyObject *__pyx_n_s_disable_image_stat_gating;
static PyObject *__pyx_kp_s_disable_image_stat_gating_not_s;
static PyObject *__pyx_n_s_distort;
static PyObject *__pyx_n_s_distorted;
static PyObject *__pyx_kp_s_distorted_1f_1f_pixel_dist_1f_ma;
static PyObject *__pyx_n_s_distorted_pixel_euclidian_distan;
static PyObject *__pyx_n_s_doing;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_ekf_observation_covariance_pixel;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_eye;
static PyObject *__pyx_n_s_fake_timestamp;
static PyObject *__pyx_n_s_fastgeom;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_filter_data;
static PyObject *__pyx_kp_s_filter_data_2;
static PyObject *__pyx_n_s_find2d;
static PyObject *__pyx_n_s_find3d;
static PyObject *__pyx_n_s_first_observation_camns;
static PyObject *__pyx_n_s_first_observation_idxs;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_flydra_core;
static PyObject *__pyx_kp_s_flydra_core__flydra_tracked_obje;
static PyObject *__pyx_n_s_flydra_core__flydra_tracked_obje_2;
static PyObject *__pyx_n_s_flydra_core_data_descriptions;
static PyObject *__pyx_n_s_flydra_core_geom;
static PyObject *__pyx_n_s_flydra_core_kalman_ekf;
static PyObject *__pyx_n_s_flydra_core_kalman_point_prob;
static PyObject *__pyx_n_s_frame;
static PyObject *__pyx_n_s_geom;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_model_with_jacobian;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_hx;
static PyObject *__pyx_kp_s_idx_d_s;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_initial_P;
static PyObject *__pyx_n_s_initial_acceleration_covariance;
static PyObject *__pyx_n_s_initial_position_covariance_esti;
static PyObject *__pyx_n_s_initial_velocity_covariance_esti;
static PyObject *__pyx_n_s_initial_x;
static PyObject *__pyx_n_s_inv;
static PyObject *__pyx_n_s_isEKF;
static PyObject *__pyx_n_s_isfinite;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_kalman;
static PyObject *__pyx_n_s_kalman_ekf;
static PyObject *__pyx_n_s_kalman_model;
static PyObject *__pyx_n_s_level;
static PyObject *__pyx_n_s_linalg;
static PyObject *__pyx_n_s_line_from_HZline;
static PyObject *__pyx_n_s_line_from_points;
static PyObject *__pyx_n_s_mahalanobis;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_math;
static PyObject *__pyx_n_s_max_frames_skipped;
static PyObject *__pyx_n_s_max_variance_dist_meters;
static PyObject *__pyx_n_s_missing_data;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_nan;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_newarr;
static PyObject *__pyx_kp_s_not_acceptable;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj_id;
static PyObject *__pyx_n_s_obs0_Lcoords;
static PyObject *__pyx_n_s_obs0_position;
static PyObject *__pyx_n_s_obs2d_hashable;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_orientation_consensus;
static PyObject *__pyx_n_s_original_cam_ids_and_points2d;
static PyObject *__pyx_n_s_original_camns_and_idxs;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pmats_and_points_cov;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_kp_s_position_MLE_used_camns_and_idxs;
static PyObject *__pyx_n_s_pprint;
static PyObject *__pyx_kp_s_predicted_2d_undistorted;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_kp_s_pt_area_f_cur_d_mean_1f_sumsqf;
static PyObject *__pyx_n_s_pt_idx;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_TrackedObject;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reconstructor;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_return_line_coords;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_kp_s_s_removing_previous_observation;
static PyObject *__pyx_n_s_save_all_data;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_skip_data_association;
static PyObject *__pyx_kp_s_so_far_the_best_taking;
static PyObject *__pyx_n_s_some_rough_negative_log_likeliho;
static PyObject *__pyx_n_s_sqrt;
static PyObject *__pyx_n_s_ss;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_step1__calculate_a_priori;
static PyObject *__pyx_n_s_step2__calculate_a_posteriori;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_to_hz;
static PyObject *__pyx_n_s_tostring;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_s_updating_for_d_frames_since_upda;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_wateri;
static PyObject *__pyx_kp_s_will_kill_next_time_because_Pmea;
static PyObject *__pyx_kp_s_will_kill_next_time_because_fram;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_xhat;
static PyObject *__pyx_n_s_xhat_k1;
static PyObject *__pyx_n_s_xhatminus;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pmats_and_points_cov, PyArrayObject *__pyx_v_xhatminus); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_2obs2d_hashable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_arr); /* proto */
static int __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_reconstructor, PyObject *__pyx_v_obj_id, long __pyx_v_frame, PyObject *__pyx_v_obs0_position, PyObject *__pyx_v_obs0_Lcoords, PyObject *__pyx_v_first_observation_camns, PyObject *__pyx_v_first_observation_idxs, PyObject *__pyx_v_kalman_model, PyObject *__pyx_v_save_all_data, double __pyx_v_area_threshold, double __pyx_v_area_threshold_for_orientation, PyObject *__pyx_v_disable_image_stat_gating, PyObject *__pyx_v_orientation_consensus, PyObject *__pyx_v_fake_timestamp); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2__repr__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_4debug_info(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_level); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6get_distance_and_nsigma(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_testx); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_8kill(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10calculate_a_posteriori_estimate(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, long __pyx_v_frame, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_v_debug1, int __pyx_v_skip_data_association, PyObject *__pyx_v_original_camns_and_idxs, PyObject *__pyx_v_original_cam_ids_and_points2d); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12remove_previous_observation(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, int __pyx_v_debug1); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_14_filter_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_xhatminus, PyObject *__pyx_v_Pminus, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_v_debug); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_16get_most_recent_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_18__reduce_cython__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_20__setstate_cython__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_4__pyx_unpickle_TrackedObject(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_11flydra_core_22_flydra_tracked_object_TrackedObject(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_224123000;
static PyObject *__pyx_slice_;
static PyObject *__pyx_slice__2;
static PyObject *__pyx_slice__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__15;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
/* Late includes */

/* "flydra_core/_flydra_tracked_object.pyx":56
 *     double sqrt(double)
 * 
 * cpdef evaluate_pmat_jacobian(object pmats_and_points_cov, np.ndarray[np.double_t, ndim=1] xhatminus):             # <<<<<<<<<<<<<<
 *     cdef int N
 *     cdef mybool missing_data
 */

static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_1evaluate_pmat_jacobian(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(PyObject *__pyx_v_pmats_and_points_cov, PyArrayObject *__pyx_v_xhatminus, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __pyx_t_11flydra_core_22_flydra_tracked_object_mybool __pyx_v_missing_data;
  int __pyx_v_i;
  int __pyx_v_ss;
  struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *__pyx_v_pinhole_model = 0;
  PyArrayObject *__pyx_v_y = 0;
  PyArrayObject *__pyx_v_C = 0;
  PyArrayObject *__pyx_v_hx = 0;
  PyArrayObject *__pyx_v_R = 0;
  PyArrayObject *__pyx_v_hx_i = 0;
  PyObject *__pyx_v_nonlin_model = NULL;
  PyObject *__pyx_v_xy2d_obs = NULL;
  PyObject *__pyx_v_cov = NULL;
  PyObject *__pyx_v_C_i = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_C;
  __Pyx_Buffer __pyx_pybuffer_C;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_R;
  __Pyx_Buffer __pyx_pybuffer_R;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hx;
  __Pyx_Buffer __pyx_pybuffer_hx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hx_i;
  __Pyx_Buffer __pyx_pybuffer_hx_i;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xhatminus;
  __Pyx_Buffer __pyx_pybuffer_xhatminus;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_y;
  __Pyx_Buffer __pyx_pybuffer_y;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  PyArrayObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyArrayObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("evaluate_pmat_jacobian", 0);
  __pyx_pybuffer_y.pybuffer.buf = NULL;
  __pyx_pybuffer_y.refcount = 0;
  __pyx_pybuffernd_y.data = NULL;
  __pyx_pybuffernd_y.rcbuffer = &__pyx_pybuffer_y;
  __pyx_pybuffer_C.pybuffer.buf = NULL;
  __pyx_pybuffer_C.refcount = 0;
  __pyx_pybuffernd_C.data = NULL;
  __pyx_pybuffernd_C.rcbuffer = &__pyx_pybuffer_C;
  __pyx_pybuffer_hx.pybuffer.buf = NULL;
  __pyx_pybuffer_hx.refcount = 0;
  __pyx_pybuffernd_hx.data = NULL;
  __pyx_pybuffernd_hx.rcbuffer = &__pyx_pybuffer_hx;
  __pyx_pybuffer_R.pybuffer.buf = NULL;
  __pyx_pybuffer_R.refcount = 0;
  __pyx_pybuffernd_R.data = NULL;
  __pyx_pybuffernd_R.rcbuffer = &__pyx_pybuffer_R;
  __pyx_pybuffer_hx_i.pybuffer.buf = NULL;
  __pyx_pybuffer_hx_i.refcount = 0;
  __pyx_pybuffernd_hx_i.data = NULL;
  __pyx_pybuffernd_hx_i.rcbuffer = &__pyx_pybuffer_hx_i;
  __pyx_pybuffer_xhatminus.pybuffer.buf = NULL;
  __pyx_pybuffer_xhatminus.refcount = 0;
  __pyx_pybuffernd_xhatminus.data = NULL;
  __pyx_pybuffernd_xhatminus.rcbuffer = &__pyx_pybuffer_xhatminus;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer, (PyObject*)__pyx_v_xhatminus, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 56, __pyx_L1_error)
  }
  __pyx_pybuffernd_xhatminus.diminfo[0].strides = __pyx_pybuffernd_xhatminus.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xhatminus.diminfo[0].shape = __pyx_pybuffernd_xhatminus.rcbuffer->pybuffer.shape[0];

  /* "flydra_core/_flydra_tracked_object.pyx":68
 *     cdef np.ndarray[np.double_t, ndim=1] hx_i
 * 
 *     ss = xhatminus.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     N = len(pmats_and_points_cov) # number of observations
 */
  __pyx_v_ss = (__pyx_v_xhatminus->dimensions[0]);

  /* "flydra_core/_flydra_tracked_object.pyx":70
 *     ss = xhatminus.shape[0]
 * 
 *     N = len(pmats_and_points_cov) # number of observations             # <<<<<<<<<<<<<<
 *     if N > 0:
 *         missing_data = False
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_pmats_and_points_cov); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_v_N = __pyx_t_1;

  /* "flydra_core/_flydra_tracked_object.pyx":71
 * 
 *     N = len(pmats_and_points_cov) # number of observations
 *     if N > 0:             # <<<<<<<<<<<<<<
 *         missing_data = False
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_N > 0) != 0);
  if (__pyx_t_2) {

    /* "flydra_core/_flydra_tracked_object.pyx":72
 *     N = len(pmats_and_points_cov) # number of observations
 *     if N > 0:
 *         missing_data = False             # <<<<<<<<<<<<<<
 *     else:
 *         missing_data = True
 */
    __pyx_v_missing_data = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":71
 * 
 *     N = len(pmats_and_points_cov) # number of observations
 *     if N > 0:             # <<<<<<<<<<<<<<
 *         missing_data = False
 *     else:
 */
    goto __pyx_L3;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":74
 *         missing_data = False
 *     else:
 *         missing_data = True             # <<<<<<<<<<<<<<
 * 
 *     # Create 2N vector of N observations.
 */
  /*else*/ {
    __pyx_v_missing_data = 1;
  }
  __pyx_L3:;

  /* "flydra_core/_flydra_tracked_object.pyx":77
 * 
 *     # Create 2N vector of N observations.
 *     y = numpy.empty((2*N,), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 * 
 *     # Create 2N x 4 observation model matrix (jacobian of h() at xhatminus).
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_long((2 * __pyx_v_N)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_9 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_y.rcbuffer->pybuffer, (PyObject*)__pyx_v_y, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;
    }
    __pyx_pybuffernd_y.diminfo[0].strides = __pyx_pybuffernd_y.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_y.diminfo[0].shape = __pyx_pybuffernd_y.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 77, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __pyx_v_y = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":80
 * 
 *     # Create 2N x 4 observation model matrix (jacobian of h() at xhatminus).
 *     C = numpy.zeros((2*N,ss), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 * 
 *     # Create 2N vector h(xhatminus) where xhatminus is the a
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_long((2 * __pyx_v_N)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_ss); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 80, __pyx_L1_error)
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_6);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_C.rcbuffer->pybuffer);
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_C.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_9 < 0)) {
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_C.rcbuffer->pybuffer, (PyObject*)__pyx_v_C, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      }
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;
    }
    __pyx_pybuffernd_C.diminfo[0].strides = __pyx_pybuffernd_C.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_C.diminfo[0].shape = __pyx_pybuffernd_C.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_C.diminfo[1].strides = __pyx_pybuffernd_C.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_C.diminfo[1].shape = __pyx_pybuffernd_C.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 80, __pyx_L1_error)
  }
  __pyx_t_13 = 0;
  __pyx_v_C = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":84
 *     # Create 2N vector h(xhatminus) where xhatminus is the a
 *     # priori estimate and h() is the observation model.
 *     hx = numpy.empty((2*N,), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 * 
 *     # Create 2N x 2N observation covariance matrix
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_long((2 * __pyx_v_N)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 84, __pyx_L1_error)
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx.rcbuffer->pybuffer);
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hx.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_9 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hx.rcbuffer->pybuffer, (PyObject*)__pyx_v_hx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;
    }
    __pyx_pybuffernd_hx.diminfo[0].strides = __pyx_pybuffernd_hx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hx.diminfo[0].shape = __pyx_pybuffernd_hx.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 84, __pyx_L1_error)
  }
  __pyx_t_14 = 0;
  __pyx_v_hx = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":87
 * 
 *     # Create 2N x 2N observation covariance matrix
 *     R = numpy.zeros((2*N,2*N), dtype=numpy.float64)             # <<<<<<<<<<<<<<
 * 
 *     # evaluate jacobian for each participating camera
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_long((2 * __pyx_v_N)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_From_long((2 * __pyx_v_N)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_t_15 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_R.rcbuffer->pybuffer);
    __pyx_t_9 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_R.rcbuffer->pybuffer, (PyObject*)__pyx_t_15, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_9 < 0)) {
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_R.rcbuffer->pybuffer, (PyObject*)__pyx_v_R, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      }
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;
    }
    __pyx_pybuffernd_R.diminfo[0].strides = __pyx_pybuffernd_R.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_R.diminfo[0].shape = __pyx_pybuffernd_R.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_R.diminfo[1].strides = __pyx_pybuffernd_R.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_R.diminfo[1].shape = __pyx_pybuffernd_R.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 87, __pyx_L1_error)
  }
  __pyx_t_15 = 0;
  __pyx_v_R = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":90
 * 
 *     # evaluate jacobian for each participating camera
 *     for i,(nonlin_model,xy2d_obs,cov) in enumerate(pmats_and_points_cov):             # <<<<<<<<<<<<<<
 *         pinhole_model = nonlin_model
 * 
 */
  __pyx_t_9 = 0;
  if (likely(PyList_CheckExact(__pyx_v_pmats_and_points_cov)) || PyTuple_CheckExact(__pyx_v_pmats_and_points_cov)) {
    __pyx_t_5 = __pyx_v_pmats_and_points_cov; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
    __pyx_t_16 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_pmats_and_points_cov); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 90, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_16)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_4); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_4); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_16(__pyx_t_5);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 90, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 90, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_17 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_18 = Py_TYPE(__pyx_t_17)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_3 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 2; __pyx_t_7 = __pyx_t_18(__pyx_t_17); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_17), 3) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
      __pyx_t_18 = NULL;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_18 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 90, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_nonlin_model, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_xy2d_obs, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_cov, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_v_i = __pyx_t_9;
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "flydra_core/_flydra_tracked_object.pyx":91
 *     # evaluate jacobian for each participating camera
 *     for i,(nonlin_model,xy2d_obs,cov) in enumerate(pmats_and_points_cov):
 *         pinhole_model = nonlin_model             # <<<<<<<<<<<<<<
 * 
 *         # fill prediction vector [ h(xhatminus) ]
 */
    if (!(likely(((__pyx_v_nonlin_model) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_nonlin_model, __pyx_ptype_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian))))) __PYX_ERR(0, 91, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_nonlin_model;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_pinhole_model, ((struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":94
 * 
 *         # fill prediction vector [ h(xhatminus) ]
 *         hx_i = pinhole_model.evaluate(xhatminus[:3])             # <<<<<<<<<<<<<<
 *         hx[2*i:2*i+2] = hx_i
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_xhatminus), __pyx_slice_); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 94, __pyx_L1_error)
    __pyx_t_7 = ((struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *)__pyx_v_pinhole_model->__pyx_vtab)->evaluate(__pyx_v_pinhole_model, ((PyArrayObject *)__pyx_t_4), 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 94, __pyx_L1_error)
    __pyx_t_19 = ((PyArrayObject *)__pyx_t_7);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx_i.rcbuffer->pybuffer);
      __pyx_t_20 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hx_i.rcbuffer->pybuffer, (PyObject*)__pyx_t_19, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
      if (unlikely(__pyx_t_20 < 0)) {
        PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hx_i.rcbuffer->pybuffer, (PyObject*)__pyx_v_hx_i, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        }
        __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;
      }
      __pyx_pybuffernd_hx_i.diminfo[0].strides = __pyx_pybuffernd_hx_i.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hx_i.diminfo[0].shape = __pyx_pybuffernd_hx_i.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 94, __pyx_L1_error)
    }
    __pyx_t_19 = 0;
    __Pyx_XDECREF_SET(__pyx_v_hx_i, ((PyArrayObject *)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":95
 *         # fill prediction vector [ h(xhatminus) ]
 *         hx_i = pinhole_model.evaluate(xhatminus[:3])
 *         hx[2*i:2*i+2] = hx_i             # <<<<<<<<<<<<<<
 * 
 *         # fill observation vector
 */
    __pyx_t_7 = __Pyx_PyInt_From_long((2 * __pyx_v_i)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyInt_From_long(((2 * __pyx_v_i) + 2)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySlice_New(__pyx_t_7, __pyx_t_4, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_hx), __pyx_t_3, ((PyObject *)__pyx_v_hx_i)) < 0)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":98
 * 
 *         # fill observation vector
 *         y[2*i:2*i+2] = xy2d_obs             # <<<<<<<<<<<<<<
 * 
 *         # fill observation model
 */
    __pyx_t_3 = __Pyx_PyInt_From_long((2 * __pyx_v_i)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_long(((2 * __pyx_v_i) + 2)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PySlice_New(__pyx_t_3, __pyx_t_4, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_y), __pyx_t_7, __pyx_v_xy2d_obs) < 0)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":101
 * 
 *         # fill observation model
 *         C_i = pinhole_model.evaluate_jacobian_at(xhatminus[:3])             # <<<<<<<<<<<<<<
 *         C[2*i:2*i+2,:3] = C_i
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_xhatminus), __pyx_slice_); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian *)__pyx_v_pinhole_model->__pyx_vtab)->evaluate_jacobian_at(__pyx_v_pinhole_model, ((PyArrayObject *)__pyx_t_7), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_C_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":102
 *         # fill observation model
 *         C_i = pinhole_model.evaluate_jacobian_at(xhatminus[:3])
 *         C[2*i:2*i+2,:3] = C_i             # <<<<<<<<<<<<<<
 * 
 *         # fill observation covariance
 */
    __pyx_t_4 = __Pyx_PyInt_From_long((2 * __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyInt_From_long(((2 * __pyx_v_i) + 2)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PySlice_New(__pyx_t_4, __pyx_t_7, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_slice_);
    __Pyx_GIVEREF(__pyx_slice_);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_slice_);
    __pyx_t_3 = 0;
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_C), __pyx_t_7, __pyx_v_C_i) < 0)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":105
 * 
 *         # fill observation covariance
 *         R[2*i:2*i+2,2*i:2*i+2]=cov             # <<<<<<<<<<<<<<
 * 
 *     ## if 1:
 */
    __pyx_t_7 = __Pyx_PyInt_From_long((2 * __pyx_v_i)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyInt_From_long(((2 * __pyx_v_i) + 2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySlice_New(__pyx_t_7, __pyx_t_3, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_long((2 * __pyx_v_i)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_From_long(((2 * __pyx_v_i) + 2)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PySlice_New(__pyx_t_3, __pyx_t_7, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_R), __pyx_t_7, __pyx_v_cov) < 0)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":90
 * 
 *     # evaluate jacobian for each participating camera
 *     for i,(nonlin_model,xy2d_obs,cov) in enumerate(pmats_and_points_cov):             # <<<<<<<<<<<<<<
 *         pinhole_model = nonlin_model
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":116
 *     ##     print 'R'
 *     ##     print R
 *     return y, hx, C, R, missing_data             # <<<<<<<<<<<<<<
 * 
 * def obs2d_hashable( arr ):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_missing_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)__pyx_v_y));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_y));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_y));
  __Pyx_INCREF(((PyObject *)__pyx_v_hx));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_hx));
  PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_v_hx));
  __Pyx_INCREF(((PyObject *)__pyx_v_C));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_C));
  PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)__pyx_v_C));
  __Pyx_INCREF(((PyObject *)__pyx_v_R));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_R));
  PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)__pyx_v_R));
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":56
 *     double sqrt(double)
 * 
 * cpdef evaluate_pmat_jacobian(object pmats_and_points_cov, np.ndarray[np.double_t, ndim=1] xhatminus):             # <<<<<<<<<<<<<<
 *     cdef int N
 *     cdef mybool missing_data
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_C.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_R.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx_i.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.evaluate_pmat_jacobian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_C.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_R.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hx_i.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_y.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pinhole_model);
  __Pyx_XDECREF((PyObject *)__pyx_v_y);
  __Pyx_XDECREF((PyObject *)__pyx_v_C);
  __Pyx_XDECREF((PyObject *)__pyx_v_hx);
  __Pyx_XDECREF((PyObject *)__pyx_v_R);
  __Pyx_XDECREF((PyObject *)__pyx_v_hx_i);
  __Pyx_XDECREF(__pyx_v_nonlin_model);
  __Pyx_XDECREF(__pyx_v_xy2d_obs);
  __Pyx_XDECREF(__pyx_v_cov);
  __Pyx_XDECREF(__pyx_v_C_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_1evaluate_pmat_jacobian(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_1evaluate_pmat_jacobian(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pmats_and_points_cov = 0;
  PyArrayObject *__pyx_v_xhatminus = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("evaluate_pmat_jacobian (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pmats_and_points_cov,&__pyx_n_s_xhatminus,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pmats_and_points_cov)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_xhatminus)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate_pmat_jacobian", 1, 2, 2, 1); __PYX_ERR(0, 56, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "evaluate_pmat_jacobian") < 0)) __PYX_ERR(0, 56, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_pmats_and_points_cov = values[0];
    __pyx_v_xhatminus = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("evaluate_pmat_jacobian", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 56, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.evaluate_pmat_jacobian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xhatminus), __pyx_ptype_5numpy_ndarray, 1, "xhatminus", 0))) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(__pyx_self, __pyx_v_pmats_and_points_cov, __pyx_v_xhatminus);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pmats_and_points_cov, PyArrayObject *__pyx_v_xhatminus) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xhatminus;
  __Pyx_Buffer __pyx_pybuffer_xhatminus;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("evaluate_pmat_jacobian", 0);
  __pyx_pybuffer_xhatminus.pybuffer.buf = NULL;
  __pyx_pybuffer_xhatminus.refcount = 0;
  __pyx_pybuffernd_xhatminus.data = NULL;
  __pyx_pybuffernd_xhatminus.rcbuffer = &__pyx_pybuffer_xhatminus;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer, (PyObject*)__pyx_v_xhatminus, &__Pyx_TypeInfo_nn___pyx_t_5numpy_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 56, __pyx_L1_error)
  }
  __pyx_pybuffernd_xhatminus.diminfo[0].strides = __pyx_pybuffernd_xhatminus.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xhatminus.diminfo[0].shape = __pyx_pybuffernd_xhatminus.rcbuffer->pybuffer.shape[0];
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(__pyx_v_pmats_and_points_cov, __pyx_v_xhatminus, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.evaluate_pmat_jacobian", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xhatminus.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":118
 *     return y, hx, C, R, missing_data
 * 
 * def obs2d_hashable( arr ):             # <<<<<<<<<<<<<<
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_3obs2d_hashable(PyObject *__pyx_self, PyObject *__pyx_v_arr); /*proto*/
static PyMethodDef __pyx_mdef_11flydra_core_22_flydra_tracked_object_3obs2d_hashable = {"obs2d_hashable", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_3obs2d_hashable, METH_O, 0};
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_3obs2d_hashable(PyObject *__pyx_self, PyObject *__pyx_v_arr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("obs2d_hashable (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_2obs2d_hashable(__pyx_self, ((PyObject *)__pyx_v_arr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_2obs2d_hashable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_arr) {
  PyObject *__pyx_v_camns = NULL;
  PyObject *__pyx_v_pt_idx = NULL;
  PyObject *__pyx_v_camn_order = NULL;
  PyObject *__pyx_v_newarr = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("obs2d_hashable", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":120
 * def obs2d_hashable( arr ):
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16             # <<<<<<<<<<<<<<
 *     assert len(arr.shape)==1
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_uint16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 120, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_flydra_tracked_object.pyx":121
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16
 *     assert len(arr.shape)==1             # <<<<<<<<<<<<<<
 * 
 *     # sort array based on camn
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!((__pyx_t_5 == 1) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 121, __pyx_L1_error)
    }
  }
  #endif

  /* "flydra_core/_flydra_tracked_object.pyx":124
 * 
 *     # sort array based on camn
 *     camns = arr[0::2]             # <<<<<<<<<<<<<<
 *     pt_idx = arr[1::2]
 *     camn_order = numpy.argsort(camns)
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_arr, __pyx_slice__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_camns = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":125
 *     # sort array based on camn
 *     camns = arr[0::2]
 *     pt_idx = arr[1::2]             # <<<<<<<<<<<<<<
 *     camn_order = numpy.argsort(camns)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_arr, __pyx_slice__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_pt_idx = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":126
 *     camns = arr[0::2]
 *     pt_idx = arr[1::2]
 *     camn_order = numpy.argsort(camns)             # <<<<<<<<<<<<<<
 * 
 *     # fill new array with sorted values
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_argsort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_camns) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_camns);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_camn_order = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":129
 * 
 *     # fill new array with sorted values
 *     newarr = numpy.zeros( (len(arr),), dtype = numpy.uint16 )             # <<<<<<<<<<<<<<
 *     newarr[ camn_order*2 ] = camns
 *     newarr[ camn_order*2+1 ] = pt_idx
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_v_arr); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_uint16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_newarr = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":130
 *     # fill new array with sorted values
 *     newarr = numpy.zeros( (len(arr),), dtype = numpy.uint16 )
 *     newarr[ camn_order*2 ] = camns             # <<<<<<<<<<<<<<
 *     newarr[ camn_order*2+1 ] = pt_idx
 * 
 */
  __pyx_t_7 = PyNumber_Multiply(__pyx_v_camn_order, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(PyObject_SetItem(__pyx_v_newarr, __pyx_t_7, __pyx_v_camns) < 0)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":131
 *     newarr = numpy.zeros( (len(arr),), dtype = numpy.uint16 )
 *     newarr[ camn_order*2 ] = camns
 *     newarr[ camn_order*2+1 ] = pt_idx             # <<<<<<<<<<<<<<
 * 
 *     val = newarr.tostring()
 */
  __pyx_t_7 = PyNumber_Multiply(__pyx_v_camn_order, __pyx_int_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_newarr, __pyx_t_3, __pyx_v_pt_idx) < 0)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":133
 *     newarr[ camn_order*2+1 ] = pt_idx
 * 
 *     val = newarr.tostring()             # <<<<<<<<<<<<<<
 *     return val
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_newarr, __pyx_n_s_tostring); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_val = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":134
 * 
 *     val = newarr.tostring()
 *     return val             # <<<<<<<<<<<<<<
 * 
 * cdef class TrackedObject:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_val);
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":118
 *     return y, hx, C, R, missing_data
 * 
 * def obs2d_hashable( arr ):             # <<<<<<<<<<<<<<
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.obs2d_hashable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_camns);
  __Pyx_XDECREF(__pyx_v_pt_idx);
  __Pyx_XDECREF(__pyx_v_camn_order);
  __Pyx_XDECREF(__pyx_v_newarr);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":160
 *     cdef object ekf_kalman_A, ekf_kalman_Q
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  reconstructor, # the Reconstructor instance
 *                  obj_id,
 */

/* Python wrapper */
static int __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__[] = "\n\n        arguments\n        =========\n        reconstructor - reconstructor instance\n        obj_id - unique identifier for each object\n        frame - frame number of first observation data\n        obs0_position - first observation (in arbitrary units)\n        kalman_model - Kalman parameters\n        area_threshold - minimum area to consider for tracking use\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__;
#endif
static int __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_reconstructor = 0;
  PyObject *__pyx_v_obj_id = 0;
  long __pyx_v_frame;
  PyObject *__pyx_v_obs0_position = 0;
  PyObject *__pyx_v_obs0_Lcoords = 0;
  PyObject *__pyx_v_first_observation_camns = 0;
  PyObject *__pyx_v_first_observation_idxs = 0;
  PyObject *__pyx_v_kalman_model = 0;
  PyObject *__pyx_v_save_all_data = 0;
  double __pyx_v_area_threshold;
  double __pyx_v_area_threshold_for_orientation;
  PyObject *__pyx_v_disable_image_stat_gating = 0;
  PyObject *__pyx_v_orientation_consensus = 0;
  PyObject *__pyx_v_fake_timestamp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reconstructor,&__pyx_n_s_obj_id,&__pyx_n_s_frame,&__pyx_n_s_obs0_position,&__pyx_n_s_obs0_Lcoords,&__pyx_n_s_first_observation_camns,&__pyx_n_s_first_observation_idxs,&__pyx_n_s_kalman_model,&__pyx_n_s_save_all_data,&__pyx_n_s_area_threshold,&__pyx_n_s_area_threshold_for_orientation,&__pyx_n_s_disable_image_stat_gating,&__pyx_n_s_orientation_consensus,&__pyx_n_s_fake_timestamp,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    /* "flydra_core/_flydra_tracked_object.pyx":168
 *                  first_observation_camns,
 *                  first_observation_idxs,
 *                  kalman_model=None,             # <<<<<<<<<<<<<<
 *                  save_all_data=False,
 *                  double area_threshold=0.0,
 */
    values[7] = ((PyObject *)Py_None);

    /* "flydra_core/_flydra_tracked_object.pyx":169
 *                  first_observation_idxs,
 *                  kalman_model=None,
 *                  save_all_data=False,             # <<<<<<<<<<<<<<
 *                  double area_threshold=0.0,
 *                  double area_threshold_for_orientation=0.0,
 */
    values[8] = ((PyObject *)Py_False);

    /* "flydra_core/_flydra_tracked_object.pyx":172
 *                  double area_threshold=0.0,
 *                  double area_threshold_for_orientation=0.0,
 *                  disable_image_stat_gating=False,             # <<<<<<<<<<<<<<
 *                  orientation_consensus=0,
 *                  fake_timestamp=None
 */
    values[11] = ((PyObject *)Py_False);
    values[12] = ((PyObject *)__pyx_int_0);

    /* "flydra_core/_flydra_tracked_object.pyx":174
 *                  disable_image_stat_gating=False,
 *                  orientation_consensus=0,
 *                  fake_timestamp=None             # <<<<<<<<<<<<<<
 *                  ):
 *         """
 */
    values[13] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reconstructor)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 1); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_frame)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 2); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obs0_position)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 3); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obs0_Lcoords)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 4); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_first_observation_camns)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 5); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_first_observation_idxs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, 6); __PYX_ERR(0, 160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kalman_model);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_save_all_data);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_area_threshold);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_area_threshold_for_orientation);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_disable_image_stat_gating);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_orientation_consensus);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fake_timestamp);
          if (value) { values[13] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_reconstructor = values[0];
    __pyx_v_obj_id = values[1];
    __pyx_v_frame = __Pyx_PyInt_As_long(values[2]); if (unlikely((__pyx_v_frame == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 163, __pyx_L3_error)
    __pyx_v_obs0_position = values[3];
    __pyx_v_obs0_Lcoords = values[4];
    __pyx_v_first_observation_camns = values[5];
    __pyx_v_first_observation_idxs = values[6];
    __pyx_v_kalman_model = values[7];
    __pyx_v_save_all_data = values[8];
    if (values[9]) {
      __pyx_v_area_threshold = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_area_threshold == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L3_error)
    } else {
      __pyx_v_area_threshold = ((double)0.0);
    }
    if (values[10]) {
      __pyx_v_area_threshold_for_orientation = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_area_threshold_for_orientation == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L3_error)
    } else {
      __pyx_v_area_threshold_for_orientation = ((double)0.0);
    }
    __pyx_v_disable_image_stat_gating = values[11];
    __pyx_v_orientation_consensus = values[12];
    __pyx_v_fake_timestamp = values[13];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 7, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), __pyx_v_reconstructor, __pyx_v_obj_id, __pyx_v_frame, __pyx_v_obs0_position, __pyx_v_obs0_Lcoords, __pyx_v_first_observation_camns, __pyx_v_first_observation_idxs, __pyx_v_kalman_model, __pyx_v_save_all_data, __pyx_v_area_threshold, __pyx_v_area_threshold_for_orientation, __pyx_v_disable_image_stat_gating, __pyx_v_orientation_consensus, __pyx_v_fake_timestamp);

  /* "flydra_core/_flydra_tracked_object.pyx":160
 *     cdef object ekf_kalman_A, ekf_kalman_Q
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  reconstructor, # the Reconstructor instance
 *                  obj_id,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_reconstructor, PyObject *__pyx_v_obj_id, long __pyx_v_frame, PyObject *__pyx_v_obs0_position, PyObject *__pyx_v_obs0_Lcoords, PyObject *__pyx_v_first_observation_camns, PyObject *__pyx_v_first_observation_idxs, PyObject *__pyx_v_kalman_model, PyObject *__pyx_v_save_all_data, double __pyx_v_area_threshold, double __pyx_v_area_threshold_for_orientation, PyObject *__pyx_v_disable_image_stat_gating, PyObject *__pyx_v_orientation_consensus, PyObject *__pyx_v_fake_timestamp) {
  PyObject *__pyx_v_PL_line3d = NULL;
  PyObject *__pyx_v_loc = NULL;
  PyObject *__pyx_v_line3d = NULL;
  PyObject *__pyx_v_ss = NULL;
  PyObject *__pyx_v_initial_x = NULL;
  PyObject *__pyx_v_P_k1 = NULL;
  long __pyx_v_i;
  PyObject *__pyx_v_first_observations_2d_pre = NULL;
  PyObject *__pyx_v_first_observations_2d = NULL;
  PyObject *__pyx_v_obs = NULL;
  PyObject *__pyx_v_camn = NULL;
  PyObject *__pyx_v_idx = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned int __pyx_t_1;
  __pyx_t_11flydra_core_22_flydra_tracked_object_mybool __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  long __pyx_t_11;
  double __pyx_t_12;
  PY_LONG_LONG __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_obs0_position);
  __Pyx_INCREF(__pyx_v_obs0_Lcoords);

  /* "flydra_core/_flydra_tracked_object.pyx":187
 *         area_threshold - minimum area to consider for tracking use
 *         """
 *         self.obj_id = obj_id             # <<<<<<<<<<<<<<
 *         self.area_threshold = area_threshold
 *         self.area_threshold_for_orientation = area_threshold_for_orientation
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_obj_id); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 187, __pyx_L1_error)
  __pyx_v_self->obj_id = __pyx_t_1;

  /* "flydra_core/_flydra_tracked_object.pyx":188
 *         """
 *         self.obj_id = obj_id
 *         self.area_threshold = area_threshold             # <<<<<<<<<<<<<<
 *         self.area_threshold_for_orientation = area_threshold_for_orientation
 *         self.save_all_data = save_all_data
 */
  __pyx_v_self->area_threshold = __pyx_v_area_threshold;

  /* "flydra_core/_flydra_tracked_object.pyx":189
 *         self.obj_id = obj_id
 *         self.area_threshold = area_threshold
 *         self.area_threshold_for_orientation = area_threshold_for_orientation             # <<<<<<<<<<<<<<
 *         self.save_all_data = save_all_data
 *         self.kill_me = False
 */
  __pyx_v_self->area_threshold_for_orientation = __pyx_v_area_threshold_for_orientation;

  /* "flydra_core/_flydra_tracked_object.pyx":190
 *         self.area_threshold = area_threshold
 *         self.area_threshold_for_orientation = area_threshold_for_orientation
 *         self.save_all_data = save_all_data             # <<<<<<<<<<<<<<
 *         self.kill_me = False
 *         self.reconstructor = reconstructor
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_save_all_data); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_v_self->save_all_data = __pyx_t_2;

  /* "flydra_core/_flydra_tracked_object.pyx":191
 *         self.area_threshold_for_orientation = area_threshold_for_orientation
 *         self.save_all_data = save_all_data
 *         self.kill_me = False             # <<<<<<<<<<<<<<
 *         self.reconstructor = reconstructor
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)
 */
  __pyx_v_self->kill_me = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":192
 *         self.save_all_data = save_all_data
 *         self.kill_me = False
 *         self.reconstructor = reconstructor             # <<<<<<<<<<<<<<
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)
 *         self.disable_image_stat_gating = disable_image_stat_gating
 */
  __Pyx_INCREF(__pyx_v_reconstructor);
  __Pyx_GIVEREF(__pyx_v_reconstructor);
  __Pyx_GOTREF(__pyx_v_self->reconstructor);
  __Pyx_DECREF(__pyx_v_self->reconstructor);
  __pyx_v_self->reconstructor = __pyx_v_reconstructor;

  /* "flydra_core/_flydra_tracked_object.pyx":193
 *         self.kill_me = False
 *         self.reconstructor = reconstructor
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)             # <<<<<<<<<<<<<<
 *         self.disable_image_stat_gating = disable_image_stat_gating
 *         self.orientation_consensus = orientation_consensus
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_kalman_model, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->distorted_pixel_euclidian_distance_accept);
  __Pyx_DECREF(__pyx_v_self->distorted_pixel_euclidian_distance_accept);
  __pyx_v_self->distorted_pixel_euclidian_distance_accept = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":194
 *         self.reconstructor = reconstructor
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)
 *         self.disable_image_stat_gating = disable_image_stat_gating             # <<<<<<<<<<<<<<
 *         self.orientation_consensus = orientation_consensus
 *         self.fake_timestamp = fake_timestamp
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_disable_image_stat_gating); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_v_self->disable_image_stat_gating = __pyx_t_5;

  /* "flydra_core/_flydra_tracked_object.pyx":195
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)
 *         self.disable_image_stat_gating = disable_image_stat_gating
 *         self.orientation_consensus = orientation_consensus             # <<<<<<<<<<<<<<
 *         self.fake_timestamp = fake_timestamp
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_orientation_consensus); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_v_self->orientation_consensus = __pyx_t_5;

  /* "flydra_core/_flydra_tracked_object.pyx":196
 *         self.disable_image_stat_gating = disable_image_stat_gating
 *         self.orientation_consensus = orientation_consensus
 *         self.fake_timestamp = fake_timestamp             # <<<<<<<<<<<<<<
 * 
 *         self.current_frameno = frame
 */
  __Pyx_INCREF(__pyx_v_fake_timestamp);
  __Pyx_GIVEREF(__pyx_v_fake_timestamp);
  __Pyx_GOTREF(__pyx_v_self->fake_timestamp);
  __Pyx_DECREF(__pyx_v_self->fake_timestamp);
  __pyx_v_self->fake_timestamp = __pyx_v_fake_timestamp;

  /* "flydra_core/_flydra_tracked_object.pyx":198
 *         self.fake_timestamp = fake_timestamp
 * 
 *         self.current_frameno = frame             # <<<<<<<<<<<<<<
 *         self.last_frameno_with_data = frame
 *         obs0_position = obs0_position
 */
  __pyx_v_self->current_frameno = __pyx_v_frame;

  /* "flydra_core/_flydra_tracked_object.pyx":199
 * 
 *         self.current_frameno = frame
 *         self.last_frameno_with_data = frame             # <<<<<<<<<<<<<<
 *         obs0_position = obs0_position
 *         if obs0_Lcoords is None:
 */
  __pyx_v_self->last_frameno_with_data = __pyx_v_frame;

  /* "flydra_core/_flydra_tracked_object.pyx":200
 *         self.current_frameno = frame
 *         self.last_frameno_with_data = frame
 *         obs0_position = obs0_position             # <<<<<<<<<<<<<<
 *         if obs0_Lcoords is None:
 *             obs0_Lcoords = NO_LCOORDS
 */
  __Pyx_INCREF(__pyx_v_obs0_position);
  __Pyx_DECREF_SET(__pyx_v_obs0_position, __pyx_v_obs0_position);

  /* "flydra_core/_flydra_tracked_object.pyx":201
 *         self.last_frameno_with_data = frame
 *         obs0_position = obs0_position
 *         if obs0_Lcoords is None:             # <<<<<<<<<<<<<<
 *             obs0_Lcoords = NO_LCOORDS
 *         else:
 */
  __pyx_t_6 = (__pyx_v_obs0_Lcoords == Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "flydra_core/_flydra_tracked_object.pyx":202
 *         obs0_position = obs0_position
 *         if obs0_Lcoords is None:
 *             obs0_Lcoords = NO_LCOORDS             # <<<<<<<<<<<<<<
 *         else:
 *             PL_line3d = flydra_core.geom.line_from_HZline(obs0_Lcoords) # PlueckerLine instance
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_NO_LCOORDS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_obs0_Lcoords, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":201
 *         self.last_frameno_with_data = frame
 *         obs0_position = obs0_position
 *         if obs0_Lcoords is None:             # <<<<<<<<<<<<<<
 *             obs0_Lcoords = NO_LCOORDS
 *         else:
 */
    goto __pyx_L3;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":204
 *             obs0_Lcoords = NO_LCOORDS
 *         else:
 *             PL_line3d = flydra_core.geom.line_from_HZline(obs0_Lcoords) # PlueckerLine instance             # <<<<<<<<<<<<<<
 *             loc = PL_line3d.closest() # closest point on line to origin
 *             line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_geom); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_line_from_HZline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_v_obs0_Lcoords) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_obs0_Lcoords);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_PL_line3d = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":205
 *         else:
 *             PL_line3d = flydra_core.geom.line_from_HZline(obs0_Lcoords) # PlueckerLine instance
 *             loc = PL_line3d.closest() # closest point on line to origin             # <<<<<<<<<<<<<<
 *             line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )
 *             obs0_Lcoords = line3d.to_hz()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_PL_line3d, __pyx_n_s_closest); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_loc = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":206
 *             PL_line3d = flydra_core.geom.line_from_HZline(obs0_Lcoords) # PlueckerLine instance
 *             loc = PL_line3d.closest() # closest point on line to origin
 *             line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )             # <<<<<<<<<<<<<<
 *             obs0_Lcoords = line3d.to_hz()
 *         ss = kalman_model['ss']
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_geom); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_line_from_points); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_PL_line3d, __pyx_n_s_direction); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Add(__pyx_v_loc, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_loc, __pyx_t_9};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_loc, __pyx_t_9};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_t_9);
      __pyx_t_9 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_line3d = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":207
 *             loc = PL_line3d.closest() # closest point on line to origin
 *             line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )
 *             obs0_Lcoords = line3d.to_hz()             # <<<<<<<<<<<<<<
 *         ss = kalman_model['ss']
 *         initial_x = numpy.zeros( (ss,) )
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_line3d, __pyx_n_s_to_hz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_obs0_Lcoords, __pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "flydra_core/_flydra_tracked_object.pyx":208
 *             line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )
 *             obs0_Lcoords = line3d.to_hz()
 *         ss = kalman_model['ss']             # <<<<<<<<<<<<<<
 *         initial_x = numpy.zeros( (ss,) )
 *         initial_x[:3] = obs0_position
 */
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_ss); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_ss = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":209
 *             obs0_Lcoords = line3d.to_hz()
 *         ss = kalman_model['ss']
 *         initial_x = numpy.zeros( (ss,) )             # <<<<<<<<<<<<<<
 *         initial_x[:3] = obs0_position
 *         P_k1=numpy.eye(ss) # initial state error covariance guess
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_ss);
  __Pyx_GIVEREF(__pyx_v_ss);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_ss);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_initial_x = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":210
 *         ss = kalman_model['ss']
 *         initial_x = numpy.zeros( (ss,) )
 *         initial_x[:3] = obs0_position             # <<<<<<<<<<<<<<
 *         P_k1=numpy.eye(ss) # initial state error covariance guess
 *         for i in range(0,3):
 */
  if (__Pyx_PyObject_SetSlice(__pyx_v_initial_x, __pyx_v_obs0_position, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1) < 0) __PYX_ERR(0, 210, __pyx_L1_error)

  /* "flydra_core/_flydra_tracked_object.pyx":211
 *         initial_x = numpy.zeros( (ss,) )
 *         initial_x[:3] = obs0_position
 *         P_k1=numpy.eye(ss) # initial state error covariance guess             # <<<<<<<<<<<<<<
 *         for i in range(0,3):
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_eye); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_v_ss) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ss);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_P_k1 = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":212
 *         initial_x[:3] = obs0_position
 *         P_k1=numpy.eye(ss) # initial state error covariance guess
 *         for i in range(0,3):             # <<<<<<<<<<<<<<
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
 *         for i in range(3,6):
 */
  for (__pyx_t_11 = 0; __pyx_t_11 < 3; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "flydra_core/_flydra_tracked_object.pyx":213
 *         P_k1=numpy.eye(ss) # initial state error covariance guess
 *         for i in range(0,3):
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']             # <<<<<<<<<<<<<<
 *         for i in range(3,6):
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
 */
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_initial_position_covariance_esti); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_10);
    __pyx_t_3 = 0;
    __pyx_t_10 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_P_k1, __pyx_t_9, __pyx_t_4) < 0)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":214
 *         for i in range(0,3):
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
 *         for i in range(3,6):             # <<<<<<<<<<<<<<
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
 *         if ss >= 9:
 */
  for (__pyx_t_11 = 3; __pyx_t_11 < 6; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "flydra_core/_flydra_tracked_object.pyx":215
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
 *         for i in range(3,6):
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)             # <<<<<<<<<<<<<<
 *         if ss >= 9:
 *             for i in range(6,9):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kalman_model, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_10);
    __pyx_t_4 = 0;
    __pyx_t_10 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_P_k1, __pyx_t_3, __pyx_t_9) < 0)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":216
 *         for i in range(3,6):
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
 *         if ss >= 9:             # <<<<<<<<<<<<<<
 *             for i in range(6,9):
 *                 P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']
 */
  __pyx_t_9 = PyObject_RichCompare(__pyx_v_ss, __pyx_int_9, Py_GE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__pyx_t_7) {

    /* "flydra_core/_flydra_tracked_object.pyx":217
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
 *         if ss >= 9:
 *             for i in range(6,9):             # <<<<<<<<<<<<<<
 *                 P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']
 * 
 */
    for (__pyx_t_11 = 6; __pyx_t_11 < 9; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "flydra_core/_flydra_tracked_object.pyx":218
 *         if ss >= 9:
 *             for i in range(6,9):
 *                 P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']             # <<<<<<<<<<<<<<
 * 
 *         self.max_variance = kalman_model['max_variance_dist_meters']**2 # square so that it is in variance units
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_initial_acceleration_covariance); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_10);
      __pyx_t_3 = 0;
      __pyx_t_10 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_P_k1, __pyx_t_4, __pyx_t_9) < 0)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":216
 *         for i in range(3,6):
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
 *         if ss >= 9:             # <<<<<<<<<<<<<<
 *             for i in range(6,9):
 *                 P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":220
 *                 P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']
 * 
 *         self.max_variance = kalman_model['max_variance_dist_meters']**2 # square so that it is in variance units             # <<<<<<<<<<<<<<
 * 
 *         self.max_frames_skipped = kalman_model['max_frames_skipped']
 */
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_max_variance_dist_meters); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = PyNumber_Power(__pyx_t_9, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->max_variance = __pyx_t_12;

  /* "flydra_core/_flydra_tracked_object.pyx":222
 *         self.max_variance = kalman_model['max_variance_dist_meters']**2 # square so that it is in variance units
 * 
 *         self.max_frames_skipped = kalman_model['max_frames_skipped']             # <<<<<<<<<<<<<<
 * 
 *         if kalman_model.get( 'isEKF', False):
 */
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_max_frames_skipped); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_13 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_4); if (unlikely((__pyx_t_13 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->max_frames_skipped = __pyx_t_13;

  /* "flydra_core/_flydra_tracked_object.pyx":224
 *         self.max_frames_skipped = kalman_model['max_frames_skipped']
 * 
 *         if kalman_model.get( 'isEKF', False):             # <<<<<<<<<<<<<<
 *             self.ekf_kalman_A = kalman_model['A']
 *             self.ekf_kalman_Q = kalman_model['Q']
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kalman_model, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__pyx_t_7) {

    /* "flydra_core/_flydra_tracked_object.pyx":225
 * 
 *         if kalman_model.get( 'isEKF', False):
 *             self.ekf_kalman_A = kalman_model['A']             # <<<<<<<<<<<<<<
 *             self.ekf_kalman_Q = kalman_model['Q']
 *             # EKF
 */
    __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_A); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __Pyx_GOTREF(__pyx_v_self->ekf_kalman_A);
    __Pyx_DECREF(__pyx_v_self->ekf_kalman_A);
    __pyx_v_self->ekf_kalman_A = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":226
 *         if kalman_model.get( 'isEKF', False):
 *             self.ekf_kalman_A = kalman_model['A']
 *             self.ekf_kalman_Q = kalman_model['Q']             # <<<<<<<<<<<<<<
 *             # EKF
 *             self.my_kalman = kalman_ekf.EKF(
 */
    __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_Q); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __Pyx_GOTREF(__pyx_v_self->ekf_kalman_Q);
    __Pyx_DECREF(__pyx_v_self->ekf_kalman_Q);
    __pyx_v_self->ekf_kalman_Q = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":228
 *             self.ekf_kalman_Q = kalman_model['Q']
 *             # EKF
 *             self.my_kalman = kalman_ekf.EKF(             # <<<<<<<<<<<<<<
 *                 initial_x=initial_x,
 *                 initial_P=P_k1,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_kalman_ekf); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_EKF); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":229
 *             # EKF
 *             self.my_kalman = kalman_ekf.EKF(
 *                 initial_x=initial_x,             # <<<<<<<<<<<<<<
 *                 initial_P=P_k1,
 *                 )
 */
    __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_initial_x, __pyx_v_initial_x) < 0) __PYX_ERR(0, 229, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":230
 *             self.my_kalman = kalman_ekf.EKF(
 *                 initial_x=initial_x,
 *                 initial_P=P_k1,             # <<<<<<<<<<<<<<
 *                 )
 *             self.ekf_observation_covariance_pixels = kalman_model['ekf_observation_covariance_pixels']
 */
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_initial_P, __pyx_v_P_k1) < 0) __PYX_ERR(0, 229, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":228
 *             self.ekf_kalman_Q = kalman_model['Q']
 *             # EKF
 *             self.my_kalman = kalman_ekf.EKF(             # <<<<<<<<<<<<<<
 *                 initial_x=initial_x,
 *                 initial_P=P_k1,
 */
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GIVEREF(__pyx_t_10);
    __Pyx_GOTREF(__pyx_v_self->my_kalman);
    __Pyx_DECREF(__pyx_v_self->my_kalman);
    __pyx_v_self->my_kalman = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":232
 *                 initial_P=P_k1,
 *                 )
 *             self.ekf_observation_covariance_pixels = kalman_model['ekf_observation_covariance_pixels']             # <<<<<<<<<<<<<<
 *         else:
 *             # non-EKF
 */
    __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_ekf_observation_covariance_pixel); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_10);
    __Pyx_GOTREF(__pyx_v_self->ekf_observation_covariance_pixels);
    __Pyx_DECREF(__pyx_v_self->ekf_observation_covariance_pixels);
    __pyx_v_self->ekf_observation_covariance_pixels = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":224
 *         self.max_frames_skipped = kalman_model['max_frames_skipped']
 * 
 *         if kalman_model.get( 'isEKF', False):             # <<<<<<<<<<<<<<
 *             self.ekf_kalman_A = kalman_model['A']
 *             self.ekf_kalman_Q = kalman_model['Q']
 */
    goto __pyx_L11;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":235
 *         else:
 *             # non-EKF
 *             self.my_kalman = kalman.KalmanFilter(kalman_model['A'],             # <<<<<<<<<<<<<<
 *                                                  kalman_model['C'],
 *                                                  kalman_model['Q'],
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_kalman); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_KalmanFilter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_A); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "flydra_core/_flydra_tracked_object.pyx":236
 *             # non-EKF
 *             self.my_kalman = kalman.KalmanFilter(kalman_model['A'],
 *                                                  kalman_model['C'],             # <<<<<<<<<<<<<<
 *                                                  kalman_model['Q'],
 *                                                  kalman_model['R'],
 */
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_C); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "flydra_core/_flydra_tracked_object.pyx":237
 *             self.my_kalman = kalman.KalmanFilter(kalman_model['A'],
 *                                                  kalman_model['C'],
 *                                                  kalman_model['Q'],             # <<<<<<<<<<<<<<
 *                                                  kalman_model['R'],
 *                                                  initial_x,
 */
    __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_Q); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "flydra_core/_flydra_tracked_object.pyx":238
 *                                                  kalman_model['C'],
 *                                                  kalman_model['Q'],
 *                                                  kalman_model['R'],             # <<<<<<<<<<<<<<
 *                                                  initial_x,
 *                                                  P_k1)
 */
    __pyx_t_14 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_R); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);

    /* "flydra_core/_flydra_tracked_object.pyx":240
 *                                                  kalman_model['R'],
 *                                                  initial_x,
 *                                                  P_k1)             # <<<<<<<<<<<<<<
 *         self.frames = [frame]
 *         self.xhats = [initial_x]
 */
    __pyx_t_15 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[7] = {__pyx_t_15, __pyx_t_9, __pyx_t_3, __pyx_t_8, __pyx_t_14, __pyx_v_initial_x, __pyx_v_P_k1};
      __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 6+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[7] = {__pyx_t_15, __pyx_t_9, __pyx_t_3, __pyx_t_8, __pyx_t_14, __pyx_v_initial_x, __pyx_v_P_k1};
      __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 6+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(6+__pyx_t_5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_15) {
        __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15); __pyx_t_15 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_5, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_5, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_5, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_14);
      PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_5, __pyx_t_14);
      __Pyx_INCREF(__pyx_v_initial_x);
      __Pyx_GIVEREF(__pyx_v_initial_x);
      PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_5, __pyx_v_initial_x);
      __Pyx_INCREF(__pyx_v_P_k1);
      __Pyx_GIVEREF(__pyx_v_P_k1);
      PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_5, __pyx_v_P_k1);
      __pyx_t_9 = 0;
      __pyx_t_3 = 0;
      __pyx_t_8 = 0;
      __pyx_t_14 = 0;
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_16, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":235
 *         else:
 *             # non-EKF
 *             self.my_kalman = kalman.KalmanFilter(kalman_model['A'],             # <<<<<<<<<<<<<<
 *                                                  kalman_model['C'],
 *                                                  kalman_model['Q'],
 */
    __Pyx_GIVEREF(__pyx_t_10);
    __Pyx_GOTREF(__pyx_v_self->my_kalman);
    __Pyx_DECREF(__pyx_v_self->my_kalman);
    __pyx_v_self->my_kalman = __pyx_t_10;
    __pyx_t_10 = 0;
  }
  __pyx_L11:;

  /* "flydra_core/_flydra_tracked_object.pyx":241
 *                                                  initial_x,
 *                                                  P_k1)
 *         self.frames = [frame]             # <<<<<<<<<<<<<<
 *         self.xhats = [initial_x]
 *         if self.fake_timestamp is None:
 */
  __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_10);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_10);
  __pyx_t_10 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->frames);
  __Pyx_DECREF(__pyx_v_self->frames);
  __pyx_v_self->frames = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":242
 *                                                  P_k1)
 *         self.frames = [frame]
 *         self.xhats = [initial_x]             # <<<<<<<<<<<<<<
 *         if self.fake_timestamp is None:
 *             self.timestamps = [time.time()]
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_initial_x);
  __Pyx_GIVEREF(__pyx_v_initial_x);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_initial_x);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->xhats);
  __Pyx_DECREF(__pyx_v_self->xhats);
  __pyx_v_self->xhats = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":243
 *         self.frames = [frame]
 *         self.xhats = [initial_x]
 *         if self.fake_timestamp is None:             # <<<<<<<<<<<<<<
 *             self.timestamps = [time.time()]
 *         else:
 */
  __pyx_t_7 = (__pyx_v_self->fake_timestamp == Py_None);
  __pyx_t_6 = (__pyx_t_7 != 0);
  if (__pyx_t_6) {

    /* "flydra_core/_flydra_tracked_object.pyx":244
 *         self.xhats = [initial_x]
 *         if self.fake_timestamp is None:
 *             self.timestamps = [time.time()]             # <<<<<<<<<<<<<<
 *         else:
 *             self.timestamps = [self.fake_timestamp]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_time); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_time); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_16);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_16, function);
      }
    }
    __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_16);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = PyList_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_16, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_16);
    __Pyx_GOTREF(__pyx_v_self->timestamps);
    __Pyx_DECREF(__pyx_v_self->timestamps);
    __pyx_v_self->timestamps = __pyx_t_16;
    __pyx_t_16 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":243
 *         self.frames = [frame]
 *         self.xhats = [initial_x]
 *         if self.fake_timestamp is None:             # <<<<<<<<<<<<<<
 *             self.timestamps = [time.time()]
 *         else:
 */
    goto __pyx_L12;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":246
 *             self.timestamps = [time.time()]
 *         else:
 *             self.timestamps = [self.fake_timestamp]             # <<<<<<<<<<<<<<
 *         self.Ps = [P_k1]
 * 
 */
  /*else*/ {
    __pyx_t_16 = PyList_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_v_self->fake_timestamp);
    __Pyx_GIVEREF(__pyx_v_self->fake_timestamp);
    PyList_SET_ITEM(__pyx_t_16, 0, __pyx_v_self->fake_timestamp);
    __Pyx_GIVEREF(__pyx_t_16);
    __Pyx_GOTREF(__pyx_v_self->timestamps);
    __Pyx_DECREF(__pyx_v_self->timestamps);
    __pyx_v_self->timestamps = __pyx_t_16;
    __pyx_t_16 = 0;
  }
  __pyx_L12:;

  /* "flydra_core/_flydra_tracked_object.pyx":247
 *         else:
 *             self.timestamps = [self.fake_timestamp]
 *         self.Ps = [P_k1]             # <<<<<<<<<<<<<<
 * 
 *         self.observations_frames = [frame]
 */
  __pyx_t_16 = PyList_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_v_P_k1);
  __Pyx_GIVEREF(__pyx_v_P_k1);
  PyList_SET_ITEM(__pyx_t_16, 0, __pyx_v_P_k1);
  __Pyx_GIVEREF(__pyx_t_16);
  __Pyx_GOTREF(__pyx_v_self->Ps);
  __Pyx_DECREF(__pyx_v_self->Ps);
  __pyx_v_self->Ps = __pyx_t_16;
  __pyx_t_16 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":249
 *         self.Ps = [P_k1]
 * 
 *         self.observations_frames = [frame]             # <<<<<<<<<<<<<<
 *         self.MLE_position = [obs0_position]
 *         self.MLE_Lcoords = [obs0_Lcoords]
 */
  __pyx_t_16 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_16);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_16);
  __pyx_t_16 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->observations_frames);
  __Pyx_DECREF(__pyx_v_self->observations_frames);
  __pyx_v_self->observations_frames = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":250
 * 
 *         self.observations_frames = [frame]
 *         self.MLE_position = [obs0_position]             # <<<<<<<<<<<<<<
 *         self.MLE_Lcoords = [obs0_Lcoords]
 * 
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_obs0_position);
  __Pyx_GIVEREF(__pyx_v_obs0_position);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_obs0_position);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->MLE_position);
  __Pyx_DECREF(__pyx_v_self->MLE_position);
  __pyx_v_self->MLE_position = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":251
 *         self.observations_frames = [frame]
 *         self.MLE_position = [obs0_position]
 *         self.MLE_Lcoords = [obs0_Lcoords]             # <<<<<<<<<<<<<<
 * 
 *         first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_obs0_Lcoords);
  __Pyx_GIVEREF(__pyx_v_obs0_Lcoords);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_obs0_Lcoords);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->MLE_Lcoords);
  __Pyx_DECREF(__pyx_v_self->MLE_Lcoords);
  __pyx_v_self->MLE_Lcoords = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":253
 *         self.MLE_Lcoords = [obs0_Lcoords]
 * 
 *         first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]             # <<<<<<<<<<<<<<
 *         first_observations_2d = []
 *         for obs in first_observations_2d_pre:
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_v_first_observation_camns);
  __Pyx_GIVEREF(__pyx_v_first_observation_camns);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_first_observation_camns);
  __Pyx_INCREF(__pyx_v_first_observation_idxs);
  __Pyx_GIVEREF(__pyx_v_first_observation_idxs);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_first_observation_idxs);
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_16, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
    __pyx_t_16 = __pyx_t_10; __Pyx_INCREF(__pyx_t_16); __pyx_t_17 = 0;
    __pyx_t_18 = NULL;
  } else {
    __pyx_t_17 = -1; __pyx_t_16 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_18 = Py_TYPE(__pyx_t_16)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 253, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  for (;;) {
    if (likely(!__pyx_t_18)) {
      if (likely(PyList_CheckExact(__pyx_t_16))) {
        if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_16)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_16, __pyx_t_17); __Pyx_INCREF(__pyx_t_10); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_16, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      } else {
        if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_16)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_16, __pyx_t_17); __Pyx_INCREF(__pyx_t_10); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_16, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      }
    } else {
      __pyx_t_10 = __pyx_t_18(__pyx_t_16);
      if (unlikely(!__pyx_t_10)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 253, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_10);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
      PyObject* sequence = __pyx_t_10;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 253, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_14 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_14 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_19 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_14 = __pyx_t_19(__pyx_t_3); if (unlikely(!__pyx_t_14)) goto __pyx_L15_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_14);
      index = 1; __pyx_t_8 = __pyx_t_19(__pyx_t_3); if (unlikely(!__pyx_t_8)) goto __pyx_L15_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_3), 2) < 0) __PYX_ERR(0, 253, __pyx_L1_error)
      __pyx_t_19 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L16_unpacking_done;
      __pyx_L15_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 253, __pyx_L1_error)
      __pyx_L16_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_camn, __pyx_t_14);
    __pyx_t_14 = 0;
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_10 = PyList_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_camn);
    __Pyx_GIVEREF(__pyx_v_camn);
    PyList_SET_ITEM(__pyx_t_10, 0, __pyx_v_camn);
    __Pyx_INCREF(__pyx_v_idx);
    __Pyx_GIVEREF(__pyx_v_idx);
    PyList_SET_ITEM(__pyx_t_10, 1, __pyx_v_idx);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_v_first_observations_2d_pre = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":254
 * 
 *         first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]
 *         first_observations_2d = []             # <<<<<<<<<<<<<<
 *         for obs in first_observations_2d_pre:
 *             first_observations_2d.extend( obs )
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_first_observations_2d = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":255
 *         first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]
 *         first_observations_2d = []
 *         for obs in first_observations_2d_pre:             # <<<<<<<<<<<<<<
 *             first_observations_2d.extend( obs )
 *         first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type
 */
  __pyx_t_4 = __pyx_v_first_observations_2d_pre; __Pyx_INCREF(__pyx_t_4); __pyx_t_17 = 0;
  for (;;) {
    if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_4)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_16 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_17); __Pyx_INCREF(__pyx_t_16); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    #else
    __pyx_t_16 = PySequence_ITEM(__pyx_t_4, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_obs, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":256
 *         first_observations_2d = []
 *         for obs in first_observations_2d_pre:
 *             first_observations_2d.extend( obs )             # <<<<<<<<<<<<<<
 *         first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_first_observations_2d, __pyx_n_s_extend); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_16 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_8, __pyx_v_obs) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_obs);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":255
 *         first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]
 *         first_observations_2d = []
 *         for obs in first_observations_2d_pre:             # <<<<<<<<<<<<<<
 *             first_observations_2d.extend( obs )
 *         first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":257
 *         for obs in first_observations_2d_pre:
 *             first_observations_2d.extend( obs )
 *         first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type             # <<<<<<<<<<<<<<
 * 
 *         self.observations_2d = [first_observations_2d]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_first_observations_2d);
  __Pyx_GIVEREF(__pyx_v_first_observations_2d);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_first_observations_2d);
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_uint16); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_dtype, __pyx_t_14) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_4, __pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v_first_observations_2d, __pyx_t_14);
  __pyx_t_14 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":259
 *         first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type
 * 
 *         self.observations_2d = [first_observations_2d]             # <<<<<<<<<<<<<<
 * 
 *         self.max_frames_skipped=kalman_model['max_frames_skipped']
 */
  __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_v_first_observations_2d);
  __Pyx_GIVEREF(__pyx_v_first_observations_2d);
  PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_first_observations_2d);
  __Pyx_GIVEREF(__pyx_t_14);
  __Pyx_GOTREF(__pyx_v_self->observations_2d);
  __Pyx_DECREF(__pyx_v_self->observations_2d);
  __pyx_v_self->observations_2d = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":261
 *         self.observations_2d = [first_observations_2d]
 * 
 *         self.max_frames_skipped=kalman_model['max_frames_skipped']             # <<<<<<<<<<<<<<
 * 
 *         # Don't run kalman filter with initial data, as this would
 */
  __pyx_t_14 = __Pyx_PyObject_Dict_GetItem(__pyx_v_kalman_model, __pyx_n_s_max_frames_skipped); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_13 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_14); if (unlikely((__pyx_t_13 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_self->max_frames_skipped = __pyx_t_13;

  /* "flydra_core/_flydra_tracked_object.pyx":160
 *     cdef object ekf_kalman_A, ekf_kalman_Q
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  reconstructor, # the Reconstructor instance
 *                  obj_id,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_PL_line3d);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XDECREF(__pyx_v_line3d);
  __Pyx_XDECREF(__pyx_v_ss);
  __Pyx_XDECREF(__pyx_v_initial_x);
  __Pyx_XDECREF(__pyx_v_P_k1);
  __Pyx_XDECREF(__pyx_v_first_observations_2d_pre);
  __Pyx_XDECREF(__pyx_v_first_observations_2d);
  __Pyx_XDECREF(__pyx_v_obs);
  __Pyx_XDECREF(__pyx_v_camn);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_obs0_position);
  __Pyx_XDECREF(__pyx_v_obs0_Lcoords);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":266
 *         # cause error estimates to drop too low.
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2__repr__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2__repr__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":267
 * 
 *     def __repr__(self):
 *         return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (             # <<<<<<<<<<<<<<
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "flydra_core/_flydra_tracked_object.pyx":268
 *     def __repr__(self):
 *         return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (
 *             self.frames[0], self.observations_2d[0], self.xhats[0])             # <<<<<<<<<<<<<<
 * 
 *     def debug_info(self,level=3):
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->frames, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_self->observations_2d, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->xhats, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":267
 * 
 *     def __repr__(self):
 *         return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (             # <<<<<<<<<<<<<<
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_TRO_frames_0_r_observations_2d, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":266
 *         # cause error estimates to drop too low.
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":270
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 * 
 *     def debug_info(self,level=3):             # <<<<<<<<<<<<<<
 *         if level > 5:
 *             sys.stdout.write('%s\n'%self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5debug_info(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5debug_info(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_level = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("debug_info (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_level,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_3);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_level);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "debug_info") < 0)) __PYX_ERR(0, 270, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_level = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("debug_info", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 270, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.debug_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_4debug_info(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), __pyx_v_level);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_4debug_info(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_level) {
  PyObject *__pyx_v_N_pts = NULL;
  PyObject *__pyx_v_start_idx = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_this_Pmean = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  long __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("debug_info", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":271
 * 
 *     def debug_info(self,level=3):
 *         if level > 5:             # <<<<<<<<<<<<<<
 *             sys.stdout.write('%s\n'%self)
 *             N_pts = len(self.xhats)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_level, __pyx_int_5, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "flydra_core/_flydra_tracked_object.pyx":272
 *     def debug_info(self,level=3):
 *         if level > 5:
 *             sys.stdout.write('%s\n'%self)             # <<<<<<<<<<<<<<
 *             N_pts = len(self.xhats)
 *             start_idx = max( N_pts-10, 0 )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":273
 *         if level > 5:
 *             sys.stdout.write('%s\n'%self)
 *             N_pts = len(self.xhats)             # <<<<<<<<<<<<<<
 *             start_idx = max( N_pts-10, 0 )
 *             for i in range(start_idx,N_pts):
 */
    __pyx_t_1 = __pyx_v_self->xhats;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_N_pts = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":274
 *             sys.stdout.write('%s\n'%self)
 *             N_pts = len(self.xhats)
 *             start_idx = max( N_pts-10, 0 )             # <<<<<<<<<<<<<<
 *             for i in range(start_idx,N_pts):
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 */
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_N_pts, __pyx_int_10, 10, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_5 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_start_idx = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":275
 *             N_pts = len(self.xhats)
 *             start_idx = max( N_pts-10, 0 )
 *             for i in range(start_idx,N_pts):             # <<<<<<<<<<<<<<
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_start_idx);
    __Pyx_GIVEREF(__pyx_v_start_idx);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_start_idx);
    __Pyx_INCREF(__pyx_v_N_pts);
    __Pyx_GIVEREF(__pyx_v_N_pts);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_N_pts);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 275, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 275, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 275, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 275, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":276
 *             start_idx = max( N_pts-10, 0 )
 *             for i in range(start_idx,N_pts):
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)             # <<<<<<<<<<<<<<
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 *                 if self.frames[i] in self.observations_frames:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_math); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_self->Ps, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_tuple__7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Power(__pyx_t_9, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_self->Ps, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_tuple__8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Power(__pyx_t_10, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyNumber_Add(__pyx_t_5, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_self->Ps, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_tuple__9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Power(__pyx_t_5, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Add(__pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_this_Pmean, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":277
 *             for i in range(start_idx,N_pts):
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )             # <<<<<<<<<<<<<<
 *                 if self.frames[i] in self.observations_frames:
 *                     j =  self.observations_frames.index(  self.frames[i] )
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_stdout); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_self->frames, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v_self->xhats, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_t_9, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyList_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_kp_s__11);
      __Pyx_GIVEREF(__pyx_kp_s__11);
      PyList_SET_ITEM(__pyx_t_9, 0, __pyx_kp_s__11);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyList_SET_ITEM(__pyx_t_9, 1, __pyx_v_i);
      __Pyx_GIVEREF(__pyx_t_5);
      PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_10);
      PyList_SET_ITEM(__pyx_t_9, 3, __pyx_t_10);
      __Pyx_INCREF(__pyx_v_this_Pmean);
      __Pyx_GIVEREF(__pyx_v_this_Pmean);
      PyList_SET_ITEM(__pyx_t_9, 4, __pyx_v_this_Pmean);
      __pyx_t_5 = 0;
      __pyx_t_10 = 0;
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
      PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyString_Join(__pyx_kp_s__10, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":278
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 *                 if self.frames[i] in self.observations_frames:             # <<<<<<<<<<<<<<
 *                     j =  self.observations_frames.index(  self.frames[i] )
 *                     sys.stdout.write( '%s\n'%self.MLE_position[j] )
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->frames, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_self->observations_frames, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = (__pyx_t_2 != 0);
      if (__pyx_t_11) {

        /* "flydra_core/_flydra_tracked_object.pyx":279
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 *                 if self.frames[i] in self.observations_frames:
 *                     j =  self.observations_frames.index(  self.frames[i] )             # <<<<<<<<<<<<<<
 *                     sys.stdout.write( '%s\n'%self.MLE_position[j] )
 *                 else:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->observations_frames, __pyx_n_s_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_self->frames, __pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":280
 *                 if self.frames[i] in self.observations_frames:
 *                     j =  self.observations_frames.index(  self.frames[i] )
 *                     sys.stdout.write( '%s\n'%self.MLE_position[j] )             # <<<<<<<<<<<<<<
 *                 else:
 *                     sys.stdout.write('\n')
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_stdout); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_self->MLE_position, __pyx_v_j); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":278
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 *                 if self.frames[i] in self.observations_frames:             # <<<<<<<<<<<<<<
 *                     j =  self.observations_frames.index(  self.frames[i] )
 *                     sys.stdout.write( '%s\n'%self.MLE_position[j] )
 */
        goto __pyx_L6;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":282
 *                     sys.stdout.write( '%s\n'%self.MLE_position[j] )
 *                 else:
 *                     sys.stdout.write('\n')             # <<<<<<<<<<<<<<
 *             sys.stdout.write('\n')
 *         elif level > 2:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_stdout); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_kp_s__12) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s__12);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_L6:;

      /* "flydra_core/_flydra_tracked_object.pyx":275
 *             N_pts = len(self.xhats)
 *             start_idx = max( N_pts-10, 0 )
 *             for i in range(start_idx,N_pts):             # <<<<<<<<<<<<<<
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":283
 *                 else:
 *                     sys.stdout.write('\n')
 *             sys.stdout.write('\n')             # <<<<<<<<<<<<<<
 *         elif level > 2:
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s__12) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s__12);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":271
 * 
 *     def debug_info(self,level=3):
 *         if level > 5:             # <<<<<<<<<<<<<<
 *             sys.stdout.write('%s\n'%self)
 *             N_pts = len(self.xhats)
 */
    goto __pyx_L3;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":284
 *                     sys.stdout.write('\n')
 *             sys.stdout.write('\n')
 *         elif level > 2:             # <<<<<<<<<<<<<<
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_level, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "flydra_core/_flydra_tracked_object.pyx":285
 *             sys.stdout.write('\n')
 *         elif level > 2:
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))             # <<<<<<<<<<<<<<
 * 
 *     def get_distance_and_nsigma( self, testx ):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_v_self->xhats;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __pyx_v_self->MLE_position;
    __Pyx_INCREF(__pyx_t_9);
    __pyx_t_6 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_10, 2, ((PyObject *)__pyx_v_self));
    __pyx_t_4 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_d_observations_d_estimates_for, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":284
 *                     sys.stdout.write('\n')
 *             sys.stdout.write('\n')
 *         elif level > 2:             # <<<<<<<<<<<<<<
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))
 * 
 */
  }
  __pyx_L3:;

  /* "flydra_core/_flydra_tracked_object.pyx":270
 *             self.frames[0], self.observations_2d[0], self.xhats[0])
 * 
 *     def debug_info(self,level=3):             # <<<<<<<<<<<<<<
 *         if level > 5:
 *             sys.stdout.write('%s\n'%self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.debug_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_N_pts);
  __Pyx_XDECREF(__pyx_v_start_idx);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_this_Pmean);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":287
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))
 * 
 *     def get_distance_and_nsigma( self, testx ):             # <<<<<<<<<<<<<<
 *         xhat = self.xhats[-1][:3]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7get_distance_and_nsigma(PyObject *__pyx_v_self, PyObject *__pyx_v_testx); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7get_distance_and_nsigma(PyObject *__pyx_v_self, PyObject *__pyx_v_testx) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_distance_and_nsigma (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6get_distance_and_nsigma(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), ((PyObject *)__pyx_v_testx));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6get_distance_and_nsigma(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_testx) {
  PyObject *__pyx_v_xhat = NULL;
  PyObject *__pyx_v_dist2 = NULL;
  PyObject *__pyx_v_dist = NULL;
  PyObject *__pyx_v_P = NULL;
  PyObject *__pyx_v_Pmean = NULL;
  PyObject *__pyx_v_sigma = NULL;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_distance_and_nsigma", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":288
 * 
 *     def get_distance_and_nsigma( self, testx ):
 *         xhat = self.xhats[-1][:3]             # <<<<<<<<<<<<<<
 * 
 *         dist2 = numpy.sum((testx-xhat)**2) # distance squared
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->xhats, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_xhat = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":290
 *         xhat = self.xhats[-1][:3]
 * 
 *         dist2 = numpy.sum((testx-xhat)**2) # distance squared             # <<<<<<<<<<<<<<
 *         dist = numpy.sqrt(dist2)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_testx, __pyx_v_xhat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Power(__pyx_t_1, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_dist2 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":291
 * 
 *         dist2 = numpy.sum((testx-xhat)**2) # distance squared
 *         dist = numpy.sqrt(dist2)             # <<<<<<<<<<<<<<
 * 
 *         # XXX This should be mahalanobis distance, probably.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_dist2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_dist2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_dist = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":294
 * 
 *         # XXX This should be mahalanobis distance, probably.
 *         P = self.Ps[-1]             # <<<<<<<<<<<<<<
 *         Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)])) # sigma squared
 *         sigma = numpy.sqrt(Pmean)
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_self->Ps, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_P = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":295
 *         # XXX This should be mahalanobis distance, probably.
 *         P = self.Ps[-1]
 *         Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)])) # sigma squared             # <<<<<<<<<<<<<<
 *         sigma = numpy.sqrt(Pmean)
 *         return dist, (dist/sigma)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sum); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  for (__pyx_t_6 = 0; __pyx_t_6 < 3; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;
    __pyx_t_7 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8);
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_P, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Power(__pyx_t_8, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_Pmean = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":296
 *         P = self.Ps[-1]
 *         Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)])) # sigma squared
 *         sigma = numpy.sqrt(Pmean)             # <<<<<<<<<<<<<<
 *         return dist, (dist/sigma)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_Pmean) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_Pmean);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sigma = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":297
 *         Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)])) # sigma squared
 *         sigma = numpy.sqrt(Pmean)
 *         return dist, (dist/sigma)             # <<<<<<<<<<<<<<
 * 
 *     def kill(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_v_dist, __pyx_v_sigma); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_dist);
  __Pyx_GIVEREF(__pyx_v_dist);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_dist);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":287
 *             sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))
 * 
 *     def get_distance_and_nsigma( self, testx ):             # <<<<<<<<<<<<<<
 *         xhat = self.xhats[-1][:3]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.get_distance_and_nsigma", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xhat);
  __Pyx_XDECREF(__pyx_v_dist2);
  __Pyx_XDECREF(__pyx_v_dist);
  __Pyx_XDECREF(__pyx_v_P);
  __Pyx_XDECREF(__pyx_v_Pmean);
  __Pyx_XDECREF(__pyx_v_sigma);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":299
 *         return dist, (dist/sigma)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         # called when killed
 *         if self.save_all_data:
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_9kill(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_9kill(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("kill (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_8kill(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_8kill(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_v_last_observation_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("kill", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":301
 *     def kill(self):
 *         # called when killed
 *         if self.save_all_data:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->save_all_data != 0);
  if (__pyx_t_1) {

    /* "flydra_core/_flydra_tracked_object.pyx":302
 *         # called when killed
 *         if self.save_all_data:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # find last data
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "flydra_core/_flydra_tracked_object.pyx":301
 *     def kill(self):
 *         # called when killed
 *         if self.save_all_data:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":305
 * 
 *         # find last data
 *         last_observation_frame = self.observations_frames[-1]             # <<<<<<<<<<<<<<
 * 
 *         # eliminate estimates past last observation
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_self->observations_frames, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_last_observation_frame = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":308
 * 
 *         # eliminate estimates past last observation
 *         while 1:             # <<<<<<<<<<<<<<
 *             if self.frames[-1] > last_observation_frame:
 *                 self.frames.pop()
 */
  while (1) {

    /* "flydra_core/_flydra_tracked_object.pyx":309
 *         # eliminate estimates past last observation
 *         while 1:
 *             if self.frames[-1] > last_observation_frame:             # <<<<<<<<<<<<<<
 *                 self.frames.pop()
 *                 self.xhats.pop()
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_self->frames, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_v_last_observation_frame, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "flydra_core/_flydra_tracked_object.pyx":310
 *         while 1:
 *             if self.frames[-1] > last_observation_frame:
 *                 self.frames.pop()             # <<<<<<<<<<<<<<
 *                 self.xhats.pop()
 *                 self.timestamps.pop()
 */
      __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->frames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":311
 *             if self.frames[-1] > last_observation_frame:
 *                 self.frames.pop()
 *                 self.xhats.pop()             # <<<<<<<<<<<<<<
 *                 self.timestamps.pop()
 *                 self.Ps.pop()
 */
      __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->xhats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":312
 *                 self.frames.pop()
 *                 self.xhats.pop()
 *                 self.timestamps.pop()             # <<<<<<<<<<<<<<
 *                 self.Ps.pop()
 *             else:
 */
      __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->timestamps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":313
 *                 self.xhats.pop()
 *                 self.timestamps.pop()
 *                 self.Ps.pop()             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
      __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->Ps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":309
 *         # eliminate estimates past last observation
 *         while 1:
 *             if self.frames[-1] > last_observation_frame:             # <<<<<<<<<<<<<<
 *                 self.frames.pop()
 *                 self.xhats.pop()
 */
      goto __pyx_L6;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":315
 *                 self.Ps.pop()
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     cpdef calculate_a_posteriori_estimate(self,
 */
    /*else*/ {
      goto __pyx_L5_break;
    }
    __pyx_L6:;
  }
  __pyx_L5_break:;

  /* "flydra_core/_flydra_tracked_object.pyx":299
 *         return dist, (dist/sigma)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         # called when killed
 *         if self.save_all_data:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.kill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_last_observation_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":317
 *                 break
 * 
 *     cpdef calculate_a_posteriori_estimate(self,             # <<<<<<<<<<<<<<
 *                                         long frame,
 *                                         object data_dict,
 */

static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11calculate_a_posteriori_estimate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, long __pyx_v_frame, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate *__pyx_optional_args) {
  int __pyx_v_debug1 = ((int)0);
  int __pyx_v_skip_data_association = ((int)0);

  /* "flydra_core/_flydra_tracked_object.pyx":323
 *                                         int debug1=0,
 *                                         int skip_data_association=0,
 *                                         object original_camns_and_idxs=None,             # <<<<<<<<<<<<<<
 *                                         object original_cam_ids_and_points2d=None,
 *                                         ):
 */
  PyObject *__pyx_v_original_camns_and_idxs = ((PyObject *)Py_None);

  /* "flydra_core/_flydra_tracked_object.pyx":324
 *                                         int skip_data_association=0,
 *                                         object original_camns_and_idxs=None,
 *                                         object original_cam_ids_and_points2d=None,             # <<<<<<<<<<<<<<
 *                                         ):
 *         # Step 1. Update Kalman state to a priori estimates for this frame.
 */
  PyObject *__pyx_v_original_cam_ids_and_points2d = ((PyObject *)Py_None);
  long __pyx_v_i;
  long __pyx_v_frames_since_update;
  long __pyx_v_frames_skipped;
  double __pyx_v_Pmean;
  PyObject *__pyx_v_xhat = NULL;
  PyObject *__pyx_v_P = NULL;
  PyObject *__pyx_v_this_observations_2d_hash = NULL;
  PyObject *__pyx_v_used_camns_and_idxs = NULL;
  PyObject *__pyx_v_all_close_camn_pt_idxs = NULL;
  PyObject *__pyx_v_xhatminus = NULL;
  PyObject *__pyx_v_Pminus = NULL;
  PyObject *__pyx_v_position_MLE = NULL;
  PyObject *__pyx_v_Lcoords = NULL;
  PyObject *__pyx_v_cam_ids_and_points2d = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_prediction_3d = NULL;
  PyObject *__pyx_v_pmats_and_points_cov = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_hx = NULL;
  PyObject *__pyx_v_C = NULL;
  PyObject *__pyx_v_R = NULL;
  PyObject *__pyx_v_missing_data = NULL;
  PyObject *__pyx_v_this_observations_2d = NULL;
  PyObject *__pyx_v_camn = NULL;
  PyObject *__pyx_v_frame_pt_idx = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_dd_idx = NULL;
  PyObject *__pyx_v_cam_id = NULL;
  PyObject *__pyx_v_value_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  long __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  int __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_t_16;
  struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  double __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_a_posteriori_estimate", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_debug1 = __pyx_optional_args->debug1;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_skip_data_association = __pyx_optional_args->skip_data_association;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_original_camns_and_idxs = __pyx_optional_args->original_camns_and_idxs;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_original_cam_ids_and_points2d = __pyx_optional_args->original_cam_ids_and_points2d;
          }
        }
      }
    }
  }

  /* "flydra_core/_flydra_tracked_object.pyx":317
 *                 break
 * 
 *     cpdef calculate_a_posteriori_estimate(self,             # <<<<<<<<<<<<<<
 *                                         long frame,
 *                                         object data_dict,
 */
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_a_posteriori_estimate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11calculate_a_posteriori_estimate)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_debug1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_skip_data_association); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_6 = __pyx_t_1; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[8] = {__pyx_t_7, __pyx_t_3, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_t_4, __pyx_t_5, __pyx_v_original_camns_and_idxs, __pyx_v_original_cam_ids_and_points2d};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 7+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[8] = {__pyx_t_7, __pyx_t_3, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_t_4, __pyx_t_5, __pyx_v_original_camns_and_idxs, __pyx_v_original_cam_ids_and_points2d};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 7+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_9 = PyTuple_New(7+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
          __Pyx_INCREF(__pyx_v_data_dict);
          __Pyx_GIVEREF(__pyx_v_data_dict);
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_data_dict);
          __Pyx_INCREF(__pyx_v_camn2cam_id);
          __Pyx_GIVEREF(__pyx_v_camn2cam_id);
          PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_camn2cam_id);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_t_5);
          __Pyx_INCREF(__pyx_v_original_camns_and_idxs);
          __Pyx_GIVEREF(__pyx_v_original_camns_and_idxs);
          PyTuple_SET_ITEM(__pyx_t_9, 5+__pyx_t_8, __pyx_v_original_camns_and_idxs);
          __Pyx_INCREF(__pyx_v_original_cam_ids_and_points2d);
          __Pyx_GIVEREF(__pyx_v_original_cam_ids_and_points2d);
          PyTuple_SET_ITEM(__pyx_t_9, 6+__pyx_t_8, __pyx_v_original_cam_ids_and_points2d);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_5 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flydra_core/_flydra_tracked_object.pyx":337
 *         cdef long frames_skipped
 *         cdef double Pmean
 *         frames_since_update = frame-self.current_frameno-1             # <<<<<<<<<<<<<<
 * 
 *         if debug1>2:
 */
  __pyx_v_frames_since_update = ((__pyx_v_frame - __pyx_v_self->current_frameno) - 1);

  /* "flydra_core/_flydra_tracked_object.pyx":339
 *         frames_since_update = frame-self.current_frameno-1
 * 
 *         if debug1>2:             # <<<<<<<<<<<<<<
 *             print 'doing',self,'============--'
 *             print 'updating for %d frames since update'%(frames_since_update,)
 */
  __pyx_t_10 = ((__pyx_v_debug1 > 2) != 0);
  if (__pyx_t_10) {

    /* "flydra_core/_flydra_tracked_object.pyx":340
 * 
 *         if debug1>2:
 *             print 'doing',self,'============--'             # <<<<<<<<<<<<<<
 *             print 'updating for %d frames since update'%(frames_since_update,)
 * 
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_doing);
    __Pyx_GIVEREF(__pyx_n_s_doing);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_doing);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
    __Pyx_INCREF(__pyx_kp_s__13);
    __Pyx_GIVEREF(__pyx_kp_s__13);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_s__13);
    if (__Pyx_Print(0, __pyx_t_1, 1) < 0) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":341
 *         if debug1>2:
 *             print 'doing',self,'============--'
 *             print 'updating for %d frames since update'%(frames_since_update,)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(frames_since_update):
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_frames_since_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_updating_for_d_frames_since_upda, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":339
 *         frames_since_update = frame-self.current_frameno-1
 * 
 *         if debug1>2:             # <<<<<<<<<<<<<<
 *             print 'doing',self,'============--'
 *             print 'updating for %d frames since update'%(frames_since_update,)
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":343
 *             print 'updating for %d frames since update'%(frames_since_update,)
 * 
 *         for i in range(frames_since_update):             # <<<<<<<<<<<<<<
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,
 */
  __pyx_t_11 = __pyx_v_frames_since_update;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "flydra_core/_flydra_tracked_object.pyx":344
 * 
 *         for i in range(frames_since_update):
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,
 *                                               self.ekf_kalman_Q)
 */
    __pyx_t_1 = __pyx_v_self->my_kalman;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_kalman_ekf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_EKF); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = PyObject_IsInstance(__pyx_t_1, __pyx_t_6); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 344, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_14 = (__pyx_t_10 != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":345
 *         for i in range(frames_since_update):
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,             # <<<<<<<<<<<<<<
 *                                               self.ekf_kalman_Q)
 *             else:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "flydra_core/_flydra_tracked_object.pyx":346
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,
 *                                               self.ekf_kalman_Q)             # <<<<<<<<<<<<<<
 *             else:
 *                 xhat, P = self.my_kalman.step()
 */
      __pyx_t_2 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self->ekf_kalman_A, __pyx_v_self->ekf_kalman_Q};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_self->ekf_kalman_A, __pyx_v_self->ekf_kalman_Q};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_self->ekf_kalman_A);
        __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_A);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_self->ekf_kalman_A);
        __Pyx_INCREF(__pyx_v_self->ekf_kalman_Q);
        __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_Q);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_self->ekf_kalman_Q);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 345, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_2)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_15(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_9 = __pyx_t_15(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_2), 2) < 0) __PYX_ERR(0, 345, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 345, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":345
 *         for i in range(frames_since_update):
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,             # <<<<<<<<<<<<<<
 *                                               self.ekf_kalman_Q)
 *             else:
 */
      __Pyx_XDECREF_SET(__pyx_v_xhat, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_P, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":344
 * 
 *         for i in range(frames_since_update):
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 xhat, P = self.my_kalman.step(self.ekf_kalman_A,
 *                                               self.ekf_kalman_Q)
 */
      goto __pyx_L6;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":348
 *                                               self.ekf_kalman_Q)
 *             else:
 *                 xhat, P = self.my_kalman.step()             # <<<<<<<<<<<<<<
 *             ############ save outputs ###############
 *             self.frames.append( self.current_frameno + i + 1 )
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 348, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_2)->tp_iternext;
        index = 0; __pyx_t_9 = __pyx_t_15(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        index = 1; __pyx_t_1 = __pyx_t_15(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_2), 2) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 348, __pyx_L1_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_xhat, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_P, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L6:;

    /* "flydra_core/_flydra_tracked_object.pyx":350
 *                 xhat, P = self.my_kalman.step()
 *             ############ save outputs ###############
 *             self.frames.append( self.current_frameno + i + 1 )             # <<<<<<<<<<<<<<
 *             self.xhats.append( xhat )
 *             self.timestamps.append( 0.0 )
 */
    __pyx_t_6 = __Pyx_PyInt_From_long(((__pyx_v_self->current_frameno + __pyx_v_i) + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->frames, __pyx_t_6); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":351
 *             ############ save outputs ###############
 *             self.frames.append( self.current_frameno + i + 1 )
 *             self.xhats.append( xhat )             # <<<<<<<<<<<<<<
 *             self.timestamps.append( 0.0 )
 *             self.Ps.append( P )
 */
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->xhats, __pyx_v_xhat); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 351, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":352
 *             self.frames.append( self.current_frameno + i + 1 )
 *             self.xhats.append( xhat )
 *             self.timestamps.append( 0.0 )             # <<<<<<<<<<<<<<
 *             self.Ps.append( P )
 * 
 */
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->timestamps, __pyx_float_0_0); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 352, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":353
 *             self.xhats.append( xhat )
 *             self.timestamps.append( 0.0 )
 *             self.Ps.append( P )             # <<<<<<<<<<<<<<
 * 
 *         self.current_frameno = frame
 */
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->Ps, __pyx_v_P); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 353, __pyx_L1_error)
  }

  /* "flydra_core/_flydra_tracked_object.pyx":355
 *             self.Ps.append( P )
 * 
 *         self.current_frameno = frame             # <<<<<<<<<<<<<<
 *         this_observations_2d_hash = None
 *         used_camns_and_idxs = []
 */
  __pyx_v_self->current_frameno = __pyx_v_frame;

  /* "flydra_core/_flydra_tracked_object.pyx":356
 * 
 *         self.current_frameno = frame
 *         this_observations_2d_hash = None             # <<<<<<<<<<<<<<
 *         used_camns_and_idxs = []
 *         all_close_camn_pt_idxs = []
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_this_observations_2d_hash = Py_None;

  /* "flydra_core/_flydra_tracked_object.pyx":357
 *         self.current_frameno = frame
 *         this_observations_2d_hash = None
 *         used_camns_and_idxs = []             # <<<<<<<<<<<<<<
 *         all_close_camn_pt_idxs = []
 *         Pmean = c_inf
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_used_camns_and_idxs = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":358
 *         this_observations_2d_hash = None
 *         used_camns_and_idxs = []
 *         all_close_camn_pt_idxs = []             # <<<<<<<<<<<<<<
 *         Pmean = c_inf
 *         if not self.kill_me:
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_all_close_camn_pt_idxs = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":359
 *         used_camns_and_idxs = []
 *         all_close_camn_pt_idxs = []
 *         Pmean = c_inf             # <<<<<<<<<<<<<<
 *         if not self.kill_me:
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 */
  __pyx_v_Pmean = __pyx_v_11flydra_core_22_flydra_tracked_object_c_inf;

  /* "flydra_core/_flydra_tracked_object.pyx":360
 *         all_close_camn_pt_idxs = []
 *         Pmean = c_inf
 *         if not self.kill_me:             # <<<<<<<<<<<<<<
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 */
  __pyx_t_14 = ((!(__pyx_v_self->kill_me != 0)) != 0);
  if (__pyx_t_14) {

    /* "flydra_core/_flydra_tracked_object.pyx":362
 *         if not self.kill_me:
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)
 */
    __pyx_t_6 = __pyx_v_self->my_kalman;
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_kalman_ekf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_EKF); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_14 = PyObject_IsInstance(__pyx_t_6, __pyx_t_9); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = (__pyx_t_14 != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":363
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(             # <<<<<<<<<<<<<<
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)
 *             else:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step1__calculate_a_priori); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "flydra_core/_flydra_tracked_object.pyx":364
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)             # <<<<<<<<<<<<<<
 *             else:
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()
 */
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self->ekf_kalman_A, __pyx_v_self->ekf_kalman_Q};
        __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_self->ekf_kalman_A, __pyx_v_self->ekf_kalman_Q};
        __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      {
        __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_self->ekf_kalman_A);
        __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_A);
        PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_v_self->ekf_kalman_A);
        __Pyx_INCREF(__pyx_v_self->ekf_kalman_Q);
        __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_Q);
        PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_v_self->ekf_kalman_Q);
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
        PyObject* sequence = __pyx_t_9;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 363, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_15(__pyx_t_1); if (unlikely(!__pyx_t_6)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_2 = __pyx_t_15(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_1), 2) < 0) __PYX_ERR(0, 363, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L14_unpacking_done;
        __pyx_L13_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 363, __pyx_L1_error)
        __pyx_L14_unpacking_done:;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":363
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(             # <<<<<<<<<<<<<<
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)
 *             else:
 */
      __pyx_v_xhatminus = __pyx_t_6;
      __pyx_t_6 = 0;
      __pyx_v_Pminus = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":362
 *         if not self.kill_me:
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)
 */
      goto __pyx_L12;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":366
 *                     self.ekf_kalman_A, self.ekf_kalman_Q)
 *             else:
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()             # <<<<<<<<<<<<<<
 *             if debug1>2:
 *                 print 'xhatminus'
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step1__calculate_a_priori); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
        PyObject* sequence = __pyx_t_9;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 366, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_15(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_6 = __pyx_t_15(__pyx_t_1); if (unlikely(!__pyx_t_6)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_1), 2) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L16_unpacking_done;
        __pyx_L15_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 366, __pyx_L1_error)
        __pyx_L16_unpacking_done:;
      }
      __pyx_v_xhatminus = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_Pminus = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L12:;

    /* "flydra_core/_flydra_tracked_object.pyx":367
 *             else:
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'xhatminus'
 *                 print xhatminus
 */
    __pyx_t_10 = ((__pyx_v_debug1 > 2) != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":368
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()
 *             if debug1>2:
 *                 print 'xhatminus'             # <<<<<<<<<<<<<<
 *                 print xhatminus
 *                 print 'Pminus'
 */
      if (__Pyx_PrintOne(0, __pyx_n_s_xhatminus) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":369
 *             if debug1>2:
 *                 print 'xhatminus'
 *                 print xhatminus             # <<<<<<<<<<<<<<
 *                 print 'Pminus'
 *                 print Pminus
 */
      if (__Pyx_PrintOne(0, __pyx_v_xhatminus) < 0) __PYX_ERR(0, 369, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":370
 *                 print 'xhatminus'
 *                 print xhatminus
 *                 print 'Pminus'             # <<<<<<<<<<<<<<
 *                 print Pminus
 *                 print
 */
      if (__Pyx_PrintOne(0, __pyx_n_s_Pminus) < 0) __PYX_ERR(0, 370, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":371
 *                 print xhatminus
 *                 print 'Pminus'
 *                 print Pminus             # <<<<<<<<<<<<<<
 *                 print
 * 
 */
      if (__Pyx_PrintOne(0, __pyx_v_Pminus) < 0) __PYX_ERR(0, 371, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":372
 *                 print 'Pminus'
 *                 print Pminus
 *                 print             # <<<<<<<<<<<<<<
 * 
 *             # Step 2. Filter incoming 2D data to use informative points (data association)
 */
      if (__Pyx_Print(0, __pyx_empty_tuple, 1) < 0) __PYX_ERR(0, 372, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":367
 *             else:
 *                 xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'xhatminus'
 *                 print xhatminus
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":375
 * 
 *             # Step 2. Filter incoming 2D data to use informative points (data association)
 *             if skip_data_association:             # <<<<<<<<<<<<<<
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip
 *                 Lcoords = None
 */
    __pyx_t_10 = (__pyx_v_skip_data_association != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":376
 *             # Step 2. Filter incoming 2D data to use informative points (data association)
 *             if skip_data_association:
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip             # <<<<<<<<<<<<<<
 *                 Lcoords = None
 *                 used_camns_and_idxs = original_camns_and_idxs
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_nan); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_9 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_tuple__14) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_tuple__14);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Multiply(__pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_position_MLE = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":377
 *             if skip_data_association:
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip
 *                 Lcoords = None             # <<<<<<<<<<<<<<
 *                 used_camns_and_idxs = original_camns_and_idxs
 *                 cam_ids_and_points2d = original_cam_ids_and_points2d
 */
      __Pyx_INCREF(Py_None);
      __pyx_v_Lcoords = Py_None;

      /* "flydra_core/_flydra_tracked_object.pyx":378
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip
 *                 Lcoords = None
 *                 used_camns_and_idxs = original_camns_and_idxs             # <<<<<<<<<<<<<<
 *                 cam_ids_and_points2d = original_cam_ids_and_points2d
 *                 all_close_camn_pt_idxs = [] # not important when skipping data association
 */
      __Pyx_INCREF(__pyx_v_original_camns_and_idxs);
      __Pyx_DECREF_SET(__pyx_v_used_camns_and_idxs, __pyx_v_original_camns_and_idxs);

      /* "flydra_core/_flydra_tracked_object.pyx":379
 *                 Lcoords = None
 *                 used_camns_and_idxs = original_camns_and_idxs
 *                 cam_ids_and_points2d = original_cam_ids_and_points2d             # <<<<<<<<<<<<<<
 *                 all_close_camn_pt_idxs = [] # not important when skipping data association
 *             else:
 */
      __Pyx_INCREF(__pyx_v_original_cam_ids_and_points2d);
      __pyx_v_cam_ids_and_points2d = __pyx_v_original_cam_ids_and_points2d;

      /* "flydra_core/_flydra_tracked_object.pyx":380
 *                 used_camns_and_idxs = original_camns_and_idxs
 *                 cam_ids_and_points2d = original_cam_ids_and_points2d
 *                 all_close_camn_pt_idxs = [] # not important when skipping data association             # <<<<<<<<<<<<<<
 *             else:
 *                 (position_MLE, Lcoords, used_camns_and_idxs,
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_all_close_camn_pt_idxs, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":375
 * 
 *             # Step 2. Filter incoming 2D data to use informative points (data association)
 *             if skip_data_association:             # <<<<<<<<<<<<<<
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip
 *                 Lcoords = None
 */
      goto __pyx_L18;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":384
 *                 (position_MLE, Lcoords, used_camns_and_idxs,
 *                  cam_ids_and_points2d,
 *                  all_close_camn_pt_idxs) = self._filter_data(xhatminus, Pminus,             # <<<<<<<<<<<<<<
 *                                                              data_dict,
 *                                                              camn2cam_id,
 */
    /*else*/ {

      /* "flydra_core/_flydra_tracked_object.pyx":387
 *                                                              data_dict,
 *                                                              camn2cam_id,
 *                                                              debug=debug1)             # <<<<<<<<<<<<<<
 *             if debug1>2:
 *                 print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs
 */
      __pyx_t_17.__pyx_n = 1;
      __pyx_t_17.debug = __pyx_v_debug1;
      __pyx_t_1 = ((struct __pyx_vtabstruct_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self->__pyx_vtab)->_filter_data(__pyx_v_self, __pyx_v_xhatminus, __pyx_v_Pminus, __pyx_v_data_dict, __pyx_v_camn2cam_id, 0, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 5)) {
          if (size > 5) __Pyx_RaiseTooManyValuesError(5);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 382, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 4); 
        } else {
          __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 4); 
        }
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[5] = {&__pyx_t_9,&__pyx_t_6,&__pyx_t_2,&__pyx_t_5,&__pyx_t_4};
          for (i=0; i < 5; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 382, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[5] = {&__pyx_t_9,&__pyx_t_6,&__pyx_t_2,&__pyx_t_5,&__pyx_t_4};
        __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_3)->tp_iternext;
        for (index=0; index < 5; index++) {
          PyObject* item = __pyx_t_15(__pyx_t_3); if (unlikely(!item)) goto __pyx_L19_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_3), 5) < 0) __PYX_ERR(0, 382, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L20_unpacking_done;
        __pyx_L19_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 382, __pyx_L1_error)
        __pyx_L20_unpacking_done:;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":382
 *                 all_close_camn_pt_idxs = [] # not important when skipping data association
 *             else:
 *                 (position_MLE, Lcoords, used_camns_and_idxs,             # <<<<<<<<<<<<<<
 *                  cam_ids_and_points2d,
 *                  all_close_camn_pt_idxs) = self._filter_data(xhatminus, Pminus,
 */
      __pyx_v_position_MLE = __pyx_t_9;
      __pyx_t_9 = 0;
      __pyx_v_Lcoords = __pyx_t_6;
      __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_used_camns_and_idxs, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_v_cam_ids_and_points2d = __pyx_t_5;
      __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_all_close_camn_pt_idxs, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L18:;

    /* "flydra_core/_flydra_tracked_object.pyx":388
 *                                                              camn2cam_id,
 *                                                              debug=debug1)
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs
 *                 print 'Lcoords (3D body orientation) : %s'%str(Lcoords)
 */
    __pyx_t_10 = ((__pyx_v_debug1 > 2) != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":389
 *                                                              debug=debug1)
 *             if debug1>2:
 *                 print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs             # <<<<<<<<<<<<<<
 *                 print 'Lcoords (3D body orientation) : %s'%str(Lcoords)
 * 
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_kp_s_position_MLE_used_camns_and_idxs);
      __Pyx_GIVEREF(__pyx_kp_s_position_MLE_used_camns_and_idxs);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_position_MLE_used_camns_and_idxs);
      __Pyx_INCREF(__pyx_v_position_MLE);
      __Pyx_GIVEREF(__pyx_v_position_MLE);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_position_MLE);
      __Pyx_INCREF(__pyx_v_used_camns_and_idxs);
      __Pyx_GIVEREF(__pyx_v_used_camns_and_idxs);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_used_camns_and_idxs);
      if (__Pyx_Print(0, __pyx_t_1, 1) < 0) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":390
 *             if debug1>2:
 *                 print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs
 *                 print 'Lcoords (3D body orientation) : %s'%str(Lcoords)             # <<<<<<<<<<<<<<
 * 
 *             # Step 3. Incorporate observation to estimate a posteriori
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_Lcoords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Lcoords_3D_body_orientation_s, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PrintOne(0, __pyx_t_4) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":388
 *                                                              camn2cam_id,
 *                                                              debug=debug1)
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs
 *                 print 'Lcoords (3D body orientation) : %s'%str(Lcoords)
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":393
 * 
 *             # Step 3. Incorporate observation to estimate a posteriori
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 prediction_3d = xhatminus[:3]
 *                 pmats_and_points_cov = [ (
 */
    __pyx_t_4 = __pyx_v_self->my_kalman;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_kalman_ekf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_EKF); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = PyObject_IsInstance(__pyx_t_4, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_14 = (__pyx_t_10 != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":394
 *             # Step 3. Incorporate observation to estimate a posteriori
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 prediction_3d = xhatminus[:3]             # <<<<<<<<<<<<<<
 *                 pmats_and_points_cov = [ (
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),
 */
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_xhatminus, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_prediction_3d = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":395
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 *                 prediction_3d = xhatminus[:3]
 *                 pmats_and_points_cov = [ (             # <<<<<<<<<<<<<<
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "flydra_core/_flydra_tracked_object.pyx":399
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
 *                                           self.ekf_observation_covariance_pixels)
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]             # <<<<<<<<<<<<<<
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(
 *                     pmats_and_points_cov,xhatminus)
 */
      if (likely(PyList_CheckExact(__pyx_v_cam_ids_and_points2d)) || PyTuple_CheckExact(__pyx_v_cam_ids_and_points2d)) {
        __pyx_t_4 = __pyx_v_cam_ids_and_points2d; __Pyx_INCREF(__pyx_t_4); __pyx_t_18 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_18 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_cam_ids_and_points2d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_19 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 399, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 399, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_19(__pyx_t_4);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 399, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 399, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_15 = Py_TYPE(__pyx_t_9)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_15(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L25_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_6 = __pyx_t_15(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L25_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_9), 2) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
          __pyx_t_15 = NULL;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L26_unpacking_done;
          __pyx_L25_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_15 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 399, __pyx_L1_error)
          __pyx_L26_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_value_tuple, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":396
 *                 prediction_3d = xhatminus[:3]
 *                 pmats_and_points_cov = [ (
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),             # <<<<<<<<<<<<<<
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
 *                                           self.ekf_observation_covariance_pixels)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_get_model_with_jacobian); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_cam_id) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_cam_id);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":397
 *                 pmats_and_points_cov = [ (
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed             # <<<<<<<<<<<<<<
 *                                           self.ekf_observation_covariance_pixels)
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]
 */
        __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_value_tuple, 0, 2, NULL, NULL, &__pyx_slice__15, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "flydra_core/_flydra_tracked_object.pyx":396
 *                 prediction_3d = xhatminus[:3]
 *                 pmats_and_points_cov = [ (
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),             # <<<<<<<<<<<<<<
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
 *                                           self.ekf_observation_covariance_pixels)
 */
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_self->ekf_observation_covariance_pixels);
        __Pyx_GIVEREF(__pyx_v_self->ekf_observation_covariance_pixels);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->ekf_observation_covariance_pixels);
        __pyx_t_1 = 0;
        __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":399
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
 *                                           self.ekf_observation_covariance_pixels)
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]             # <<<<<<<<<<<<<<
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(
 *                     pmats_and_points_cov,xhatminus)
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_pmats_and_points_cov = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":401
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(
 *                     pmats_and_points_cov,xhatminus)             # <<<<<<<<<<<<<<
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 *                     xhatminus, Pminus, y=y,hx=hx,
 */
      if (!(likely(((__pyx_v_xhatminus) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_xhatminus, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 401, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":400
 *                                           self.ekf_observation_covariance_pixels)
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(             # <<<<<<<<<<<<<<
 *                     pmats_and_points_cov,xhatminus)
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 */
      __pyx_t_5 = __pyx_f_11flydra_core_22_flydra_tracked_object_evaluate_pmat_jacobian(__pyx_v_pmats_and_points_cov, ((PyArrayObject *)__pyx_v_xhatminus), 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 5)) {
          if (size > 5) __Pyx_RaiseTooManyValuesError(5);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 400, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 3); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 4); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 3); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 4); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[5] = {&__pyx_t_4,&__pyx_t_2,&__pyx_t_6,&__pyx_t_1,&__pyx_t_9};
          for (i=0; i < 5; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 400, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[5] = {&__pyx_t_4,&__pyx_t_2,&__pyx_t_6,&__pyx_t_1,&__pyx_t_9};
        __pyx_t_3 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_3)->tp_iternext;
        for (index=0; index < 5; index++) {
          PyObject* item = __pyx_t_15(__pyx_t_3); if (unlikely(!item)) goto __pyx_L27_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_3), 5) < 0) __PYX_ERR(0, 400, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L28_unpacking_done;
        __pyx_L27_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 400, __pyx_L1_error)
        __pyx_L28_unpacking_done:;
      }
      __pyx_v_y = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v_hx = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_C = __pyx_t_6;
      __pyx_t_6 = 0;
      __pyx_v_R = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_missing_data = __pyx_t_9;
      __pyx_t_9 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":402
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(
 *                     pmats_and_points_cov,xhatminus)
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(             # <<<<<<<<<<<<<<
 *                     xhatminus, Pminus, y=y,hx=hx,
 *                     C=C,R=R,missing_data=missing_data)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step2__calculate_a_posteriori); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "flydra_core/_flydra_tracked_object.pyx":403
 *                     pmats_and_points_cov,xhatminus)
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 *                     xhatminus, Pminus, y=y,hx=hx,             # <<<<<<<<<<<<<<
 *                     C=C,R=R,missing_data=missing_data)
 *             else:
 */
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_xhatminus);
      __Pyx_GIVEREF(__pyx_v_xhatminus);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_xhatminus);
      __Pyx_INCREF(__pyx_v_Pminus);
      __Pyx_GIVEREF(__pyx_v_Pminus);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_Pminus);
      __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_y, __pyx_v_y) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hx, __pyx_v_hx) < 0) __PYX_ERR(0, 403, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":404
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 *                     xhatminus, Pminus, y=y,hx=hx,
 *                     C=C,R=R,missing_data=missing_data)             # <<<<<<<<<<<<<<
 *             else:
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 */
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_C, __pyx_v_C) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_R, __pyx_v_R) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_missing_data, __pyx_v_missing_data) < 0) __PYX_ERR(0, 403, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":402
 *                 y,hx,C,R,missing_data = evaluate_pmat_jacobian(
 *                     pmats_and_points_cov,xhatminus)
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(             # <<<<<<<<<<<<<<
 *                     xhatminus, Pminus, y=y,hx=hx,
 *                     C=C,R=R,missing_data=missing_data)
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 402, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_5)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_15(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L29_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_9 = __pyx_t_15(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L29_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_5), 2) < 0) __PYX_ERR(0, 402, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L30_unpacking_done;
        __pyx_L29_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 402, __pyx_L1_error)
        __pyx_L30_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_xhat, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_P, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":393
 * 
 *             # Step 3. Incorporate observation to estimate a posteriori
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):             # <<<<<<<<<<<<<<
 *                 prediction_3d = xhatminus[:3]
 *                 pmats_and_points_cov = [ (
 */
      goto __pyx_L22;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":406
 *                     C=C,R=R,missing_data=missing_data)
 *             else:
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(             # <<<<<<<<<<<<<<
 *                     xhatminus, Pminus,
 *                     y=position_MLE)
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_step2__calculate_a_posteriori); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "flydra_core/_flydra_tracked_object.pyx":407
 *             else:
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 *                     xhatminus, Pminus,             # <<<<<<<<<<<<<<
 *                     y=position_MLE)
 * 
 */
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_xhatminus);
      __Pyx_GIVEREF(__pyx_v_xhatminus);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_xhatminus);
      __Pyx_INCREF(__pyx_v_Pminus);
      __Pyx_GIVEREF(__pyx_v_Pminus);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_Pminus);

      /* "flydra_core/_flydra_tracked_object.pyx":408
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(
 *                     xhatminus, Pminus,
 *                     y=position_MLE)             # <<<<<<<<<<<<<<
 * 
 *             # calculate mean variance of x y z position (assumes first three components of state vector are position)
 */
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_y, __pyx_v_position_MLE) < 0) __PYX_ERR(0, 408, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":406
 *                     C=C,R=R,missing_data=missing_data)
 *             else:
 *                 xhat, P = self.my_kalman.step2__calculate_a_posteriori(             # <<<<<<<<<<<<<<
 *                     xhatminus, Pminus,
 *                     y=position_MLE)
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 406, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 406, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_15 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_15(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L31_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_9 = __pyx_t_15(__pyx_t_6); if (unlikely(!__pyx_t_9)) goto __pyx_L31_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_6), 2) < 0) __PYX_ERR(0, 406, __pyx_L1_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L32_unpacking_done;
        __pyx_L31_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 406, __pyx_L1_error)
        __pyx_L32_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_xhat, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_P, __pyx_t_9);
      __pyx_t_9 = 0;
    }
    __pyx_L22:;

    /* "flydra_core/_flydra_tracked_object.pyx":413
 * 
 *             # XXX should probably use trace/N (mean of variances) or determinant (volume of variance)
 *             Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)]))             # <<<<<<<<<<<<<<
 * 
 *             if debug1>2:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_numpy); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    for (__pyx_t_11 = 0; __pyx_t_11 < 3; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;
      __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_P, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_Power(__pyx_t_3, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_9 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_20 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_20 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_Pmean = __pyx_t_20;

    /* "flydra_core/_flydra_tracked_object.pyx":415
 *             Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)]))
 * 
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'xhat'
 *                 print xhat
 */
    __pyx_t_14 = ((__pyx_v_debug1 > 2) != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":416
 * 
 *             if debug1>2:
 *                 print 'xhat'             # <<<<<<<<<<<<<<
 *                 print xhat
 *                 print 'P'
 */
      if (__Pyx_PrintOne(0, __pyx_n_s_xhat) < 0) __PYX_ERR(0, 416, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":417
 *             if debug1>2:
 *                 print 'xhat'
 *                 print xhat             # <<<<<<<<<<<<<<
 *                 print 'P'
 *                 print P
 */
      if (__Pyx_PrintOne(0, __pyx_v_xhat) < 0) __PYX_ERR(0, 417, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":418
 *                 print 'xhat'
 *                 print xhat
 *                 print 'P'             # <<<<<<<<<<<<<<
 *                 print P
 *                 print 'Pmean',Pmean
 */
      if (__Pyx_PrintOne(0, __pyx_n_s_P) < 0) __PYX_ERR(0, 418, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":419
 *                 print xhat
 *                 print 'P'
 *                 print P             # <<<<<<<<<<<<<<
 *                 print 'Pmean',Pmean
 * 
 */
      if (__Pyx_PrintOne(0, __pyx_v_P) < 0) __PYX_ERR(0, 419, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":420
 *                 print 'P'
 *                 print P
 *                 print 'Pmean',Pmean             # <<<<<<<<<<<<<<
 * 
 *             # XXX Need to test if error for this object has grown beyond a
 */
      __pyx_t_5 = PyFloat_FromDouble(__pyx_v_Pmean); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_n_s_Pmean);
      __Pyx_GIVEREF(__pyx_n_s_Pmean);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Pmean);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      if (__Pyx_Print(0, __pyx_t_1, 1) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":415
 *             Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)]))
 * 
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print 'xhat'
 *                 print xhat
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":424
 *             # XXX Need to test if error for this object has grown beyond a
 *             # threshold at which it should be terminated.
 *             if Pmean > self.max_variance:             # <<<<<<<<<<<<<<
 *                 self.kill_me = True
 *                 if debug1>=1:
 */
    __pyx_t_14 = ((__pyx_v_Pmean > __pyx_v_self->max_variance) != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":425
 *             # threshold at which it should be terminated.
 *             if Pmean > self.max_variance:
 *                 self.kill_me = True             # <<<<<<<<<<<<<<
 *                 if debug1>=1:
 *                     print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)
 */
      __pyx_v_self->kill_me = 1;

      /* "flydra_core/_flydra_tracked_object.pyx":426
 *             if Pmean > self.max_variance:
 *                 self.kill_me = True
 *                 if debug1>=1:             # <<<<<<<<<<<<<<
 *                     print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)
 * 
 */
      __pyx_t_14 = ((__pyx_v_debug1 >= 1) != 0);
      if (__pyx_t_14) {

        /* "flydra_core/_flydra_tracked_object.pyx":427
 *                 self.kill_me = True
 *                 if debug1>=1:
 *                     print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)             # <<<<<<<<<<<<<<
 * 
 *             frames_skipped = frame - self.last_frameno_with_data
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_Pmean); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->max_variance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
        __pyx_t_1 = 0;
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_will_kill_next_time_because_Pmea, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__Pyx_PrintOne(0, __pyx_t_5) < 0) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":426
 *             if Pmean > self.max_variance:
 *                 self.kill_me = True
 *                 if debug1>=1:             # <<<<<<<<<<<<<<
 *                     print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)
 * 
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":424
 *             # XXX Need to test if error for this object has grown beyond a
 *             # threshold at which it should be terminated.
 *             if Pmean > self.max_variance:             # <<<<<<<<<<<<<<
 *                 self.kill_me = True
 *                 if debug1>=1:
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":429
 *                     print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)
 * 
 *             frames_skipped = frame - self.last_frameno_with_data             # <<<<<<<<<<<<<<
 *             if frames_skipped > self.max_frames_skipped:
 *                 self.kill_me = True
 */
    __pyx_v_frames_skipped = (__pyx_v_frame - __pyx_v_self->last_frameno_with_data);

    /* "flydra_core/_flydra_tracked_object.pyx":430
 * 
 *             frames_skipped = frame - self.last_frameno_with_data
 *             if frames_skipped > self.max_frames_skipped:             # <<<<<<<<<<<<<<
 *                 self.kill_me = True
 *                 if debug1>=1:
 */
    __pyx_t_14 = ((__pyx_v_frames_skipped > __pyx_v_self->max_frames_skipped) != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":431
 *             frames_skipped = frame - self.last_frameno_with_data
 *             if frames_skipped > self.max_frames_skipped:
 *                 self.kill_me = True             # <<<<<<<<<<<<<<
 *                 if debug1>=1:
 *                     print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(
 */
      __pyx_v_self->kill_me = 1;

      /* "flydra_core/_flydra_tracked_object.pyx":432
 *             if frames_skipped > self.max_frames_skipped:
 *                 self.kill_me = True
 *                 if debug1>=1:             # <<<<<<<<<<<<<<
 *                     print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(
 *                         frames_skipped, self.max_frames_skipped,frame)
 */
      __pyx_t_14 = ((__pyx_v_debug1 >= 1) != 0);
      if (__pyx_t_14) {

        /* "flydra_core/_flydra_tracked_object.pyx":434
 *                 if debug1>=1:
 *                     print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(
 *                         frames_skipped, self.max_frames_skipped,frame)             # <<<<<<<<<<<<<<
 * 
 *             ############ save outputs ###############
 */
        __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_frames_skipped); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_9 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->max_frames_skipped); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
        __pyx_t_5 = 0;
        __pyx_t_9 = 0;
        __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":433
 *                 self.kill_me = True
 *                 if debug1>=1:
 *                     print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(             # <<<<<<<<<<<<<<
 *                         frames_skipped, self.max_frames_skipped,frame)
 * 
 */
        __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_will_kill_next_time_because_fram, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":432
 *             if frames_skipped > self.max_frames_skipped:
 *                 self.kill_me = True
 *                 if debug1>=1:             # <<<<<<<<<<<<<<
 *                     print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(
 *                         frames_skipped, self.max_frames_skipped,frame)
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":430
 * 
 *             frames_skipped = frame - self.last_frameno_with_data
 *             if frames_skipped > self.max_frames_skipped:             # <<<<<<<<<<<<<<
 *                 self.kill_me = True
 *                 if debug1>=1:
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":437
 * 
 *             ############ save outputs ###############
 *             self.frames.append( frame )             # <<<<<<<<<<<<<<
 *             self.xhats.append( xhat )
 *             if self.fake_timestamp is None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->frames, __pyx_t_1); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":438
 *             ############ save outputs ###############
 *             self.frames.append( frame )
 *             self.xhats.append( xhat )             # <<<<<<<<<<<<<<
 *             if self.fake_timestamp is None:
 *                 self.timestamps.append(time.time())
 */
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->xhats, __pyx_v_xhat); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 438, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":439
 *             self.frames.append( frame )
 *             self.xhats.append( xhat )
 *             if self.fake_timestamp is None:             # <<<<<<<<<<<<<<
 *                 self.timestamps.append(time.time())
 *             else:
 */
    __pyx_t_14 = (__pyx_v_self->fake_timestamp == Py_None);
    __pyx_t_10 = (__pyx_t_14 != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":440
 *             self.xhats.append( xhat )
 *             if self.fake_timestamp is None:
 *                 self.timestamps.append(time.time())             # <<<<<<<<<<<<<<
 *             else:
 *                 self.timestamps.append(self.fake_timestamp)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->timestamps, __pyx_t_1); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":439
 *             self.frames.append( frame )
 *             self.xhats.append( xhat )
 *             if self.fake_timestamp is None:             # <<<<<<<<<<<<<<
 *                 self.timestamps.append(time.time())
 *             else:
 */
      goto __pyx_L40;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":442
 *                 self.timestamps.append(time.time())
 *             else:
 *                 self.timestamps.append(self.fake_timestamp)             # <<<<<<<<<<<<<<
 *             self.Ps.append( P )
 * 
 */
    /*else*/ {
      __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->timestamps, __pyx_v_self->fake_timestamp); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 442, __pyx_L1_error)
    }
    __pyx_L40:;

    /* "flydra_core/_flydra_tracked_object.pyx":443
 *             else:
 *                 self.timestamps.append(self.fake_timestamp)
 *             self.Ps.append( P )             # <<<<<<<<<<<<<<
 * 
 *             if position_MLE is not None:
 */
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->Ps, __pyx_v_P); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 443, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":445
 *             self.Ps.append( P )
 * 
 *             if position_MLE is not None:             # <<<<<<<<<<<<<<
 *                 self.last_frameno_with_data = frame
 *                 self.observations_frames.append( frame )
 */
    __pyx_t_10 = (__pyx_v_position_MLE != Py_None);
    __pyx_t_14 = (__pyx_t_10 != 0);
    if (__pyx_t_14) {

      /* "flydra_core/_flydra_tracked_object.pyx":446
 * 
 *             if position_MLE is not None:
 *                 self.last_frameno_with_data = frame             # <<<<<<<<<<<<<<
 *                 self.observations_frames.append( frame )
 *                 self.MLE_position.append( position_MLE )
 */
      __pyx_v_self->last_frameno_with_data = __pyx_v_frame;

      /* "flydra_core/_flydra_tracked_object.pyx":447
 *             if position_MLE is not None:
 *                 self.last_frameno_with_data = frame
 *                 self.observations_frames.append( frame )             # <<<<<<<<<<<<<<
 *                 self.MLE_position.append( position_MLE )
 *                 if Lcoords is None:
 */
      __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_frame); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->observations_frames, __pyx_t_1); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":448
 *                 self.last_frameno_with_data = frame
 *                 self.observations_frames.append( frame )
 *                 self.MLE_position.append( position_MLE )             # <<<<<<<<<<<<<<
 *                 if Lcoords is None:
 *                     self.MLE_Lcoords.append( NO_LCOORDS )
 */
      __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->MLE_position, __pyx_v_position_MLE); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 448, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":449
 *                 self.observations_frames.append( frame )
 *                 self.MLE_position.append( position_MLE )
 *                 if Lcoords is None:             # <<<<<<<<<<<<<<
 *                     self.MLE_Lcoords.append( NO_LCOORDS )
 *                 else:
 */
      __pyx_t_14 = (__pyx_v_Lcoords == Py_None);
      __pyx_t_10 = (__pyx_t_14 != 0);
      if (__pyx_t_10) {

        /* "flydra_core/_flydra_tracked_object.pyx":450
 *                 self.MLE_position.append( position_MLE )
 *                 if Lcoords is None:
 *                     self.MLE_Lcoords.append( NO_LCOORDS )             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.MLE_Lcoords.append( Lcoords )
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_LCOORDS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->MLE_Lcoords, __pyx_t_1); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 450, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":449
 *                 self.observations_frames.append( frame )
 *                 self.MLE_position.append( position_MLE )
 *                 if Lcoords is None:             # <<<<<<<<<<<<<<
 *                     self.MLE_Lcoords.append( NO_LCOORDS )
 *                 else:
 */
        goto __pyx_L42;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":452
 *                     self.MLE_Lcoords.append( NO_LCOORDS )
 *                 else:
 *                     self.MLE_Lcoords.append( Lcoords )             # <<<<<<<<<<<<<<
 *                 this_observations_2d = []
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:
 */
      /*else*/ {
        __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->MLE_Lcoords, __pyx_v_Lcoords); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 452, __pyx_L1_error)
      }
      __pyx_L42:;

      /* "flydra_core/_flydra_tracked_object.pyx":453
 *                 else:
 *                     self.MLE_Lcoords.append( Lcoords )
 *                 this_observations_2d = []             # <<<<<<<<<<<<<<
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_this_observations_2d = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":454
 *                     self.MLE_Lcoords.append( Lcoords )
 *                 this_observations_2d = []
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:             # <<<<<<<<<<<<<<
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
 */
      if (likely(PyList_CheckExact(__pyx_v_used_camns_and_idxs)) || PyTuple_CheckExact(__pyx_v_used_camns_and_idxs)) {
        __pyx_t_1 = __pyx_v_used_camns_and_idxs; __Pyx_INCREF(__pyx_t_1); __pyx_t_18 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_18 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_used_camns_and_idxs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 454, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_18); __Pyx_INCREF(__pyx_t_9); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 454, __pyx_L1_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 454, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          } else {
            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_18); __Pyx_INCREF(__pyx_t_9); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 454, __pyx_L1_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 454, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          }
        } else {
          __pyx_t_9 = __pyx_t_19(__pyx_t_1);
          if (unlikely(!__pyx_t_9)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 454, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_9);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
          PyObject* sequence = __pyx_t_9;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 454, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_7 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_15 = Py_TYPE(__pyx_t_7)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_15(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L45_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_5 = __pyx_t_15(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L45_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          index = 2; __pyx_t_6 = __pyx_t_15(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L45_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_7), 3) < 0) __PYX_ERR(0, 454, __pyx_L1_error)
          __pyx_t_15 = NULL;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L46_unpacking_done;
          __pyx_L45_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_15 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 454, __pyx_L1_error)
          __pyx_L46_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_camn, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_frame_pt_idx, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_dd_idx, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":455
 *                 this_observations_2d = []
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )             # <<<<<<<<<<<<<<
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
 *                 self.observations_2d.append( this_observations_2d )
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_this_observations_2d, __pyx_n_s_extend); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_camn);
        __Pyx_GIVEREF(__pyx_v_camn);
        PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_camn);
        __Pyx_INCREF(__pyx_v_frame_pt_idx);
        __Pyx_GIVEREF(__pyx_v_frame_pt_idx);
        PyList_SET_ITEM(__pyx_t_5, 1, __pyx_v_frame_pt_idx);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_9 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":454
 *                     self.MLE_Lcoords.append( Lcoords )
 *                 this_observations_2d = []
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:             # <<<<<<<<<<<<<<
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":456
 *                 for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy             # <<<<<<<<<<<<<<
 *                 self.observations_2d.append( this_observations_2d )
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_this_observations_2d);
      __Pyx_GIVEREF(__pyx_v_this_observations_2d);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_this_observations_2d);
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_uint16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_this_observations_2d, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":457
 *                     this_observations_2d.extend( [camn,frame_pt_idx] )
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
 *                 self.observations_2d.append( this_observations_2d )             # <<<<<<<<<<<<<<
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )
 *             if debug1>2:
 */
      __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_self->observations_2d, __pyx_v_this_observations_2d); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 457, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":458
 *                 this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
 *                 self.observations_2d.append( this_observations_2d )
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )             # <<<<<<<<<<<<<<
 *             if debug1>2:
 *                 print
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_obs2d_hashable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_v_this_observations_2d) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_this_observations_2d);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_this_observations_2d_hash, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":445
 *             self.Ps.append( P )
 * 
 *             if position_MLE is not None:             # <<<<<<<<<<<<<<
 *                 self.last_frameno_with_data = frame
 *                 self.observations_frames.append( frame )
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":459
 *                 self.observations_2d.append( this_observations_2d )
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print
 *         return (used_camns_and_idxs, this_observations_2d_hash,
 */
    __pyx_t_10 = ((__pyx_v_debug1 > 2) != 0);
    if (__pyx_t_10) {

      /* "flydra_core/_flydra_tracked_object.pyx":460
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )
 *             if debug1>2:
 *                 print             # <<<<<<<<<<<<<<
 *         return (used_camns_and_idxs, this_observations_2d_hash,
 *                 Pmean, all_close_camn_pt_idxs)
 */
      if (__Pyx_Print(0, __pyx_empty_tuple, 1) < 0) __PYX_ERR(0, 460, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":459
 *                 self.observations_2d.append( this_observations_2d )
 *                 this_observations_2d_hash = obs2d_hashable( this_observations_2d )
 *             if debug1>2:             # <<<<<<<<<<<<<<
 *                 print
 *         return (used_camns_and_idxs, this_observations_2d_hash,
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":360
 *         all_close_camn_pt_idxs = []
 *         Pmean = c_inf
 *         if not self.kill_me:             # <<<<<<<<<<<<<<
 *             # Step 1.B. Update Kalman to provide a priori estimates for this frame
 *             if isinstance(self.my_kalman, kalman_ekf.EKF):
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":461
 *             if debug1>2:
 *                 print
 *         return (used_camns_and_idxs, this_observations_2d_hash,             # <<<<<<<<<<<<<<
 *                 Pmean, all_close_camn_pt_idxs)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "flydra_core/_flydra_tracked_object.pyx":462
 *                 print
 *         return (used_camns_and_idxs, this_observations_2d_hash,
 *                 Pmean, all_close_camn_pt_idxs)             # <<<<<<<<<<<<<<
 * 
 *     def remove_previous_observation(self, int debug1=0):
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_Pmean); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "flydra_core/_flydra_tracked_object.pyx":461
 *             if debug1>2:
 *                 print
 *         return (used_camns_and_idxs, this_observations_2d_hash,             # <<<<<<<<<<<<<<
 *                 Pmean, all_close_camn_pt_idxs)
 * 
 */
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_used_camns_and_idxs);
  __Pyx_GIVEREF(__pyx_v_used_camns_and_idxs);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_used_camns_and_idxs);
  __Pyx_INCREF(__pyx_v_this_observations_2d_hash);
  __Pyx_GIVEREF(__pyx_v_this_observations_2d_hash);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_this_observations_2d_hash);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_all_close_camn_pt_idxs);
  __Pyx_GIVEREF(__pyx_v_all_close_camn_pt_idxs);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_all_close_camn_pt_idxs);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":317
 *                 break
 * 
 *     cpdef calculate_a_posteriori_estimate(self,             # <<<<<<<<<<<<<<
 *                                         long frame,
 *                                         object data_dict,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.calculate_a_posteriori_estimate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xhat);
  __Pyx_XDECREF(__pyx_v_P);
  __Pyx_XDECREF(__pyx_v_this_observations_2d_hash);
  __Pyx_XDECREF(__pyx_v_used_camns_and_idxs);
  __Pyx_XDECREF(__pyx_v_all_close_camn_pt_idxs);
  __Pyx_XDECREF(__pyx_v_xhatminus);
  __Pyx_XDECREF(__pyx_v_Pminus);
  __Pyx_XDECREF(__pyx_v_position_MLE);
  __Pyx_XDECREF(__pyx_v_Lcoords);
  __Pyx_XDECREF(__pyx_v_cam_ids_and_points2d);
  __Pyx_XDECREF(__pyx_v_prediction_3d);
  __Pyx_XDECREF(__pyx_v_pmats_and_points_cov);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_hx);
  __Pyx_XDECREF(__pyx_v_C);
  __Pyx_XDECREF(__pyx_v_R);
  __Pyx_XDECREF(__pyx_v_missing_data);
  __Pyx_XDECREF(__pyx_v_this_observations_2d);
  __Pyx_XDECREF(__pyx_v_camn);
  __Pyx_XDECREF(__pyx_v_frame_pt_idx);
  __Pyx_XDECREF(__pyx_v_dd_idx);
  __Pyx_XDECREF(__pyx_v_cam_id);
  __Pyx_XDECREF(__pyx_v_value_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11calculate_a_posteriori_estimate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11calculate_a_posteriori_estimate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  long __pyx_v_frame;
  PyObject *__pyx_v_data_dict = 0;
  PyObject *__pyx_v_camn2cam_id = 0;
  int __pyx_v_debug1;
  int __pyx_v_skip_data_association;
  PyObject *__pyx_v_original_camns_and_idxs = 0;
  PyObject *__pyx_v_original_cam_ids_and_points2d = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_a_posteriori_estimate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_frame,&__pyx_n_s_data_dict,&__pyx_n_s_camn2cam_id,&__pyx_n_s_debug1,&__pyx_n_s_skip_data_association,&__pyx_n_s_original_camns_and_idxs,&__pyx_n_s_original_cam_ids_and_points2d,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "flydra_core/_flydra_tracked_object.pyx":323
 *                                         int debug1=0,
 *                                         int skip_data_association=0,
 *                                         object original_camns_and_idxs=None,             # <<<<<<<<<<<<<<
 *                                         object original_cam_ids_and_points2d=None,
 *                                         ):
 */
    values[5] = ((PyObject *)Py_None);

    /* "flydra_core/_flydra_tracked_object.pyx":324
 *                                         int skip_data_association=0,
 *                                         object original_camns_and_idxs=None,
 *                                         object original_cam_ids_and_points2d=None,             # <<<<<<<<<<<<<<
 *                                         ):
 *         # Step 1. Update Kalman state to a priori estimates for this frame.
 */
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_frame)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calculate_a_posteriori_estimate", 0, 3, 7, 1); __PYX_ERR(0, 317, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_camn2cam_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calculate_a_posteriori_estimate", 0, 3, 7, 2); __PYX_ERR(0, 317, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_debug1);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_skip_data_association);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_original_camns_and_idxs);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_original_cam_ids_and_points2d);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calculate_a_posteriori_estimate") < 0)) __PYX_ERR(0, 317, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_frame = __Pyx_PyInt_As_long(values[0]); if (unlikely((__pyx_v_frame == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L3_error)
    __pyx_v_data_dict = values[1];
    __pyx_v_camn2cam_id = values[2];
    if (values[3]) {
      __pyx_v_debug1 = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_debug1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 321, __pyx_L3_error)
    } else {
      __pyx_v_debug1 = ((int)0);
    }
    if (values[4]) {
      __pyx_v_skip_data_association = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_skip_data_association == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    } else {
      __pyx_v_skip_data_association = ((int)0);
    }
    __pyx_v_original_camns_and_idxs = values[5];
    __pyx_v_original_cam_ids_and_points2d = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calculate_a_posteriori_estimate", 0, 3, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 317, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.calculate_a_posteriori_estimate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10calculate_a_posteriori_estimate(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), __pyx_v_frame, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_v_debug1, __pyx_v_skip_data_association, __pyx_v_original_camns_and_idxs, __pyx_v_original_cam_ids_and_points2d);

  /* "flydra_core/_flydra_tracked_object.pyx":317
 *                 break
 * 
 *     cpdef calculate_a_posteriori_estimate(self,             # <<<<<<<<<<<<<<
 *                                         long frame,
 *                                         object data_dict,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10calculate_a_posteriori_estimate(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, long __pyx_v_frame, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_v_debug1, int __pyx_v_skip_data_association, PyObject *__pyx_v_original_camns_and_idxs, PyObject *__pyx_v_original_cam_ids_and_points2d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_a_posteriori_estimate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 4;
  __pyx_t_2.debug1 = __pyx_v_debug1;
  __pyx_t_2.skip_data_association = __pyx_v_skip_data_association;
  __pyx_t_2.original_camns_and_idxs = __pyx_v_original_camns_and_idxs;
  __pyx_t_2.original_cam_ids_and_points2d = __pyx_v_original_cam_ids_and_points2d;
  __pyx_t_1 = __pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject->calculate_a_posteriori_estimate(__pyx_v_self, __pyx_v_frame, __pyx_v_data_dict, __pyx_v_camn2cam_id, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.calculate_a_posteriori_estimate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":464
 *                 Pmean, all_close_camn_pt_idxs)
 * 
 *     def remove_previous_observation(self, int debug1=0):             # <<<<<<<<<<<<<<
 * 
 *         # This will remove the just-done observation from my
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_13remove_previous_observation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_13remove_previous_observation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_debug1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_previous_observation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_debug1,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_debug1);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "remove_previous_observation") < 0)) __PYX_ERR(0, 464, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_debug1 = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_debug1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    } else {
      __pyx_v_debug1 = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_previous_observation", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.remove_previous_observation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12remove_previous_observation(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), __pyx_v_debug1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12remove_previous_observation(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, int __pyx_v_debug1) {
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_previous_observation", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":470
 *         # one observation.
 * 
 *         self.current_frameno -= 1             # <<<<<<<<<<<<<<
 * 
 *         # Remove most recent information
 */
  __pyx_v_self->current_frameno = (__pyx_v_self->current_frameno - 1);

  /* "flydra_core/_flydra_tracked_object.pyx":473
 * 
 *         # Remove most recent information
 *         frame = self.frames.pop()             # <<<<<<<<<<<<<<
 *         if debug1>2:
 *             sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))
 */
  __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->frames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_frame = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":474
 *         # Remove most recent information
 *         frame = self.frames.pop()
 *         if debug1>2:             # <<<<<<<<<<<<<<
 *             sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))
 * 
 */
  __pyx_t_2 = ((__pyx_v_debug1 > 2) != 0);
  if (__pyx_t_2) {

    /* "flydra_core/_flydra_tracked_object.pyx":475
 *         frame = self.frames.pop()
 *         if debug1>2:
 *             sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))             # <<<<<<<<<<<<<<
 * 
 *         self.xhats.pop()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self));
    __Pyx_INCREF(__pyx_v_frame);
    __Pyx_GIVEREF(__pyx_v_frame);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_frame);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_removing_previous_observation, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":474
 *         # Remove most recent information
 *         frame = self.frames.pop()
 *         if debug1>2:             # <<<<<<<<<<<<<<
 *             sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))
 * 
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":477
 *             sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))
 * 
 *         self.xhats.pop()             # <<<<<<<<<<<<<<
 *         self.timestamps.pop()
 *         self.Ps.pop()
 */
  __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->xhats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":478
 * 
 *         self.xhats.pop()
 *         self.timestamps.pop()             # <<<<<<<<<<<<<<
 *         self.Ps.pop()
 *         if self.observations_frames[-1] == frame:
 */
  __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->timestamps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":479
 *         self.xhats.pop()
 *         self.timestamps.pop()
 *         self.Ps.pop()             # <<<<<<<<<<<<<<
 *         if self.observations_frames[-1] == frame:
 *             self.observations_frames.pop()
 */
  __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->Ps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":480
 *         self.timestamps.pop()
 *         self.Ps.pop()
 *         if self.observations_frames[-1] == frame:             # <<<<<<<<<<<<<<
 *             self.observations_frames.pop()
 *             self.MLE_position.pop()
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->observations_frames, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_frame, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "flydra_core/_flydra_tracked_object.pyx":481
 *         self.Ps.pop()
 *         if self.observations_frames[-1] == frame:
 *             self.observations_frames.pop()             # <<<<<<<<<<<<<<
 *             self.MLE_position.pop()
 *             self.MLE_Lcoords.pop()
 */
    __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->observations_frames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":482
 *         if self.observations_frames[-1] == frame:
 *             self.observations_frames.pop()
 *             self.MLE_position.pop()             # <<<<<<<<<<<<<<
 *             self.MLE_Lcoords.pop()
 *             self.observations_2d.pop()
 */
    __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->MLE_position); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":483
 *             self.observations_frames.pop()
 *             self.MLE_position.pop()
 *             self.MLE_Lcoords.pop()             # <<<<<<<<<<<<<<
 *             self.observations_2d.pop()
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->MLE_Lcoords); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":484
 *             self.MLE_position.pop()
 *             self.MLE_Lcoords.pop()
 *             self.observations_2d.pop()             # <<<<<<<<<<<<<<
 * 
 *         # reset self.my_kalman
 */
    __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->observations_2d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":480
 *         self.timestamps.pop()
 *         self.Ps.pop()
 *         if self.observations_frames[-1] == frame:             # <<<<<<<<<<<<<<
 *             self.observations_frames.pop()
 *             self.MLE_position.pop()
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":487
 * 
 *         # reset self.my_kalman
 *         self.my_kalman.xhat_k1 = self.xhats[-1]             # <<<<<<<<<<<<<<
 *         self.my_kalman.P_k1 = self.Ps[-1]
 * 
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->xhats, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_xhat_k1, __pyx_t_3) < 0) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":488
 *         # reset self.my_kalman
 *         self.my_kalman.xhat_k1 = self.xhats[-1]
 *         self.my_kalman.P_k1 = self.Ps[-1]             # <<<<<<<<<<<<<<
 * 
 *         self.last_frameno_with_data = self.observations_frames[-1]
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->Ps, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->my_kalman, __pyx_n_s_P_k1, __pyx_t_3) < 0) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":490
 *         self.my_kalman.P_k1 = self.Ps[-1]
 * 
 *         self.last_frameno_with_data = self.observations_frames[-1]             # <<<<<<<<<<<<<<
 * 
 *     cpdef _filter_data(self, object xhatminus, object Pminus,
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->observations_frames, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_6 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->last_frameno_with_data = __pyx_t_6;

  /* "flydra_core/_flydra_tracked_object.pyx":464
 *                 Pmean, all_close_camn_pt_idxs)
 * 
 *     def remove_previous_observation(self, int debug1=0):             # <<<<<<<<<<<<<<
 * 
 *         # This will remove the just-done observation from my
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.remove_previous_observation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":492
 *         self.last_frameno_with_data = self.observations_frames[-1]
 * 
 *     cpdef _filter_data(self, object xhatminus, object Pminus,             # <<<<<<<<<<<<<<
 *                        object data_dict, object camn2cam_id,
 *                        int debug=0):
 */

static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15_filter_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_xhatminus, PyObject *__pyx_v_Pminus, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data *__pyx_optional_args) {
  int __pyx_v_debug = ((int)0);
  double __pyx_v_least_nll;
  double __pyx_v_nll_this_point;
  double __pyx_v_dist2;
  double __pyx_v_dist;
  double __pyx_v_p_y_x;
  int __pyx_v_gated_in;
  int __pyx_v_pixel_dist_criterion_passed;
  double __pyx_v_pt_area;
  double __pyx_v_mean_val;
  double __pyx_v_sumsqf_val;
  double __pyx_v_area;
  int __pyx_v_cur_val;
  int __pyx_v_camn;
  int __pyx_v_frame_pt_idx;
  struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine *__pyx_v_projected_line = 0;
  struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *__pyx_v_best_3d_location = 0;
  PyObject *__pyx_v_all_close_camn_pt_idxs = NULL;
  PyObject *__pyx_v_prediction_3d = NULL;
  struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *__pyx_v_fast_prediction_3d = 0;
  PyObject *__pyx_v_pixel_dist_cmp = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_neg_predicted_3d = NULL;
  PyObject *__pyx_v_cam_ids_and_points2d = NULL;
  PyObject *__pyx_v_used_camns_and_idxs = NULL;
  PyObject *__pyx_v_candidate_point_list = NULL;
  PyObject *__pyx_v_cam_id = NULL;
  PyObject *__pyx_v_predicted_2d_distorted = NULL;
  PyObject *__pyx_v_predicted_2d_undistorted = NULL;
  PyObject *__pyx_v_closest_idx = NULL;
  PyObject *__pyx_v_Pminus_inv = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_pt_undistorted = NULL;
  PyObject *__pyx_v_pt_x_undist = NULL;
  PyObject *__pyx_v_pt_y_undist = NULL;
  PyObject *__pyx_v_pt_x_dist = NULL;
  PyObject *__pyx_v_pt_y_dist = NULL;
  PyObject *__pyx_v_pixel_dist = NULL;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_extra_print = NULL;
  PyObject *__pyx_v_observed_2d = NULL;
  PyObject *__pyx_v_Lcoords = NULL;
  PyObject *__pyx_v_position_MLE = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  double __pyx_t_17;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_filter_data", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_debug = __pyx_optional_args->debug;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filter_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15_filter_data)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[6] = {__pyx_t_5, __pyx_v_xhatminus, __pyx_v_Pminus, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[6] = {__pyx_t_5, __pyx_v_xhatminus, __pyx_v_Pminus, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 5+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(5+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 492, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_xhatminus);
          __Pyx_GIVEREF(__pyx_v_xhatminus);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_xhatminus);
          __Pyx_INCREF(__pyx_v_Pminus);
          __Pyx_GIVEREF(__pyx_v_Pminus);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_Pminus);
          __Pyx_INCREF(__pyx_v_data_dict);
          __Pyx_GIVEREF(__pyx_v_data_dict);
          PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_data_dict);
          __Pyx_INCREF(__pyx_v_camn2cam_id);
          __Pyx_GIVEREF(__pyx_v_camn2cam_id);
          PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_camn2cam_id);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 4+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flydra_core/_flydra_tracked_object.pyx":520
 *         cdef _fastgeom.ThreeTuple best_3d_location
 * 
 *         all_close_camn_pt_idxs = [] # store all "maybes"             # <<<<<<<<<<<<<<
 * 
 *         prediction_3d = xhatminus[:3]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_all_close_camn_pt_idxs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":522
 *         all_close_camn_pt_idxs = [] # store all "maybes"
 * 
 *         prediction_3d = xhatminus[:3]             # <<<<<<<<<<<<<<
 *         cdef _fastgeom.ThreeTuple fast_prediction_3d = _fastgeom.ThreeTuple(xhatminus[:3])
 *         pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_xhatminus, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_prediction_3d = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":523
 * 
 *         prediction_3d = xhatminus[:3]
 *         cdef _fastgeom.ThreeTuple fast_prediction_3d = _fastgeom.ThreeTuple(xhatminus[:3])             # <<<<<<<<<<<<<<
 *         pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept
 *         neg_predicted_3d = -geom.ThreeTuple( prediction_3d )
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_xhatminus, 0, 3, NULL, NULL, &__pyx_slice_, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_11flydra_core_9_fastgeom_ThreeTuple), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fast_prediction_3d = ((struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":524
 *         prediction_3d = xhatminus[:3]
 *         cdef _fastgeom.ThreeTuple fast_prediction_3d = _fastgeom.ThreeTuple(xhatminus[:3])
 *         pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept             # <<<<<<<<<<<<<<
 *         neg_predicted_3d = -geom.ThreeTuple( prediction_3d )
 *         cam_ids_and_points2d = []
 */
  __pyx_t_2 = __pyx_v_self->distorted_pixel_euclidian_distance_accept;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pixel_dist_cmp = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":525
 *         cdef _fastgeom.ThreeTuple fast_prediction_3d = _fastgeom.ThreeTuple(xhatminus[:3])
 *         pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept
 *         neg_predicted_3d = -geom.ThreeTuple( prediction_3d )             # <<<<<<<<<<<<<<
 *         cam_ids_and_points2d = []
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_geom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ThreeTuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_prediction_3d) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_prediction_3d);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Negative(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_neg_predicted_3d = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":526
 *         pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept
 *         neg_predicted_3d = -geom.ThreeTuple( prediction_3d )
 *         cam_ids_and_points2d = []             # <<<<<<<<<<<<<<
 * 
 *         used_camns_and_idxs = []
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_cam_ids_and_points2d = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":528
 *         cam_ids_and_points2d = []
 * 
 *         used_camns_and_idxs = []             # <<<<<<<<<<<<<<
 *         if debug>2:
 *             print '_filter_data():'
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_used_camns_and_idxs = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":529
 * 
 *         used_camns_and_idxs = []
 *         if debug>2:             # <<<<<<<<<<<<<<
 *             print '_filter_data():'
 *         for camn, candidate_point_list in data_dict.items():
 */
  __pyx_t_8 = ((__pyx_v_debug > 2) != 0);
  if (__pyx_t_8) {

    /* "flydra_core/_flydra_tracked_object.pyx":530
 *         used_camns_and_idxs = []
 *         if debug>2:
 *             print '_filter_data():'             # <<<<<<<<<<<<<<
 *         for camn, candidate_point_list in data_dict.items():
 *             cam_id = camn2cam_id[camn]
 */
    if (__Pyx_PrintOne(0, __pyx_kp_s_filter_data_2) < 0) __PYX_ERR(0, 530, __pyx_L1_error)

    /* "flydra_core/_flydra_tracked_object.pyx":529
 * 
 *         used_camns_and_idxs = []
 *         if debug>2:             # <<<<<<<<<<<<<<
 *             print '_filter_data():'
 *         for camn, candidate_point_list in data_dict.items():
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":531
 *         if debug>2:
 *             print '_filter_data():'
 *         for camn, candidate_point_list in data_dict.items():             # <<<<<<<<<<<<<<
 *             cam_id = camn2cam_id[camn]
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_dict, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_2 = __pyx_t_4; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 531, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_10(__pyx_t_2);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 531, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 531, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_7 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_3), 2) < 0) __PYX_ERR(0, 531, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 531, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_camn = __pyx_t_6;
    __Pyx_XDECREF_SET(__pyx_v_candidate_point_list, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":532
 *             print '_filter_data():'
 *         for camn, candidate_point_list in data_dict.items():
 *             cam_id = camn2cam_id[camn]             # <<<<<<<<<<<<<<
 * 
 *             if pixel_dist_cmp is not None:
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_camn2cam_id, __pyx_v_camn, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_cam_id, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":534
 *             cam_id = camn2cam_id[camn]
 * 
 *             if pixel_dist_cmp is not None:             # <<<<<<<<<<<<<<
 *                 predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)
 * 
 */
    __pyx_t_8 = (__pyx_v_pixel_dist_cmp != Py_None);
    __pyx_t_12 = (__pyx_t_8 != 0);
    if (__pyx_t_12) {

      /* "flydra_core/_flydra_tracked_object.pyx":535
 * 
 *             if pixel_dist_cmp is not None:
 *                 predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)             # <<<<<<<<<<<<<<
 * 
 *             if debug>2:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_find2d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_cam_id);
      __Pyx_GIVEREF(__pyx_v_cam_id);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_cam_id);
      __Pyx_INCREF(__pyx_v_prediction_3d);
      __Pyx_GIVEREF(__pyx_v_prediction_3d);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_prediction_3d);
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_distorted, Py_True) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_predicted_2d_distorted, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":534
 *             cam_id = camn2cam_id[camn]
 * 
 *             if pixel_dist_cmp is not None:             # <<<<<<<<<<<<<<
 *                 predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)
 * 
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":537
 *                 predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)
 * 
 *             if debug>2:             # <<<<<<<<<<<<<<
 *                 predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)
 *                 print '  cam_id',cam_id,'camn',camn,'--------'
 */
    __pyx_t_12 = ((__pyx_v_debug > 2) != 0);
    if (__pyx_t_12) {

      /* "flydra_core/_flydra_tracked_object.pyx":538
 * 
 *             if debug>2:
 *                 predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)             # <<<<<<<<<<<<<<
 *                 print '  cam_id',cam_id,'camn',camn,'--------'
 *                 print '    predicted_2d (undistorted)',predicted_2d_undistorted
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_find2d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_cam_id);
      __Pyx_GIVEREF(__pyx_v_cam_id);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_cam_id);
      __Pyx_INCREF(__pyx_v_prediction_3d);
      __Pyx_GIVEREF(__pyx_v_prediction_3d);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_prediction_3d);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_distorted, Py_False) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_predicted_2d_undistorted, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":539
 *             if debug>2:
 *                 predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)
 *                 print '  cam_id',cam_id,'camn',camn,'--------'             # <<<<<<<<<<<<<<
 *                 print '    predicted_2d (undistorted)',predicted_2d_undistorted
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_camn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_kp_s_cam_id);
      __Pyx_GIVEREF(__pyx_kp_s_cam_id);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_cam_id);
      __Pyx_INCREF(__pyx_v_cam_id);
      __Pyx_GIVEREF(__pyx_v_cam_id);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_cam_id);
      __Pyx_INCREF(__pyx_n_s_camn);
      __Pyx_GIVEREF(__pyx_n_s_camn);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_n_s_camn);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
      __Pyx_INCREF(__pyx_kp_s__16);
      __Pyx_GIVEREF(__pyx_kp_s__16);
      PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_kp_s__16);
      __pyx_t_4 = 0;
      if (__Pyx_Print(0, __pyx_t_7, 1) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":540
 *                 predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)
 *                 print '  cam_id',cam_id,'camn',camn,'--------'
 *                 print '    predicted_2d (undistorted)',predicted_2d_undistorted             # <<<<<<<<<<<<<<
 * 
 *             # Iterate over all 2D data. Find the 2D point which minimizes::
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_kp_s_predicted_2d_undistorted);
      __Pyx_GIVEREF(__pyx_kp_s_predicted_2d_undistorted);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_s_predicted_2d_undistorted);
      __Pyx_INCREF(__pyx_v_predicted_2d_undistorted);
      __Pyx_GIVEREF(__pyx_v_predicted_2d_undistorted);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_predicted_2d_undistorted);
      if (__Pyx_Print(0, __pyx_t_7, 1) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":537
 *                 predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)
 * 
 *             if debug>2:             # <<<<<<<<<<<<<<
 *                 predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)
 *                 print '  cam_id',cam_id,'camn',camn,'--------'
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":559
 *             # normal.
 * 
 *             least_nll = c_inf             # <<<<<<<<<<<<<<
 *             closest_idx = None
 * 
 */
    __pyx_v_least_nll = __pyx_v_11flydra_core_22_flydra_tracked_object_c_inf;

    /* "flydra_core/_flydra_tracked_object.pyx":560
 * 
 *             least_nll = c_inf
 *             closest_idx = None             # <<<<<<<<<<<<<<
 * 
 *             Pminus_inv = None # defer inverting Pminus until necessary.
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_closest_idx, Py_None);

    /* "flydra_core/_flydra_tracked_object.pyx":562
 *             closest_idx = None
 * 
 *             Pminus_inv = None # defer inverting Pminus until necessary.             # <<<<<<<<<<<<<<
 * 
 *             for idx,(pt_undistorted,projected_line) in enumerate(
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_Pminus_inv, Py_None);

    /* "flydra_core/_flydra_tracked_object.pyx":564
 *             Pminus_inv = None # defer inverting Pminus until necessary.
 * 
 *             for idx,(pt_undistorted,projected_line) in enumerate(             # <<<<<<<<<<<<<<
 *                 candidate_point_list):
 * 
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_7 = __pyx_int_0;

    /* "flydra_core/_flydra_tracked_object.pyx":565
 * 
 *             for idx,(pt_undistorted,projected_line) in enumerate(
 *                 candidate_point_list):             # <<<<<<<<<<<<<<
 * 
 *                 # Iterate over each candidate point. Each point has:
 */
    if (likely(PyList_CheckExact(__pyx_v_candidate_point_list)) || PyTuple_CheckExact(__pyx_v_candidate_point_list)) {
      __pyx_t_4 = __pyx_v_candidate_point_list; __Pyx_INCREF(__pyx_t_4); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_candidate_point_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 564, __pyx_L1_error)
    }
    for (;;) {

      /* "flydra_core/_flydra_tracked_object.pyx":564
 *             Pminus_inv = None # defer inverting Pminus until necessary.
 * 
 *             for idx,(pt_undistorted,projected_line) in enumerate(             # <<<<<<<<<<<<<<
 *                 candidate_point_list):
 * 
 */
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 564, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 564, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_14(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 564, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 564, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_15 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_15)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_11(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_5 = __pyx_t_11(__pyx_t_15); if (unlikely(!__pyx_t_5)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_15), 2) < 0) __PYX_ERR(0, 564, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 564, __pyx_L1_error)
        __pyx_L13_unpacking_done:;
      }
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_11flydra_core_9_fastgeom_PlueckerLine))))) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_pt_undistorted, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_projected_line, ((struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine *)__pyx_t_5));
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_7);
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7);
      __pyx_t_7 = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":574
 *                 #    image plane.
 * 
 *                 gated_in = False             # <<<<<<<<<<<<<<
 * 
 *                 # First, a quick gating based on image plane distance.
 */
      __pyx_v_gated_in = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":578
 *                 # First, a quick gating based on image plane distance.
 * 
 *                 pixel_dist_criterion_passed = True             # <<<<<<<<<<<<<<
 *                 if pixel_dist_cmp is not None:
 *                     # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
 */
      __pyx_v_pixel_dist_criterion_passed = 1;

      /* "flydra_core/_flydra_tracked_object.pyx":579
 * 
 *                 pixel_dist_criterion_passed = True
 *                 if pixel_dist_cmp is not None:             # <<<<<<<<<<<<<<
 *                     # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
 *                     pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]
 */
      __pyx_t_12 = (__pyx_v_pixel_dist_cmp != Py_None);
      __pyx_t_8 = (__pyx_t_12 != 0);
      if (__pyx_t_8) {

        /* "flydra_core/_flydra_tracked_object.pyx":581
 *                 if pixel_dist_cmp is not None:
 *                     # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
 *                     pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]             # <<<<<<<<<<<<<<
 *                     pt_y_undist =  pt_undistorted[PT_TUPLE_IDX_Y]
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_X); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_pt_x_undist, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":582
 *                     # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
 *                     pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]
 *                     pt_y_undist =  pt_undistorted[PT_TUPLE_IDX_Y]             # <<<<<<<<<<<<<<
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PT_TUPLE_IDX_Y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_pt_y_undist, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":583
 *                     pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]
 *                     pt_y_undist =  pt_undistorted[PT_TUPLE_IDX_Y]
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )             # <<<<<<<<<<<<<<
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
 *                     if pixel_dist > pixel_dist_cmp:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_distort); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_pt_x_undist);
        __Pyx_GIVEREF(__pyx_v_pt_x_undist);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pt_x_undist);
        __Pyx_INCREF(__pyx_v_pt_y_undist);
        __Pyx_GIVEREF(__pyx_v_pt_y_undist);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_pt_y_undist);
        __pyx_t_15 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_cam_id, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_cam_id, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_15) {
            __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15); __pyx_t_15 = NULL;
          }
          __Pyx_INCREF(__pyx_v_cam_id);
          __Pyx_GIVEREF(__pyx_v_cam_id);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_6, __pyx_v_cam_id);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 583, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_16);
          #else
          __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext;
          index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          index = 1; __pyx_t_16 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_16)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_16);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_3), 2) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
          __pyx_t_11 = NULL;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L16_unpacking_done;
          __pyx_L15_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_11 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 583, __pyx_L1_error)
          __pyx_L16_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_pt_x_dist, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_pt_y_dist, __pyx_t_16);
        __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":584
 *                     pt_y_undist =  pt_undistorted[PT_TUPLE_IDX_Y]
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)             # <<<<<<<<<<<<<<
 *                     if pixel_dist > pixel_dist_cmp:
 *                         pixel_dist_criterion_passed = False
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_numpy); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_v_predicted_2d_distorted)) { __Pyx_RaiseUnboundLocalError("predicted_2d_distorted"); __PYX_ERR(0, 584, __pyx_L1_error) }
        __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_predicted_2d_distorted, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_3 = PyNumber_Subtract(__pyx_t_16, __pyx_v_pt_x_dist); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = PyNumber_Power(__pyx_t_3, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_v_predicted_2d_distorted)) { __Pyx_RaiseUnboundLocalError("predicted_2d_distorted"); __PYX_ERR(0, 584, __pyx_L1_error) }
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_predicted_2d_distorted, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_15 = PyNumber_Subtract(__pyx_t_3, __pyx_v_pt_y_dist); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Power(__pyx_t_15, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = PyNumber_Add(__pyx_t_16, __pyx_t_3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_pixel_dist, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":585
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
 *                     if pixel_dist > pixel_dist_cmp:             # <<<<<<<<<<<<<<
 *                         pixel_dist_criterion_passed = False
 * 
 */
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_pixel_dist, __pyx_v_pixel_dist_cmp, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_8) {

          /* "flydra_core/_flydra_tracked_object.pyx":586
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
 *                     if pixel_dist > pixel_dist_cmp:
 *                         pixel_dist_criterion_passed = False             # <<<<<<<<<<<<<<
 * 
 *                 if pixel_dist_criterion_passed:
 */
          __pyx_v_pixel_dist_criterion_passed = 0;

          /* "flydra_core/_flydra_tracked_object.pyx":585
 *                     pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
 *                     pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
 *                     if pixel_dist > pixel_dist_cmp:             # <<<<<<<<<<<<<<
 *                         pixel_dist_criterion_passed = False
 * 
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":579
 * 
 *                 pixel_dist_criterion_passed = True
 *                 if pixel_dist_cmp is not None:             # <<<<<<<<<<<<<<
 *                     # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
 *                     pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":588
 *                         pixel_dist_criterion_passed = False
 * 
 *                 if pixel_dist_criterion_passed:             # <<<<<<<<<<<<<<
 * 
 *                     # Second, a quick gating based on image attributes.
 */
      __pyx_t_8 = (__pyx_v_pixel_dist_criterion_passed != 0);
      if (__pyx_t_8) {

        /* "flydra_core/_flydra_tracked_object.pyx":592
 *                     # Second, a quick gating based on image attributes.
 * 
 *                     pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]             # <<<<<<<<<<<<<<
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                     if tmp is None:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_AREA); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 592, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 592, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_pt_area = __pyx_t_17;

        /* "flydra_core/_flydra_tracked_object.pyx":593
 * 
 *                     pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]             # <<<<<<<<<<<<<<
 *                     if tmp is None:
 *                         if not self.disable_image_stat_gating:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 593, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_tmp, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":594
 *                     pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                     if tmp is None:             # <<<<<<<<<<<<<<
 *                         if not self.disable_image_stat_gating:
 *                             raise ValueError(
 */
        __pyx_t_8 = (__pyx_v_tmp == Py_None);
        __pyx_t_12 = (__pyx_t_8 != 0);
        if (__pyx_t_12) {

          /* "flydra_core/_flydra_tracked_object.pyx":595
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                     if tmp is None:
 *                         if not self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                             raise ValueError(
 *                                 '--disable-image-stat-gating not specified and '
 */
          __pyx_t_12 = ((!(__pyx_v_self->disable_image_stat_gating != 0)) != 0);
          if (unlikely(__pyx_t_12)) {

            /* "flydra_core/_flydra_tracked_object.pyx":596
 *                     if tmp is None:
 *                         if not self.disable_image_stat_gating:
 *                             raise ValueError(             # <<<<<<<<<<<<<<
 *                                 '--disable-image-stat-gating not specified and '
 *                                 'no image statistics were saved.')
 */
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_Raise(__pyx_t_1, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_ERR(0, 596, __pyx_L1_error)

            /* "flydra_core/_flydra_tracked_object.pyx":595
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                     if tmp is None:
 *                         if not self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                             raise ValueError(
 *                                 '--disable-image-stat-gating not specified and '
 */
          }

          /* "flydra_core/_flydra_tracked_object.pyx":594
 *                     pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                     tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                     if tmp is None:             # <<<<<<<<<<<<<<
 *                         if not self.disable_image_stat_gating:
 *                             raise ValueError(
 */
          goto __pyx_L19;
        }

        /* "flydra_core/_flydra_tracked_object.pyx":600
 *                                 'no image statistics were saved.')
 *                     else:
 *                         cur_val = tmp             # <<<<<<<<<<<<<<
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 */
        /*else*/ {
          __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_tmp); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 600, __pyx_L1_error)
          __pyx_v_cur_val = __pyx_t_6;

          /* "flydra_core/_flydra_tracked_object.pyx":601
 *                     else:
 *                         cur_val = tmp
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]             # <<<<<<<<<<<<<<
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 601, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 601, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_mean_val = __pyx_t_17;

          /* "flydra_core/_flydra_tracked_object.pyx":602
 *                         cur_val = tmp
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]             # <<<<<<<<<<<<<<
 * 
 *                     if cur_val is None or self.disable_image_stat_gating:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 602, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_sumsqf_val = __pyx_t_17;
        }
        __pyx_L19:;

        /* "flydra_core/_flydra_tracked_object.pyx":604
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 * 
 *                     if cur_val is None or self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                         p_y_x = 0.0 # no penalty
 *                     else:
 */
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_cur_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = (__pyx_t_1 == Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_18 = (__pyx_t_8 != 0);
        if (!__pyx_t_18) {
        } else {
          __pyx_t_12 = __pyx_t_18;
          goto __pyx_L22_bool_binop_done;
        }
        __pyx_t_18 = (__pyx_v_self->disable_image_stat_gating != 0);
        __pyx_t_12 = __pyx_t_18;
        __pyx_L22_bool_binop_done:;
        if (__pyx_t_12) {

          /* "flydra_core/_flydra_tracked_object.pyx":605
 * 
 *                     if cur_val is None or self.disable_image_stat_gating:
 *                         p_y_x = 0.0 # no penalty             # <<<<<<<<<<<<<<
 *                     else:
 *                         # this could even depend on 3d geometry
 */
          __pyx_v_p_y_x = 0.0;

          /* "flydra_core/_flydra_tracked_object.pyx":604
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 * 
 *                     if cur_val is None or self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                         p_y_x = 0.0 # no penalty
 *                     else:
 */
          goto __pyx_L21;
        }

        /* "flydra_core/_flydra_tracked_object.pyx":608
 *                     else:
 *                         # this could even depend on 3d geometry
 *                         p_y_x = some_rough_negative_log_likelihood(             # <<<<<<<<<<<<<<
 *                             pt_area, cur_val, mean_val, sumsqf_val )
 * 
 */
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_some_rough_negative_log_likeliho); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "flydra_core/_flydra_tracked_object.pyx":609
 *                         # this could even depend on 3d geometry
 *                         p_y_x = some_rough_negative_log_likelihood(
 *                             pt_area, cur_val, mean_val, sumsqf_val )             # <<<<<<<<<<<<<<
 * 
 *                     if pt_area < self.area_threshold:
 */
          __pyx_t_15 = PyFloat_FromDouble(__pyx_v_pt_area); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 609, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cur_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_16 = PyFloat_FromDouble(__pyx_v_mean_val); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 609, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_19 = PyFloat_FromDouble(__pyx_v_sumsqf_val); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 609, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_20 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_20)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_20);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[5] = {__pyx_t_20, __pyx_t_15, __pyx_t_3, __pyx_t_16, __pyx_t_19};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[5] = {__pyx_t_20, __pyx_t_15, __pyx_t_3, __pyx_t_16, __pyx_t_19};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          } else
          #endif
          {
            __pyx_t_21 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 608, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_21);
            if (__pyx_t_20) {
              __Pyx_GIVEREF(__pyx_t_20); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_20); __pyx_t_20 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_6, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_6, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_21, 2+__pyx_t_6, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_19);
            PyTuple_SET_ITEM(__pyx_t_21, 3+__pyx_t_6, __pyx_t_19);
            __pyx_t_15 = 0;
            __pyx_t_3 = 0;
            __pyx_t_16 = 0;
            __pyx_t_19 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "flydra_core/_flydra_tracked_object.pyx":608
 *                     else:
 *                         # this could even depend on 3d geometry
 *                         p_y_x = some_rough_negative_log_likelihood(             # <<<<<<<<<<<<<<
 *                             pt_area, cur_val, mean_val, sumsqf_val )
 * 
 */
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_p_y_x = __pyx_t_17;
        }
        __pyx_L21:;

        /* "flydra_core/_flydra_tracked_object.pyx":611
 *                             pt_area, cur_val, mean_val, sumsqf_val )
 * 
 *                     if pt_area < self.area_threshold:             # <<<<<<<<<<<<<<
 *                         # This should not fall under
 *                         # "disable_image_stat_gating" -- it is a separate test.
 */
        __pyx_t_12 = ((__pyx_v_pt_area < __pyx_v_self->area_threshold) != 0);
        if (__pyx_t_12) {

          /* "flydra_core/_flydra_tracked_object.pyx":614
 *                         # This should not fall under
 *                         # "disable_image_stat_gating" -- it is a separate test.
 *                         p_y_x = np.inf             # <<<<<<<<<<<<<<
 * 
 *                     if np.isfinite(p_y_x):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_inf); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 614, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_p_y_x = __pyx_t_17;

          /* "flydra_core/_flydra_tracked_object.pyx":611
 *                             pt_area, cur_val, mean_val, sumsqf_val )
 * 
 *                     if pt_area < self.area_threshold:             # <<<<<<<<<<<<<<
 *                         # This should not fall under
 *                         # "disable_image_stat_gating" -- it is a separate test.
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":616
 *                         p_y_x = np.inf
 * 
 *                     if np.isfinite(p_y_x):             # <<<<<<<<<<<<<<
 *                         gated_in = True
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_isfinite); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_p_y_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_19 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
          __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_21);
          if (likely(__pyx_t_19)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
            __Pyx_INCREF(__pyx_t_19);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_21, function);
          }
        }
        __pyx_t_5 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_19, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_12) {

          /* "flydra_core/_flydra_tracked_object.pyx":617
 * 
 *                     if np.isfinite(p_y_x):
 *                         gated_in = True             # <<<<<<<<<<<<<<
 * 
 *                         # Find point on ray with closest Mahalanobis distance.
 */
          __pyx_v_gated_in = 1;

          /* "flydra_core/_flydra_tracked_object.pyx":621
 *                         # Find point on ray with closest Mahalanobis distance.
 * 
 *                         if Pminus_inv is None:             # <<<<<<<<<<<<<<
 *                             Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )
 *                         best_3d_location = _mahalanobis.line_fit_3d(
 */
          __pyx_t_12 = (__pyx_v_Pminus_inv == Py_None);
          __pyx_t_18 = (__pyx_t_12 != 0);
          if (__pyx_t_18) {

            /* "flydra_core/_flydra_tracked_object.pyx":622
 * 
 *                         if Pminus_inv is None:
 *                             Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )             # <<<<<<<<<<<<<<
 *                         best_3d_location = _mahalanobis.line_fit_3d(
 *                             projected_line, fast_prediction_3d, Pminus_inv )
 */
            __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_numpy); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_21);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_n_s_linalg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_inv); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_21);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_Pminus, __pyx_tuple__18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_19 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_21))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_21);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_21, function);
              }
            }
            __pyx_t_5 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_19, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_t_1);
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 622, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_DECREF_SET(__pyx_v_Pminus_inv, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "flydra_core/_flydra_tracked_object.pyx":621
 *                         # Find point on ray with closest Mahalanobis distance.
 * 
 *                         if Pminus_inv is None:             # <<<<<<<<<<<<<<
 *                             Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )
 *                         best_3d_location = _mahalanobis.line_fit_3d(
 */
          }

          /* "flydra_core/_flydra_tracked_object.pyx":623
 *                         if Pminus_inv is None:
 *                             Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )
 *                         best_3d_location = _mahalanobis.line_fit_3d(             # <<<<<<<<<<<<<<
 *                             projected_line, fast_prediction_3d, Pminus_inv )
 * 
 */
          __pyx_t_5 = ((PyObject *)__pyx_f_11flydra_core_12_mahalanobis_line_fit_3d(__pyx_v_projected_line, __pyx_v_fast_prediction_3d, __pyx_v_Pminus_inv, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_v_best_3d_location, ((struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *)__pyx_t_5));
          __pyx_t_5 = 0;

          /* "flydra_core/_flydra_tracked_object.pyx":628
 *                         # find closest distance between projected_line and predicted position for each 2d point
 *                         #   squared distance between prediction and camera ray
 *                         dist2=_mahalanobis.dist2( best_3d_location,             # <<<<<<<<<<<<<<
 *                                                   fast_prediction_3d,
 *                                                   Pminus_inv )
 */
          __pyx_v_dist2 = __pyx_f_11flydra_core_12_mahalanobis_dist2(__pyx_v_best_3d_location, __pyx_v_fast_prediction_3d, __pyx_v_Pminus_inv, 0);

          /* "flydra_core/_flydra_tracked_object.pyx":631
 *                                                   fast_prediction_3d,
 *                                                   Pminus_inv )
 *                         dist = sqrt(dist2)             # <<<<<<<<<<<<<<
 *                         nll_this_point = p_y_x + dist # negative log likelihood of this point
 * 
 */
          __pyx_v_dist = sqrt(__pyx_v_dist2);

          /* "flydra_core/_flydra_tracked_object.pyx":632
 *                                                   Pminus_inv )
 *                         dist = sqrt(dist2)
 *                         nll_this_point = p_y_x + dist # negative log likelihood of this point             # <<<<<<<<<<<<<<
 * 
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 */
          __pyx_v_nll_this_point = (__pyx_v_p_y_x + __pyx_v_dist);

          /* "flydra_core/_flydra_tracked_object.pyx":616
 *                         p_y_x = np.inf
 * 
 *                     if np.isfinite(p_y_x):             # <<<<<<<<<<<<<<
 *                         gated_in = True
 * 
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":588
 *                         pixel_dist_criterion_passed = False
 * 
 *                 if pixel_dist_criterion_passed:             # <<<<<<<<<<<<<<
 * 
 *                     # Second, a quick gating based on image attributes.
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":634
 *                         nll_this_point = p_y_x + dist # negative log likelihood of this point
 * 
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]             # <<<<<<<<<<<<<<
 *                 if debug>2:
 *                     extra_print = []
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_21 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_5); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_21); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_v_frame_pt_idx = __pyx_t_6;

      /* "flydra_core/_flydra_tracked_object.pyx":635
 * 
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 if debug>2:             # <<<<<<<<<<<<<<
 *                     extra_print = []
 *                     if not self.disable_image_stat_gating:
 */
      __pyx_t_18 = ((__pyx_v_debug > 2) != 0);
      if (__pyx_t_18) {

        /* "flydra_core/_flydra_tracked_object.pyx":636
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 if debug>2:
 *                     extra_print = []             # <<<<<<<<<<<<<<
 *                     if not self.disable_image_stat_gating:
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 */
        __pyx_t_21 = PyList_New(0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 636, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_XDECREF_SET(__pyx_v_extra_print, ((PyObject*)__pyx_t_21));
        __pyx_t_21 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":637
 *                 if debug>2:
 *                     extra_print = []
 *                     if not self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 */
        __pyx_t_18 = ((!(__pyx_v_self->disable_image_stat_gating != 0)) != 0);
        if (__pyx_t_18) {

          /* "flydra_core/_flydra_tracked_object.pyx":638
 *                     extra_print = []
 *                     if not self.disable_image_stat_gating:
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]             # <<<<<<<<<<<<<<
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 */
          __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 638, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_21); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 638, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 638, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_cur_val = __pyx_t_6;

          /* "flydra_core/_flydra_tracked_object.pyx":639
 *                     if not self.disable_image_stat_gating:
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]             # <<<<<<<<<<<<<<
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 *                         if pixel_dist_criterion_passed:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_21 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_5); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 639, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_21); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 639, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __pyx_v_mean_val = __pyx_t_17;

          /* "flydra_core/_flydra_tracked_object.pyx":640
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]             # <<<<<<<<<<<<<<
 *                         if pixel_dist_criterion_passed:
 *                             extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(
 */
          __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 640, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_21); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 640, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 640, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_sumsqf_val = __pyx_t_17;

          /* "flydra_core/_flydra_tracked_object.pyx":641
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 *                         if pixel_dist_criterion_passed:             # <<<<<<<<<<<<<<
 *                             extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(
 *                                 pt_area,cur_val,mean_val,sumsqf_val))
 */
          __pyx_t_18 = (__pyx_v_pixel_dist_criterion_passed != 0);
          if (__pyx_t_18) {

            /* "flydra_core/_flydra_tracked_object.pyx":643
 *                         if pixel_dist_criterion_passed:
 *                             extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(
 *                                 pt_area,cur_val,mean_val,sumsqf_val))             # <<<<<<<<<<<<<<
 * 
 *                     if gated_in:
 */
            __pyx_t_5 = PyFloat_FromDouble(__pyx_v_pt_area); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_21 = __Pyx_PyInt_From_int(__pyx_v_cur_val); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_21);
            __pyx_t_1 = PyFloat_FromDouble(__pyx_v_mean_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_19 = PyFloat_FromDouble(__pyx_v_sumsqf_val); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_16 = PyTuple_New(4); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 643, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_21);
            PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_21);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_19);
            PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_19);
            __pyx_t_5 = 0;
            __pyx_t_21 = 0;
            __pyx_t_1 = 0;
            __pyx_t_19 = 0;

            /* "flydra_core/_flydra_tracked_object.pyx":642
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 *                         if pixel_dist_criterion_passed:
 *                             extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(             # <<<<<<<<<<<<<<
 *                                 pt_area,cur_val,mean_val,sumsqf_val))
 * 
 */
            __pyx_t_19 = __Pyx_PyString_Format(__pyx_kp_s_pt_area_f_cur_d_mean_1f_sumsqf, __pyx_t_16); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 642, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_extra_print, __pyx_t_19); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 642, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

            /* "flydra_core/_flydra_tracked_object.pyx":641
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 *                         sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
 *                         if pixel_dist_criterion_passed:             # <<<<<<<<<<<<<<
 *                             extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(
 *                                 pt_area,cur_val,mean_val,sumsqf_val))
 */
          }

          /* "flydra_core/_flydra_tracked_object.pyx":637
 *                 if debug>2:
 *                     extra_print = []
 *                     if not self.disable_image_stat_gating:             # <<<<<<<<<<<<<<
 *                         cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
 *                         mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":645
 *                                 pt_area,cur_val,mean_val,sumsqf_val))
 * 
 *                     if gated_in:             # <<<<<<<<<<<<<<
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 */
        __pyx_t_18 = (__pyx_v_gated_in != 0);
        if (__pyx_t_18) {

          /* "flydra_core/_flydra_tracked_object.pyx":647
 *                     if gated_in:
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))             # <<<<<<<<<<<<<<
 *                     print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(
 *                         frame_pt_idx,' '.join(extra_print))
 */
          if (unlikely(!__pyx_v_pt_x_dist)) { __Pyx_RaiseUnboundLocalError("pt_x_dist"); __PYX_ERR(0, 647, __pyx_L1_error) }
          if (unlikely(!__pyx_v_pt_y_dist)) { __Pyx_RaiseUnboundLocalError("pt_y_dist"); __PYX_ERR(0, 647, __pyx_L1_error) }
          if (unlikely(!__pyx_v_pixel_dist)) { __Pyx_RaiseUnboundLocalError("pixel_dist"); __PYX_ERR(0, 647, __pyx_L1_error) }
          __pyx_t_19 = PyFloat_FromDouble(__pyx_v_dist); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 647, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_pixel_dist_criterion_passed); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 647, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_16 = PyTuple_New(5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 647, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_INCREF(__pyx_v_pt_x_dist);
          __Pyx_GIVEREF(__pyx_v_pt_x_dist);
          PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_pt_x_dist);
          __Pyx_INCREF(__pyx_v_pt_y_dist);
          __Pyx_GIVEREF(__pyx_v_pt_y_dist);
          PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_pt_y_dist);
          __Pyx_INCREF(__pyx_v_pixel_dist);
          __Pyx_GIVEREF(__pyx_v_pixel_dist);
          PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_v_pixel_dist);
          __Pyx_GIVEREF(__pyx_t_19);
          PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_19);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_16, 4, __pyx_t_1);
          __pyx_t_19 = 0;
          __pyx_t_1 = 0;

          /* "flydra_core/_flydra_tracked_object.pyx":646
 * 
 *                     if gated_in:
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(             # <<<<<<<<<<<<<<
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 *                     print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(
 */
          __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_distorted_1f_1f_pixel_dist_1f_ma, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_extra_print, __pyx_t_1); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "flydra_core/_flydra_tracked_object.pyx":645
 *                                 pt_area,cur_val,mean_val,sumsqf_val))
 * 
 *                     if gated_in:             # <<<<<<<<<<<<<<
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":648
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 *                     print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(             # <<<<<<<<<<<<<<
 *                         frame_pt_idx,' '.join(extra_print))
 * 
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_dist2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_16 = __Pyx_PyObject_GetSlice(__pyx_v_pt_undistorted, 0, 2, NULL, NULL, &__pyx_slice__15, 0, 1, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);

        /* "flydra_core/_flydra_tracked_object.pyx":649
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 *                     print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(
 *                         frame_pt_idx,' '.join(extra_print))             # <<<<<<<<<<<<<<
 * 
 *                 if gated_in:
 */
        __pyx_t_19 = __Pyx_PyInt_From_int(__pyx_v_frame_pt_idx); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        __pyx_t_21 = __Pyx_PyString_Join(__pyx_kp_s__10, __pyx_v_extra_print); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_19);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_19);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_21);
        __pyx_t_19 = 0;
        __pyx_t_21 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":648
 *                         extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
 *                             pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
 *                     print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(             # <<<<<<<<<<<<<<
 *                         frame_pt_idx,' '.join(extra_print))
 * 
 */
        __pyx_t_21 = __Pyx_PyString_Format(__pyx_kp_s_idx_d_s, __pyx_t_5); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_kp_s__19);
        __Pyx_GIVEREF(__pyx_kp_s__19);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_s__19);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_21);
        __pyx_t_1 = 0;
        __pyx_t_16 = 0;
        __pyx_t_21 = 0;
        if (__Pyx_Print(0, __pyx_t_5, 1) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":635
 * 
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 if debug>2:             # <<<<<<<<<<<<<<
 *                     extra_print = []
 *                     if not self.disable_image_stat_gating:
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":651
 *                         frame_pt_idx,' '.join(extra_print))
 * 
 *                 if gated_in:             # <<<<<<<<<<<<<<
 *                     if debug>2:
 *                         print '       (acceptable)'
 */
      __pyx_t_18 = (__pyx_v_gated_in != 0);
      if (__pyx_t_18) {

        /* "flydra_core/_flydra_tracked_object.pyx":652
 * 
 *                 if gated_in:
 *                     if debug>2:             # <<<<<<<<<<<<<<
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:
 */
        __pyx_t_18 = ((__pyx_v_debug > 2) != 0);
        if (__pyx_t_18) {

          /* "flydra_core/_flydra_tracked_object.pyx":653
 *                 if gated_in:
 *                     if debug>2:
 *                         print '       (acceptable)'             # <<<<<<<<<<<<<<
 *                     if nll_this_point < least_nll:
 *                         if debug>2:
 */
          if (__Pyx_PrintOne(0, __pyx_kp_s_acceptable) < 0) __PYX_ERR(0, 653, __pyx_L1_error)

          /* "flydra_core/_flydra_tracked_object.pyx":652
 * 
 *                 if gated_in:
 *                     if debug>2:             # <<<<<<<<<<<<<<
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":654
 *                     if debug>2:
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:             # <<<<<<<<<<<<<<
 *                         if debug>2:
 *                             print '       (so far the best -- taking)'
 */
        __pyx_t_18 = ((__pyx_v_nll_this_point < __pyx_v_least_nll) != 0);
        if (__pyx_t_18) {

          /* "flydra_core/_flydra_tracked_object.pyx":655
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:
 *                         if debug>2:             # <<<<<<<<<<<<<<
 *                             print '       (so far the best -- taking)'
 *                         least_nll = nll_this_point
 */
          __pyx_t_18 = ((__pyx_v_debug > 2) != 0);
          if (__pyx_t_18) {

            /* "flydra_core/_flydra_tracked_object.pyx":656
 *                     if nll_this_point < least_nll:
 *                         if debug>2:
 *                             print '       (so far the best -- taking)'             # <<<<<<<<<<<<<<
 *                         least_nll = nll_this_point
 *                         closest_idx = idx
 */
            if (__Pyx_PrintOne(0, __pyx_kp_s_so_far_the_best_taking) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

            /* "flydra_core/_flydra_tracked_object.pyx":655
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:
 *                         if debug>2:             # <<<<<<<<<<<<<<
 *                             print '       (so far the best -- taking)'
 *                         least_nll = nll_this_point
 */
          }

          /* "flydra_core/_flydra_tracked_object.pyx":657
 *                         if debug>2:
 *                             print '       (so far the best -- taking)'
 *                         least_nll = nll_this_point             # <<<<<<<<<<<<<<
 *                         closest_idx = idx
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
 */
          __pyx_v_least_nll = __pyx_v_nll_this_point;

          /* "flydra_core/_flydra_tracked_object.pyx":658
 *                             print '       (so far the best -- taking)'
 *                         least_nll = nll_this_point
 *                         closest_idx = idx             # <<<<<<<<<<<<<<
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
 *                 elif debug>2:
 */
          __Pyx_INCREF(__pyx_v_idx);
          __Pyx_DECREF_SET(__pyx_v_closest_idx, __pyx_v_idx);

          /* "flydra_core/_flydra_tracked_object.pyx":654
 *                     if debug>2:
 *                         print '       (acceptable)'
 *                     if nll_this_point < least_nll:             # <<<<<<<<<<<<<<
 *                         if debug>2:
 *                             print '       (so far the best -- taking)'
 */
        }

        /* "flydra_core/_flydra_tracked_object.pyx":659
 *                         least_nll = nll_this_point
 *                         closest_idx = idx
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )             # <<<<<<<<<<<<<<
 *                 elif debug>2:
 *                     print '       (not acceptable)'
 */
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_camn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_21 = __Pyx_PyInt_From_int(__pyx_v_frame_pt_idx); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 659, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_21);
        __pyx_t_5 = 0;
        __pyx_t_21 = 0;
        __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_all_close_camn_pt_idxs, __pyx_t_16); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 659, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":651
 *                         frame_pt_idx,' '.join(extra_print))
 * 
 *                 if gated_in:             # <<<<<<<<<<<<<<
 *                     if debug>2:
 *                         print '       (acceptable)'
 */
        goto __pyx_L31;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":660
 *                         closest_idx = idx
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
 *                 elif debug>2:             # <<<<<<<<<<<<<<
 *                     print '       (not acceptable)'
 * 
 */
      __pyx_t_18 = ((__pyx_v_debug > 2) != 0);
      if (__pyx_t_18) {

        /* "flydra_core/_flydra_tracked_object.pyx":661
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
 *                 elif debug>2:
 *                     print '       (not acceptable)'             # <<<<<<<<<<<<<<
 * 
 *             if closest_idx is not None:
 */
        if (__Pyx_PrintOne(0, __pyx_kp_s_not_acceptable) < 0) __PYX_ERR(0, 661, __pyx_L1_error)

        /* "flydra_core/_flydra_tracked_object.pyx":660
 *                         closest_idx = idx
 *                     all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
 *                 elif debug>2:             # <<<<<<<<<<<<<<
 *                     print '       (not acceptable)'
 * 
 */
      }
      __pyx_L31:;

      /* "flydra_core/_flydra_tracked_object.pyx":564
 *             Pminus_inv = None # defer inverting Pminus until necessary.
 * 
 *             for idx,(pt_undistorted,projected_line) in enumerate(             # <<<<<<<<<<<<<<
 *                 candidate_point_list):
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":663
 *                     print '       (not acceptable)'
 * 
 *             if closest_idx is not None:             # <<<<<<<<<<<<<<
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]
 */
    __pyx_t_18 = (__pyx_v_closest_idx != Py_None);
    __pyx_t_12 = (__pyx_t_18 != 0);
    if (__pyx_t_12) {

      /* "flydra_core/_flydra_tracked_object.pyx":664
 * 
 *             if closest_idx is not None:
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]             # <<<<<<<<<<<<<<
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                 if area >= self.area_threshold_for_orientation:
 */
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_candidate_point_list, __pyx_v_closest_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 664, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_16);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_21 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_21)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_11(__pyx_t_21); if (unlikely(!__pyx_t_4)) goto __pyx_L36_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_16 = __pyx_t_11(__pyx_t_21); if (unlikely(!__pyx_t_16)) goto __pyx_L36_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_16);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_21), 2) < 0) __PYX_ERR(0, 664, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        goto __pyx_L37_unpacking_done;
        __pyx_L36_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 664, __pyx_L1_error)
        __pyx_L37_unpacking_done:;
      }
      if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_11flydra_core_9_fastgeom_PlueckerLine))))) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_pt_undistorted, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_projected_line, ((struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine *)__pyx_t_16));
      __pyx_t_16 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":665
 *             if closest_idx is not None:
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]             # <<<<<<<<<<<<<<
 *                 if area >= self.area_threshold_for_orientation:
 *                     # with orientation
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PT_TUPLE_IDX_AREA); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_16 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_7); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_16); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_v_area = __pyx_t_17;

      /* "flydra_core/_flydra_tracked_object.pyx":666
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                 if area >= self.area_threshold_for_orientation:             # <<<<<<<<<<<<<<
 *                     # with orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 */
      __pyx_t_12 = ((__pyx_v_area >= __pyx_v_self->area_threshold_for_orientation) != 0);
      if (__pyx_t_12) {

        /* "flydra_core/_flydra_tracked_object.pyx":668
 *                 if area >= self.area_threshold_for_orientation:
 *                     # with orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_X); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":669
 *                     # with orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    pt_undistorted[PT_TUPLE_IDX_SLOPE],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_Y); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":670
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_SLOPE],
 *                                    pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_AREA); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_21 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":671
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    pt_undistorted[PT_TUPLE_IDX_SLOPE],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],
 *                                    pt_undistorted[PT_TUPLE_IDX_P1],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_SLOPE); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 671, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 671, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":672
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    pt_undistorted[PT_TUPLE_IDX_SLOPE],
 *                                    pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_P1],
 *                                    pt_undistorted[PT_TUPLE_IDX_P2],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_ECCENTRICITY); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 672, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":673
 *                                    pt_undistorted[PT_TUPLE_IDX_SLOPE],
 *                                    pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],
 *                                    pt_undistorted[PT_TUPLE_IDX_P1],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_P2],
 *                                    pt_undistorted[PT_TUPLE_IDX_P3],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_P1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_19 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":674
 *                                    pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],
 *                                    pt_undistorted[PT_TUPLE_IDX_P1],
 *                                    pt_undistorted[PT_TUPLE_IDX_P2],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_P3],
 *                                    pt_undistorted[PT_TUPLE_IDX_P4])
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_P2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":675
 *                                    pt_undistorted[PT_TUPLE_IDX_P1],
 *                                    pt_undistorted[PT_TUPLE_IDX_P2],
 *                                    pt_undistorted[PT_TUPLE_IDX_P3],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_P4])
 *                 else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_P3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":676
 *                                    pt_undistorted[PT_TUPLE_IDX_P2],
 *                                    pt_undistorted[PT_TUPLE_IDX_P3],
 *                                    pt_undistorted[PT_TUPLE_IDX_P4])             # <<<<<<<<<<<<<<
 *                 else:
 *                     # with no orientation
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_P4); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 676, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_20 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 676, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":668
 *                 if area >= self.area_threshold_for_orientation:
 *                     # with orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 */
        __pyx_t_16 = PyTuple_New(9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_t_21);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_16, 4, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_19);
        PyTuple_SET_ITEM(__pyx_t_16, 5, __pyx_t_19);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_16, 6, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_15);
        PyTuple_SET_ITEM(__pyx_t_16, 7, __pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_20);
        PyTuple_SET_ITEM(__pyx_t_16, 8, __pyx_t_20);
        __pyx_t_7 = 0;
        __pyx_t_4 = 0;
        __pyx_t_21 = 0;
        __pyx_t_5 = 0;
        __pyx_t_1 = 0;
        __pyx_t_19 = 0;
        __pyx_t_3 = 0;
        __pyx_t_15 = 0;
        __pyx_t_20 = 0;
        __Pyx_XDECREF_SET(__pyx_v_observed_2d, ((PyObject*)__pyx_t_16));
        __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":666
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]
 *                 if area >= self.area_threshold_for_orientation:             # <<<<<<<<<<<<<<
 *                     # with orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 */
        goto __pyx_L38;
      }

      /* "flydra_core/_flydra_tracked_object.pyx":679
 *                 else:
 *                     # with no orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_X); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_20 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":680
 *                     # with no orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    cnan,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_Y); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 680, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 680, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":681
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],             # <<<<<<<<<<<<<<
 *                                    cnan,
 *                                    cnan,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_PT_TUPLE_IDX_AREA); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":682
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    cnan,             # <<<<<<<<<<<<<<
 *                                    cnan,
 *                                    cnan,
 */
        __pyx_t_16 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);

        /* "flydra_core/_flydra_tracked_object.pyx":683
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 *                                    cnan,
 *                                    cnan,             # <<<<<<<<<<<<<<
 *                                    cnan,
 *                                    cnan,
 */
        __pyx_t_19 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 683, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);

        /* "flydra_core/_flydra_tracked_object.pyx":684
 *                                    cnan,
 *                                    cnan,
 *                                    cnan,             # <<<<<<<<<<<<<<
 *                                    cnan,
 *                                    cnan,
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "flydra_core/_flydra_tracked_object.pyx":685
 *                                    cnan,
 *                                    cnan,
 *                                    cnan,             # <<<<<<<<<<<<<<
 *                                    cnan,
 *                                    cnan)
 */
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "flydra_core/_flydra_tracked_object.pyx":686
 *                                    cnan,
 *                                    cnan,
 *                                    cnan,             # <<<<<<<<<<<<<<
 *                                    cnan)
 *                 cam_ids_and_points2d.append( (cam_id,observed_2d) )
 */
        __pyx_t_21 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 686, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);

        /* "flydra_core/_flydra_tracked_object.pyx":687
 *                                    cnan,
 *                                    cnan,
 *                                    cnan)             # <<<<<<<<<<<<<<
 *                 cam_ids_and_points2d.append( (cam_id,observed_2d) )
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 */
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_11flydra_core_22_flydra_tracked_object_cnan); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 687, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "flydra_core/_flydra_tracked_object.pyx":679
 *                 else:
 *                     # with no orientation
 *                     observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],             # <<<<<<<<<<<<<<
 *                                    pt_undistorted[PT_TUPLE_IDX_Y],
 *                                    pt_undistorted[PT_TUPLE_IDX_AREA],
 */
        __pyx_t_7 = PyTuple_New(9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_20);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_20);
        __Pyx_GIVEREF(__pyx_t_15);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_19);
        PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_19);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_t_21);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_7, 8, __pyx_t_4);
        __pyx_t_20 = 0;
        __pyx_t_15 = 0;
        __pyx_t_3 = 0;
        __pyx_t_16 = 0;
        __pyx_t_19 = 0;
        __pyx_t_1 = 0;
        __pyx_t_5 = 0;
        __pyx_t_21 = 0;
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_observed_2d, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;
      }
      __pyx_L38:;

      /* "flydra_core/_flydra_tracked_object.pyx":688
 *                                    cnan,
 *                                    cnan)
 *                 cam_ids_and_points2d.append( (cam_id,observed_2d) )             # <<<<<<<<<<<<<<
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_cam_id);
      __Pyx_GIVEREF(__pyx_v_cam_id);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_cam_id);
      __Pyx_INCREF(__pyx_v_observed_2d);
      __Pyx_GIVEREF(__pyx_v_observed_2d);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_observed_2d);
      __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_cam_ids_and_points2d, __pyx_t_7); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":689
 *                                    cnan)
 *                 cam_ids_and_points2d.append( (cam_id,observed_2d) )
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]             # <<<<<<<<<<<<<<
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
 *                 if debug>2:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_pt_undistorted, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_frame_pt_idx = __pyx_t_6;

      /* "flydra_core/_flydra_tracked_object.pyx":690
 *                 cam_ids_and_points2d.append( (cam_id,observed_2d) )
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )             # <<<<<<<<<<<<<<
 *                 if debug>2:
 *                     print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_camn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_frame_pt_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_21 = PyTuple_New(3); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_7);
      __Pyx_INCREF(__pyx_v_closest_idx);
      __Pyx_GIVEREF(__pyx_v_closest_idx);
      PyTuple_SET_ITEM(__pyx_t_21, 2, __pyx_v_closest_idx);
      __pyx_t_4 = 0;
      __pyx_t_7 = 0;
      __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_used_camns_and_idxs, __pyx_t_21); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":691
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
 *                 if debug>2:             # <<<<<<<<<<<<<<
 *                     print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))
 * 
 */
      __pyx_t_12 = ((__pyx_v_debug > 2) != 0);
      if (__pyx_t_12) {

        /* "flydra_core/_flydra_tracked_object.pyx":692
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
 *                 if debug>2:
 *                     print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))             # <<<<<<<<<<<<<<
 * 
 *         Lcoords = None # default to no line coordinates
 */
        __pyx_t_21 = __Pyx_PyObject_GetSlice(__pyx_v_pt_undistorted, 0, 2, NULL, NULL, &__pyx_slice__15, 0, 1, 1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_21); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_INCREF(__pyx_v_closest_idx);
        __Pyx_GIVEREF(__pyx_v_closest_idx);
        PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_v_closest_idx);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_best_match_idx_d_s, __pyx_t_21); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        if (__Pyx_PrintOne(0, __pyx_t_7) < 0) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "flydra_core/_flydra_tracked_object.pyx":691
 *                 frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
 *                 used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
 *                 if debug>2:             # <<<<<<<<<<<<<<
 *                     print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))
 * 
 */
      }

      /* "flydra_core/_flydra_tracked_object.pyx":663
 *                     print '       (not acceptable)'
 * 
 *             if closest_idx is not None:             # <<<<<<<<<<<<<<
 *                 pt_undistorted, projected_line = candidate_point_list[closest_idx]
 *                 area = pt_undistorted[PT_TUPLE_IDX_AREA]
 */
    }

    /* "flydra_core/_flydra_tracked_object.pyx":531
 *         if debug>2:
 *             print '_filter_data():'
 *         for camn, candidate_point_list in data_dict.items():             # <<<<<<<<<<<<<<
 *             cam_id = camn2cam_id[camn]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":694
 *                     print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))
 * 
 *         Lcoords = None # default to no line coordinates             # <<<<<<<<<<<<<<
 *         # Now cam_ids_and_points2d has just the 2d points we'll use for this reconstruction
 *         if len(cam_ids_and_points2d)==1:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_Lcoords = Py_None;

  /* "flydra_core/_flydra_tracked_object.pyx":696
 *         Lcoords = None # default to no line coordinates
 *         # Now cam_ids_and_points2d has just the 2d points we'll use for this reconstruction
 *         if len(cam_ids_and_points2d)==1:             # <<<<<<<<<<<<<<
 *             # keep 3D "observation" because we need to save 2d observations
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 */
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_cam_ids_and_points2d); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 696, __pyx_L1_error)
  __pyx_t_12 = ((__pyx_t_9 == 1) != 0);
  if (__pyx_t_12) {

    /* "flydra_core/_flydra_tracked_object.pyx":698
 *         if len(cam_ids_and_points2d)==1:
 *             # keep 3D "observation" because we need to save 2d observations
 *             position_MLE = numpy.nan*numpy.ones( (3,))             # <<<<<<<<<<<<<<
 *         elif len(cam_ids_and_points2d)>=2:
 *             if self.reconstructor.wateri is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_numpy); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_n_s_ones); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_21)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_21);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_21, __pyx_tuple__14) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_tuple__14);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_position_MLE = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "flydra_core/_flydra_tracked_object.pyx":696
 *         Lcoords = None # default to no line coordinates
 *         # Now cam_ids_and_points2d has just the 2d points we'll use for this reconstruction
 *         if len(cam_ids_and_points2d)==1:             # <<<<<<<<<<<<<<
 *             # keep 3D "observation" because we need to save 2d observations
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 */
    goto __pyx_L40;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":699
 *             # keep 3D "observation" because we need to save 2d observations
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 *         elif len(cam_ids_and_points2d)>=2:             # <<<<<<<<<<<<<<
 *             if self.reconstructor.wateri is not None:
 *                 # do not attempt this if we have refractive boundary
 */
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_cam_ids_and_points2d); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 699, __pyx_L1_error)
  __pyx_t_12 = ((__pyx_t_9 >= 2) != 0);
  if (__pyx_t_12) {

    /* "flydra_core/_flydra_tracked_object.pyx":700
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 *         elif len(cam_ids_and_points2d)>=2:
 *             if self.reconstructor.wateri is not None:             # <<<<<<<<<<<<<<
 *                 # do not attempt this if we have refractive boundary
 *                 # keep 3D "observation" because we need to save 2d observations
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_wateri); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = (__pyx_t_4 != Py_None);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_18 = (__pyx_t_12 != 0);
    if (__pyx_t_18) {

      /* "flydra_core/_flydra_tracked_object.pyx":703
 *                 # do not attempt this if we have refractive boundary
 *                 # keep 3D "observation" because we need to save 2d observations
 *                 position_MLE = numpy.nan*numpy.ones( (3,))             # <<<<<<<<<<<<<<
 *             else:
 *                 position_MLE, Lcoords = self.reconstructor.find3d(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_nan); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ones); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_21);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_21, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_7, __pyx_tuple__14) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_tuple__14);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_21 = PyNumber_Multiply(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_position_MLE = __pyx_t_21;
      __pyx_t_21 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":700
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 *         elif len(cam_ids_and_points2d)>=2:
 *             if self.reconstructor.wateri is not None:             # <<<<<<<<<<<<<<
 *                 # do not attempt this if we have refractive boundary
 *                 # keep 3D "observation" because we need to save 2d observations
 */
      goto __pyx_L41;
    }

    /* "flydra_core/_flydra_tracked_object.pyx":705
 *                 position_MLE = numpy.nan*numpy.ones( (3,))
 *             else:
 *                 position_MLE, Lcoords = self.reconstructor.find3d(             # <<<<<<<<<<<<<<
 *                     cam_ids_and_points2d, return_line_coords = True,
 *                     orientation_consensus=self.orientation_consensus)
 */
    /*else*/ {
      __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->reconstructor, __pyx_n_s_find3d); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);

      /* "flydra_core/_flydra_tracked_object.pyx":706
 *             else:
 *                 position_MLE, Lcoords = self.reconstructor.find3d(
 *                     cam_ids_and_points2d, return_line_coords = True,             # <<<<<<<<<<<<<<
 *                     orientation_consensus=self.orientation_consensus)
 *         else:
 */
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_cam_ids_and_points2d);
      __Pyx_GIVEREF(__pyx_v_cam_ids_and_points2d);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_cam_ids_and_points2d);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return_line_coords, Py_True) < 0) __PYX_ERR(0, 706, __pyx_L1_error)

      /* "flydra_core/_flydra_tracked_object.pyx":707
 *                 position_MLE, Lcoords = self.reconstructor.find3d(
 *                     cam_ids_and_points2d, return_line_coords = True,
 *                     orientation_consensus=self.orientation_consensus)             # <<<<<<<<<<<<<<
 *         else:
 *             position_MLE = None
 */
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->orientation_consensus); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_orientation_consensus, __pyx_t_7) < 0) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "flydra_core/_flydra_tracked_object.pyx":705
 *                 position_MLE = numpy.nan*numpy.ones( (3,))
 *             else:
 *                 position_MLE, Lcoords = self.reconstructor.find3d(             # <<<<<<<<<<<<<<
 *                     cam_ids_and_points2d, return_line_coords = True,
 *                     orientation_consensus=self.orientation_consensus)
 */
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 705, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_21 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 705, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_21)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_11(__pyx_t_21); if (unlikely(!__pyx_t_2)) goto __pyx_L42_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_4 = __pyx_t_11(__pyx_t_21); if (unlikely(!__pyx_t_4)) goto __pyx_L42_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_21), 2) < 0) __PYX_ERR(0, 705, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        goto __pyx_L43_unpacking_done;
        __pyx_L42_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 705, __pyx_L1_error)
        __pyx_L43_unpacking_done:;
      }
      __pyx_v_position_MLE = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_Lcoords, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L41:;

    /* "flydra_core/_flydra_tracked_object.pyx":699
 *             # keep 3D "observation" because we need to save 2d observations
 *             position_MLE = numpy.nan*numpy.ones( (3,))
 *         elif len(cam_ids_and_points2d)>=2:             # <<<<<<<<<<<<<<
 *             if self.reconstructor.wateri is not None:
 *                 # do not attempt this if we have refractive boundary
 */
    goto __pyx_L40;
  }

  /* "flydra_core/_flydra_tracked_object.pyx":709
 *                     orientation_consensus=self.orientation_consensus)
 *         else:
 *             position_MLE = None             # <<<<<<<<<<<<<<
 *         return (position_MLE, Lcoords, used_camns_and_idxs,
 *                 cam_ids_and_points2d, all_close_camn_pt_idxs)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_position_MLE = Py_None;
  }
  __pyx_L40:;

  /* "flydra_core/_flydra_tracked_object.pyx":710
 *         else:
 *             position_MLE = None
 *         return (position_MLE, Lcoords, used_camns_and_idxs,             # <<<<<<<<<<<<<<
 *                 cam_ids_and_points2d, all_close_camn_pt_idxs)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "flydra_core/_flydra_tracked_object.pyx":711
 *             position_MLE = None
 *         return (position_MLE, Lcoords, used_camns_and_idxs,
 *                 cam_ids_and_points2d, all_close_camn_pt_idxs)             # <<<<<<<<<<<<<<
 * 
 *     def get_most_recent_data(self):
 */
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_position_MLE);
  __Pyx_GIVEREF(__pyx_v_position_MLE);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_position_MLE);
  __Pyx_INCREF(__pyx_v_Lcoords);
  __Pyx_GIVEREF(__pyx_v_Lcoords);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_Lcoords);
  __Pyx_INCREF(__pyx_v_used_camns_and_idxs);
  __Pyx_GIVEREF(__pyx_v_used_camns_and_idxs);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_used_camns_and_idxs);
  __Pyx_INCREF(__pyx_v_cam_ids_and_points2d);
  __Pyx_GIVEREF(__pyx_v_cam_ids_and_points2d);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_cam_ids_and_points2d);
  __Pyx_INCREF(__pyx_v_all_close_camn_pt_idxs);
  __Pyx_GIVEREF(__pyx_v_all_close_camn_pt_idxs);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_all_close_camn_pt_idxs);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":492
 *         self.last_frameno_with_data = self.observations_frames[-1]
 * 
 *     cpdef _filter_data(self, object xhatminus, object Pminus,             # <<<<<<<<<<<<<<
 *                        object data_dict, object camn2cam_id,
 *                        int debug=0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject._filter_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_projected_line);
  __Pyx_XDECREF((PyObject *)__pyx_v_best_3d_location);
  __Pyx_XDECREF(__pyx_v_all_close_camn_pt_idxs);
  __Pyx_XDECREF(__pyx_v_prediction_3d);
  __Pyx_XDECREF((PyObject *)__pyx_v_fast_prediction_3d);
  __Pyx_XDECREF(__pyx_v_pixel_dist_cmp);
  __Pyx_XDECREF(__pyx_v_neg_predicted_3d);
  __Pyx_XDECREF(__pyx_v_cam_ids_and_points2d);
  __Pyx_XDECREF(__pyx_v_used_camns_and_idxs);
  __Pyx_XDECREF(__pyx_v_candidate_point_list);
  __Pyx_XDECREF(__pyx_v_cam_id);
  __Pyx_XDECREF(__pyx_v_predicted_2d_distorted);
  __Pyx_XDECREF(__pyx_v_predicted_2d_undistorted);
  __Pyx_XDECREF(__pyx_v_closest_idx);
  __Pyx_XDECREF(__pyx_v_Pminus_inv);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_pt_undistorted);
  __Pyx_XDECREF(__pyx_v_pt_x_undist);
  __Pyx_XDECREF(__pyx_v_pt_y_undist);
  __Pyx_XDECREF(__pyx_v_pt_x_dist);
  __Pyx_XDECREF(__pyx_v_pt_y_dist);
  __Pyx_XDECREF(__pyx_v_pixel_dist);
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_extra_print);
  __Pyx_XDECREF(__pyx_v_observed_2d);
  __Pyx_XDECREF(__pyx_v_Lcoords);
  __Pyx_XDECREF(__pyx_v_position_MLE);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15_filter_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11flydra_core_22_flydra_tracked_object_13TrackedObject_14_filter_data[] = "given state estimate, select useful incoming data and make new observation\n\n        This function 'solves' the data association problem. 2D\n        observations are associated with a kalman object if they are\n        with some distance between a predicted image of the object on\n        the camera plane. The distance estimation is made in 3D as the\n        Euclidian distance between the ray in 3D space corresponding\n        to the image point observed and the predicted 3D location of\n        the kalman object. (Other distance metrics could also be\n        implemented, such as 2D Euclidian distance on the image\n        plane. Note that a raw threshold for this is implemented with\n        distorted_pixel_euclidian_distance_accept.)\n\n        ";
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15_filter_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_xhatminus = 0;
  PyObject *__pyx_v_Pminus = 0;
  PyObject *__pyx_v_data_dict = 0;
  PyObject *__pyx_v_camn2cam_id = 0;
  int __pyx_v_debug;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_filter_data (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_xhatminus,&__pyx_n_s_Pminus,&__pyx_n_s_data_dict,&__pyx_n_s_camn2cam_id,&__pyx_n_s_debug,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_xhatminus)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Pminus)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_filter_data", 0, 4, 5, 1); __PYX_ERR(0, 492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_filter_data", 0, 4, 5, 2); __PYX_ERR(0, 492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_camn2cam_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_filter_data", 0, 4, 5, 3); __PYX_ERR(0, 492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_debug);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_filter_data") < 0)) __PYX_ERR(0, 492, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_xhatminus = values[0];
    __pyx_v_Pminus = values[1];
    __pyx_v_data_dict = values[2];
    __pyx_v_camn2cam_id = values[3];
    if (values[4]) {
      __pyx_v_debug = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_debug == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L3_error)
    } else {
      __pyx_v_debug = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_filter_data", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 492, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject._filter_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_14_filter_data(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), __pyx_v_xhatminus, __pyx_v_Pminus, __pyx_v_data_dict, __pyx_v_camn2cam_id, __pyx_v_debug);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_14_filter_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v_xhatminus, PyObject *__pyx_v_Pminus, PyObject *__pyx_v_data_dict, PyObject *__pyx_v_camn2cam_id, int __pyx_v_debug) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_filter_data", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.debug = __pyx_v_debug;
  __pyx_t_1 = __pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject->_filter_data(__pyx_v_self, __pyx_v_xhatminus, __pyx_v_Pminus, __pyx_v_data_dict, __pyx_v_camn2cam_id, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject._filter_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":713
 *                 cam_ids_and_points2d, all_close_camn_pt_idxs)
 * 
 *     def get_most_recent_data(self):             # <<<<<<<<<<<<<<
 *         if not len(self.xhats):
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_17get_most_recent_data(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_17get_most_recent_data(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_most_recent_data (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_16get_most_recent_data(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_16get_most_recent_data(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_v_xhat = NULL;
  PyObject *__pyx_v_P = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_most_recent_data", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":714
 * 
 *     def get_most_recent_data(self):
 *         if not len(self.xhats):             # <<<<<<<<<<<<<<
 *             return
 *         xhat = self.xhats[-1]
 */
  __pyx_t_1 = __pyx_v_self->xhats;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "flydra_core/_flydra_tracked_object.pyx":715
 *     def get_most_recent_data(self):
 *         if not len(self.xhats):
 *             return             # <<<<<<<<<<<<<<
 *         xhat = self.xhats[-1]
 *         P = self.Ps[-1]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "flydra_core/_flydra_tracked_object.pyx":714
 * 
 *     def get_most_recent_data(self):
 *         if not len(self.xhats):             # <<<<<<<<<<<<<<
 *             return
 *         xhat = self.xhats[-1]
 */
  }

  /* "flydra_core/_flydra_tracked_object.pyx":716
 *         if not len(self.xhats):
 *             return
 *         xhat = self.xhats[-1]             # <<<<<<<<<<<<<<
 *         P = self.Ps[-1]
 *         return self.obj_id, xhat,P
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->xhats, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_xhat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":717
 *             return
 *         xhat = self.xhats[-1]
 *         P = self.Ps[-1]             # <<<<<<<<<<<<<<
 *         return self.obj_id, xhat,P
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->Ps, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_P = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":718
 *         xhat = self.xhats[-1]
 *         P = self.Ps[-1]
 *         return self.obj_id, xhat,P             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->obj_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_xhat);
  __Pyx_GIVEREF(__pyx_v_xhat);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_xhat);
  __Pyx_INCREF(__pyx_v_P);
  __Pyx_GIVEREF(__pyx_v_P);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_P);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "flydra_core/_flydra_tracked_object.pyx":713
 *                 cam_ids_and_points2d, all_close_camn_pt_idxs)
 * 
 *     def get_most_recent_data(self):             # <<<<<<<<<<<<<<
 *         if not len(self.xhats):
 *             return
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.get_most_recent_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xhat);
  __Pyx_XDECREF(__pyx_v_P);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":141
 * 
 *     """
 *     cdef readonly long current_frameno             # <<<<<<<<<<<<<<
 *     cdef readonly long last_frameno_with_data
 *     cdef long long max_frames_skipped
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->current_frameno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.current_frameno.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":142
 *     """
 *     cdef readonly long current_frameno
 *     cdef readonly long last_frameno_with_data             # <<<<<<<<<<<<<<
 *     cdef long long max_frames_skipped
 *     cdef readonly mybool kill_me
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->last_frameno_with_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.last_frameno_with_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":144
 *     cdef readonly long last_frameno_with_data
 *     cdef long long max_frames_skipped
 *     cdef readonly mybool kill_me             # <<<<<<<<<<<<<<
 *     cdef mybool save_all_data
 *     cdef double area_threshold, area_threshold_for_orientation
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->kill_me); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.kill_me.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":153
 *     cdef object ekf_observation_covariance_pixels
 * 
 *     cdef readonly object frames, xhats, timestamps, Ps, MLE_position, MLE_Lcoords             # <<<<<<<<<<<<<<
 *     cdef readonly object observations_frames, observations_2d
 *     cdef int disable_image_stat_gating, orientation_consensus
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->frames);
  __pyx_r = __pyx_v_self->frames;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->xhats);
  __pyx_r = __pyx_v_self->xhats;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->timestamps);
  __pyx_r = __pyx_v_self->timestamps;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->Ps);
  __pyx_r = __pyx_v_self->Ps;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->MLE_position);
  __pyx_r = __pyx_v_self->MLE_position;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->MLE_Lcoords);
  __pyx_r = __pyx_v_self->MLE_Lcoords;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":154
 * 
 *     cdef readonly object frames, xhats, timestamps, Ps, MLE_position, MLE_Lcoords
 *     cdef readonly object observations_frames, observations_2d             # <<<<<<<<<<<<<<
 *     cdef int disable_image_stat_gating, orientation_consensus
 *     cdef object fake_timestamp
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->observations_frames);
  __pyx_r = __pyx_v_self->observations_frames;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->observations_2d);
  __pyx_r = __pyx_v_self->observations_2d;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flydra_core/_flydra_tracked_object.pyx":157
 *     cdef int disable_image_stat_gating, orientation_consensus
 *     cdef object fake_timestamp
 *     cdef readonly unsigned int obj_id             # <<<<<<<<<<<<<<
 *     cdef object ekf_kalman_A, ekf_kalman_Q
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id___get__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id___get__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->obj_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.obj_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_18__reduce_cython__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_18__reduce_cython__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.MLE_Lcoords, self.MLE_position, self.Ps, self.area_threshold, self.area_threshold_for_orientation, self.current_frameno, self.disable_image_stat_gating, self.distorted_pixel_euclidian_distance_accept, self.ekf_kalman_A, self.ekf_kalman_Q, self.ekf_observation_covariance_pixels, self.fake_timestamp, self.frames, self.kill_me, self.last_frameno_with_data, self.max_frames_skipped, self.max_variance, self.my_kalman, self.obj_id, self.observations_2d, self.observations_frames, self.orientation_consensus, self.reconstructor, self.save_all_data, self.timestamps, self.xhats)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->area_threshold); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->area_threshold_for_orientation); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_self->current_frameno); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->disable_image_stat_gating); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->kill_me); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_v_self->last_frameno_with_data); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->max_frames_skipped); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_self->max_variance); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->obj_id); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->orientation_consensus); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->save_all_data); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(26); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_v_self->MLE_Lcoords);
  __Pyx_GIVEREF(__pyx_v_self->MLE_Lcoords);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_self->MLE_Lcoords);
  __Pyx_INCREF(__pyx_v_self->MLE_position);
  __Pyx_GIVEREF(__pyx_v_self->MLE_position);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_self->MLE_position);
  __Pyx_INCREF(__pyx_v_self->Ps);
  __Pyx_GIVEREF(__pyx_v_self->Ps);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_self->Ps);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_12, 6, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->distorted_pixel_euclidian_distance_accept);
  __Pyx_GIVEREF(__pyx_v_self->distorted_pixel_euclidian_distance_accept);
  PyTuple_SET_ITEM(__pyx_t_12, 7, __pyx_v_self->distorted_pixel_euclidian_distance_accept);
  __Pyx_INCREF(__pyx_v_self->ekf_kalman_A);
  __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_A);
  PyTuple_SET_ITEM(__pyx_t_12, 8, __pyx_v_self->ekf_kalman_A);
  __Pyx_INCREF(__pyx_v_self->ekf_kalman_Q);
  __Pyx_GIVEREF(__pyx_v_self->ekf_kalman_Q);
  PyTuple_SET_ITEM(__pyx_t_12, 9, __pyx_v_self->ekf_kalman_Q);
  __Pyx_INCREF(__pyx_v_self->ekf_observation_covariance_pixels);
  __Pyx_GIVEREF(__pyx_v_self->ekf_observation_covariance_pixels);
  PyTuple_SET_ITEM(__pyx_t_12, 10, __pyx_v_self->ekf_observation_covariance_pixels);
  __Pyx_INCREF(__pyx_v_self->fake_timestamp);
  __Pyx_GIVEREF(__pyx_v_self->fake_timestamp);
  PyTuple_SET_ITEM(__pyx_t_12, 11, __pyx_v_self->fake_timestamp);
  __Pyx_INCREF(__pyx_v_self->frames);
  __Pyx_GIVEREF(__pyx_v_self->frames);
  PyTuple_SET_ITEM(__pyx_t_12, 12, __pyx_v_self->frames);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_12, 13, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_12, 14, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_12, 15, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_12, 16, __pyx_t_8);
  __Pyx_INCREF(__pyx_v_self->my_kalman);
  __Pyx_GIVEREF(__pyx_v_self->my_kalman);
  PyTuple_SET_ITEM(__pyx_t_12, 17, __pyx_v_self->my_kalman);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_12, 18, __pyx_t_9);
  __Pyx_INCREF(__pyx_v_self->observations_2d);
  __Pyx_GIVEREF(__pyx_v_self->observations_2d);
  PyTuple_SET_ITEM(__pyx_t_12, 19, __pyx_v_self->observations_2d);
  __Pyx_INCREF(__pyx_v_self->observations_frames);
  __Pyx_GIVEREF(__pyx_v_self->observations_frames);
  PyTuple_SET_ITEM(__pyx_t_12, 20, __pyx_v_self->observations_frames);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_12, 21, __pyx_t_10);
  __Pyx_INCREF(__pyx_v_self->reconstructor);
  __Pyx_GIVEREF(__pyx_v_self->reconstructor);
  PyTuple_SET_ITEM(__pyx_t_12, 22, __pyx_v_self->reconstructor);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_12, 23, __pyx_t_11);
  __Pyx_INCREF(__pyx_v_self->timestamps);
  __Pyx_GIVEREF(__pyx_v_self->timestamps);
  PyTuple_SET_ITEM(__pyx_t_12, 24, __pyx_v_self->timestamps);
  __Pyx_INCREF(__pyx_v_self->xhats);
  __Pyx_GIVEREF(__pyx_v_self->xhats);
  PyTuple_SET_ITEM(__pyx_t_12, 25, __pyx_v_self->xhats);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.MLE_Lcoords, self.MLE_position, self.Ps, self.area_threshold, self.area_threshold_for_orientation, self.current_frameno, self.disable_image_stat_gating, self.distorted_pixel_euclidian_distance_accept, self.ekf_kalman_A, self.ekf_kalman_Q, self.ekf_observation_covariance_pixels, self.fake_timestamp, self.frames, self.kill_me, self.last_frameno_with_data, self.max_frames_skipped, self.max_variance, self.my_kalman, self.obj_id, self.observations_2d, self.observations_frames, self.orientation_consensus, self.reconstructor, self.save_all_data, self.timestamps, self.xhats)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_12 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_v__dict = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "(tree fragment)":7
 *     state = (self.MLE_Lcoords, self.MLE_position, self.Ps, self.area_threshold, self.area_threshold_for_orientation, self.current_frameno, self.disable_image_stat_gating, self.distorted_pixel_euclidian_distance_accept, self.ekf_kalman_A, self.ekf_kalman_Q, self.ekf_observation_covariance_pixels, self.fake_timestamp, self.frames, self.kill_me, self.last_frameno_with_data, self.max_frames_skipped, self.max_variance, self.my_kalman, self.obj_id, self.observations_2d, self.observations_frames, self.orientation_consensus, self.reconstructor, self.save_all_data, self.timestamps, self.xhats)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_13 = (__pyx_v__dict != Py_None);
  __pyx_t_14 = (__pyx_t_13 != 0);
  if (__pyx_t_14) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v__dict);
    __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_11));
    __pyx_t_11 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.MLE_Lcoords is not None or self.MLE_position is not None or self.Ps is not None or self.distorted_pixel_euclidian_distance_accept is not None or self.ekf_kalman_A is not None or self.ekf_kalman_Q is not None or self.ekf_observation_covariance_pixels is not None or self.fake_timestamp is not None or self.frames is not None or self.my_kalman is not None or self.observations_2d is not None or self.observations_frames is not None or self.reconstructor is not None or self.timestamps is not None or self.xhats is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.MLE_Lcoords, self.MLE_position, self.Ps, self.area_threshold, self.area_threshold_for_orientation, self.current_frameno, self.disable_image_stat_gating, self.distorted_pixel_euclidian_distance_accept, self.ekf_kalman_A, self.ekf_kalman_Q, self.ekf_observation_covariance_pixels, self.fake_timestamp, self.frames, self.kill_me, self.last_frameno_with_data, self.max_frames_skipped, self.max_variance, self.my_kalman, self.obj_id, self.observations_2d, self.observations_frames, self.orientation_consensus, self.reconstructor, self.save_all_data, self.timestamps, self.xhats)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.MLE_Lcoords is not None or self.MLE_position is not None or self.Ps is not None or self.distorted_pixel_euclidian_distance_accept is not None or self.ekf_kalman_A is not None or self.ekf_kalman_Q is not None or self.ekf_observation_covariance_pixels is not None or self.fake_timestamp is not None or self.frames is not None or self.my_kalman is not None or self.observations_2d is not None or self.observations_frames is not None or self.reconstructor is not None or self.timestamps is not None or self.xhats is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, None), state
 */
  /*else*/ {
    __pyx_t_13 = (__pyx_v_self->MLE_Lcoords != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->MLE_position != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->Ps != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->distorted_pixel_euclidian_distance_accept != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->ekf_kalman_A != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->ekf_kalman_Q != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->ekf_observation_covariance_pixels != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->fake_timestamp != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->frames != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->my_kalman != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->observations_2d != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->observations_frames != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->reconstructor != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->timestamps != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->xhats != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    __pyx_t_14 = __pyx_t_15;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_14;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.MLE_Lcoords is not None or self.MLE_position is not None or self.Ps is not None or self.distorted_pixel_euclidian_distance_accept is not None or self.ekf_kalman_A is not None or self.ekf_kalman_Q is not None or self.ekf_observation_covariance_pixels is not None or self.fake_timestamp is not None or self.frames is not None or self.my_kalman is not None or self.observations_2d is not None or self.observations_frames is not None or self.reconstructor is not None or self.timestamps is not None or self.xhats is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, None), state
 *     else:
 */
  __pyx_t_14 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_14) {

    /* "(tree fragment)":13
 *         use_setstate = self.MLE_Lcoords is not None or self.MLE_position is not None or self.Ps is not None or self.distorted_pixel_euclidian_distance_accept is not None or self.ekf_kalman_A is not None or self.ekf_kalman_Q is not None or self.ekf_observation_covariance_pixels is not None or self.fake_timestamp is not None or self.frames is not None or self.my_kalman is not None or self.observations_2d is not None or self.observations_frames is not None or self.reconstructor is not None or self.timestamps is not None or self.xhats is not None
 *     if use_setstate:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_pyx_unpickle_TrackedObject); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_224123000);
    __Pyx_GIVEREF(__pyx_int_224123000);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_224123000);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_12, 2, Py_None);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_12);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state);
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.MLE_Lcoords is not None or self.MLE_position is not None or self.Ps is not None or self.distorted_pixel_euclidian_distance_accept is not None or self.ekf_kalman_A is not None or self.ekf_kalman_Q is not None or self.ekf_observation_covariance_pixels is not None or self.fake_timestamp is not None or self.frames is not None or self.my_kalman is not None or self.observations_2d is not None or self.observations_frames is not None or self.reconstructor is not None or self.timestamps is not None or self.xhats is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, None), state
 *     else:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TrackedObject__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pyx_unpickle_TrackedObject); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_224123000);
    __Pyx_GIVEREF(__pyx_int_224123000);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_224123000);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_state);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_12);
    __pyx_t_10 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TrackedObject__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_20__setstate_cython__(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_13TrackedObject_20__setstate_cython__(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TrackedObject__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_11flydra_core_22_flydra_tracked_object___pyx_unpickle_TrackedObject__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TrackedObject, (type(self), 0xd5bd878, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TrackedObject__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.TrackedObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TrackedObject(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_5__pyx_unpickle_TrackedObject(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_11flydra_core_22_flydra_tracked_object_5__pyx_unpickle_TrackedObject = {"__pyx_unpickle_TrackedObject", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_5__pyx_unpickle_TrackedObject, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_11flydra_core_22_flydra_tracked_object_5__pyx_unpickle_TrackedObject(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TrackedObject (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TrackedObject", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TrackedObject", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_TrackedObject") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TrackedObject", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.__pyx_unpickle_TrackedObject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11flydra_core_22_flydra_tracked_object_4__pyx_unpickle_TrackedObject(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11flydra_core_22_flydra_tracked_object_4__pyx_unpickle_TrackedObject(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TrackedObject", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd5bd878:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd5bd878) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd5bd878:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 *     __pyx_result = TrackedObject.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd5bd878:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = TrackedObject.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd5, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd5bd878:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 *     __pyx_result = TrackedObject.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_11flydra_core_22_flydra_tracked_object_TrackedObject), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 *     __pyx_result = TrackedObject.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = TrackedObject.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_11flydra_core_22_flydra_tracked_object___pyx_unpickle_TrackedObject__set_state(((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd5bd878 = (MLE_Lcoords, MLE_position, Ps, area_threshold, area_threshold_for_orientation, current_frameno, disable_image_stat_gating, distorted_pixel_euclidian_distance_accept, ekf_kalman_A, ekf_kalman_Q, ekf_observation_covariance_pixels, fake_timestamp, frames, kill_me, last_frameno_with_data, max_frames_skipped, max_variance, my_kalman, obj_id, observations_2d, observations_frames, orientation_consensus, reconstructor, save_all_data, timestamps, xhats))" % __pyx_checksum)
 *     __pyx_result = TrackedObject.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TrackedObject(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.__pyx_unpickle_TrackedObject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_11flydra_core_22_flydra_tracked_object___pyx_unpickle_TrackedObject__set_state(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  long __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_11flydra_core_22_flydra_tracked_object_mybool __pyx_t_5;
  PY_LONG_LONG __pyx_t_6;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TrackedObject__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[26])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->MLE_Lcoords);
  __Pyx_DECREF(__pyx_v___pyx_result->MLE_Lcoords);
  __pyx_v___pyx_result->MLE_Lcoords = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->MLE_position);
  __Pyx_DECREF(__pyx_v___pyx_result->MLE_position);
  __pyx_v___pyx_result->MLE_position = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->Ps);
  __Pyx_DECREF(__pyx_v___pyx_result->Ps);
  __pyx_v___pyx_result->Ps = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->area_threshold = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->area_threshold_for_orientation = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->current_frameno = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->disable_image_stat_gating = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->distorted_pixel_euclidian_distance_accept);
  __Pyx_DECREF(__pyx_v___pyx_result->distorted_pixel_euclidian_distance_accept);
  __pyx_v___pyx_result->distorted_pixel_euclidian_distance_accept = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->ekf_kalman_A);
  __Pyx_DECREF(__pyx_v___pyx_result->ekf_kalman_A);
  __pyx_v___pyx_result->ekf_kalman_A = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->ekf_kalman_Q);
  __Pyx_DECREF(__pyx_v___pyx_result->ekf_kalman_Q);
  __pyx_v___pyx_result->ekf_kalman_Q = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->ekf_observation_covariance_pixels);
  __Pyx_DECREF(__pyx_v___pyx_result->ekf_observation_covariance_pixels);
  __pyx_v___pyx_result->ekf_observation_covariance_pixels = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fake_timestamp);
  __Pyx_DECREF(__pyx_v___pyx_result->fake_timestamp);
  __pyx_v___pyx_result->fake_timestamp = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->frames);
  __Pyx_DECREF(__pyx_v___pyx_result->frames);
  __pyx_v___pyx_result->frames = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->kill_me = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->last_frameno_with_data = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_1); if (unlikely((__pyx_t_6 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_frames_skipped = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_variance = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 17, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->my_kalman);
  __Pyx_DECREF(__pyx_v___pyx_result->my_kalman);
  __pyx_v___pyx_result->my_kalman = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 18, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_unsigned_int(__pyx_t_1); if (unlikely((__pyx_t_7 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->obj_id = __pyx_t_7;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 19, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->observations_2d);
  __Pyx_DECREF(__pyx_v___pyx_result->observations_2d);
  __pyx_v___pyx_result->observations_2d = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 20, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->observations_frames);
  __Pyx_DECREF(__pyx_v___pyx_result->observations_frames);
  __pyx_v___pyx_result->observations_frames = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 21, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->orientation_consensus = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 22, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->reconstructor);
  __Pyx_DECREF(__pyx_v___pyx_result->reconstructor);
  __pyx_v___pyx_result->reconstructor = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 23, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->save_all_data = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 24, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->timestamps);
  __Pyx_DECREF(__pyx_v___pyx_result->timestamps);
  __pyx_v___pyx_result->timestamps = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 25, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->xhats);
  __Pyx_DECREF(__pyx_v___pyx_result->xhats);
  __pyx_v___pyx_result->xhats = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[26])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_9 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_10 = ((__pyx_t_9 > 26) != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_8 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_11 = (__pyx_t_10 != 0);
  __pyx_t_8 = __pyx_t_11;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_8) {

    /* "(tree fragment)":14
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[26])             # <<<<<<<<<<<<<<
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_update); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_12 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 26, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[26])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TrackedObject__set_state(<TrackedObject> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TrackedObject__set_state(TrackedObject __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.MLE_Lcoords = __pyx_state[0]; __pyx_result.MLE_position = __pyx_state[1]; __pyx_result.Ps = __pyx_state[2]; __pyx_result.area_threshold = __pyx_state[3]; __pyx_result.area_threshold_for_orientation = __pyx_state[4]; __pyx_result.current_frameno = __pyx_state[5]; __pyx_result.disable_image_stat_gating = __pyx_state[6]; __pyx_result.distorted_pixel_euclidian_distance_accept = __pyx_state[7]; __pyx_result.ekf_kalman_A = __pyx_state[8]; __pyx_result.ekf_kalman_Q = __pyx_state[9]; __pyx_result.ekf_observation_covariance_pixels = __pyx_state[10]; __pyx_result.fake_timestamp = __pyx_state[11]; __pyx_result.frames = __pyx_state[12]; __pyx_result.kill_me = __pyx_state[13]; __pyx_result.last_frameno_with_data = __pyx_state[14]; __pyx_result.max_frames_skipped = __pyx_state[15]; __pyx_result.max_variance = __pyx_state[16]; __pyx_result.my_kalman = __pyx_state[17]; __pyx_result.obj_id = __pyx_state[18]; __pyx_result.observations_2d = __pyx_state[19]; __pyx_result.observations_frames = __pyx_state[20]; __pyx_result.orientation_consensus = __pyx_state[21]; __pyx_result.reconstructor = __pyx_state[22]; __pyx_result.save_all_data = __pyx_state[23]; __pyx_result.timestamps = __pyx_state[24]; __pyx_result.xhats = __pyx_state[25]
 *     if len(__pyx_state) > 26 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("flydra_core._flydra_tracked_object.__pyx_unpickle_TrackedObject__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":931
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":932
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":933
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":931
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":935
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":936
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":937
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":938
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":937
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":939
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":935
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":943
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":944
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":945
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 945, __pyx_L3_error)

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":944
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":946
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 946, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":947
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 947, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 947, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":944
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":943
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":949
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":950
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":951
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 951, __pyx_L3_error)

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":950
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":952
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 952, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":953
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 953, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 953, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":950
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":949
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":955
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":956
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":957
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 957, __pyx_L3_error)

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":956
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":958
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 958, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":959
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 959, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 959, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":956
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":955
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":969
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":981
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":969
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":984
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":996
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":984
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":999
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1006
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":999
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1009
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1013
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1009
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1016
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1020
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1016
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}
static struct __pyx_vtabstruct_11flydra_core_22_flydra_tracked_object_TrackedObject __pyx_vtable_11flydra_core_22_flydra_tracked_object_TrackedObject;

static PyObject *__pyx_tp_new_11flydra_core_22_flydra_tracked_object_TrackedObject(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)o);
  p->__pyx_vtab = __pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject;
  p->reconstructor = Py_None; Py_INCREF(Py_None);
  p->my_kalman = Py_None; Py_INCREF(Py_None);
  p->distorted_pixel_euclidian_distance_accept = Py_None; Py_INCREF(Py_None);
  p->ekf_observation_covariance_pixels = Py_None; Py_INCREF(Py_None);
  p->frames = Py_None; Py_INCREF(Py_None);
  p->xhats = Py_None; Py_INCREF(Py_None);
  p->timestamps = Py_None; Py_INCREF(Py_None);
  p->Ps = Py_None; Py_INCREF(Py_None);
  p->MLE_position = Py_None; Py_INCREF(Py_None);
  p->MLE_Lcoords = Py_None; Py_INCREF(Py_None);
  p->observations_frames = Py_None; Py_INCREF(Py_None);
  p->observations_2d = Py_None; Py_INCREF(Py_None);
  p->fake_timestamp = Py_None; Py_INCREF(Py_None);
  p->ekf_kalman_A = Py_None; Py_INCREF(Py_None);
  p->ekf_kalman_Q = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_11flydra_core_22_flydra_tracked_object_TrackedObject(PyObject *o) {
  struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *p = (struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reconstructor);
  Py_CLEAR(p->my_kalman);
  Py_CLEAR(p->distorted_pixel_euclidian_distance_accept);
  Py_CLEAR(p->ekf_observation_covariance_pixels);
  Py_CLEAR(p->frames);
  Py_CLEAR(p->xhats);
  Py_CLEAR(p->timestamps);
  Py_CLEAR(p->Ps);
  Py_CLEAR(p->MLE_position);
  Py_CLEAR(p->MLE_Lcoords);
  Py_CLEAR(p->observations_frames);
  Py_CLEAR(p->observations_2d);
  Py_CLEAR(p->fake_timestamp);
  Py_CLEAR(p->ekf_kalman_A);
  Py_CLEAR(p->ekf_kalman_Q);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_11flydra_core_22_flydra_tracked_object_TrackedObject(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *p = (struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)o;
  if (p->reconstructor) {
    e = (*v)(p->reconstructor, a); if (e) return e;
  }
  if (p->my_kalman) {
    e = (*v)(p->my_kalman, a); if (e) return e;
  }
  if (p->distorted_pixel_euclidian_distance_accept) {
    e = (*v)(p->distorted_pixel_euclidian_distance_accept, a); if (e) return e;
  }
  if (p->ekf_observation_covariance_pixels) {
    e = (*v)(p->ekf_observation_covariance_pixels, a); if (e) return e;
  }
  if (p->frames) {
    e = (*v)(p->frames, a); if (e) return e;
  }
  if (p->xhats) {
    e = (*v)(p->xhats, a); if (e) return e;
  }
  if (p->timestamps) {
    e = (*v)(p->timestamps, a); if (e) return e;
  }
  if (p->Ps) {
    e = (*v)(p->Ps, a); if (e) return e;
  }
  if (p->MLE_position) {
    e = (*v)(p->MLE_position, a); if (e) return e;
  }
  if (p->MLE_Lcoords) {
    e = (*v)(p->MLE_Lcoords, a); if (e) return e;
  }
  if (p->observations_frames) {
    e = (*v)(p->observations_frames, a); if (e) return e;
  }
  if (p->observations_2d) {
    e = (*v)(p->observations_2d, a); if (e) return e;
  }
  if (p->fake_timestamp) {
    e = (*v)(p->fake_timestamp, a); if (e) return e;
  }
  if (p->ekf_kalman_A) {
    e = (*v)(p->ekf_kalman_A, a); if (e) return e;
  }
  if (p->ekf_kalman_Q) {
    e = (*v)(p->ekf_kalman_Q, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_11flydra_core_22_flydra_tracked_object_TrackedObject(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *p = (struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *)o;
  tmp = ((PyObject*)p->reconstructor);
  p->reconstructor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->my_kalman);
  p->my_kalman = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->distorted_pixel_euclidian_distance_accept);
  p->distorted_pixel_euclidian_distance_accept = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ekf_observation_covariance_pixels);
  p->ekf_observation_covariance_pixels = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->frames);
  p->frames = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->xhats);
  p->xhats = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->timestamps);
  p->timestamps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ps);
  p->Ps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->MLE_position);
  p->MLE_position = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->MLE_Lcoords);
  p->MLE_Lcoords = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->observations_frames);
  p->observations_frames = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->observations_2d);
  p->observations_2d = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fake_timestamp);
  p->fake_timestamp = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ekf_kalman_A);
  p->ekf_kalman_A = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ekf_kalman_Q);
  p->ekf_kalman_Q = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_current_frameno(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15current_frameno_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_last_frameno_with_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_22last_frameno_with_data_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_kill_me(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7kill_me_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_frames(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6frames_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_xhats(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5xhats_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_timestamps(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_10timestamps_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_Ps(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_2Ps_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_MLE_position(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_12MLE_position_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_MLE_Lcoords(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11MLE_Lcoords_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_observations_frames(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19observations_frames_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_observations_2d(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15observations_2d_1__get__(o);
}

static PyObject *__pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_obj_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_6obj_id_1__get__(o);
}

static PyMethodDef __pyx_methods_11flydra_core_22_flydra_tracked_object_TrackedObject[] = {
  {"debug_info", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_5debug_info, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_distance_and_nsigma", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_7get_distance_and_nsigma, METH_O, 0},
  {"kill", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_9kill, METH_NOARGS, 0},
  {"calculate_a_posteriori_estimate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_11calculate_a_posteriori_estimate, METH_VARARGS|METH_KEYWORDS, 0},
  {"remove_previous_observation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_13remove_previous_observation, METH_VARARGS|METH_KEYWORDS, 0},
  {"_filter_data", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_15_filter_data, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11flydra_core_22_flydra_tracked_object_13TrackedObject_14_filter_data},
  {"get_most_recent_data", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_17get_most_recent_data, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_11flydra_core_22_flydra_tracked_object_TrackedObject[] = {
  {(char *)"current_frameno", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_current_frameno, 0, (char *)0, 0},
  {(char *)"last_frameno_with_data", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_last_frameno_with_data, 0, (char *)0, 0},
  {(char *)"kill_me", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_kill_me, 0, (char *)0, 0},
  {(char *)"frames", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_frames, 0, (char *)0, 0},
  {(char *)"xhats", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_xhats, 0, (char *)0, 0},
  {(char *)"timestamps", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_timestamps, 0, (char *)0, 0},
  {(char *)"Ps", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_Ps, 0, (char *)0, 0},
  {(char *)"MLE_position", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_MLE_position, 0, (char *)0, 0},
  {(char *)"MLE_Lcoords", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_MLE_Lcoords, 0, (char *)0, 0},
  {(char *)"observations_frames", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_observations_frames, 0, (char *)0, 0},
  {(char *)"observations_2d", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_observations_2d, 0, (char *)0, 0},
  {(char *)"obj_id", __pyx_getprop_11flydra_core_22_flydra_tracked_object_13TrackedObject_obj_id, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject = {
  PyVarObject_HEAD_INIT(0, 0)
  "flydra_core._flydra_tracked_object.TrackedObject", /*tp_name*/
  sizeof(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Track one object using a Kalman filter.\n\n    ", /*tp_doc*/
  __pyx_tp_traverse_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_traverse*/
  __pyx_tp_clear_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_11flydra_core_22_flydra_tracked_object_13TrackedObject_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_11flydra_core_22_flydra_tracked_object_TrackedObject, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"evaluate_pmat_jacobian", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_11flydra_core_22_flydra_tracked_object_1evaluate_pmat_jacobian, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__flydra_tracked_object(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__flydra_tracked_object},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_flydra_tracked_object",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_n_s_EKF, __pyx_k_EKF, sizeof(__pyx_k_EKF), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd5, __pyx_k_Incompatible_checksums_s_vs_0xd5, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd5), 0, 0, 1, 0},
  {&__pyx_n_s_KalmanFilter, __pyx_k_KalmanFilter, sizeof(__pyx_k_KalmanFilter), 0, 0, 1, 1},
  {&__pyx_kp_s_Lcoords_3D_body_orientation_s, __pyx_k_Lcoords_3D_body_orientation_s, sizeof(__pyx_k_Lcoords_3D_body_orientation_s), 0, 0, 1, 0},
  {&__pyx_n_s_NO_LCOORDS, __pyx_k_NO_LCOORDS, sizeof(__pyx_k_NO_LCOORDS), 0, 0, 1, 1},
  {&__pyx_n_s_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_AREA, __pyx_k_PT_TUPLE_IDX_AREA, sizeof(__pyx_k_PT_TUPLE_IDX_AREA), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX, __pyx_k_PT_TUPLE_IDX_CUR_VAL_IDX, sizeof(__pyx_k_PT_TUPLE_IDX_CUR_VAL_IDX), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_ECCENTRICITY, __pyx_k_PT_TUPLE_IDX_ECCENTRICITY, sizeof(__pyx_k_PT_TUPLE_IDX_ECCENTRICITY), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX, __pyx_k_PT_TUPLE_IDX_FRAME_PT_IDX, sizeof(__pyx_k_PT_TUPLE_IDX_FRAME_PT_IDX), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX, __pyx_k_PT_TUPLE_IDX_MEAN_VAL_IDX, sizeof(__pyx_k_PT_TUPLE_IDX_MEAN_VAL_IDX), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_P1, __pyx_k_PT_TUPLE_IDX_P1, sizeof(__pyx_k_PT_TUPLE_IDX_P1), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_P2, __pyx_k_PT_TUPLE_IDX_P2, sizeof(__pyx_k_PT_TUPLE_IDX_P2), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_P3, __pyx_k_PT_TUPLE_IDX_P3, sizeof(__pyx_k_PT_TUPLE_IDX_P3), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_P4, __pyx_k_PT_TUPLE_IDX_P4, sizeof(__pyx_k_PT_TUPLE_IDX_P4), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_SLOPE, __pyx_k_PT_TUPLE_IDX_SLOPE, sizeof(__pyx_k_PT_TUPLE_IDX_SLOPE), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX, __pyx_k_PT_TUPLE_IDX_SUMSQF_VAL_IDX, sizeof(__pyx_k_PT_TUPLE_IDX_SUMSQF_VAL_IDX), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_X, __pyx_k_PT_TUPLE_IDX_X, sizeof(__pyx_k_PT_TUPLE_IDX_X), 0, 0, 1, 1},
  {&__pyx_n_s_PT_TUPLE_IDX_Y, __pyx_k_PT_TUPLE_IDX_Y, sizeof(__pyx_k_PT_TUPLE_IDX_Y), 0, 0, 1, 1},
  {&__pyx_n_s_P_k1, __pyx_k_P_k1, sizeof(__pyx_k_P_k1), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pmean, __pyx_k_Pmean, sizeof(__pyx_k_Pmean), 0, 0, 1, 1},
  {&__pyx_n_s_Pminus, __pyx_k_Pminus, sizeof(__pyx_k_Pminus), 0, 0, 1, 1},
  {&__pyx_n_s_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 0, 1, 1},
  {&__pyx_n_s_R, __pyx_k_R, sizeof(__pyx_k_R), 0, 0, 1, 1},
  {&__pyx_kp_s_TRO_frames_0_r_observations_2d, __pyx_k_TRO_frames_0_r_observations_2d, sizeof(__pyx_k_TRO_frames_0_r_observations_2d), 0, 0, 1, 0},
  {&__pyx_n_s_ThreeTuple, __pyx_k_ThreeTuple, sizeof(__pyx_k_ThreeTuple), 0, 0, 1, 1},
  {&__pyx_n_s_TrackedObject, __pyx_k_TrackedObject, sizeof(__pyx_k_TrackedObject), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 1, 0},
  {&__pyx_kp_s__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 0, 1, 0},
  {&__pyx_kp_s__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 0, 1, 0},
  {&__pyx_kp_s__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 1, 0},
  {&__pyx_kp_s__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 0, 1, 0},
  {&__pyx_kp_s__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 0, 1, 0},
  {&__pyx_n_s__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 1, 1},
  {&__pyx_kp_s_acceptable, __pyx_k_acceptable, sizeof(__pyx_k_acceptable), 0, 0, 1, 0},
  {&__pyx_n_s_adskalman_adskalman, __pyx_k_adskalman_adskalman, sizeof(__pyx_k_adskalman_adskalman), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_area_threshold, __pyx_k_area_threshold, sizeof(__pyx_k_area_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_area_threshold_for_orientation, __pyx_k_area_threshold_for_orientation, sizeof(__pyx_k_area_threshold_for_orientation), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_arr, __pyx_k_arr, sizeof(__pyx_k_arr), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_kp_s_best_match_idx_d_s, __pyx_k_best_match_idx_d_s, sizeof(__pyx_k_best_match_idx_d_s), 0, 0, 1, 0},
  {&__pyx_n_s_calculate_a_posteriori_estimate, __pyx_k_calculate_a_posteriori_estimate, sizeof(__pyx_k_calculate_a_posteriori_estimate), 0, 0, 1, 1},
  {&__pyx_kp_s_cam_id, __pyx_k_cam_id, sizeof(__pyx_k_cam_id), 0, 0, 1, 0},
  {&__pyx_n_s_camn, __pyx_k_camn, sizeof(__pyx_k_camn), 0, 0, 1, 1},
  {&__pyx_n_s_camn2cam_id, __pyx_k_camn2cam_id, sizeof(__pyx_k_camn2cam_id), 0, 0, 1, 1},
  {&__pyx_n_s_camn_order, __pyx_k_camn_order, sizeof(__pyx_k_camn_order), 0, 0, 1, 1},
  {&__pyx_n_s_camns, __pyx_k_camns, sizeof(__pyx_k_camns), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_closest, __pyx_k_closest, sizeof(__pyx_k_closest), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_kp_s_d_observations_d_estimates_for, __pyx_k_d_observations_d_estimates_for, sizeof(__pyx_k_d_observations_d_estimates_for), 0, 0, 1, 0},
  {&__pyx_n_s_data_descriptions, __pyx_k_data_descriptions, sizeof(__pyx_k_data_descriptions), 0, 0, 1, 1},
  {&__pyx_n_s_data_dict, __pyx_k_data_dict, sizeof(__pyx_k_data_dict), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_debug1, __pyx_k_debug1, sizeof(__pyx_k_debug1), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 0, 1, 1},
  {&__pyx_n_s_disable_image_stat_gating, __pyx_k_disable_image_stat_gating, sizeof(__pyx_k_disable_image_stat_gating), 0, 0, 1, 1},
  {&__pyx_kp_s_disable_image_stat_gating_not_s, __pyx_k_disable_image_stat_gating_not_s, sizeof(__pyx_k_disable_image_stat_gating_not_s), 0, 0, 1, 0},
  {&__pyx_n_s_distort, __pyx_k_distort, sizeof(__pyx_k_distort), 0, 0, 1, 1},
  {&__pyx_n_s_distorted, __pyx_k_distorted, sizeof(__pyx_k_distorted), 0, 0, 1, 1},
  {&__pyx_kp_s_distorted_1f_1f_pixel_dist_1f_ma, __pyx_k_distorted_1f_1f_pixel_dist_1f_ma, sizeof(__pyx_k_distorted_1f_1f_pixel_dist_1f_ma), 0, 0, 1, 0},
  {&__pyx_n_s_distorted_pixel_euclidian_distan, __pyx_k_distorted_pixel_euclidian_distan, sizeof(__pyx_k_distorted_pixel_euclidian_distan), 0, 0, 1, 1},
  {&__pyx_n_s_doing, __pyx_k_doing, sizeof(__pyx_k_doing), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_ekf_observation_covariance_pixel, __pyx_k_ekf_observation_covariance_pixel, sizeof(__pyx_k_ekf_observation_covariance_pixel), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_eye, __pyx_k_eye, sizeof(__pyx_k_eye), 0, 0, 1, 1},
  {&__pyx_n_s_fake_timestamp, __pyx_k_fake_timestamp, sizeof(__pyx_k_fake_timestamp), 0, 0, 1, 1},
  {&__pyx_n_s_fastgeom, __pyx_k_fastgeom, sizeof(__pyx_k_fastgeom), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_filter_data, __pyx_k_filter_data, sizeof(__pyx_k_filter_data), 0, 0, 1, 1},
  {&__pyx_kp_s_filter_data_2, __pyx_k_filter_data_2, sizeof(__pyx_k_filter_data_2), 0, 0, 1, 0},
  {&__pyx_n_s_find2d, __pyx_k_find2d, sizeof(__pyx_k_find2d), 0, 0, 1, 1},
  {&__pyx_n_s_find3d, __pyx_k_find3d, sizeof(__pyx_k_find3d), 0, 0, 1, 1},
  {&__pyx_n_s_first_observation_camns, __pyx_k_first_observation_camns, sizeof(__pyx_k_first_observation_camns), 0, 0, 1, 1},
  {&__pyx_n_s_first_observation_idxs, __pyx_k_first_observation_idxs, sizeof(__pyx_k_first_observation_idxs), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core, __pyx_k_flydra_core, sizeof(__pyx_k_flydra_core), 0, 0, 1, 1},
  {&__pyx_kp_s_flydra_core__flydra_tracked_obje, __pyx_k_flydra_core__flydra_tracked_obje, sizeof(__pyx_k_flydra_core__flydra_tracked_obje), 0, 0, 1, 0},
  {&__pyx_n_s_flydra_core__flydra_tracked_obje_2, __pyx_k_flydra_core__flydra_tracked_obje_2, sizeof(__pyx_k_flydra_core__flydra_tracked_obje_2), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core_data_descriptions, __pyx_k_flydra_core_data_descriptions, sizeof(__pyx_k_flydra_core_data_descriptions), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core_geom, __pyx_k_flydra_core_geom, sizeof(__pyx_k_flydra_core_geom), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core_kalman_ekf, __pyx_k_flydra_core_kalman_ekf, sizeof(__pyx_k_flydra_core_kalman_ekf), 0, 0, 1, 1},
  {&__pyx_n_s_flydra_core_kalman_point_prob, __pyx_k_flydra_core_kalman_point_prob, sizeof(__pyx_k_flydra_core_kalman_point_prob), 0, 0, 1, 1},
  {&__pyx_n_s_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 0, 1, 1},
  {&__pyx_n_s_geom, __pyx_k_geom, sizeof(__pyx_k_geom), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_model_with_jacobian, __pyx_k_get_model_with_jacobian, sizeof(__pyx_k_get_model_with_jacobian), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hx, __pyx_k_hx, sizeof(__pyx_k_hx), 0, 0, 1, 1},
  {&__pyx_kp_s_idx_d_s, __pyx_k_idx_d_s, sizeof(__pyx_k_idx_d_s), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_initial_P, __pyx_k_initial_P, sizeof(__pyx_k_initial_P), 0, 0, 1, 1},
  {&__pyx_n_s_initial_acceleration_covariance, __pyx_k_initial_acceleration_covariance, sizeof(__pyx_k_initial_acceleration_covariance), 0, 0, 1, 1},
  {&__pyx_n_s_initial_position_covariance_esti, __pyx_k_initial_position_covariance_esti, sizeof(__pyx_k_initial_position_covariance_esti), 0, 0, 1, 1},
  {&__pyx_n_s_initial_velocity_covariance_esti, __pyx_k_initial_velocity_covariance_esti, sizeof(__pyx_k_initial_velocity_covariance_esti), 0, 0, 1, 1},
  {&__pyx_n_s_initial_x, __pyx_k_initial_x, sizeof(__pyx_k_initial_x), 0, 0, 1, 1},
  {&__pyx_n_s_inv, __pyx_k_inv, sizeof(__pyx_k_inv), 0, 0, 1, 1},
  {&__pyx_n_s_isEKF, __pyx_k_isEKF, sizeof(__pyx_k_isEKF), 0, 0, 1, 1},
  {&__pyx_n_s_isfinite, __pyx_k_isfinite, sizeof(__pyx_k_isfinite), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_kalman, __pyx_k_kalman, sizeof(__pyx_k_kalman), 0, 0, 1, 1},
  {&__pyx_n_s_kalman_ekf, __pyx_k_kalman_ekf, sizeof(__pyx_k_kalman_ekf), 0, 0, 1, 1},
  {&__pyx_n_s_kalman_model, __pyx_k_kalman_model, sizeof(__pyx_k_kalman_model), 0, 0, 1, 1},
  {&__pyx_n_s_level, __pyx_k_level, sizeof(__pyx_k_level), 0, 0, 1, 1},
  {&__pyx_n_s_linalg, __pyx_k_linalg, sizeof(__pyx_k_linalg), 0, 0, 1, 1},
  {&__pyx_n_s_line_from_HZline, __pyx_k_line_from_HZline, sizeof(__pyx_k_line_from_HZline), 0, 0, 1, 1},
  {&__pyx_n_s_line_from_points, __pyx_k_line_from_points, sizeof(__pyx_k_line_from_points), 0, 0, 1, 1},
  {&__pyx_n_s_mahalanobis, __pyx_k_mahalanobis, sizeof(__pyx_k_mahalanobis), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_max_frames_skipped, __pyx_k_max_frames_skipped, sizeof(__pyx_k_max_frames_skipped), 0, 0, 1, 1},
  {&__pyx_n_s_max_variance_dist_meters, __pyx_k_max_variance_dist_meters, sizeof(__pyx_k_max_variance_dist_meters), 0, 0, 1, 1},
  {&__pyx_n_s_missing_data, __pyx_k_missing_data, sizeof(__pyx_k_missing_data), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_newarr, __pyx_k_newarr, sizeof(__pyx_k_newarr), 0, 0, 1, 1},
  {&__pyx_kp_s_not_acceptable, __pyx_k_not_acceptable, sizeof(__pyx_k_not_acceptable), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj_id, __pyx_k_obj_id, sizeof(__pyx_k_obj_id), 0, 0, 1, 1},
  {&__pyx_n_s_obs0_Lcoords, __pyx_k_obs0_Lcoords, sizeof(__pyx_k_obs0_Lcoords), 0, 0, 1, 1},
  {&__pyx_n_s_obs0_position, __pyx_k_obs0_position, sizeof(__pyx_k_obs0_position), 0, 0, 1, 1},
  {&__pyx_n_s_obs2d_hashable, __pyx_k_obs2d_hashable, sizeof(__pyx_k_obs2d_hashable), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_orientation_consensus, __pyx_k_orientation_consensus, sizeof(__pyx_k_orientation_consensus), 0, 0, 1, 1},
  {&__pyx_n_s_original_cam_ids_and_points2d, __pyx_k_original_cam_ids_and_points2d, sizeof(__pyx_k_original_cam_ids_and_points2d), 0, 0, 1, 1},
  {&__pyx_n_s_original_camns_and_idxs, __pyx_k_original_camns_and_idxs, sizeof(__pyx_k_original_camns_and_idxs), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pmats_and_points_cov, __pyx_k_pmats_and_points_cov, sizeof(__pyx_k_pmats_and_points_cov), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_kp_s_position_MLE_used_camns_and_idxs, __pyx_k_position_MLE_used_camns_and_idxs, sizeof(__pyx_k_position_MLE_used_camns_and_idxs), 0, 0, 1, 0},
  {&__pyx_n_s_pprint, __pyx_k_pprint, sizeof(__pyx_k_pprint), 0, 0, 1, 1},
  {&__pyx_kp_s_predicted_2d_undistorted, __pyx_k_predicted_2d_undistorted, sizeof(__pyx_k_predicted_2d_undistorted), 0, 0, 1, 0},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_kp_s_pt_area_f_cur_d_mean_1f_sumsqf, __pyx_k_pt_area_f_cur_d_mean_1f_sumsqf, sizeof(__pyx_k_pt_area_f_cur_d_mean_1f_sumsqf), 0, 0, 1, 0},
  {&__pyx_n_s_pt_idx, __pyx_k_pt_idx, sizeof(__pyx_k_pt_idx), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_TrackedObject, __pyx_k_pyx_unpickle_TrackedObject, sizeof(__pyx_k_pyx_unpickle_TrackedObject), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reconstructor, __pyx_k_reconstructor, sizeof(__pyx_k_reconstructor), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_return_line_coords, __pyx_k_return_line_coords, sizeof(__pyx_k_return_line_coords), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_removing_previous_observation, __pyx_k_s_removing_previous_observation, sizeof(__pyx_k_s_removing_previous_observation), 0, 0, 1, 0},
  {&__pyx_n_s_save_all_data, __pyx_k_save_all_data, sizeof(__pyx_k_save_all_data), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_skip_data_association, __pyx_k_skip_data_association, sizeof(__pyx_k_skip_data_association), 0, 0, 1, 1},
  {&__pyx_kp_s_so_far_the_best_taking, __pyx_k_so_far_the_best_taking, sizeof(__pyx_k_so_far_the_best_taking), 0, 0, 1, 0},
  {&__pyx_n_s_some_rough_negative_log_likeliho, __pyx_k_some_rough_negative_log_likeliho, sizeof(__pyx_k_some_rough_negative_log_likeliho), 0, 0, 1, 1},
  {&__pyx_n_s_sqrt, __pyx_k_sqrt, sizeof(__pyx_k_sqrt), 0, 0, 1, 1},
  {&__pyx_n_s_ss, __pyx_k_ss, sizeof(__pyx_k_ss), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_step1__calculate_a_priori, __pyx_k_step1__calculate_a_priori, sizeof(__pyx_k_step1__calculate_a_priori), 0, 0, 1, 1},
  {&__pyx_n_s_step2__calculate_a_posteriori, __pyx_k_step2__calculate_a_posteriori, sizeof(__pyx_k_step2__calculate_a_posteriori), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_to_hz, __pyx_k_to_hz, sizeof(__pyx_k_to_hz), 0, 0, 1, 1},
  {&__pyx_n_s_tostring, __pyx_k_tostring, sizeof(__pyx_k_tostring), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_s_updating_for_d_frames_since_upda, __pyx_k_updating_for_d_frames_since_upda, sizeof(__pyx_k_updating_for_d_frames_since_upda), 0, 0, 1, 0},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_wateri, __pyx_k_wateri, sizeof(__pyx_k_wateri), 0, 0, 1, 1},
  {&__pyx_kp_s_will_kill_next_time_because_Pmea, __pyx_k_will_kill_next_time_because_Pmea, sizeof(__pyx_k_will_kill_next_time_because_Pmea), 0, 0, 1, 0},
  {&__pyx_kp_s_will_kill_next_time_because_fram, __pyx_k_will_kill_next_time_because_fram, sizeof(__pyx_k_will_kill_next_time_because_fram), 0, 0, 1, 0},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_xhat, __pyx_k_xhat, sizeof(__pyx_k_xhat), 0, 0, 1, 1},
  {&__pyx_n_s_xhat_k1, __pyx_k_xhat_k1, sizeof(__pyx_k_xhat_k1), 0, 0, 1, 1},
  {&__pyx_n_s_xhatminus, __pyx_k_xhatminus, sizeof(__pyx_k_xhatminus), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 253, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 596, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 947, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "flydra_core/_flydra_tracked_object.pyx":94
 * 
 *         # fill prediction vector [ h(xhatminus) ]
 *         hx_i = pinhole_model.evaluate(xhatminus[:3])             # <<<<<<<<<<<<<<
 *         hx[2*i:2*i+2] = hx_i
 * 
 */
  __pyx_slice_ = PySlice_New(Py_None, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);

  /* "flydra_core/_flydra_tracked_object.pyx":124
 * 
 *     # sort array based on camn
 *     camns = arr[0::2]             # <<<<<<<<<<<<<<
 *     pt_idx = arr[1::2]
 *     camn_order = numpy.argsort(camns)
 */
  __pyx_slice__2 = PySlice_New(__pyx_int_0, Py_None, __pyx_int_2); if (unlikely(!__pyx_slice__2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__2);
  __Pyx_GIVEREF(__pyx_slice__2);

  /* "flydra_core/_flydra_tracked_object.pyx":125
 *     # sort array based on camn
 *     camns = arr[0::2]
 *     pt_idx = arr[1::2]             # <<<<<<<<<<<<<<
 *     camn_order = numpy.argsort(camns)
 * 
 */
  __pyx_slice__3 = PySlice_New(__pyx_int_1, Py_None, __pyx_int_2); if (unlikely(!__pyx_slice__3)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__3);
  __Pyx_GIVEREF(__pyx_slice__3);

  /* "flydra_core/_flydra_tracked_object.pyx":193
 *         self.kill_me = False
 *         self.reconstructor = reconstructor
 *         self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)             # <<<<<<<<<<<<<<
 *         self.disable_image_stat_gating = disable_image_stat_gating
 *         self.orientation_consensus = orientation_consensus
 */
  __pyx_tuple__4 = PyTuple_Pack(2, __pyx_n_s_distorted_pixel_euclidian_distan, Py_None); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "flydra_core/_flydra_tracked_object.pyx":215
 *             P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
 *         for i in range(3,6):
 *             P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)             # <<<<<<<<<<<<<<
 *         if ss >= 9:
 *             for i in range(6,9):
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_n_s_initial_velocity_covariance_esti, __pyx_int_0); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "flydra_core/_flydra_tracked_object.pyx":224
 *         self.max_frames_skipped = kalman_model['max_frames_skipped']
 * 
 *         if kalman_model.get( 'isEKF', False):             # <<<<<<<<<<<<<<
 *             self.ekf_kalman_A = kalman_model['A']
 *             self.ekf_kalman_Q = kalman_model['Q']
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_n_s_isEKF, Py_False); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "flydra_core/_flydra_tracked_object.pyx":276
 *             start_idx = max( N_pts-10, 0 )
 *             for i in range(start_idx,N_pts):
 *                 this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)             # <<<<<<<<<<<<<<
 *                 sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
 *                 if self.frames[i] in self.observations_frames:
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_2); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "flydra_core/_flydra_tracked_object.pyx":376
 *             # Step 2. Filter incoming 2D data to use informative points (data association)
 *             if skip_data_association:
 *                 position_MLE = numpy.nan*numpy.ones( (3,)) # skip             # <<<<<<<<<<<<<<
 *                 Lcoords = None
 *                 used_camns_and_idxs = original_camns_and_idxs
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "flydra_core/_flydra_tracked_object.pyx":397
 *                 pmats_and_points_cov = [ (
 *                                           self.reconstructor.get_model_with_jacobian(cam_id),
 *                                           value_tuple[:2],#just first 2 components (x,y) become xy2d_observed             # <<<<<<<<<<<<<<
 *                                           self.ekf_observation_covariance_pixels)
 *                                          for (cam_id,value_tuple) in cam_ids_and_points2d]
 */
  __pyx_slice__15 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__15)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__15);
  __Pyx_GIVEREF(__pyx_slice__15);

  /* "flydra_core/_flydra_tracked_object.pyx":596
 *                     if tmp is None:
 *                         if not self.disable_image_stat_gating:
 *                             raise ValueError(             # <<<<<<<<<<<<<<
 *                                 '--disable-image-stat-gating not specified and '
 *                                 'no image statistics were saved.')
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_disable_image_stat_gating_not_s); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "flydra_core/_flydra_tracked_object.pyx":622
 * 
 *                         if Pminus_inv is None:
 *                             Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )             # <<<<<<<<<<<<<<
 *                         best_3d_location = _mahalanobis.line_fit_3d(
 *                             projected_line, fast_prediction_3d, Pminus_inv )
 */
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_slice_, __pyx_slice_); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":947
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":953
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "flydra_core/_flydra_tracked_object.pyx":118
 *     return y, hx, C, R, missing_data
 * 
 * def obs2d_hashable( arr ):             # <<<<<<<<<<<<<<
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16
 */
  __pyx_tuple__23 = PyTuple_Pack(6, __pyx_n_s_arr, __pyx_n_s_camns, __pyx_n_s_pt_idx, __pyx_n_s_camn_order, __pyx_n_s_newarr, __pyx_n_s_val); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_flydra_core__flydra_tracked_obje, __pyx_n_s_obs2d_hashable, 118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 118, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_TrackedObject(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__25 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TrackedObject, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_224123000 = PyInt_FromLong(224123000L); if (unlikely(!__pyx_int_224123000)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject = &__pyx_vtable_11flydra_core_22_flydra_tracked_object_TrackedObject;
  __pyx_vtable_11flydra_core_22_flydra_tracked_object_TrackedObject.calculate_a_posteriori_estimate = (PyObject *(*)(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *, long, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate *__pyx_optional_args))__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject_calculate_a_posteriori_estimate;
  __pyx_vtable_11flydra_core_22_flydra_tracked_object_TrackedObject._filter_data = (PyObject *(*)(struct __pyx_obj_11flydra_core_22_flydra_tracked_object_TrackedObject *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data *__pyx_optional_args))__pyx_f_11flydra_core_22_flydra_tracked_object_13TrackedObject__filter_data;
  if (PyType_Ready(&__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject.tp_dictoffset && __pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 136, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__.doc = __pyx_doc_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_11flydra_core_22_flydra_tracked_object_13TrackedObject___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject.tp_dict, __pyx_vtabptr_11flydra_core_22_flydra_tracked_object_TrackedObject) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TrackedObject, (PyObject *)&__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __pyx_ptype_11flydra_core_22_flydra_tracked_object_TrackedObject = &__pyx_type_11flydra_core_22_flydra_tracked_object_TrackedObject;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flydra_core._fastgeom"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_11flydra_core_9_fastgeom_ThreeTuple = __Pyx_ImportType(__pyx_t_1, "flydra_core._fastgeom", "ThreeTuple", sizeof(struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_11flydra_core_9_fastgeom_ThreeTuple) __PYX_ERR(4, 4, __pyx_L1_error)
  __pyx_ptype_11flydra_core_9_fastgeom_PlueckerLine = __Pyx_ImportType(__pyx_t_1, "flydra_core._fastgeom", "PlueckerLine", sizeof(struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_11flydra_core_9_fastgeom_PlueckerLine) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flydra_core._pmat_jacobian"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian = __Pyx_ImportType(__pyx_t_1, "flydra_core._pmat_jacobian", "PinholeCameraModelWithJacobian", sizeof(struct __pyx_obj_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian) __PYX_ERR(5, 5, __pyx_L1_error)
  __pyx_vtabptr_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian = (struct __pyx_vtabstruct_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian*)__Pyx_GetVtable(__pyx_ptype_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian->tp_dict); if (unlikely(!__pyx_vtabptr_11flydra_core_14_pmat_jacobian_PinholeCameraModelWithJacobian)) __PYX_ERR(5, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flydra_core._mahalanobis"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "line_fit_3d", (void (**)(void))&__pyx_f_11flydra_core_12_mahalanobis_line_fit_3d, "struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *(struct __pyx_obj_11flydra_core_9_fastgeom_PlueckerLine *, struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "dist2", (void (**)(void))&__pyx_f_11flydra_core_12_mahalanobis_dist2, "double (struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, struct __pyx_obj_11flydra_core_9_fastgeom_ThreeTuple *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_flydra_tracked_object(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_flydra_tracked_object(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__flydra_tracked_object(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__flydra_tracked_object(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__flydra_tracked_object(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_flydra_tracked_object' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__flydra_tracked_object(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_flydra_tracked_object", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flydra_core___flydra_tracked_object) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flydra_core._flydra_tracked_object")) {
      if (unlikely(PyDict_SetItemString(modules, "flydra_core._flydra_tracked_object", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "flydra_core/_flydra_tracked_object.pyx":4
 * #cython: language_level=2
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * cimport _fastgeom
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":10
 * cimport _pmat_jacobian
 * 
 * import numpy             # <<<<<<<<<<<<<<
 * import numpy as np
 * import time, sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":11
 * 
 * import numpy
 * import numpy as np             # <<<<<<<<<<<<<<
 * import time, sys
 * import adskalman.adskalman as kalman
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":12
 * import numpy
 * import numpy as np
 * import time, sys             # <<<<<<<<<<<<<<
 * import adskalman.adskalman as kalman
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":13
 * import numpy as np
 * import time, sys
 * import adskalman.adskalman as kalman             # <<<<<<<<<<<<<<
 * 
 * import flydra_core.kalman.ekf as kalman_ekf
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__22);
  __Pyx_GIVEREF(__pyx_n_s__22);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__22);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_adskalman_adskalman, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_kalman, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":15
 * import adskalman.adskalman as kalman
 * 
 * import flydra_core.kalman.ekf as kalman_ekf             # <<<<<<<<<<<<<<
 * #import flydra_core.geom as geom
 * import _fastgeom as geom
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__22);
  __Pyx_GIVEREF(__pyx_n_s__22);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__22);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_flydra_core_kalman_ekf, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_kalman_ekf, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":17
 * import flydra_core.kalman.ekf as kalman_ekf
 * #import flydra_core.geom as geom
 * import _fastgeom as geom             # <<<<<<<<<<<<<<
 * import flydra_core.geom
 * import _mahalanobis
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fastgeom, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_geom, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":18
 * #import flydra_core.geom as geom
 * import _fastgeom as geom
 * import flydra_core.geom             # <<<<<<<<<<<<<<
 * import _mahalanobis
 * import math, struct
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_flydra_core_geom, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flydra_core, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":19
 * import _fastgeom as geom
 * import flydra_core.geom
 * import _mahalanobis             # <<<<<<<<<<<<<<
 * import math, struct
 * import flydra_core.data_descriptions
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_mahalanobis, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mahalanobis, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":20
 * import flydra_core.geom
 * import _mahalanobis
 * import math, struct             # <<<<<<<<<<<<<<
 * import flydra_core.data_descriptions
 * from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_math, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":21
 * import _mahalanobis
 * import math, struct
 * import flydra_core.data_descriptions             # <<<<<<<<<<<<<<
 * from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood
 * import collections
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_flydra_core_data_descriptions, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flydra_core, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":22
 * import math, struct
 * import flydra_core.data_descriptions
 * from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood             # <<<<<<<<<<<<<<
 * import collections
 * from pprint import pprint
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_some_rough_negative_log_likeliho);
  __Pyx_GIVEREF(__pyx_n_s_some_rough_negative_log_likeliho);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_some_rough_negative_log_likeliho);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_flydra_core_kalman_point_prob, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_some_rough_negative_log_likeliho); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_some_rough_negative_log_likeliho, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":23
 * import flydra_core.data_descriptions
 * from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood
 * import collections             # <<<<<<<<<<<<<<
 * from pprint import pprint
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":24
 * from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood
 * import collections
 * from pprint import pprint             # <<<<<<<<<<<<<<
 * 
 * cdef double c_inf
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_pprint);
  __Pyx_GIVEREF(__pyx_n_s_pprint);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_pprint);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pprint, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_pprint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pprint, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":27
 * 
 * cdef double c_inf
 * c_inf = np.inf             # <<<<<<<<<<<<<<
 * 
 * __all__ = ['TrackedObject']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_11flydra_core_22_flydra_tracked_object_c_inf = __pyx_t_3;

  /* "flydra_core/_flydra_tracked_object.pyx":29
 * c_inf = np.inf
 * 
 * __all__ = ['TrackedObject']             # <<<<<<<<<<<<<<
 * 
 * PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_TrackedObject);
  __Pyx_GIVEREF(__pyx_n_s_TrackedObject);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_TrackedObject);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":31
 * __all__ = ['TrackedObject']
 * 
 * PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
 * PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_X); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_X, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":32
 * 
 * PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
 * PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
 * PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_Y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_Y, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":33
 * PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
 * PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
 * PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
 * PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_AREA); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_AREA, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":34
 * PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
 * PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
 * PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
 * PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_SLOPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_SLOPE, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":35
 * PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
 * PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
 * PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
 * PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_ECCENTRICITY); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_ECCENTRICITY, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":36
 * PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
 * PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
 * PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
 * PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_P1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_P1, __pyx_t_2) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":37
 * PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
 * PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
 * PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3
 * PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_P2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_P2, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":38
 * PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
 * PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
 * PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_P3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_P3, __pyx_t_2) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":39
 * PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
 * PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3
 * PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4             # <<<<<<<<<<<<<<
 * 
 * PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_P4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_P4, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":41
 * PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4
 * 
 * PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
 * PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_FRAME_PT_IDX, __pyx_t_2) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":42
 * 
 * PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
 * PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
 * PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_CUR_VAL_IDX, __pyx_t_2) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":43
 * PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
 * PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
 * PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX             # <<<<<<<<<<<<<<
 * PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_MEAN_VAL_IDX, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":44
 * PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
 * PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
 * PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX             # <<<<<<<<<<<<<<
 * 
 * NO_LCOORDS = numpy.nan,numpy.nan,numpy.nan,  numpy.nan,numpy.nan,numpy.nan
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_flydra_core); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_data_descriptions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PT_TUPLE_IDX_SUMSQF_VAL_IDX, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":46
 * PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX
 * 
 * NO_LCOORDS = numpy.nan,numpy.nan,numpy.nan,  numpy.nan,numpy.nan,numpy.nan             # <<<<<<<<<<<<<<
 * 
 * cdef double cnan
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NO_LCOORDS, __pyx_t_2) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":49
 * 
 * cdef double cnan
 * cnan = np.nan             # <<<<<<<<<<<<<<
 * 
 * ctypedef int mybool
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_nan); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_8); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_11flydra_core_22_flydra_tracked_object_cnan = __pyx_t_3;

  /* "flydra_core/_flydra_tracked_object.pyx":118
 *     return y, hx, C, R, missing_data
 * 
 * def obs2d_hashable( arr ):             # <<<<<<<<<<<<<<
 *     # XXX FIXME: this could likely be sped up
 *     assert arr.dtype == numpy.uint16
 */
  __pyx_t_8 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_22_flydra_tracked_object_3obs2d_hashable, NULL, __pyx_n_s_flydra_core__flydra_tracked_obje_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_obs2d_hashable, __pyx_t_8) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TrackedObject(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_8 = PyCFunction_NewEx(&__pyx_mdef_11flydra_core_22_flydra_tracked_object_5__pyx_unpickle_TrackedObject, NULL, __pyx_n_s_flydra_core__flydra_tracked_obje_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TrackedObject, __pyx_t_8) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "flydra_core/_flydra_tracked_object.pyx":1
 * #emacs, this is -*-Python-*- mode             # <<<<<<<<<<<<<<
 * #cython: language_level=2
 * 
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_8) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "../../../../AppData/Local/Packages/PythonSoftwareFoundation.Python.3.8_qbz5n2kfra8p0/LocalCache/local-packages/Python38/site-packages/numpy/__init__.pxd":1016
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init flydra_core._flydra_tracked_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flydra_core._flydra_tracked_object");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseArgTupleInvalid */
  static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
  static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
  static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
  static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SliceObject */
  static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* StringJoin */
  #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* SliceObject */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
  static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
  static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
  static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* PyObjectCallMethod1 */
  static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
  static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyObjectSetAttrStr */
  #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) ((PY_LONG_LONG) 0 - (PY_LONG_LONG) 1), const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) ((PY_LONG_LONG) 0 - (PY_LONG_LONG) 1), const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (PY_LONG_LONG) -1;
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
